/*
 * This file is part of firk's Common C Library
 *  charset/charset.c - charset conversions
 *
 * (c) firk 2019; firk@cantconnect.ru
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. The origin of this software must not be misrepresented; you must
 *    not claim that you wrote the original software.
 * 4. Altered versions in any form must be plainly marked as such, and
 *    must not be misinterpreted as being the original software.
 *
 * This software is provided by the author and contributors `as is'
 * without any express or implied warranty.
 */

#include <sys/types.h>
#include "../types.h"
#include "../string.h"
#include "charset.h"

typedef struct fcl_charset_struct {
  char const * name;
  int is_alias;
  uint32 t[256];
} t_charset;

static t_charset charsets[] = {
#include "charsets.inc"
  { NULL }
};

static t_charset * find_charset(uchar const * name) {
  t_charset * cs;
  cs = charsets;
  while(cs->name && strcmp((char const*)name,cs->name)) cs++;
  while(cs->name && cs->is_alias) cs++;
  return (cs->name?cs:NULL);
}

/*
   1 = non-characters (FDD0..FDEF FFFE, FFFF, xxFFFE, xxFFFF)
   2 = surrogate characters (D800 .. DFFF)
   4 = private-use (E000..F8FF F0000..FFFFD 100000..10FFFD)
   8 = out-of-range (>10FFFF)
 */
static int disallowed_unc(int flags, uint32 uc) {
  if(flags&1) {
    if(uc>=0xFDD0 && uc<=0xFDEF || (uc&0xFFFE)==0xFFFE) return 1;
  }
  if(flags&2) {
    if(uc>=0xD800 && uc<=0xDFFF) return 1;
  }
  if(flags&4) {
    if(uc>=0xE000 && uc<=0xF8FF
     || uc>=0xF0000 && uc<=0xFFFFD
     || uc>=0x100000 && uc<=0x10FFFD) return 1;
  }
  if(flags&8) {
    if(uc>=0x110000) return 1;
  }
  return 0;
}

/*
  invmode:
   0 = keep invalid input "as is" (most likely producing invalid output)
   1 = replace invalid input with U+FFFD (UTF-8: EF BF BD)
   2 = skip invalid input
   in all cases, return value changed from 0 to 3 when invalid input occured
  return:
   0 = ok
   1 = unknown charset
   2 = not enough place in dest
   3 = invalid characters for specified charset, but tried to recover
      example: >=128 for us-ascii ot wrong utf-8 sequence
 */
#define NEED_SPACE(n) if(dsz-p<n) return 2
#define UNICODE_REPLACER() do { NEED_SPACE(3); dest[p++]=0xEF; \
                                dest[p++]=0xBF; dest[p++]=0xBD; } while(0)
#define SINGLE_BYTE(c) do { if(p>=dsz) return 2; dest[p++]=c; } while(0)

#define UC(c) (((c)&0xC0)==0x80)

#define IS_BL(c) (ascii_bl && ((ascii_bl[c/8] >> (c&7)) & 1))

extern unsigned charset_to_utf8(uchar const * src, size_t slen, uchar * dest,
                size_t dsz, size_t * rsz, uchar const * charset, int invmode,
                int u_flags, uchar const * ascii_bl) {
  uchar c;
  size_t p = 0;
  unsigned inv = 0;
  size_t ulen, u;
  t_charset * cs;
  uint32 uc;
  uint16 ca, cb;
  
  if(!strcmp((char*)charset,"us-ascii")) {
    while(slen--) {
      c = *(src++);
      if(c>127) {
        inv = 1;
        if(!invmode) goto inv_tail_as_is;
        if(invmode==1) UNICODE_REPLACER();
        continue;
      }
      if(IS_BL(c)) {
        NEED_SPACE(2);
        dest[p++] = 0xC0|(c>>6);
        dest[p++] = 0x80|(c&63);
      } else SINGLE_BYTE(c);
    }
    *rsz = p;
    return (inv?3:0);
  }
  if(!strcmp((char*)charset,"utf-8")) { if(!u_flags) {
    while(slen--) {
      c = *(src++);
      if(c<=0x7F) {
        if(IS_BL(c)) {
          NEED_SPACE(2);
          dest[p++] = 0xC0|(c>>6);
          dest[p++] = 0x80|(c&63);
        } else SINGLE_BYTE(c);
        continue;
      }
      if(c<=0xBF) {
        inv = 1;
        if(!invmode) goto inv_tail_as_is;
        if(invmode==1) UNICODE_REPLACER();
        continue;
      }
      if(c<=0xDF) ulen=1;
      else if(c<=0xEF) ulen=2;
      else if(c<=0xF7) ulen=3;
      else if(c<=0xFB) ulen=4;
      else if(c<=0xFD) ulen=5;
      else {
uc_wrong:
        inv = 1;
        if(!invmode) goto inv_tail_as_is;
        if(invmode==1) UNICODE_REPLACER();
        while(slen && UC(*src)) { slen--; src++; }
        continue;
      }
      if(slen<ulen) goto uc_wrong;
      for(u=0; u<ulen; u++) if(!UC(src[u])) goto uc_wrong;
      NEED_SPACE(ulen+1);
      dest[p++] = c;
      for(u=0; u<ulen; u++) dest[p++] = *(src++);
    }
    *rsz = p;
    return (inv?3:0);
  } else { /* here u_flags != 0 */
    while(slen--) {
      c = *(src++);
      if(c<=0x7F) {
        if(IS_BL(c)) {
          NEED_SPACE(2);
          dest[p++] = 0xC0|(c>>6);
          dest[p++] = 0x80|(c&63);
        } else SINGLE_BYTE(c);
        continue;
      }
      if(c<=0xBF) {
        inv = 1;
        if(!invmode) SINGLE_BYTE(c);
        else if(invmode==1) UNICODE_REPLACER();
        continue;
      }
      if(c<=0xDF) ulen=1;
      else if(c<=0xEF) ulen=2;
      else if(c<=0xF7) ulen=3;
      else if(c<=0xFB) ulen=4;
      else if(c<=0xFD) ulen=5;
      else {
uc_wrong_f:
        inv = 1;
        if(!invmode) { SINGLE_BYTE(c); continue; }
        if(invmode==1) UNICODE_REPLACER();
        while(slen && UC(*src)) { slen--; src++; }
        continue;
      }
      if(slen<ulen) goto uc_wrong_f;
      for(u=0; u<ulen; u++) if(!UC(src[u])) goto uc_wrong_f;
      uc = (c & (63 >> ulen));
      for(u=0; u<ulen; u++) uc = (uc<<6)|(src[u]&63);
      if(disallowed_unc(u_flags,uc)) goto uc_wrong_f;
      if(u_flags&16) {
        if(uc<=0x7F || uc<=0x7FF && ulen>1 || uc<=0xFFFF && ulen>2 ||
           uc<=0x1FFFFF && ulen>3 || uc<=0x3FFFFFF && ulen>4) goto uc_wrong_f;
      }
      if(!(u_flags&32)) {
        NEED_SPACE(ulen+1);
        dest[p++] = c;
        for(u=0; u<ulen; u++) dest[p++] = *(src++);
      } else {
        if(uc<=0x7F && !IS_BL(uc)) {
          SINGLE_BYTE((uchar)uc);
        } else if(uc<=0x7FF) {
          NEED_SPACE(2);
          dest[p++] = 0xC0 | (uc>>6);
          dest[p++] = 0x80 | (uc&63);
        } else if(uc<=0xFFFF) {
          NEED_SPACE(3);
          dest[p++] = 0xE0 | (uc>>12);
          dest[p++] = 0x80 | ((uc>>6)&63);
          dest[p++] = 0x80 | (uc&63);
        } else if(uc<=0x1FFFFF) {
          NEED_SPACE(4);
          dest[p++] = 0xF0 | (uc>>18);
          dest[p++] = 0x80 | ((uc>>12)&63);
          dest[p++] = 0x80 | ((uc>>6)&63);
          dest[p++] = 0x80 | (uc&63);
        } else if(uc<=0x3FFFFFF) {
          NEED_SPACE(5);
          dest[p++] = 0xF8 | (uc>>24);
          dest[p++] = 0x80 | ((uc>>18)&63);
          dest[p++] = 0x80 | ((uc>>12)&63);
          dest[p++] = 0x80 | ((uc>>6)&63);
          dest[p++] = 0x80 | (uc&63);
        } else if(uc<=0x7FFFFFFF) {
          NEED_SPACE(6);
          dest[p++] = 0xFC | (uc>>30);
          dest[p++] = 0x80 | ((uc>>24)&63);
          dest[p++] = 0x80 | ((uc>>18)&63);
          dest[p++] = 0x80 | ((uc>>12)&63);
          dest[p++] = 0x80 | ((uc>>6)&63);
          dest[p++] = 0x80 | (uc&63);
        } else goto uc_wrong_f;
      }
    }
    *rsz = p;
    return (inv?3:0);
  } }
/* TODO
  if(!strcmp((char*)charset,"utf-16le")) {
    slen/=2;
    while(slen--) {
      ca = ((uint16)(src[0])) | (((uint16)(src[0]))<<8);
      src += 2;
    }
  }
*/
  if(cs = find_charset(charset)) {
    while(slen--) {
      c = *(src++);
      uc = cs->t[c];
      if(uc==0xFFFF) goto cs_wrong;
      if(uc<=0x7F && !IS_BL(uc)) {
        SINGLE_BYTE((uchar)uc);
      } else if(uc<=0x7FF) {
        NEED_SPACE(2);
        dest[p++] = 0xC0 | (uc>>6);
        dest[p++] = 0x80 | (uc&63);
      } else if(uc<=0xFFFF) {
        NEED_SPACE(3);
        dest[p++] = 0xE0 | (uc>>12);
        dest[p++] = 0x80 | ((uc>>6)&63);
        dest[p++] = 0x80 | (uc&63);
      } else if(uc<=0x1FFFFF) {
        NEED_SPACE(4);
        dest[p++] = 0xF0 | (uc>>18);
        dest[p++] = 0x80 | ((uc>>12)&63);
        dest[p++] = 0x80 | ((uc>>6)&63);
        dest[p++] = 0x80 | (uc&63);
      } else if(uc<=0x3FFFFFF) {
        NEED_SPACE(5);
        dest[p++] = 0xF8 | (uc>>24);
        dest[p++] = 0x80 | ((uc>>18)&63);
        dest[p++] = 0x80 | ((uc>>12)&63);
        dest[p++] = 0x80 | ((uc>>6)&63);
        dest[p++] = 0x80 | (uc&63);
      } else if(uc<=0x7FFFFFFF) {
        NEED_SPACE(6);
        dest[p++] = 0xFC | (uc>>30);
        dest[p++] = 0x80 | ((uc>>24)&63);
        dest[p++] = 0x80 | ((uc>>18)&63);
        dest[p++] = 0x80 | ((uc>>12)&63);
        dest[p++] = 0x80 | ((uc>>6)&63);
        dest[p++] = 0x80 | (uc&63);
      } else {
cs_wrong:
        inv = 1;
        if(invmode==1) UNICODE_REPLACER();
        else if(!invmode) SINGLE_BYTE(c);
      }
    }
    *rsz = p;
    return (inv?3:0);
  }
  return 1;
inv_tail_as_is:
  /* store "c" and raw copy tail */
  /* used only by us-ascii and utf-8 charsets as they does not need
     actual conversion */
  NEED_SPACE(slen+1);
  dest[p++] = c;
  bcopy(src, dest+p, slen);
  *rsz = p+slen;
  return 3;
}

