/*
 * This file is part of firk's Common C Library
 *  charset/gen.c - generator for bundled charsets file
 *
 * (c) firk 2019; firk@cantconnect.ru
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. The origin of this software must not be misrepresented; you must
 *    not claim that you wrote the original software.
 * 4. Altered versions in any form must be plainly marked as such, and
 *    must not be misinterpreted as being the original software.
 *
 * This software is provided by the author and contributors `as is'
 * without any express or implied warranty.
 */

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <errno.h>

static void log_skip(unsigned ln, unsigned first, unsigned last) {
  if(first==last) fprintf(stderr, "> line %u - warning: skipping 0x%02X character\n", ln, first);
  else fprintf(stderr, "> line %u - warning: skipping 0x%02X...0x%02X characters\n", ln, first, last);
}

static int parse_file(char const * fn) {
  FILE * fp;
  char buf[256];
  size_t l;
  unsigned ln = 0, idx = 0, skip1;
  char c, *b, *u;
  unsigned long bb, uu;
  
  if(!(fp = fopen(fn,"rb"))) {
    fprintf(stderr, "can't open %s\n", fn);
    return -1;
  }
  while(fgets(buf, sizeof(buf), fp)) {
    ln++;
    l = strlen(buf);
    if(!l) {
      fprintf(stderr, "> line %u - fgets() returned zero-length string\n", ln);
      return -1;
    }
    if(l==1 && *buf==26) break; /* support DOS ^Z end of file mark */
    if(buf[l-1]!='\n') {
      fprintf(stderr, "> line %u - missing EOLN or line too long\n", ln);
      return -1;
    }
    buf[l-1] = 0;
    for(b=buf; (c=*b)==32 || c==9; b++);
    if(!c || c=='#') continue;
    errno=0; bb=strtoul(b,&u,0);
    if(errno || u==b || (c=*u) && c!=9 && c!=32 && c!='#') {
      fprintf(stderr, "> line %u - syntax error\n", ln);
      return -1;
    }
    skip1 = 0xFFFF;
    while(bb>idx && idx<0x100) {
      if(skip1==0xFFFF) skip1=idx;
      printf(" 0xFFFF");
      idx++;
      if(idx!=0x100) printf(",");
      if(!(idx&7)) printf("\n");
    }
    if(skip1!=0xFFFF) log_skip(ln, skip1, idx-1);
    if(idx>0xFF) {
      fprintf(stderr, "> line %u - extra line after 0xFF character\n", ln);
      return -1;
    }
    if(bb!=idx) {
      fprintf(stderr, "> line %u - idx 0x%02X expected, got 0x%02X\n", ln, (unsigned)idx, (unsigned)bb);
      return -1;
    }

    while((c=*u)==32 || c==9) u++;
    if(!c || c=='#') {
      uu = 0xFFFF;
    } else {
      errno=0; uu=strtoul(u,&b,0);
      if(errno || b==u || (c=*b) && c!=9 && c!=32 && c!='#') {
        fprintf(stderr, "> line %u - syntax error\n", ln);
        return -1;
      }
    }

    printf(" 0x%04X", (unsigned)uu);
    idx++;
    if(idx!=0x100) printf(",");
    if(!(idx&7)) printf("\n");
  }
  fclose(fp);
  if(idx<0x100) {
    skip1 = idx;
    while(idx<0x100) {
      printf(" 0xFFFF");
      idx++;
      if(idx!=0x100) printf(",");
      if(!(idx&7)) printf("\n");
    }
    log_skip(ln, skip1, idx-1);
  }
  return 0;
}

int main(int argc, char * * argv) {
  int work = 0;
  size_t l;
  unsigned ln = 0;
  char buf[256];
  char c, *b, *fn, *t;
  
  while(fgets(buf, sizeof(buf), stdin)) {
    work = 1;
    ln++;
    l = strlen(buf);
    if(!l) {
      fprintf(stderr, "line %u - fgets() returned zero-length string\n", ln);
      return -1;
    }
    if(buf[l-1]!='\n') {
      fprintf(stderr, "line %u - missing EOLN or line too long\n", ln);
      return -1;
    }
    buf[l-1] = 0;
    for(b=buf; (c=*b)==32 || c==9; b++);
    if(!c || c=='#') continue;
    for(fn=b; (c=*fn)!=32 && c!=9; fn++) {
      if(!c) {
        fprintf(stderr, "line %u - syntax error\n", ln);
        return -1;
      }
      if((c<'a' || c>'z') && (c<'0' || c>'9') && c!='-') {
        fprintf(stderr, "line %u - bad characters in charset name, "
                        "allowed are: a-z, 0-9, -\n", ln);
        return -1;
      }
    }
    *(fn++) = 0;
    while((c=*fn)==32 || c==9) fn++;
    if(!c || c=='#') {
      fprintf(stderr, "line %u - syntax error\n", ln);
      return -1;
    }
    for(t=fn; (c=*t) && c!=32 && c!=9 && c!='#'; t++) {
      if((c<33 || c>126) && !(c&0x80)) {
        fprintf(stderr, "line %u - bad characters in filename\n", ln);
        return -1;
      }
    }
    *t = 0;
    if(!strcmp(fn,"=>")) {
      fprintf(stderr, "line %u charset-alias %s\n", ln, b);
      printf(" { \"%s\", 1 },\n", b);
    } else {
      fprintf(stderr, "line %u charset %s file %s\n", ln, b, fn);
      printf(" { \"%s\", 0, {\n", b);
      if(parse_file(fn)<0) return -1;
      printf(" } },\n");
    }
  }
  if(!work) {
    fprintf(stderr, "Usage:\n"
                    "  ./gen < charset_list.txt > charsets.inc\n"
                    "Input syntax is:\n"
                    "  # this is comment\n"
                    "  charset-name-12345-alias =>\n"
                    "  charset-name-12345 charset_file.txt\n"
                    "Alias charset names should be before main entry\n"
                    "All charset names should be lowercase\n");
    return -1;
  }
  return 0;
}
