/*
 * This file is part of firk's Common C Library
 *  conio/conio.c - interactive console i/o
 *
 * (c) firk 2015; firk@cantconnect.ru
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. The origin of this software must not be misrepresented; you must
 *    not claim that you wrote the original software.
 * 4. Altered versions in any form must be plainly marked as such, and
 *    must not be misinterpreted as being the original software.
 *
 * This software is provided by the author and contributors `as is'
 * without any express or implied warranty.
 */

#include <stdarg.h>
#include <string.h>
#include <termios.h>
#include <unistd.h>
#include <errno.h>
#include <sys/select.h>
#include <sys/ioctl.h>
#include "conio.h"

#ifndef STDIN_FILENO
#define STDIN_FILENO 0
#endif
#ifndef STDOUT_FILENO
#define STDOUT_FILENO 1
#endif

static unsigned int cbreak_mode = 0;
static unsigned int current_text_color = 7;
static unsigned int current_text_background = 0;
static struct termios saved_termios;
static int conio_initialized = 0;

static void refresh_textattr(void);
int conio_init(void) {
  /* set character mode */
  /* should be called at program startup
     and every time when terminal has
     a chance to be messed up by another
     process */
  struct termios t;
  if(tcgetattr(STDIN_FILENO, &t)<0) return -1;
  memcpy(&saved_termios, &t, sizeof(struct termios));
  t.c_iflag &= ~(IXON|IXOFF|ICRNL|IGNCR|INLCR|ISTRIP);
  t.c_iflag |= IMAXBEL;
  t.c_lflag &= ~(ICANON|ECHO|ISIG|IEXTEN);
  if(cbreak_mode) t.c_lflag |= ISIG;
  t.c_oflag &= ~OPOST;
  t.c_cflag &= ~CSIZE;
  t.c_cflag |= CS8;
  t.c_cc[VMIN] = 1;
  t.c_cc[VTIME] = 0;
  while(tcsetattr(STDIN_FILENO, TCSAFLUSH, &t)<0) {
    if(errno!=EINTR) return -1;
  }
  putch_raw(0x1B); putch_raw('7');
  refresh_textattr();
  conio_initialized = 1;
  return 0;
}

int conio_restore(void) {
  putch_raw(0x1B); putch_raw('8');
  while(tcsetattr(STDIN_FILENO, TCSAFLUSH, &saved_termios)<0) {
    if(errno!=EINTR) return -1;
  }
  return 0;
}

int setcbreak(int enable) {
  struct termios t;
  if(tcgetattr(STDIN_FILENO, &t)<0) return -1;
  if(enable) t.c_lflag |= ISIG; else t.c_lflag &= ~ISIG;
  cbreak_mode = (unsigned int)enable;
  while(tcsetattr(STDIN_FILENO, TCSANOW, &t)<0) {
    if(errno!=EINTR) return -1;
  }
  return 0;
}

int get_screensize(unsigned short * x, unsigned short * y) {
  struct winsize wsz;
  if(ioctl(STDOUT_FILENO, TIOCGWINSZ, &wsz)<0) return -1;
  *x = wsz.ws_col;
  *y = wsz.ws_row;
  return 0;
}

int set_screensize(unsigned short x, unsigned short y) {
  struct winsize wsz;
  wsz.ws_col = x;
  wsz.ws_row = y;
  if(ioctl(STDOUT_FILENO, TIOCSWINSZ, &wsz)<0) return -1;
  return 0;
}

int putch_raw(int c) {
  unsigned char ch;
  int res;
  ch = (unsigned char)c;
again:
  res = write(STDOUT_FILENO, &ch, 1);
  if(res==1) return c;
  if(res==-1 && (errno==EAGAIN || errno==EWOULDBLOCK || errno==EINTR)) {
    usleep(1000);
    goto again;
  }
  return EOF;
}

int putch(int c) {
  /* TODO: print special characters like ESC as characters,
     somehow avoid their special interpretation by terminal */
  if(c<0 || c>255) return -1;
  if(c==0x1B) return -1;
  return putch_raw(c);
}

int getche(void) {
  int c;
  c = getch();
  if(c>=0 && c<=255) putch(c);
  return c;
}

int cputs(char const * s) {
  unsigned char c;
  while(c=(unsigned char)(*s)) {
    if(putch((int)c)<0) return -1;
    s++;
  }
  return 0;
}

static void put_ushort(unsigned short u) {
  if(u>=10) put_ushort(u/10);
  putch_raw((int)(u%10+'0'));
}

static void put_escape(char const * fmt, ...) {
  unsigned char c;
  int i;
  va_list arg;
  va_start(arg, fmt);
  while(1) {
    c = (unsigned char)(*fmt);
    fmt++;
    if(!c) break;
    if(c!='\r') putch_raw(c);
    else {
      i = va_arg(arg, int);
      put_ushort((unsigned short)i);
    }
  }
}

void gotoxy(unsigned short x, unsigned short y) {
  if(x<1) x=1; if(y<1) y=1;
  if(x>1000) x=1000; if(y>1000) y=1000;
  put_escape("\x1B" "[\r;\rf", (int)y, (int)x);
}

void gotox(unsigned short x) {
  if(x<1) x=1; if(x>1000) x=1000;
  put_escape("\x1B" "[\rG", (int)x);
}

void clrscr(void) {
  put_escape("\x1B" "[2J");
}

void clreol(void) {
  put_escape("\x1B" "[K");
}

void clrline(void) {
  put_escape("\x1B" "[2K");
}

static int ansicolors[8]={0,4,2,6,1,5,3,7};

void textcolor(int color) {
  current_text_color = (unsigned int)color;
  refresh_textattr();
}

void textbackground(int color) {
  current_text_background = (unsigned int)color;
  refresh_textattr();
}

void textcolors(int fg, int bg) {
  current_text_color = (unsigned int)fg;
  current_text_background = (unsigned int)bg;
  refresh_textattr();
}

static void refresh_textattr(void) {
  put_escape("\x1B" "[0;");
  if(current_text_color&8) put_escape("1;");
  if((current_text_background&8) || (current_text_color&Blink)) {
    put_escape("5;");
  }
  put_escape("3\r;4\rm", ansicolors[current_text_color&7],
    ansicolors[current_text_background&7]);
}



