/*
 * This file is part of firk's Common C Library
 *  conio/getch.c - special keys escape sequences parser
 *
 * (c) firk 2015, 2022; firk@cantconnect.ru
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. The origin of this software must not be misrepresented; you must
 *    not claim that you wrote the original software.
 * 4. Altered versions in any form must be plainly marked as such, and
 *    must not be misinterpreted as being the original software.
 *
 * This software is provided by the author and contributors `as is'
 * without any express or implied warranty.
 */

#include <unistd.h>
#include <assert.h>
#include <sys/select.h>
#include "../types.h"
#include "../recvbuf.h"
#include "keycodes.h"
#include "vt100key.h"
#include "conio.h"

static unsigned char idata[16];
static struct receive_buffer inputbuf = {idata, 16, 0, 0};
static int _sync_input(void) {
  size_t nb, sz;
  int res;
  assert(inputbuf.head>=inputbuf.tail);
  if(inputbuf.tail) {
    sz = inputbuf.head - inputbuf.tail;
    for(nb=0; nb<sz; nb++) idata[nb] = idata[nb+inputbuf.tail];
    inputbuf.tail = 0;
    inputbuf.head = sz;
  }
  if(receive_buffer_set_nonblock_fd(STDIN_FILENO)<0) return -1;
  res = receive_buffer_sync_all(&inputbuf, STDIN_FILENO, &nb);
  return res;
}

static int _wait_input(void) {
  fd_set rfds;
  int res;
  FD_ZERO(&rfds);
  FD_SET(STDIN_FILENO, &rfds);
  res = select(STDIN_FILENO+1, &rfds, NULL, NULL, NULL);
  if(res<=0) return res;
  return FD_ISSET(STDIN_FILENO, &rfds);
}

static int pressed_key = -1;
static vt100key_parser_ctx pctx;

int keypressed(void) {
  struct escseq_struct * cur;
  char const * cc;
  size_t pos;
  int32 rr;
  if(pressed_key>=0) return 1;
  _sync_input();
  assert(!inputbuf.tail);
  if(inputbuf.head<1) return 0;
  if(idata[0]!=0x1B && !pctx.state) {
    pressed_key = idata[0];
    inputbuf.tail++;
    return 1;
  }
  for(pos=0; pos<inputbuf.head; ) {
    rr = vt100key_parse_byte(&pctx, idata[pos++]);
    if(rr>=0) { pressed_key = rr; inputbuf.tail = pos; return 1; }
  }
  inputbuf.head = inputbuf.tail = 0;
  return 0;
}

int getch(void) {
  int res;
  while(!keypressed()) _wait_input();
  /* -1 keypressed() leads to return of pressed_key equal to -1 */
  res = pressed_key;
  pressed_key = -1;
  return res;
}


