/*
 * This file is part of firk's Common C Library
 *  conio/vt100key.c - vt100+ keyboard seqs parser
 *
 * (c) firk 2021, 2022; firk@cantconnect.ru
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. The origin of this software must not be misrepresented; you must
 *    not claim that you wrote the original software.
 * 4. Altered versions in any form must be plainly marked as such, and
 *    must not be misinterpreted as being the original software.
 *
 * This software is provided by the author and contributors `as is'
 * without any express or implied warranty.
 */

#include "../types.h"
#include "keycodes.h"
#include "vt100key.h"

#define ISTATE_NONE 0
#define ISTATE_ESC  1
#define ISTATE_CSI  2
#define ISTATE_SS3  3

static uint32 Mod_to_key(uint16 m) {
  uint32 r;
  if(m<2) return 0;
  m-=1;
  r = 0;
  if(m&1) r |= MKeyShift;
  if(m&2) r |= MKeyAlt;
  if(m&4) r |= MKeyCtrl;
  return r;
}

static uint32 SS3_to_key(uchar c) {
  switch(c) {
  case 32: return 32;
  case 'A': return KeyUp;     /* [1] */
  case 'B': return KeyDown;   /* [1] */
  case 'C': return KeyRight;  /* [1] */
  case 'D': return KeyLeft;   /* [1] */
  case 'E': return KeyCenter; /* [1] */
  case 'F': return KeyEnd;    /* [1]; sometimes CSI 8 ~ */
  /* 'G' */
  case 'H': return KeyHome;   /* [1]; sometimes CSI 7 ~ */
  case 'I': return KeyTab;    /* [1] */
  /* 'J' */
  /* 'K' */
  /* 'L' */
  case 'M': return KeyKEnter;  /* numpad enter with [66] */
  /* 'N' */
  /* 'O' */
  case 'P': return KeyF1;
  case 'Q': return KeyF2;
  case 'R': return KeyF3;
  case 'S': return KeyF4;
  
  case 'X': return KeyKEqual;
  
  case 'j': return KeyKMul;    /* numpad mul with [66] */
  case 'k': return KeyKPlus;   /* numpad plus with [66] */
  case 'l': return KeyKComma;
  case 'm': return KeyKMinus;  /* numpad minus with [66] */
  case 'n': return KeyDelete;  /* this looks never used; instead: CSI 3 ~ */
  case 'o': return KeyKDiv;    /* numpad div with [66] */
  default: return 0;
  }
}

static uint32 CSI_to_key(uint16 n) {
  switch(n) {
  case 11: return KeyF1;
  case 12: return KeyF2;
  case 13: return KeyF3;
  case 14: return KeyF4;
  case 15: return KeyF5;
  case 17: return KeyF6;
  case 18: return KeyF7;
  case 19: return KeyF8;
  case 20: return KeyF9;
  case 21: return KeyF10;
  case 23: return KeyF11;
  case 24: return KeyF12;
  case 25: return KeyF13;
  case 26: return KeyF14;
  case 28: return KeyF15;
  case 29: return KeyF16;
  case 31: return KeyF17;
  case 32: return KeyF18;
  case 33: return KeyF19;
  case 34: return KeyF20;
  case 2: return KeyInsert;
  case 3: return KeyDelete;
  case 5: return KeyPgUp;
  case 6: return KeyPgDn;
  case 7: return KeyHome;
  case 8: return KeyEnd;
  default: return 0;
  }
}

extern int32 vt100key_parse_byte(vt100key_parser_ctx * this, uchar c) {
  uint32 k;
  switch(this->state) {
  case ISTATE_NONE:
    if(c==27) { this->state = ISTATE_ESC; return -1; }
    if(c==10) return KeyEnter;
    /* Ctrl+Enter is always same code as just Enter;
      mc detects Ctrl via:
      1) connection to X11 when run under X11
      2) ioctl(0,TIOCLINUX) when run in native Linux console
      3) some other ways for QNX
      see lib/tty/key.c/get_modifier()
      there is no way to do it over ssh
     */
    return c;
  case ISTATE_ESC:
    switch(c) {
    case 'O': this->state = ISTATE_SS3; this->a[0]=0; return -1;
    case '[': this->state = ISTATE_CSI; this->la=this->na=0; return -1;
    /* we do not try to implement real VT100 ctlseq syntax parser;
       the only VT100 ctlseqs parsed here is CSI (ESC [) and SS3 (ESC O) */
    case 27: k = KeyQuit; break;
    case '1': k = KeyF1; break;
    case '2': k = KeyF2; break;
    case '3': k = KeyF3; break;
    case '4': k = KeyF4; break;
    case '5': k = KeyF5; break;
    case '6': k = KeyF6; break;
    case '7': k = KeyF7; break;
    case '8': k = KeyF8; break;
    case '9': k = KeyF9; break;
    case '0': k = KeyF10; break;
    case '-': k = KeyF11; break; /* invalid for VT100 but it is Alt-"-" */
    case '=': k = KeyF12; break;
    case 10: k = /* NO // MKeyCtrl | */KeyEnter | MKeyAlt; break;
    default: k = c | MKeyAlt; break;
    }
    this->state = ISTATE_NONE; return k;
  case ISTATE_SS3:
    /* SS3 key codes are used mostly when ESC[?1h (DECCKM)
       and/or ESC[?66h (DECNKM) enabled; exact mode-code linkages
       are dependent on terminal model ot terminal emulator
       implementation, but we don't care and just support them all,
       luckily they doesn't overlap */
    if(c>='0' && c<='9') { this->a[0] = this->a[0]*10+(c-'0'); return -1; }
    if(k = SS3_to_key(c)) k = k | Mod_to_key(this->a[0]);
    break;
  case ISTATE_CSI:
    if(c>='0' && c<='9') {
      if(this->na<10) {
        if(this->la==this->na) this->a[this->la++] = 0;
        this->a[this->na] = this->a[this->na]*10+(c-'0');
      }
      return -1;
    }
    if(c==';') {
      if(this->na<10) {
        if(this->la==this->na) this->a[this->la++] = 0;
        this->na++;
      }
      return -1;
    }
    if(c==' ') {
      k = ' ' | MKeyCtrl;
      if(this->la && this->a[0]==1) k |= MKeyAlt;
    } else if(c=='Z') {
      k = KeyTab | MKeyShift;
    } else if(k = SS3_to_key(c)) {
      if(this->la>=1 && this->a[0]!=1) k = 0;
      else if(this->la>=2) k |= Mod_to_key(this->a[1]);
    } else if(c=='~') {
      if(this->la>=1 && (k=CSI_to_key(this->a[0]))) {
        if(this->la>=2) k |= Mod_to_key(this->a[1]);
      }
    }
    break;
  }
  this->state = ISTATE_NONE;
  if(k) return k;
  return -1;
}

