/*
 * This file is part of firk's Common C Library
 *  dynbufs.c - various dynamically allocated buffers
 *
 * (c) firk 2018, 2019, 2020, 2021; firk@cantconnect.ru
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. The origin of this software must not be misrepresented; you must
 *    not claim that you wrote the original software.
 * 4. Altered versions in any form must be plainly marked as such, and
 *    must not be misinterpreted as being the original software.
 *
 * This software is provided by the author and contributors `as is'
 * without any express or implied warranty.
 */

#include <stdio.h>
#include <stdlib.h> /* for abort() after failed vsnprintf(); TODO */
#include <stdarg.h>
#include <assert.h>
#include <errno.h>
#include "types.h"
#include "string.h"
#include "malloc.h"
#include "qlog.h"
#include "dynbufs.h"

extern void dynbuf_init(t_dynbuf * dynbuf,
                        size_t init, size_t step, size_t max) {
  if(init<8) init=8;
  if(step<1) step=1;
  if(max<init) max=init;
  dynbuf->len = 0;
  dynbuf->alloc = init;
  dynbuf->step = step;
  dynbuf->max = max;
  dynbuf->overflowed = 0;
  if(!(dynbuf->buf = fcl_malloc(init))) dynbuf->overflowed = 2;
}

extern void dynbuf_init_data(t_dynbuf * dynbuf, void const * src, size_t sz) {
  size_t a;
  a = sz?sz:1;
  if(!(dynbuf->buf = fcl_malloc(a))) {
    dynbuf->overflowed = 2;
  } else {
    if(sz) bcopy(src, dynbuf->buf, sz);
    dynbuf->overflowed = 0;
  }
  dynbuf->len = sz;
  dynbuf->alloc = a;
  dynbuf->step = 1;
  dynbuf->max = a;
}

extern void dynbuf_init_check(t_dynbuf * dynbuf,
                              size_t init, size_t step, size_t max) {
  if(init<8) init=8;
  if(step<1) step=1;
  if(max<init) max=init;
  dynbuf->len = 0;
  dynbuf->alloc = init;
  dynbuf->step = step;
  dynbuf->max = max;
  dynbuf->overflowed = 0;
  dynbuf->buf = fcl_malloc_check(init);
}

extern void dynbuf_init_data_check(t_dynbuf * dynbuf,
                                   void const * src, size_t sz) {
  size_t a;
  a = sz?sz:1;
  dynbuf->buf = fcl_malloc_check(a);
  if(sz) bcopy(src, dynbuf->buf, sz);
  dynbuf->overflowed = 0;
  dynbuf->len = sz;
  dynbuf->alloc = a;
  dynbuf->step = 1;
  dynbuf->max = a;
}

extern void dynbuf_clear(t_dynbuf * dynbuf) {
  if(dynbuf->buf) fcl_free(dynbuf->buf);
  bzero(dynbuf, sizeof(*dynbuf));
}

extern int dynbuf_write(t_dynbuf * dynbuf, void const * data, size_t sz) {
  size_t cur, alloc, step, max, new;
  unsigned char * newbuf;
  if(dynbuf->overflowed) { errno=ENOSPC; return -1; }
  if(!sz) return 0;
  cur = dynbuf->len;
  alloc = dynbuf->alloc;
  step = dynbuf->step;
  max = dynbuf->max;
  assert(cur<=alloc && alloc<=max);
  if(sz<=alloc-cur) {
    bcopy(data, dynbuf->buf+cur, sz);
    dynbuf->len = cur+sz;
    return 0;
  }
  if(sz>max-cur) { dynbuf->overflowed=1; errno=ENOSPC; return -1; }
  new = alloc;
  if(alloc/10 >= max-new) new=max; else new+=alloc/10;
  if(step > max-new) new=max; else new+=step;
  if(sz > new-cur) new=cur+sz;
  assert(new>cur && new>alloc && new-cur>=sz && new<=max);
  newbuf = fcl_realloc(dynbuf->buf, new);
  if(!newbuf) { dynbuf->overflowed=2; errno=ENOSPC; return -1; }
  dynbuf->buf = newbuf;
  dynbuf->alloc = new;
  bcopy(data, newbuf+cur, sz);
  dynbuf->len = cur+sz;
  return 0;
}

extern int dynbuf_repeat(t_dynbuf * dynbuf, char c, size_t sz) {
  size_t cur, alloc, step, max, new;
  unsigned char * newbuf;
  if(dynbuf->overflowed) { errno=ENOSPC; return -1; }
  if(!sz) return 0;
  cur = dynbuf->len;
  alloc = dynbuf->alloc;
  step = dynbuf->step;
  max = dynbuf->max;
  assert(cur<=alloc && alloc<=max);
  if(sz<=alloc-cur) {
    memset(dynbuf->buf+cur, c, sz);
    dynbuf->len = cur+sz;
    return 0;
  }
  if(sz>max-cur) { dynbuf->overflowed=1; errno=ENOSPC; return -1; }
  new = alloc;
  if(alloc/10 >= max-new) new=max; else new+=alloc/10;
  if(step > max-new) new=max; else new+=step;
  if(sz > new-cur) new=cur+sz;
  assert(new>cur && new>alloc && new-cur>=sz && new<=max);
  newbuf = fcl_realloc(dynbuf->buf, new);
  if(!newbuf) { dynbuf->overflowed=2; errno=ENOSPC; return -1; }
  dynbuf->buf = newbuf;
  dynbuf->alloc = new;
  memset(newbuf+cur, c, sz);
  dynbuf->len = cur+sz;
  return 0;
}

extern int dynbuf_printstr(t_dynbuf * dynbuf, char const * str) {
  return dynbuf_write(dynbuf, str, strlen(str));
}

#define MODE_HSC  0
#define MODE_C    1
#define MODE_URL  2
#define MODE_MOPT 3
static char const h[]="0123456789ABCDEF";
static int dynbuf_printbytes_esc(t_dynbuf * dynbuf,
                                 void const * data, size_t sz, int mode) {
  size_t csz;
  size_t cur, alloc, step, max, new;
  unsigned char * b;
  char const * str;
  char c;
  str = data;
  if(dynbuf->overflowed) { errno=ENOSPC; return -1; }
  while(sz) {
    cur = dynbuf->len;
    alloc = dynbuf->alloc;
    step = dynbuf->step;
    max = dynbuf->max;
    assert(cur<=alloc && alloc<=max);
    csz = sz;
    if(csz<16) csz=16;
    if(csz>alloc-cur) {
      if(csz>max-cur) { dynbuf->overflowed=1; errno=ENOSPC; return -1; }
      new = alloc;
      if(alloc/10 > max-new) new=max; else new+=alloc/10;
      if(step > max-new) new=max; else new+=step;
      if(csz > new-cur) new=cur+csz;
      assert(new>cur && new>alloc && new-cur>=csz && new<=max);
      b = fcl_realloc(dynbuf->buf, new);
      if(!b) { dynbuf->overflowed=2; errno=ENOSPC; return -1; }
      dynbuf->buf = b;
      dynbuf->alloc = alloc = new;
    } else b = dynbuf->buf;
    assert(csz<=alloc-cur);
    c = *(str++);
    sz--;
    if(mode==MODE_HSC) { /* htmlspecialchars() */
      switch(c) {
      case '&': b[cur++]='&';b[cur++]='a';b[cur++]='m';b[cur++]='p';
                b[cur++]=';';break;
      case '<': b[cur++]='&';b[cur++]='l';b[cur++]='t';b[cur++]=';';break;
      case '>': b[cur++]='&';b[cur++]='g';b[cur++]='t';b[cur++]=';';break;
      case '"': b[cur++]='&';b[cur++]='q';b[cur++]='u';b[cur++]='o';
                b[cur++]='t';b[cur++]=';';break;
      default:  b[cur++] = c; break;
      }
    } else if(mode==MODE_URL) { /* urlencode() */
      if(c==' ') b[cur++]='+';
      else if(c<=32 || c>=127 || c=='+' || c=='&' || c=='?' || c=='='
          || c=='#' || c=='%' || c=='\\' || c=='`') {
          b[cur++] = '%';
          b[cur++] = h[(c>>4)&15];
          b[cur++] = h[c&15];
      } else b[cur++] = c;
    } else if(mode==MODE_MOPT) { /* mount -o options */
      if(c=='\\' || c==',') { b[cur++]='\\'; b[cur++] = c; }
      else if(c>=0 && c<=32 || c==127) {
          b[cur++] = '\\';
          b[cur++] = h[(c>>6)&3];
          b[cur++] = h[(c>>3)&7];
          b[cur++] = h[c&7];
      } else b[cur++] = c;
    } else { /* escape like C string literal;
                but really hex may differ from C standard */
      switch(c) {
      case '\\': b[cur++]='\\';b[cur++]='\\';break;
      case '"':  b[cur++]='\\';b[cur++]='"';break;
      case '\n': b[cur++]='\\';b[cur++]='n';break;
      case '\r': b[cur++]='\\';b[cur++]='r';break;
      case '\t': b[cur++]='\\';b[cur++]='t';break;
      default:
        if(c>=0 && c<=31 || c==127) {
          b[cur++] = '\\';
          b[cur++] = 'x';
          b[cur++] = h[(c>>4)&15];
          b[cur++] = h[c&15];
          /* unfortunately C standard allows another hex digits after these
             two, and next 0..9 A..F a..f character may conflict, but this
             function is just to escape dangerous chars for logging and not to
             construct real C strings, so ok */
          /* octal escape sequences is far less readable in logs so don't use
             them */
        } else b[cur++] = c;
      }
    }
    dynbuf->len = cur;
  }
  return 0;
}

extern int dynbuf_printbytes_hsc(t_dynbuf * dynbuf,
                                 void const * data, size_t sz) {
  return dynbuf_printbytes_esc(dynbuf, data, sz, MODE_HSC);
}

extern int dynbuf_printstr_hsc(t_dynbuf * dynbuf, char const * str) {
  if(!str) str="(null)";
  return dynbuf_printbytes_esc(dynbuf, str, strlen(str), MODE_HSC);
}

extern int dynbuf_printbytes_cesc(t_dynbuf * dynbuf,
                                  void const * data, size_t sz) {
  return dynbuf_printbytes_esc(dynbuf, data, sz, MODE_C);
}

extern int dynbuf_printstr_cesc(t_dynbuf * dynbuf, char const * str) {
  if(!str) str="(null)";
  return dynbuf_printbytes_esc(dynbuf, str, strlen(str), MODE_C);
}

extern int dynbuf_printbytes_uesc(t_dynbuf * dynbuf,
                                  void const * data, size_t sz) {
  return dynbuf_printbytes_esc(dynbuf, data, sz, MODE_URL);
}

extern int dynbuf_printstr_uesc(t_dynbuf * dynbuf, char const * str) {
  if(!str) str="(null)";
  return dynbuf_printbytes_esc(dynbuf, str, strlen(str), MODE_URL);
}

extern int dynbuf_printbytes_mesc(t_dynbuf * dynbuf,
                                  void const * data, size_t sz) {
  return dynbuf_printbytes_esc(dynbuf, data, sz, MODE_MOPT);
}

extern int dynbuf_printstr_mesc(t_dynbuf * dynbuf, char const * str) {
  if(!str) str="(null)";
  return dynbuf_printbytes_esc(dynbuf, str, strlen(str), MODE_MOPT);
}

extern void dynbuf_nullterm(t_dynbuf * dynbuf) {
  size_t cur, alloc, step, max, new;
  unsigned char *buf, *newbuf;
  buf = dynbuf->buf;
  cur = dynbuf->len;
  alloc = dynbuf->alloc;
  step = dynbuf->step;
  max = dynbuf->max;
  assert(buf && alloc>=1 && cur<=alloc && alloc<=max && max>0);
  if(cur && !buf[cur-1]) return;
  if(cur<alloc) { buf[cur] = 0; dynbuf->len = cur+1; return; }
  if(cur==max) { buf[cur-1] = 0; return; }
  if(!(newbuf=fcl_realloc(buf, ++alloc))) { buf[cur-1] = 0; return; }
  newbuf[cur++] = 0;
  dynbuf->buf = newbuf;
  dynbuf->alloc = alloc;
  dynbuf->len = cur;
  assert(cur==alloc && cur<=max);
}

extern void dynbuf_drain(t_dynbuf * dynbuf, size_t sz) {
  size_t cur;
  unsigned char * buf;
  cur = dynbuf->len;
  buf = dynbuf->buf;
  assert(cur>=sz);
  cur -= sz;
  if(cur) bcopy(buf+sz, buf, cur);
  dynbuf->len = cur;
}

static int dynbuf_step(t_dynbuf * dynbuf) {
  size_t cur, alloc, step, max, new;
  unsigned char * newbuf;
  if(dynbuf->overflowed) return -1;
  cur = dynbuf->len;
  alloc = dynbuf->alloc;
  step = dynbuf->step;
  max = dynbuf->max;
  assert(cur<=alloc && alloc<=max);
  if(alloc==max) { dynbuf->overflowed=1; return -1; }
  new = alloc;
  if(alloc/10 >= max-new) new=max; else new+=alloc/10;
  if(step > max-new) new=max; else new+=step;
  assert(new>cur && new>alloc && new<=max);
  newbuf = fcl_realloc(dynbuf->buf, new);
  if(!newbuf) { dynbuf->overflowed=2; return -1; }
  dynbuf->buf = newbuf;
  dynbuf->alloc = new;
  return 0;
}

extern int dynbuf_printf(t_dynbuf * dynbuf, char const * fmt, ...) {
  size_t cur, alloc, max;
  int r;
  va_list arg;
  
  if(dynbuf->overflowed) { errno=ENOSPC; return -1; }
  cur = dynbuf->len;
  alloc = dynbuf->alloc;
  max = dynbuf->max;
  assert(cur<=alloc && alloc<=max);
  if(cur==max) { dynbuf->overflowed=1; nospc: errno=ENOSPC; return -1; }
  if(alloc-cur<1 && dynbuf_step(dynbuf)<0) goto nospc;
  while(1) {
    alloc = dynbuf->alloc;
    assert(cur<alloc && alloc<=max);
    /* TODO: custom snprintf which returns size_t as number of bytes */
    /* see also: http://demin.ws/blog/russian/2013/01/28
                                /use-snprintf-on-different-platforms/
        it seems this func incompatible with HP-UX */
    va_start(arg,fmt);
    r = vsnprintf((char*)dynbuf->buf+cur, alloc-cur, fmt, arg);
    va_end(arg);
    if(r<0) { qlog("vsnprintf() returned error; this may be result "
                             "of attempt to print >2GB string"); abort(); }
    if(((unsigned)r)<alloc-cur) {
      dynbuf->len = cur+r;
      return 0;
    }
    if(dynbuf_step(dynbuf)<0) goto nospc;
  }
}


typedef struct {
  t_dynbuf * dynbuf;
  size_t rl, ml, mlp;
} print_ctx;
#define PA print_ctx * ctx,
#define CA ctx,
#define CHECK_STOP if(ctx->dynbuf->overflowed) return 0;
#define remlen ctx->rl
#define maxlen ctx->ml
#define maxlen_present ctx->mlp
static void E_write(PA char const * s, size_t len) {
  if(remlen>len) remlen -= len; else remlen = 0;
  if(maxlen_present) { if(len>maxlen) len=maxlen; maxlen -= len; }
  dynbuf_write(ctx->dynbuf, s, len);
}

static void E_repeat(PA char c, size_t len) {
  if(remlen>len) remlen -= len; else remlen = 0;
  if(maxlen_present) { if(len>maxlen) len=maxlen; maxlen -= len; }
  dynbuf_repeat(ctx->dynbuf, c, len);
}

static void E_print_char(PA char c) {
  if(remlen) remlen--;
  if(maxlen_present) { if(maxlen) maxlen--; else return; }
  dynbuf_write(ctx->dynbuf, &c, 1);
}

#include "strfmt.inc"

extern int dynbuf_qprintf(t_dynbuf * dynbuf, char const * fmt, ...) {
  size_t cur, alloc, max;
  va_list arg;
  print_ctx ctx;
  
  if(dynbuf->overflowed) { errno=ENOSPC; return -1; }
  cur = dynbuf->len;
  alloc = dynbuf->alloc;
  max = dynbuf->max;
  assert(cur<=alloc && alloc<=max);

  ctx.dynbuf = dynbuf;
  va_start(arg,fmt);
  if(E_vprint(&ctx, fmt, arg)<0) {
    qlog("dynbuf_qprintf: bad format string: \"%S\"", fmt);
    exit(-1);
  }
  va_end(arg);
  if(dynbuf->overflowed) { errno=ENOSPC; return -1; }
  return 0;
}
