/*
 * This file is part of firk's Common C Library
 *  dynbufs.h - various dynamically-allocated buffers
 *
 * (c) firk 2018, 2019, 2020; firk@cantconnect.ru
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. The origin of this software must not be misrepresented; you must
 *    not claim that you wrote the original software.
 * 4. Altered versions in any form must be plainly marked as such, and
 *    must not be misinterpreted as being the original software.
 *
 * This software is provided by the author and contributors `as is'
 * without any express or implied warranty.
 */
#ifndef FCL_INCLUDED_DYNBUFS_H
#define FCL_INCLUDED_DYNBUFS_H

/* t_dynbuf is mostly copypasted from sslproxy util.c ~ 2018
 ***************** dynbuf *****************
 initializing funcs may return ->buf=NULL and ->overflowed=2
   in case of memory allocation failure
 initializing funcs with _check suffix uses _check-suffixed mallocs
 all other funcs will
   return -1, errno=ENOSPC, dynbuf->overflowed!=0 in case of errors
   (1 = "size limit reached" 2 = "memory allocation failure")
   TODO: return ENOMEM on overflowed=2 ?
 dynbuf->overflowed!=0 prevents any other writing functions
   except dynbuf_nullterm() from working (they returns -1/ENOSPC immediately)
 dynbuf_nullterm() guarantes that dynbuf->buf will be null-terminated
   it may eat last character of existing bytes in buffer if there is no another
   way to null-terminate it
   it will not add extra zero byte if last byte already zero
   it expects ->buf!=NULL so it should be checked before
 dynbuf_printf may got a error (retval -1) from vsnprintf() and will log about
    that and abort()
    vsnprintf() will return -1 if we try to print more than INT_MAX bytes at
    once (2GB on all usual 32/64 platforms)
    so dynbuf_printf() should not be used to print unknown and possibly very
    large amounts of data
    also it seems will work wrongly on HP-UX, see comments in source
  
 dynbuf_drain will throw assertion failure if sz>dynbuf->len
 
 all funcs may throw some assertion failures on damaged dynbuf struct
 */


typedef struct {
  size_t len, alloc, step, max;
  int overflowed;
  unsigned char * buf;
} t_dynbuf;

extern void dynbuf_init(t_dynbuf * db, size_t init, size_t step, size_t max);
extern void dynbuf_init_data(t_dynbuf * db, void const * src, size_t sz);
extern void dynbuf_init_check
                       (t_dynbuf * db, size_t init, size_t step, size_t max);
extern void dynbuf_init_data_check(t_dynbuf * db, void const * src, size_t sz);
extern void dynbuf_clear(t_dynbuf * db);
extern int dynbuf_write(t_dynbuf * db, void const * data, size_t sz);
extern int dynbuf_repeat(t_dynbuf * db, char c, size_t sz);
extern int dynbuf_printstr(t_dynbuf * db, char const * str);
extern int dynbuf_printbytes_hsc(t_dynbuf * db, void const * data, size_t sz);
extern int dynbuf_printstr_hsc(t_dynbuf * db, char const * str);
extern int dynbuf_printbytes_cesc(t_dynbuf * db, void const * data, size_t sz);
extern int dynbuf_printstr_cesc(t_dynbuf * db, char const * str);
extern int dynbuf_printbytes_uesc(t_dynbuf * db, void const * data, size_t sz);
extern int dynbuf_printstr_uesc(t_dynbuf * db, char const * str);
extern int dynbuf_printbytes_mesc(t_dynbuf * db, void const * data, size_t sz);
extern int dynbuf_printstr_mesc(t_dynbuf * db, char const * str);
extern void dynbuf_nullterm(t_dynbuf * db);
extern void dynbuf_drain(t_dynbuf * db, size_t sz);
extern int dynbuf_printf(t_dynbuf * db, char const * fmt, ...); /* TODO */
extern int dynbuf_qprintf(t_dynbuf * db, char const * fmt, ...);

#endif

