/*
 * This file is part of firk's Common C Library
 *  hashtbl.c - commonly used hashtables
 *
 * (c) firk 2021; firk@cantconnect.ru
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. The origin of this software must not be misrepresented; you must
 *    not claim that you wrote the original software.
 * 4. Altered versions in any form must be plainly marked as such, and
 *    must not be misinterpreted as being the original software.
 *
 * This software is provided by the author and contributors `as is'
 * without any express or implied warranty.
 */

#include <assert.h>
#include "types.h"
#include "malloc.h"
#include "string.h"
#include "strhash.h"
#include "hashtbl.h"

#ifdef HT_STATIC
#define HT fcl_ht_static
#define FN(n) fcl_ht_static_ ## n
#endif
#ifdef HT_MALLOC
#define HT fcl_ht_malloc
#define FN(n) fcl_ht_malloc_ ## n
#endif
#ifdef HT_CUSTOM
#define HT fcl_ht_custom
#define FN(n) fcl_ht_custom_ ## n
#endif

extern int FN(init)(HT * ht, fcl_size w
#ifdef HT_CUSTOM
  , fcl_free_cb cb
#endif
) {
  ht->w = w;
  ht->n = 0;
  if(!(ht->d = fcl_mallocz_a(sizeof(*ht->d), w))) return -1;
#ifdef HT_CUSTOM
  ht->cb = cb;
#endif
  return 0;
}

extern fcl_ht_value * FN(get)(HT * ht, char const * name) {
  fcl_ht_item * hi;
  ulong h;
  h = strhash(name);
  for(hi=ht->d[h%ht->w]; hi; hi=hi->next) {
    if(hi->h==h && !strcmp(hi->name,name)) {
      return &hi->v;
    }
  }
  return NULL;
}

extern fcl_ht_value * FN(getl)(HT * ht, char const * name, size_t l) {
  fcl_ht_item * hi;
  ulong h;
  h = memhash((uchar*)name,l);
  for(hi=ht->d[h%ht->w]; hi; hi=hi->next) {
    if(hi->h==h && !strncmp(hi->name,name,l) && !hi->name[l]) {
      return &hi->v;
    }
  }
  return NULL;
}

extern int FN(unset)(HT * ht, char const * name) {
  fcl_ht_item *hi, **phi;
  ulong h;
  h = strhash(name);
  for(phi=&ht->d[h%ht->w]; hi=*phi; phi=&hi->next) {
    if(hi->h==h && !strcmp(hi->name,name)) {
      *phi = hi->next;
#ifdef HT_MALLOC
      fcl_free(hi->v.ptr);
#endif
#ifdef HT_CUSTOM
      ht->cb(hi->v.ptr);
#endif
      fcl_free_c_check(hi->name);
      fcl_free(hi);
      ht->n--;
      return 1;
    }
  }
  return 0;
}

/*
  is_old = 1 - overwrite existing entry;
               ret. 1=ok, 0=wasn't exist, -1=out of memory
  is_old = 0 - add new entry;
               ret. 0=ok, 1=was exist, -1=out of memory
  is_old = -1 - add or overwrite
               ret. 1=overwriten, 0=added, -1=out of memory
 */
extern int FN(set)(HT * ht, char const * name, fcl_ht_value v, int is_old) {
  fcl_ht_item * hi;
  ulong h, hmod;
  h = strhash(name);
  hmod = h%ht->w;
  for(hi=ht->d[hmod]; hi; hi=hi->next) {
    if(hi->h==h && !strcmp(hi->name,name)) {
      if(is_old==0) return 1;
#ifdef HT_MALLOC
      fcl_free(hi->v.ptr);
#endif
#ifdef HT_CUSTOM
      ht->cb(hi->v.ptr);
#endif
      hi->v = v;
      return 1;
    }
  }
  if(is_old==1) return 0;
  if(!(hi = fcl_mallocz(sizeof(*hi)))) return -1;
  hi->h = h;
  if(!(hi->name = fcl_strdup(name))) { fcl_free_check(hi); return -1; }
  hi->next = ht->d[hmod];
  hi->v = v;
  ht->d[hmod] = hi;
  return 0;
}

extern void FN(clear)(HT * ht) {
  fcl_ht_item *hi, *next;
  ulong h;
  for(h=0; h<ht->w; h++) {
    next = ht->d[h];
    while(hi=next) {
      next = hi->next;
#ifdef HT_MALLOC
      fcl_free(hi->v.ptr);
#endif
#ifdef HT_CUSTOM
      ht->cb(hi->v.ptr);
#endif
      fcl_free_c_check(hi->name);
      fcl_free(hi);
      ht->n--;
    }
    ht->d[h] = NULL;
  }
  assert(ht->n==0);
}

extern void FN(free)(HT * ht) {
  FN(clear)(ht);
  fcl_free_check(ht->d);
  bzero(ht, sizeof(*ht));
}


