/*
 * This file is part of firk's Common C Library
 *  inifile.c - .ini files reader
 *
 * (c) firk 2015; firk@cantconnect.ru
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. The origin of this software must not be misrepresented; you must
 *    not claim that you wrote the original software.
 * 4. Altered versions in any form must be plainly marked as such, and
 *    must not be misinterpreted as being the original software.
 *
 * This software is provided by the author and contributors `as is'
 * without any express or implied warranty.
 */

#include <assert.h>
#include <errno.h>
#include <unistd.h>
#include <fcntl.h>
#include "inifile.h"



static int _allowed_inifile_char(unsigned char c) {
  if(c==9 || c>=32 && c<=126 || c>=0x80) return 1;
  return 0;
}

static int _allowed_ininame_char(unsigned char c) {
  if(c>='A' && c<='Z' || c>='a' && c<='z' || c>='0' && c<='9' ||
     c=='.' || c=='-' || c=='_' || c=='[' || c==']' || c>=0x80) return 1;
  return 0;
}

int fcl_inifile_read(char const * filename, void * udata,
                     fcl_inifile_newsection_func newsec,
                     fcl_inifile_option_func option,
                     unsigned int * ln) {
  int fd, err_save, got_eof, got_err;
  char buf[1024], sbuf[1024], secname[1024], name[1024], value[1024];
  unsigned int bufsz, idx, idxs, j, k;
  int res;
  char c;
  unsigned int line;

  if(ln) *ln = 0;
  got_eof = got_err = bufsz = 0;
  secname[0] = 0;
  fd = open(filename, O_RDONLY);
  if(fd<0) return -1;
  line = 1;
  while(1) {
    if(got_eof==0 && got_err==0 && bufsz<sizeof(buf)) {
      res = read(fd, buf+bufsz, sizeof(buf)-bufsz);
      if(res==0) { got_eof=1; continue; }
      if(res>0) { bufsz+=res; continue; }
      if(errno==EINTR || errno==EAGAIN || errno==EWOULDBLOCK) continue;
      err_save = errno;
      got_err = 1;
      continue;
    }
    if(!bufsz) break;
    idx = 0;
    while(1) {
      assert(idx<=1000 && idx<=bufsz);
      if(bufsz==idx) { if(got_err) goto err; idxs=idx; break; }
      if(bufsz>idx && buf[idx]=='\n') { idxs=idx+1; break; }
      if(bufsz>idx+1 && buf[idx]=='\r' && buf[idx+1]=='\n'
        ) { idxs=idx+2; break; }
      if(idx==1000) {
        if(bufsz==1001 && got_err && buf[1000]=='\r') goto err;
        goto syntax_err;
      }
      c = buf[idx];
      if(!_allowed_inifile_char((unsigned char)c)) goto syntax_err;
      sbuf[idx] = c;
      idx++; 
    }
    while(idx && (sbuf[idx-1]==32 || sbuf[idx-1]==9)) idx--;
    sbuf[idx] = 0;
    bufsz -= idxs;
    for(j=0; j<bufsz; j++) buf[j] = buf[idxs+j];
    if(idx<1) goto next;
    for(j=0; sbuf[j]==9 || sbuf[j]==32; j++);
    assert(sbuf[j]!=0);
    if(sbuf[j]==0 || sbuf[j]==';' || sbuf[j]=='#') goto next;
    if(sbuf[j]=='[') {
      if(sbuf[idx-1]!=']') goto syntax_err;
      j++; idx--;
      for(k=j; k<idx; k++) secname[k-j] = sbuf[k];
      secname[idx-j] = 0;
      if(newsec) if(newsec(filename, line, udata, secname)<0) goto err3;
    } else {
      k=0;
      while(1) {
        c = sbuf[j];
        if(c==9 || c==32 || c=='=') { assert(k!=0); name[k]=0; break; }
        if(!_allowed_ininame_char((unsigned char)c)) goto syntax_err;
        name[k] = c;
        j++; k++;
      }
      while(1) {
        c = sbuf[j++];
        if(c=='=') break;
        if(c!=9 && c!=32) goto syntax_err;
      }
      while(1) {
        c = sbuf[j];
        if(c!=0 && c!=32) break;
        j++;
      }
      if(sbuf[j]!='"') {
        for(k=j; k<idx; k++) value[k-j] = sbuf[k];
        value[idx-j] = 0;
      } else {
        if(j==idx-1 || sbuf[idx-1]!='"') goto syntax_err;
        j++; idx--;
        for(k=j; k<idx; k++) {
          c = sbuf[k];
          if(c=='\\') {
            k++;
            if(k==idx) goto syntax_err;
            c = sbuf[k];
            switch(c) {
            default: goto syntax_err;
            case 'n': c = '\n'; break;
            case 'r': c = '\r'; break;
            case '\\': break;
            }
          }
          value[k-j] = c;
        }
        value[idx-j] = 0;
      }
      if(option) if(option(filename, line, udata, secname, name, value)<0) goto err3;
    }
next:
    line++;
  }
err:
  close(fd);
  if(ln) *ln = line;
  if(got_err) { errno = err_save; return -1; }
  return 0;
syntax_err:
  close(fd);
  if(ln) *ln = line;
  return -2;
err3:
  close(fd);
  if(ln) *ln = line;
  return -3;
}


