/*
 * This file is part of firk's Common C Library
 *  io.c - signal-safe buffered file i/o
 *
 * (c) firk 2020; firk@cantconnect.ru
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. The origin of this software must not be misrepresented; you must
 *    not claim that you wrote the original software.
 * 4. Altered versions in any form must be plainly marked as such, and
 *    must not be misinterpreted as being the original software.
 *
 * This software is provided by the author and contributors `as is'
 * without any express or implied warranty.
 */

#include <stdio.h>
#include <stdlib.h>
#include <errno.h>
#include <unistd.h>
#include <assert.h>
#include "string.h"
#include "malloc.h"
#include "types.h"
#include "io.h"

#error missing fcl_fflush() calls from fcl_fgets() and fcl_puts()
#error fcl_fflush() does not prevents from being called again because
#error                  w_pos remains not null

/*
  allow_intr behaviour:
    1) !allow_intr && fp->allow_intr && *fp->allow_intr
    2) allow_intr && *allow_intr
    (1) or (2) means EINTR will be passed to caller
    any other case means EINTR will be hidden inside wrapper,
    and syscall automatically restarted
*/
#define AI char const volatile * allow_intr

/*
  saved error behaviour:
    when error occurs, it saved via FCL_FERROR flag and fcl->err
    all consecutive i/o calls will return this error
 */

/*
  1) if saved error present: return it
  2) if file not opened for reading: return EBADF
  3) if write buffer not empty: flush it, return if error
  4) if sz==0: return 1, *rsz=0
  5) if read buffer not empty: return data entirely from read buffer,
       do not try to read() underlying file descriptor
  6) if there is no underlying file descriptor: return EOF
  7) if FCL_READ_AHEAD unset: read(sz), and pass results to caller
  TODO: description of logic
 returns:
   *rsz: always number of read bytes (0 if error or EOF)
   1 = read something OR sz=0
   0 = EOF
  -1 = got error from read() syscall, errno set
  -1/ENOMEM = failed to allocate read buffer
  -1/EBADF = fp not open for reading
  -2 = (errno set) pending error from previous operation
       or "pending" error from flushing write buffer before reading
  */

int fcl_fread(FCL_FILE * fp, void * buf, size_t sz, size_t * rsz, AI) {
  size_t p;
  ssize_t r;
  
  if(fp->flags & FCL_FERROR) { errno = fp->err; *rsz = 0; return -2; }
  if(!(fp->flags & FCL_READ)) { errno = fp->err = EBADF; goto err2; }
  if(fp->w_pos && fcl_fflush(fp,allow_intr)<0) { *rsz = 0; return -2; }
  if(!sz) { *rsz = 0; return 1; }
  
  if(fp->r_pos && fp->r_pos!=fp->r_end) {
  retbuf:
    p = fp->r_end - fp->r_pos;
    if(p>sz) {
      bcopy(fp->r_pos, buf, sz);
      fp->r_pos += sz;
      *rsz = sz;
      return 1;
    }
    bcopy(fp->r_pos, buf, p);
    fp->r_pos = fp->r_end;
    *rsz = p;
    return 1;
  }
  if(fp->fd<0) { *rsz = 0; return 0; }
  if(!allow_intr) allow_intr = fp->allow_intr;
  if(fp->flags & FCL_READ_AHEAD) {
    if(!fp->r_size) fp->r_size=FCL_DEFAULT_BUFSIZE;
  }
  if(!(fp->flags & FCL_READ_AHEAD) || sz>=fp->r_size) {
    while(((ssize_t)sz)<0) sz /= 2;
    while((r = read(fp->fd, buf, sz))<0)
      if(errno!=EINTR || !allow_intr || !*allow_intr) goto err;
    *rsz = r;
    return (r?1:0);
  }
  if(!fp->r_buf && !(fp->r_buf=fcl_malloc(fp->r_size)))
    { errno = fp->err = ENOMEM; goto err2; }
  while((r = read(fp->fd, fp->r_buf, fp->r_size))<0)
    if(errno!=EINTR || !allow_intr || !*allow_intr) goto err;
  if(!r) { *rsz = 0; return 0; }
  fp->r_pos = fp->r_buf;
  fp->r_end = fp->r_buf + r;
  goto retbuf;
err: fp->err = errno;
err2: fp->flags |= FCL_FERROR; *rsz = 0; return -1;
}

/* same as fcl_fread(), but repeat read() call until:
  1) error occured
  2) EOF
  3) sz bytes stored to buf
  return values: same, but *rsz can be non-zero even on error
  */
int fcl_fread_all(FCL_FILE * fp, void * buf, size_t sz, size_t * rsz, AI) {
  size_t p, done;
  ssize_t r;
  
  done = 0;
  if(fp->flags & FCL_FERROR) { errno = fp->err; *rsz = 0; return -2; }
  if(!(fp->flags & FCL_READ)) { errno = fp->err = EBADF; goto err2; }
  if(fp->w_pos && fcl_fflush(fp,allow_intr)<0) { *rsz = 0; return -2; }
  if(!sz) { *rsz = 0; return 1; }
  
  if(fp->r_pos && fp->r_pos!=fp->r_end) {
  retbuf:
    p = fp->r_end - fp->r_pos;
    if(p>=sz) {
      bcopy(fp->r_pos, buf, sz);
      fp->r_pos += sz;
      *rsz = done+sz;
      return 1;
    }
    bcopy(fp->r_pos, buf, p);
    fp->r_pos = fp->r_end;
    done += p;
    buf = ((char*)buf) + p;
    sz -= p;
  }
  if(fp->fd<0) { *rsz = done; return 0; }
  if(!allow_intr) allow_intr = fp->allow_intr;
  if(fp->flags & FCL_READ_AHEAD) {
    if(!fp->r_size) fp->r_size=FCL_DEFAULT_BUFSIZE;
  }
  while(!(fp->flags & FCL_READ_AHEAD) || sz>=fp->r_size) {
    p = sz;
    while(((ssize_t)p)<0) p /= 2;
    while((r = read(fp->fd, buf, p))<0)
      if(errno!=EINTR || !allow_intr || !*allow_intr) goto err;
    if(!r) { *rsz = done; return 0; }
    if(sz==(size_t)r) { *rsz = done+r; return 1; }
    done += r;
    buf = ((char*)buf) + r;
    sz -= r;
  }
  if(!fp->r_buf && !(fp->r_buf=fcl_malloc(fp->r_size)))
    { errno = fp->err = ENOMEM; goto err2; }
  while((r = read(fp->fd, fp->r_buf, fp->r_size))<0)
    if(errno!=EINTR || !allow_intr || !*allow_intr) goto err;
  if(!r) { *rsz = done; return 0; }
  fp->r_pos = fp->r_buf;
  fp->r_end = fp->r_buf + r;
  goto retbuf;
err: fp->err = errno;
err2: fp->flags |= FCL_FERROR; *rsz = done; return -1;
}


/*
  make i/o buffers empty
  returns:
    -2/errno = pending error
    -1/EBADF = there is no underlying file descriptor
    -1/errno = error from lseek() while flushing read-ahead
    -1/errno = error from write() while flushing write buffer
       write buffer may be partially flushed
    0 = ok
    NOTE: file position is undefined after failed fcl_fflush with
       non-zero fp->r_pos after it
  */
int fcl_fflush(FCL_FILE * fp, AI) {
  size_t readlen, p;
  ssize_t r;
  off_t seekpos;
  uint8 * wp;
  
  if(fp->flags & FCL_FERROR) { errno = fp->err; return -2; }
  if(fp->fd<0) { errno = fp->err = EBADF; fp->flags |= FCL_FERROR; return -1; }

  if(!fp->r_pos) {
    assert(!fp->r_end);
  } else {
    assert(fp->r_end && fp->r_end>=fp->r_pos);
    if(!(fp->flags & FCL_NOSEEK)
        && (readlen = fp->r_end-fp->r_pos)) {
      assert(!fp->w_pos);
      seekpos = readlen;
      while(lseek(fp->fd, -seekpos, SEEK_CUR)<0) {
        assert(errno!=EINTR);
        /* idk how to handle lseek() EINTR;
           it should not happen according to manpage */
        fp->err = errno; fp->flags |= FCL_FERROR; return -1;
      }
      fp->r_pos = fp->r_end = NULL;
      return 0;
    }
  }
  
  if(!(wp=fp->w_buf)) {
    assert(!fp->w_pos);
    return 0;
  }
  if(!fp->w_pos) return 0;
  assert(fp->w_pos>=wp);
  assert(fp->flags & FCL_WRITE_BUF);
  if(!allow_intr) allow_intr = fp->allow_intr;
  for( ; wp<fp->w_pos; wp+=r) {
    p = fp->w_pos - wp;
    while(((ssize_t)p)<0) p/=2;
    while((r = write(fp->fd, wp, p))<0) {
      if(errno!=EINTR || !allow_intr || !*allow_intr) {
        fp->err = errno;
        if((p=fp->w_pos-wp) && wp!=fp->w_buf) bcopy(wp, fp->w_buf, p);
        fp->w_pos = fp->w_buf+p;
        errno = fp->err; fp->flags |= FCL_FERROR; return -1;
      }
    }
    assert(r>0 && ((size_t)r)<=p);
  }
  fp->w_pos = fp->w_buf;
  return 0;
}




/* return:
  NOTE: unlike fcl_getc(), pending errors and FCL_READ flag are checked even
        when reading can be done from buffer
  1/*c = ok
  0 = EOF
  0 = there is no more data in buffer and there is no underlying fd
  -1/EBADF = file not opened for reading
  -1/ENOMEM = can't allocate read buffer
  -1/errno = got error
  -2/errno = pending error
 */
int fcl_fgetc(FCL_FILE * fp, char * c, AI) {
  ssize_t r;
  if(!(fp->flags & FCL_FERROR)) {
    if(fp->flags & FCL_READ) {
      if(fp->r_pos && fp->r_pos<fp->r_end) {*c = *(fp->r_pos++); return 1;}
      if(fp->fd<0) return 0;
      if(fp->flags & FCL_READ_AHEAD) {
        if(!fp->r_buf) {
          if(!fp->r_size) fp->r_size = FCL_DEFAULT_BUFSIZE;
          if(!(fp->r_buf = fcl_malloc(fp->r_size)))
            { errno = fp->err = ENOMEM; fp->flags |= FCL_FERROR; return -1; }
        }
        assert(fp->r_size);
        while((r = read(fp->fd, fp->r_buf, fp->r_size))<0)
          if(errno!=EINTR || !allow_intr || !*allow_intr)
            { fp->err = errno; fp->flags |= FCL_FERROR; return -1; }
        if(!r) return 0;
        *c = *fp->r_buf;
        fp->r_pos = fp->r_buf + 1;
        fp->r_end = fp->r_buf + r;
        return 1;
      } else {
        while((r = read(fp->fd, c, 1))<0)
          if(errno!=EINTR || !allow_intr || !*allow_intr)
            { fp->err = errno; fp->flags |= FCL_FERROR; return -1; }
        if(!r) return 0;
        return 1;
      }
    } else { errno = fp->err = EBADF; fp->flags |= FCL_FERROR; return -1; }
  } else { errno = fp->err; return -2; }
}


/*
  return:
    1 = ok
    -1/EBADF = file not opened for writing
    -1/EBADF = buffer overflowed and there is not underlying fd
    -1/errno = got error
    -2/errno = pending error
 */
int fcl_fputc(FCL_FILE * fp, char c, AI) {
  ssize_t r;
  uint8 *p, *pp;
  size_t ppp;
  if(!(fp->flags & FCL_FERROR)) {
    if(fp->flags & FCL_WRITE) {
      if(pp = fp->w_pos) {
        *(pp++) = c;
        if(pp<=fp->w_end) {
          if(c!='\n' ||
             !(fp->flags & FCL_WRITE_LINEBUF)) { fp->w_pos = pp; return 1; }
        } else {
          assert(pp==fp->w_end+1);
        }
        if(fp->fd<0) {
          errno = fp->err = EBADF; fp->flags |= FCL_FERROR; return -1;
        }
        for(p=fp->w_buf; p<pp; p+=r) {
          ppp = pp-p;
          while(((ssize_t)ppp)<0) ppp/=2;
          while((r = write(fp->fd, p, ppp))<0)
            if(errno!=EINTR || !allow_intr || !*allow_intr) {
              fp->err = errno;
              if(p!=fp->w_buf) {
                if(ppp=fp->w_pos-p) bcopy(p, fp->w_buf, ppp);
                fp->w_pos = fp->w_buf+ppp;
                /* we could store 'c' here but we have not a way to
                   both store 'c' and report a error;
                   and it is bad idea to hide the error */
                errno = fp->err;
              }
              fp->flags |= FCL_FERROR; return -1;
            }
          assert(r>0 && ((size_t)r)<=ppp);
        }
        fp->w_pos = fp->w_buf;
        return 1;
      }
      if(fp->flags & FCL_WRITE_BUF) {
        if(!fp->w_buf) {
          if(!fp->w_size) fp->w_size = FCL_DEFAULT_BUFSIZE;
          if(!(fp->w_buf = fcl_malloc(fp->w_size)))
            { errno = fp->err = ENOMEM; fp->flags |= FCL_FERROR; return -1; }
        }
        assert(fp->w_size);
        fp->w_end = fp->w_buf + fp->w_size - 1;
        fp->w_pos = fp->w_buf;
        if(fp->w_size>1 && (c!='\n' || !(fp->flags & FCL_WRITE_LINEBUF))) {
          *(fp->w_pos++) = c; return 1;
        }
      }
      if(fp->fd<0) {
        errno = fp->err = EBADF; fp->flags |= FCL_FERROR; return -1;
      }
      while((r = write(fp->fd, &c, 1))<0)
        if(errno!=EINTR || !allow_intr || !*allow_intr)
          { fp->err = errno; fp->flags |= FCL_FERROR; return -1; }
      assert(r==1);
      return 1;
    } else { errno = fp->err = EBADF; fp->flags |= FCL_FERROR; return -1; }
  } else { errno = fp->err; return -2; }
}

/*
  rsz is IN/OUT! should be set to 0 before call; used to restart after EINTR
  return:
    -2 pending error
    -1 got error
     0 got EOF
     1 sz fully read
     2 stopped at '\n'
 */
int fcl_fgets(FCL_FILE * fp, char * buf, size_t sz, size_t * rsz, AI) {
  size_t p, done;
  ssize_t r;
  char c;
  
  done = *rsz;
  assert(done<sz);
  if(fp->flags & FCL_FERROR) { errno = fp->err; return -2; }
  if(!(fp->flags & FCL_READ)) { errno = fp->err = EBADF; goto err2; }
  if(fp->w_pos && fcl_fflush(fp,allow_intr)<0) return -2;
  
  if(fp->r_pos) {
  retbuf:
    while(fp->r_pos<fp->r_end) {
      c = *(fp->r_pos++);
      if(c=='\n') { *rsz = done; return 2; }
      buf[done++] = c;
      if(done==sz) { *rsz = done; return 1; }
    }
  }
  if(fp->fd<0) { *rsz = done; return 0; }
  if(!allow_intr) allow_intr = fp->allow_intr;
  if(fp->flags & FCL_READ_AHEAD) {
    if(!fp->r_size) fp->r_size=FCL_DEFAULT_BUFSIZE;
    if(!fp->r_buf && !(fp->r_buf=fcl_malloc(fp->r_size)))
      { errno = fp->err = ENOMEM; goto err2; }
    while((r = read(fp->fd, fp->r_buf, fp->r_size))<0)
      if(errno!=EINTR || !allow_intr || !*allow_intr) goto err;
    if(!r) { *rsz = done; return 0; }
    fp->r_pos = fp->r_buf;
    fp->r_end = fp->r_buf + r;
    goto retbuf;
  }
  while(1) {
    while((r = read(fp->fd, &c, 1))<0)
      if(errno!=EINTR || !allow_intr || !*allow_intr) goto err;
    if(!r) { *rsz = done; return 0; }
    assert(r==1);
    if(c=='\n') { *rsz = done; return 2; }
    buf[done++] = c;
    if(done==sz) { *rsz = done; return 1; }
  }
err: fp->err = errno;
err2: fp->flags |= FCL_FERROR; *rsz = done; return -1;
}

int fcl_fputs(FCL_FILE * fp, char * st, size_t * cursz, AI) {
  return -1000;
}


