/*
 * This file is part of firk's Common C Library
 *  malloc.c - overflow-safe versions of memory allocation funcs
 *
 * (c) firk 2019, 2020; firk@cantconnect.ru
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. The origin of this software must not be misrepresented; you must
 *    not claim that you wrote the original software.
 * 4. Altered versions in any form must be plainly marked as such, and
 *    must not be misinterpreted as being the original software.
 *
 * This software is provided by the author and contributors `as is'
 * without any express or implied warranty.
 */

#include <stdlib.h>
#include <errno.h>
#include "types.h"
#include "string.h"
#include "qlog.h"
#include "malloc.h"


/************* memory allocation - return NULL on fail *************/
extern void * fcl_malloc(size_t sz) {
  if(!sz) sz = 1;
  return malloc(sz);
}

extern void * fcl_mallocz(size_t sz) {
  if(!sz) sz = 1;
  return calloc(sz, 1);
}

extern void * fcl_realloc(void * p, size_t sz) {
  if(!sz) sz = 1;
  return p?realloc(p, sz):malloc(sz);
}

extern void * fcl_malloc_a(size_t sz, size_t n) {
  size_t tsz;
  if(!sz || !n) tsz = 1;
  else {
    tsz = sz*n;
    if(tsz/sz!=n) { errno=ENOMEM; return NULL; }
  }
  return malloc(tsz);
}

extern void * fcl_mallocz_a(size_t sz, size_t n) {
  /* there are some buggy calloc() implementations
     that doesn't check for overflow */
  size_t tsz;
  if(!sz || !n) tsz = 1;
  else {
    tsz = sz*n;
    if(tsz/sz!=n) { errno=ENOMEM; return NULL; }
  }
  return calloc(tsz, 1);
}

extern void * fcl_realloc_a(void * p, size_t sz, size_t n) {
  size_t tsz;
  if(!sz || !n) tsz = 1;
  else {
    tsz = sz*n;
    if(tsz/sz!=n) { errno=ENOMEM; return NULL; }
  }
  return p?realloc(p, tsz):malloc(tsz);
}

extern void fcl_free(void * p) {
  if(p) free(p);
}

extern void fcl_free_c(void const * p) {
  if(p) free((void*)p);
}

extern void * fcl_reallocfree(void * p, size_t sz) {
  void * np;
  if(!sz) sz = 1;
  np = p?realloc(p,sz):malloc(sz);
  if(!np) { if(p) free(p); errno=ENOMEM; return NULL; }
  return np;
}

extern void * fcl_reallocfree_a(void * p, size_t sz, size_t n) {
  size_t tsz;
  void * np;
  if(!sz || !n) tsz = 1;
  else {
    tsz = sz*n;
    if(tsz/sz!=n) { if(p) free(p); errno=ENOMEM; return NULL; }
  }
  np = p?realloc(p,tsz):malloc(tsz);
  if(!np) { if(p) free(p); errno=ENOMEM; return NULL; }
  return np;
}

#ifndef MALLOC_ABORT_ON_ERR
#define QUIT() exit(-1)
#else
#define QUIT() abort()
#endif

/* avoid using library functions when ENOMEM as they could use malloc() too */
static size_t print_size(char * dest, size_t dsz, size_t v) {
  size_t i, j;
  char tmp[100];
  i = 0;
  do { tmp[i++] = (v%10) + '0'; v /= 10; } while(v && i<sizeof(tmp));
  if(i>dsz) i=dsz;
  for(j=i; j; ) *(dest++) = tmp[--j];
  return i;
}

static void print_err_1(char const * name, size_t sz, char const * err) {
  char tmp[250];
  size_t i;
  tmp[0] = 'f'; tmp[1] = 'c'; tmp[2] = 'l'; tmp[3] = ':'; tmp[4] = ' ';
  for(i=0; i<20 && name[i]; i++) tmp[5+i] = name[i];
  i += 5;
  tmp[i++] = '(';
  i += print_size(tmp+i, 50, sz);
  tmp[i++] = ')';
  tmp[i++] = ' ';
  while(*err && i+2<sizeof(tmp)) tmp[i++] = *(err++);
  tmp[i++] = 0;
  qlog_puts(tmp);
  qlog_end_line();
}

static void print_err_2(char const * name, size_t sz1, size_t sz2,
                        char const * err) {
  char tmp[250];
  size_t i;
  tmp[0] = 'f'; tmp[1] = 'c'; tmp[2] = 'l'; tmp[3] = ':'; tmp[4] = ' ';
  for(i=0; i<20 && name[i]; i++) tmp[5+i] = name[i];
  i += 5;
  tmp[i++] = '(';
  i += print_size(tmp+i, 50, sz1);
  tmp[i++] = ',';
  i += print_size(tmp+i, 50, sz2);
  tmp[i++] = ')';
  tmp[i++] = ' ';
  while(*err && i+2<sizeof(tmp)) tmp[i++] = *(err++);
  tmp[i++] = 0;
  qlog_puts(tmp);
  qlog_end_line();
}

#define MNMEM1(NAME, SZ1) print_err_1(NAME, SZ1, "not enough memory")
#define MNMEM2(NAME, SZ1, SZ2) print_err_2(NAME, SZ1, SZ2, "not enough memory")
#define MOVRF2(NAME, SZ1, SZ2) print_err_2(NAME, SZ1, SZ2, "overflow")

/************* memory allocation - exit(-1) on fail *************/
extern void * fcl_malloc_check(size_t sz) {
  void * p;
  if(!sz) sz = 1;
  p = malloc(sz);
  if(!p) { MNMEM1("malloc", sz); QUIT(); }
  return p;
}

extern void * fcl_mallocz_check(size_t sz) {
  void * p;
  if(!sz) sz = 1;
  p = calloc(sz, 1);
  if(!p) { MNMEM1("mallocz", sz); QUIT(); }
  return p;
}

extern void * fcl_realloc_check(void * p, size_t sz) {
  p = p?realloc(p,sz):malloc(sz);
  if(!p) { MNMEM1("realloc", sz); QUIT(); }
  return p;
}

extern void * fcl_malloc_a_check(size_t sz, size_t n) {
  void * p;
  size_t tsz;
  if(!sz || !n) tsz = 1;
  else {
    tsz = sz*n;
    if(tsz/sz!=n) { MOVRF2("malloc_a", sz, n); QUIT(); }
  }
  p = malloc(tsz);
  if(!p) { MNMEM2("malloc_a", sz, n); QUIT(); }
  return p;
}

extern void * fcl_mallocz_a_check(size_t sz, size_t n) {
  void * p;
  size_t tsz;
  if(!sz || !n) tsz = 1;
  else {
    tsz = sz*n;
    if(tsz/sz!=n) { MOVRF2("mallocz_a", sz, n); QUIT(); }
  }
  p = calloc(tsz, 1);
  if(!p) { MNMEM2("mallocz_a", sz, n); QUIT(); }
  return p;
}

extern void * fcl_realloc_a_check(void * p, size_t sz, size_t n) {
  size_t tsz;
  if(!sz || !n) tsz = 1;
  else {
    tsz = sz*n;
    if(tsz/sz!=n) { MOVRF2("realloc_a", sz, n); QUIT(); }
  }
  p = p?realloc(p,tsz):malloc(tsz);
  if(!p) { MNMEM2("realloc_a", sz, n); QUIT(); }
  return p;
}

extern void fcl_free_check(void * p) {
  if(!p) { qlog_puts("fcl: free(NULL)"); qlog_end_line(); QUIT(); }
  free(p);
}

extern void fcl_free_c_check(void const * p) {
  if(!p) { qlog_puts("fcl: free(NULL)"); qlog_end_line(); QUIT(); }
  free((void*)p);
}


/************* memory duplication - return NULL on fail *************/
extern char * fcl_strdup(char const * st) {
  char * r;
  size_t len, a;
  if(!st) { errno=EFAULT; return NULL; }
  len = strlen(st);
  a = len+1;
  if(a<=len || !(r = malloc(a))) { errno=ENOMEM; return NULL; }
  if(len) bcopy(st, r, len);
  r[len] = 0;
  return r;
}

extern char * fcl_strndup(char const * st, size_t sz) {
  char * r;
  size_t len, a;
  if(!sz) {
    if(!(r = malloc(1))) { errno=ENOMEM; return NULL; }
    r[0] = 0;
    return r;
  }
  if(!st) { errno=EFAULT; return NULL; }
  for(len=0; len<sz && st[len]; len++);
  a = len+1;
  if(a<=len || !(r = malloc(a))) { errno=ENOMEM; return NULL; }
  if(len) bcopy(st, r, len);
  r[len] = 0;
  return r;
}

extern char * fcl_substrdup(char const * st, size_t sz) {
  char * r;
  size_t a;
  if(!sz) {
    if(!(r = malloc(1))) { errno=ENOMEM; return NULL; }
    r[0] = 0;
    return r;
  }
  if(!st) { errno=EFAULT; return NULL; }
  a = sz+1;
  if(a<=sz || !(r = malloc(a))) { errno=ENOMEM; return NULL; }
  if(sz) bcopy(st, r, sz);
  r[sz] = 0;
  return r;
}

extern void * fcl_memdup(void const * p0, size_t sz) {
  void * p;
  if(!sz) {
    if(!(p = malloc(1))) { errno=ENOMEM; return NULL; }
    return p;
  }
  if(!p0) { errno=EFAULT; return NULL; }
  if(!(p = malloc(sz))) { errno=ENOMEM; return NULL; }
  bcopy(p0, p, sz);
  return p;
}


/************* memory duplication - exit(-1) on fail *************/
/* still may return NULL if input pointer==NULL */
extern char * fcl_strdup_check(char const * st) {
  char * r;
  size_t len, a;
  if(!st) { errno=EFAULT; return NULL; }
  len = strlen(st);
  a = len+1;
  if(a<=len || !(r = malloc(a))) { MNMEM1("strdup", len); QUIT(); }
  if(len) bcopy(st, r, len);
  r[len] = 0;
  return r;
}

extern char * fcl_strndup_check(char const * st, size_t sz) {
  char * r;
  size_t len, a;
  if(!sz) {
    if(!(r = malloc(1))) { MNMEM1("strndup", 0); QUIT(); }
    r[0] = 0;
    return r;
  }
  if(!st) { errno=EFAULT; return NULL; }
  for(len=0; len<sz && st[len]; len++);
  a = len+1;
  if(a<=len || !(r = malloc(a))) { MNMEM1("strndup", len); QUIT(); }
  if(len) bcopy(st, r, len);
  r[len] = 0;
  return r;
}

extern char * fcl_substrdup_check(char const * st, size_t sz) {
  char * r;
  size_t a;
  if(!sz) {
    if(!(r = malloc(1))) { MNMEM1("substrdup", 0); QUIT(); }
    r[0] = 0;
    return r;
  }
  if(!st) { errno=EFAULT; return NULL; }
  a = sz+1;
  if(a<=sz || !(r = malloc(a))) { MNMEM1("substrdup", sz); QUIT(); }
  if(sz) bcopy(st, r, sz);
  r[sz] = 0;
  return r;
}

extern void * fcl_memdup_check(void const * p0, size_t sz) {
  void * p;
  if(!sz) {
    if(!(p = malloc(1))) { MNMEM1("memdup", 0); QUIT(); }
    return p;
  }
  if(!p0) { errno=EFAULT; return NULL; }
  if(!(p = malloc(sz))) { MNMEM1("memdup", sz); QUIT(); }
  bcopy(p0, p, sz);
  return p;
}


