/*
 * This file is part of firk's Common C Library
 *  malloc.c - overflow-safe versions of memory allocation funcs
 *
 * (c) firk 2019, 2020; firk@cantconnect.ru
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. The origin of this software must not be misrepresented; you must
 *    not claim that you wrote the original software.
 * 4. Altered versions in any form must be plainly marked as such, and
 *    must not be misinterpreted as being the original software.
 *
 * This software is provided by the author and contributors `as is'
 * without any express or implied warranty.
 */

#include <stdlib.h>
#include <errno.h>
#include "types.h"
#include "string.h"
#include "qlog.h"
#include "malloc_debug.h"

#define FNLN char const * fn, unsigned int ln

typedef struct {
  void * p;
  size_t sz;
  char const * fn;
  unsigned int ln;
} alloc_log;
#define MAX_ALLOC_LOG 100000
static alloc_log aa[MAX_ALLOC_LOG];
static size_t    na;

static void add_alloc_log(void * p, size_t sz, FNLN) {
  if(na<MAX_ALLOC_LOG) {
    aa[na].p = p;
    aa[na].sz = sz;
    aa[na].fn = fn;
    aa[na].ln = ln;
    na++;
    return;
  }
  qlog_q("heap debug log overflowed at %s:%u", fn, ln); abort();
}

static alloc_log * find_alloc_log(void * p, FNLN) {
  size_t n;
  for(n=0; n<na; n++) if(aa[n].p==p) return aa+n;
  qlog_q("try to use non-heap pointer %p at %s:%u", p, fn, ln); abort();
}

static void * malloc_D(size_t sz, FNLN) {
  void * r;
  if(r = malloc(sz)) add_alloc_log(r,sz,fn,ln);
  return r;
}

static void * calloc_D(size_t sz, FNLN) {
  void * r;
  if(r = calloc(sz,1)) add_alloc_log(r,sz,fn,ln);
  return r;
}

static void * realloc_D(void * p, size_t sz, FNLN) {
  void * r;
  alloc_log * a;
  a = find_alloc_log(p,fn,ln);
  if(r = realloc(p,sz)) { a->p = r; a->sz = sz; }
  return r;
}

static void free_D(void * p, FNLN) {
  alloc_log * a;
  a = find_alloc_log(p,fn,ln);
  free(p);
  if(!na) { qlog_q("na=0"); abort(); }
  na--;
  *a = aa[na];
}

extern void fcl_malloc_debug_dump(void) {
  size_t n;
  unsigned long total;
  total = 0;
  for(n=0; n<na; n++) {
    qlog_q("%zu bytes at %s:%u", aa[n].sz, aa[n].fn, aa[n].ln);
    total += aa[n].sz;
  }
  qlog_q("total alloc: %lu bytes, %lu chunks", total, (unsigned long)na);
}

/************* memory allocation - return NULL on fail *************/
extern void * fcl_malloc_debug(size_t sz, FNLN) {
  if(!sz) sz = 1;
  return malloc_D(sz,fn,ln);
}

extern void * fcl_mallocz_debug(size_t sz, FNLN) {
  if(!sz) sz = 1;
  return calloc_D(sz,fn,ln);
}

extern void * fcl_realloc_debug(void * p, size_t sz, FNLN) {
  if(!sz) sz = 1;
  return p?realloc_D(p,sz,fn,ln):malloc_D(sz,fn,ln);
}

extern void * fcl_malloc_a_debug(size_t sz, size_t n, FNLN) {
  size_t tsz;
  if(!sz || !n) tsz = 1;
  else {
    tsz = sz*n;
    if(tsz/sz!=n) { errno=ENOMEM; return NULL; }
  }
  return malloc_D(tsz,fn,ln);
}

extern void * fcl_mallocz_a_debug(size_t sz, size_t n, FNLN) {
  /* there are some buggy calloc() implementations
     that doesn't check for overflow */
  size_t tsz;
  if(!sz || !n) tsz = 1;
  else {
    tsz = sz*n;
    if(tsz/sz!=n) { errno=ENOMEM; return NULL; }
  }
  return calloc_D(tsz,fn,ln);
}

extern void * fcl_realloc_a_debug(void * p, size_t sz, size_t n, FNLN) {
  size_t tsz;
  if(!sz || !n) tsz = 1;
  else {
    tsz = sz*n;
    if(tsz/sz!=n) { errno=ENOMEM; return NULL; }
  }
  return p?realloc_D(p,tsz,fn,ln):malloc_D(tsz,fn,ln);
}

extern void fcl_free_debug(void * p, FNLN) {
  if(p) free_D(p,fn,ln);
}

extern void fcl_free_c_debug(void const * p, FNLN) {
  if(p) free_D((void*)p,fn,ln);
}

extern void * fcl_reallocfree_debug(void * p, size_t sz, FNLN) {
  void * np;
  if(!sz) sz = 1;
  np = p?realloc_D(p,sz,fn,ln):malloc_D(sz,fn,ln);
  if(!np) { if(p) free_D(p,fn,ln); errno=ENOMEM; return NULL; }
  return np;
}

extern void * fcl_reallocfree_a_debug(void * p, size_t sz, size_t n, FNLN) {
  size_t tsz;
  void * np;
  if(!sz || !n) tsz = 1;
  else {
    tsz = sz*n;
    if(tsz/sz!=n) { if(p) free_D(p,fn,ln); errno=ENOMEM; return NULL; }
  }
  np = p?realloc_D(p,tsz,fn,ln):malloc_D(tsz,fn,ln);
  if(!np) { if(p) free_D(p,fn,ln); errno=ENOMEM; return NULL; }
  return np;
}

#ifndef MALLOC_ABORT_ON_ERR
#define QUIT() exit(-1)
#else
#define QUIT() abort()
#endif

/* avoid using library functions when ENOMEM as they could use malloc() too */
static size_t print_size(char * dest, size_t dsz, size_t v) {
  size_t i, j;
  char tmp[100];
  i = 0;
  do { tmp[i++] = (v%10) + '0'; v /= 10; } while(v && i<sizeof(tmp));
  if(i>dsz) i=dsz;
  for(j=i; j; ) *(dest++) = tmp[--j];
  return i;
}

static void print_err_1(char const * name, size_t sz, char const * err, FNLN) {
  char tmp[250];
  size_t i;
  tmp[0] = 'f'; tmp[1] = 'c'; tmp[2] = 'l'; tmp[3] = ':'; tmp[4] = ' ';
  for(i=0; i<20 && name[i]; i++) tmp[5+i] = name[i];
  i += 5;
  tmp[i++] = '(';
  i += print_size(tmp+i, 50, sz);
  tmp[i++] = ')';
  tmp[i++] = ' ';
  while(*err && i+20<sizeof(tmp)) tmp[i++] = *(err++);
  tmp[i++] = ' ';
  tmp[i++] = 'a';
  tmp[i++] = 't';
  tmp[i++] = ' ';
  while(*fn && i+15<sizeof(tmp)) tmp[i++] = *(fn++);
  tmp[i++] = ':';
  i += print_size(tmp+i, 10, ln);
  tmp[i++] = 0;
  qlog_puts(tmp);
  qlog_end_line();
}

static void print_err_2(char const * name, size_t sz1, size_t sz2,
                        char const * err, FNLN) {
  char tmp[250];
  size_t i;
  tmp[0] = 'f'; tmp[1] = 'c'; tmp[2] = 'l'; tmp[3] = ':'; tmp[4] = ' ';
  for(i=0; i<20 && name[i]; i++) tmp[5+i] = name[i];
  i += 5;
  tmp[i++] = '(';
  i += print_size(tmp+i, 50, sz1);
  tmp[i++] = ',';
  i += print_size(tmp+i, 50, sz2);
  tmp[i++] = ')';
  tmp[i++] = ' ';
  while(*err && i+20<sizeof(tmp)) tmp[i++] = *(err++);
  tmp[i++] = ' ';
  tmp[i++] = 'a';
  tmp[i++] = 't';
  tmp[i++] = ' ';
  while(*fn && i+15<sizeof(tmp)) tmp[i++] = *(fn++);
  tmp[i++] = ':';
  i += print_size(tmp+i, 10, ln);
  tmp[i++] = 0;
  qlog_puts(tmp);
  qlog_end_line();
}

#define MNMEM1(NAME, SZ1) print_err_1(NAME, SZ1, "not enough memory", fn, ln)
#define MNMEM2(NAME, SZ1, SZ2) \
  print_err_2(NAME, SZ1, SZ2, "not enough memory", fn, ln)
#define MOVRF2(NAME, SZ1, SZ2) print_err_2(NAME, SZ1, SZ2, "overflow", fn, ln)

/************* memory allocation - exit(-1) on fail *************/
extern void * fcl_malloc_check_debug(size_t sz, FNLN) {
  void * p;
  if(!sz) sz = 1;
  p = malloc_D(sz,fn,ln);
  if(!p) { MNMEM1("malloc", sz); QUIT(); }
  return p;
}

extern void * fcl_mallocz_check_debug(size_t sz, FNLN) {
  void * p;
  if(!sz) sz = 1;
  p = calloc_D(sz,fn,ln);
  if(!p) { MNMEM1("mallocz", sz); QUIT(); }
  return p;
}

extern void * fcl_realloc_check_debug(void * p, size_t sz, FNLN) {
  p = p?realloc_D(p,sz,fn,ln):malloc_D(sz,fn,ln);
  if(!p) { MNMEM1("realloc", sz); QUIT(); }
  return p;
}

extern void * fcl_malloc_a_check_debug(size_t sz, size_t n, FNLN) {
  void * p;
  size_t tsz;
  if(!sz || !n) tsz = 1;
  else {
    tsz = sz*n;
    if(tsz/sz!=n) { MOVRF2("malloc_a", sz, n); QUIT(); }
  }
  p = malloc_D(tsz,fn,ln);
  if(!p) { MNMEM2("malloc_a", sz, n); QUIT(); }
  return p;
}

extern void * fcl_mallocz_a_check_debug(size_t sz, size_t n, FNLN) {
  void * p;
  size_t tsz;
  if(!sz || !n) tsz = 1;
  else {
    tsz = sz*n;
    if(tsz/sz!=n) { MOVRF2("mallocz_a", sz, n); QUIT(); }
  }
  p = calloc_D(tsz,fn,ln);
  if(!p) { MNMEM2("mallocz_a", sz, n); QUIT(); }
  return p;
}

extern void * fcl_realloc_a_check_debug(void * p, size_t sz, size_t n, FNLN) {
  size_t tsz;
  if(!sz || !n) tsz = 1;
  else {
    tsz = sz*n;
    if(tsz/sz!=n) { MOVRF2("realloc_a", sz, n); QUIT(); }
  }
  p = p?realloc_D(p,tsz,fn,ln):malloc_D(tsz,fn,ln);
  if(!p) { MNMEM2("realloc_a", sz, n); QUIT(); }
  return p;
}

extern void fcl_free_check_debug(void * p, FNLN) {
  if(!p) { qlog_puts("fcl: free(NULL)"); qlog_end_line(); QUIT(); }
  free_D(p,fn,ln);
}

extern void fcl_free_c_check_debug(void const * p, FNLN) {
  if(!p) { qlog_puts("fcl: free(NULL)"); qlog_end_line(); QUIT(); }
  free_D((void*)p,fn,ln);
}


/************* memory duplication - return NULL on fail *************/
extern char * fcl_strdup_debug(char const * st, FNLN) {
  char * r;
  size_t len, a;
  if(!st) { errno=EFAULT; return NULL; }
  len = strlen(st);
  a = len+1;
  if(a<=len || !(r = malloc_D(a,fn,ln))) { errno=ENOMEM; return NULL; }
  if(len) bcopy(st, r, len);
  r[len] = 0;
  return r;
}

extern char * fcl_strndup_debug(char const * st, size_t sz, FNLN) {
  char * r;
  size_t len, a;
  if(!sz) {
    if(!(r = malloc_D(1,fn,ln))) { errno=ENOMEM; return NULL; }
    r[0] = 0;
    return r;
  }
  if(!st) { errno=EFAULT; return NULL; }
  for(len=0; len<sz && st[len]; len++);
  a = len+1;
  if(a<=len || !(r = malloc_D(a,fn,ln))) { errno=ENOMEM; return NULL; }
  if(len) bcopy(st, r, len);
  r[len] = 0;
  return r;
}

extern char * fcl_substrdup_debug(char const * st, size_t sz, FNLN) {
  char * r;
  size_t a;
  if(!sz) {
    if(!(r = malloc_D(1,fn,ln))) { errno=ENOMEM; return NULL; }
    r[0] = 0;
    return r;
  }
  if(!st) { errno=EFAULT; return NULL; }
  a = sz+1;
  if(a<=sz || !(r = malloc_D(a,fn,ln))) { errno=ENOMEM; return NULL; }
  if(sz) bcopy(st, r, sz);
  r[sz] = 0;
  return r;
}

extern void * fcl_memdup_debug(void const * p0, size_t sz, FNLN) {
  void * p;
  if(!sz) {
    if(!(p = malloc_D(1,fn,ln))) { errno=ENOMEM; return NULL; }
    return p;
  }
  if(!p0) { errno=EFAULT; return NULL; }
  if(!(p = malloc_D(sz,fn,ln))) { errno=ENOMEM; return NULL; }
  bcopy(p0, p, sz);
  return p;
}


/************* memory duplication - exit(-1) on fail *************/
/* still may return NULL if input pointer==NULL */
extern char * fcl_strdup_check_debug(char const * st, FNLN) {
  char * r;
  size_t len, a;
  if(!st) { errno=EFAULT; return NULL; }
  len = strlen(st);
  a = len+1;
  if(a<=len || !(r = malloc_D(a,fn,ln))) { MNMEM1("strdup", len); QUIT(); }
  if(len) bcopy(st, r, len);
  r[len] = 0;
  return r;
}

extern char * fcl_strndup_check_debug(char const * st, size_t sz, FNLN) {
  char * r;
  size_t len, a;
  if(!sz) {
    if(!(r = malloc_D(1,fn,ln))) { MNMEM1("strndup", 0); QUIT(); }
    r[0] = 0;
    return r;
  }
  if(!st) { errno=EFAULT; return NULL; }
  for(len=0; len<sz && st[len]; len++);
  a = len+1;
  if(a<=len || !(r = malloc_D(a,fn,ln))) { MNMEM1("strndup", len); QUIT(); }
  if(len) bcopy(st, r, len);
  r[len] = 0;
  return r;
}

extern char * fcl_substrdup_check_debug(char const * st, size_t sz, FNLN) {
  char * r;
  size_t a;
  if(!sz) {
    if(!(r = malloc_D(1,fn,ln))) { MNMEM1("substrdup", 0); QUIT(); }
    r[0] = 0;
    return r;
  }
  if(!st) { errno=EFAULT; return NULL; }
  a = sz+1;
  if(a<=sz || !(r = malloc_D(a,fn,ln))) { MNMEM1("substrdup", sz); QUIT(); }
  if(sz) bcopy(st, r, sz);
  r[sz] = 0;
  return r;
}

extern void * fcl_memdup_check_debug(void const * p0, size_t sz, FNLN) {
  void * p;
  if(!sz) {
    if(!(p = malloc_D(1,fn,ln))) { MNMEM1("memdup", 0); QUIT(); }
    return p;
  }
  if(!p0) { errno=EFAULT; return NULL; }
  if(!(p = malloc_D(sz,fn,ln))) { MNMEM1("memdup", sz); QUIT(); }
  bcopy(p0, p, sz);
  return p;
}


