/*
 * This file is part of firk's Common C Library
 *  qlog.c - quick logging
 *
 * (c) firk 2019-2022; firk@cantconnect.ru
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. The origin of this software must not be misrepresented; you must
 *    not claim that you wrote the original software.
 * 4. Altered versions in any form must be plainly marked as such, and
 *    must not be misinterpreted as being the original software.
 *
 * This software is provided by the author and contributors `as is'
 * without any express or implied warranty.
 */

#include <stdlib.h>
#include <unistd.h>
#include <stdarg.h>
#include <errno.h>
#include <time.h>
#include <sys/time.h>
#include "types.h"
#include "string.h"
#include "strerror.h"
#define FCL_WITH_VA_LIST
#include "qlog.h"

/* do not use library assert because it may use heap */
#define assert(X) do { if(!(X)) abort(); } while(0)

static int errh;

static void write_all(int fd, char const * buf, size_t nb) {
  size_t csz;
  ssize_t wsz;
  while(nb) {
    csz = nb;
    while(((ssize_t)csz)<=0) csz/=2;
    wsz = write(fd, buf, csz);
    if(wsz<0) {
      if(errno==EINTR) continue;
      if(errh & QLOG_ERRH_WAITREC) {
        if(errno==ENOSPC || errno==ENOBUFS || errno==ENOMEM || errno==EDQUOT
         || errno==EAGAIN || errno==EINPROGRESS || errno==EWOULDBLOCK
         ) { sleep(1); continue; }
      }
      if(errh & QLOG_ERRH_IGNORE) return;
      exit(-1);
    }
    assert(wsz>=1 && ((size_t)wsz)<=csz && csz<=nb);
    buf += wsz;
    nb -= wsz;
  }
}

static int errfd = 2, pending_newline, use_time;
static char const * log_prefix;
static char errbuf[FCL_QLOG_BUFSIZE];
static size_t errlen, remlen, maxlen, maxlen_present;

static void E_flush(void) {
  if(errlen) { write_all(errfd, errbuf, errlen); errlen = 0; }
}

static void E_write(char const * s, size_t len) {
  size_t add;
  assert(errlen<sizeof(errbuf));
  if(remlen>len) remlen -= len; else remlen = 0;
  if(maxlen_present) { if(len>maxlen) len=maxlen; maxlen -= len; }
  if(!len) return;
  if(len<=sizeof(errbuf)-errlen) {
    bcopy(s, errbuf+errlen, len);
    errlen += len;
    if(errlen==sizeof(errbuf)) E_flush();
  } else if(len<sizeof(errbuf)/2) {
    bcopy(s, errbuf+errlen, add=sizeof(errbuf)-errlen);
    errlen = sizeof(errbuf);
    E_flush();
    bcopy(s+add, errbuf, errlen=len-add);
  } else {
    E_flush();
    write_all(errfd, s, len);
  }
}

static void E_repeat(char c, size_t len) {
  size_t add;
  assert(errlen<sizeof(errbuf));
  if(remlen>len) remlen -= len; else remlen = 0;
  if(maxlen_present) { if(len>maxlen) len=maxlen; maxlen -= len; }
  if(!len) return;
  if(len<=sizeof(errbuf)-errlen) {
    memset(errbuf+errlen, c, len);
    errlen += len;
    if(errlen==sizeof(errbuf)) E_flush();
  } else {
    memset(errbuf+errlen, c, add=sizeof(errbuf)-errlen);
    errlen = sizeof(errbuf);
    E_flush();
    errlen = len-add;
    if(errlen<sizeof(errbuf)) {
      memset(errbuf, c, errlen);
    } else {
      memset(errbuf, c, sizeof(errbuf));
      while(errlen>=sizeof(errbuf)) {
        write_all(errfd, errbuf, sizeof(errbuf));
        errlen -= sizeof(errbuf);
      }
    }
  }
}

static void E_print_char(char c) {
  assert(errlen<sizeof(errbuf));
  if(remlen) remlen--;
  if(maxlen_present) { if(maxlen) maxlen--; else return; }
  errbuf[errlen++] = c;
  if(errlen==sizeof(errbuf)) E_flush();
}

#define PA
#define CA
#define CHECK_STOP
#define strerror fcl_strerror
#include "strfmt.inc"
#undef strerror
#undef CHECK_STOP
#undef CA
#undef PA

static void E_print_timestr(time_t now) {
  struct tm * t;
  if(!(t = localtime(&now))) {
    E_print_ull(now, 1, 0);
  } else {
    E_print_uint(t->tm_year+1900, 4, F_ZERO_PAD);
    E_print_char('-');
    E_print_uint(t->tm_mon+1, 2, F_ZERO_PAD);
    E_print_char('-');
    E_print_uint(t->tm_mday, 2, F_ZERO_PAD);
    E_print_char(' ');
    E_print_uint(t->tm_hour, 2, F_ZERO_PAD);
    E_print_char(':');
    E_print_uint(t->tm_min, 2, F_ZERO_PAD);
    E_print_char(':');
    E_print_uint(t->tm_sec, 2, F_ZERO_PAD);
  }
}

extern void qlog_set_fd(int fd) { errfd = fd; }
extern void qlog_set_times(int flag) {
  char const * e;
  if(flag>=100) {
    if(e = getenv("FCL_QLOG_TIMES")) {
      if(!strcmp(e,"0") || !strcasecmp(e,"no")) use_time = 0;
      else if(!strcmp(e,"2") || !strcasecmp(e,"us") || !strcasecmp(e,"usec"))
      use_time = 2; else use_time = 1; return;
    }
    flag -= 100;
  }
  use_time = flag;
}
extern void qlog_set_prefix(char const * pre) { log_prefix = pre; }
extern void qlog_set_errh(int flags) {
  char const * e;
  if(flags>=100) {
    if(e = getenv("FCL_QLOG_ERRH")) {
      if(!strcmp(e,"1")) errh = 1;
      else if(!strcmp(e,"2")) errh = 2;
      else if(!strcmp(e,"3")) errh = 3;
      else if(!strcmp(e,"4")) errh = 4;
      else if(!strcmp(e,"5")) errh = 5;
      else if(!strcmp(e,"6")) errh = 6;
      else if(!strcmp(e,"7")) errh = 7;
      else errh = 0;
      return;
    }
    flags -= 100;
  }
  errh = flags;
}


extern void qlog_pending_newline(int flag) {
  pending_newline = flag;
}

extern void qlog_begin_line(void) {
  struct timeval tv;
  if(pending_newline) { E_print_char('\n'); pending_newline = 0; }
  if(use_time>=2 && gettimeofday(&tv,NULL)>=0) {
    E_print_char('[');
    E_print_timestr(tv.tv_sec);
    E_print_char('.');
    E_print_ulong(tv.tv_usec, 6, F_ZERO_PAD);
    E_write("] ", 2);
  } else if(use_time>=1) {
    E_print_char('[');
    E_print_timestr(time(NULL));
    E_write("] ", 2);
  }
  if(log_prefix) {
    E_write(log_prefix, strlen(log_prefix));
    E_write(": ", 2);
  }
}

extern void qlog_end_line(void) {
  E_print_char('\n');
  E_flush();
}

extern void qlog_flush(void) {
  E_flush();
}

extern void qlog(char const * fmt, ...) {
  va_list arg;
  qlog_begin_line();
  va_start(arg, fmt);
  if(E_vprint(fmt, arg)<0) {
    E_print_str("\nqlog: bad format string: \"");
    E_print_escaped_str(fmt);
    E_print_str("\"\n");
    E_flush();
    if(!(errh & QLOG_ERRH_IGNFMT)) exit(-1);
  }
  va_end(arg);
  qlog_end_line();
}

extern void qlog_q(char const * fmt, ...) {
  va_list arg;
  if(pending_newline) { E_print_char('\n'); pending_newline = 0; }
  va_start(arg, fmt);
  if(E_vprint(fmt, arg)<0) {
    E_print_str("\nqlog: bad format string: \"");
    E_print_escaped_str(fmt);
    E_print_str("\"\n");
    E_flush();
    if(!(errh & QLOG_ERRH_IGNFMT)) exit(-1);
  }
  va_end(arg);
  qlog_end_line();
}

extern void qlog_printf(char const * fmt, ...) {
  va_list arg;
  va_start(arg, fmt);
  if(E_vprint(fmt, arg)<0) {
    E_print_str("\nqlog: bad format string: \"");
    E_print_escaped_str(fmt);
    E_print_str("\"\n");
    E_flush();
    if(!(errh & QLOG_ERRH_IGNFMT)) exit(-1);
  }
  va_end(arg);
}

extern void qlog_vprintf(char const * fmt, va_list arg) {
  if(E_vprint(fmt, arg)<0) {
    E_print_str("\nqlog: bad format string: \"");
    E_print_escaped_str(fmt);
    E_print_str("\"\n");
    E_flush();
    if(!(errh & QLOG_ERRH_IGNFMT)) exit(-1);
  }
}

extern void qlog_putc(char c) {
  E_print_char(c);
}

extern void qlog_putce(char c) {
  E_print_escaped_char(c);
}

extern void qlog_puts(char const * s) {
  E_print_str(s);
}

extern void qlog_putse(char const * s) {
  E_print_escaped_str(s);
}

extern void qlog_putb(char const * s, size_t len) {
  E_write(s, len);
}

extern void qlog_putbe(char const * s, size_t len) {
  E_print_escaped_bytes(s, len);
}

