/*
 * This file is part of firk's Common C Library
 *  qprintfd.c - open() wrapper for qprint.c functions
 *
 * (c) firk 2021; firk@cantconnect.ru
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. The origin of this software must not be misrepresented; you must
 *    not claim that you wrote the original software.
 * 4. Altered versions in any form must be plainly marked as such, and
 *    must not be misinterpreted as being the original software.
 *
 * This software is provided by the author and contributors `as is'
 * without any express or implied warranty.
 */

#include <unistd.h>
#include <errno.h>
#include <fcntl.h>
#include "types.h"
#include "qprint.h"

static int mode_to_flags(unsigned int mode) {
  int flags;
  flags = O_WRONLY|O_APPEND|O_NOCTTY;
  if(mode & QP_FOPEN_NOFOLLOW) flags |= O_NOFOLLOW;
  switch(mode & 0x3000) {
  default: return (flags | O_CREAT);
  case QP_FOPEN_EXISTING: return (flags);
  case QP_FOPEN_NEW: return (flags | O_CREAT | O_EXCL);
  case QP_FOPEN_TRUNC: return (flags | O_CREAT | O_TRUNC);
  }
}

extern int qp_fopen(qprint_ctx * ctx, char const * name, unsigned int mode) {
  int fd, flags;
  if(ctx->fd>=0) { errno = EINVAL; return -1; }
  flags = mode_to_flags(mode);
  if(mode & QP_FOPEN_NOINTR) {
    do { fd = open(name, flags, mode & 0xFFF); } while(fd<0 && errno==EINTR);
  } else {
    fd = open(name, flags, mode & 0xFFF);
  }
  if(fd<0) return -1;
  ctx->fd = fd;
  return 0;
}

extern int qp_fopenat(qprint_ctx * ctx, int dirfd, char const * name,
                      unsigned int mode) {
  int fd, flags;
  if(ctx->fd>=0) { errno = EINVAL; return -1; }
  flags = mode_to_flags(mode);
  if(mode & QP_FOPEN_NOINTR) {
    mode &= 0xFFF;
    do { fd = openat(dirfd, name, flags, mode); } while(fd<0 && errno==EINTR);
  } else {
    fd = openat(dirfd, name, flags, mode & 0xFFF);
  }
  if(fd<0) return -1;
  ctx->fd = fd;
  return 0;
}

extern int qp_fclose(qprint_ctx * ctx) {
  if(ctx->fd<0) { errno = EINVAL; return -1; }
  if(ctx->buf && ctx->bufpos) qp_flush(ctx);
  close(ctx->fd);
  if(ctx->fail) { errno = (ctx->fail<0)?0:ctx->fail; return -1; }
  return 0;
}

