/*
 * This file is part of firk's Common C Library
 *  qscan.c - file/stream reader
 *
 * (c) firk 2021-2022; firk@cantconnect.ru
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. The origin of this software must not be misrepresented; you must
 *    not claim that you wrote the original software.
 * 4. Altered versions in any form must be plainly marked as such, and
 *    must not be misinterpreted as being the original software.
 *
 * This software is provided by the author and contributors `as is'
 * without any express or implied warranty.
 */

#include <errno.h>
#include <assert.h>
#include <unistd.h>
#include "types.h"
#include "malloc.h"
#include "qscan.h"


static int check_room(qscan_ctx * ctx) {
  fcl_size a, inc;
  char * r;
  if(ctx->b && ctx->bpos<ctx->bsz) return 1;
  if(ctx->bpos>ctx->bsz) { errno = EINVAL; return -1; }
  if(!(ctx->flags & QS_ALLOC)) {
    if(ctx->b && ctx->bsz && ctx->bpos==ctx->bsz) errno = ENOSPC;
    else errno = EINVAL;
    return -1;
  }
  if(!ctx->b) {
    if(ctx->bpos || ctx->bsz || !ctx->bmax) { errno = EINVAL; return -1; }
    a = 16;
    if(a>ctx->bmax) a = ctx->bmax;
    if(!(ctx->b = fcl_malloc(a))) { errno = ENOMEM; return -1; }
    ctx->bsz = a;
  } else {
    assert(ctx->b && ctx->bpos==ctx->bsz);
    if(ctx->bsz>=ctx->bmax) { errno = ENOSPC; return -1; }
    a = ctx->bsz;
    inc = a/3 + 11;
    if(ctx->bmax-a>inc) a += inc; else a = ctx->bmax;
    if(!(r = fcl_realloc(ctx->b, a))) { errno = ENOMEM; return -1; }
    ctx->b = r;
    ctx->bsz = a;
  }
  assert(ctx->b && ctx->bpos<ctx->bsz);
  return 1;
}

static int qscan_unbuffered(qscan_ctx * ctx) {
  fcl_ssize r;
  char c;
  while(check_room(ctx)>0) {
    r = read(ctx->fd, &c, 1);
    if(r<0) {
      if(errno==EINTR && (ctx->flags & QS_NOINTR)) continue;
      return -1;
    }
    if(!r) {
      ctx->b[ctx->bpos] = 0;
      return 0;
    }
    assert(r==1);
    if(c!='\n') {
      ctx->b[ctx->bpos++] = c;
      continue;
    }
    if((ctx->flags & QS_DOSNL) && ctx->bpos && ctx->b[ctx->bpos-1]=='\r')
      ctx->bpos--;
    ctx->b[ctx->bpos] = 0;
    return 1;
  }
  return -1;
}

extern int qscan(qscan_ctx * ctx) {
  fcl_ssize r;
  char c;
  if(!ctx->buf || !ctx->bufsz) return qscan_unbuffered(ctx);
  while(check_room(ctx)>0) {
    if(ctx->buflen==ctx->bufpos) {
      while(1) {
        r = read(ctx->fd, ctx->buf, ctx->bufsz);
        if(r<0) {
          if(errno==EINTR && (ctx->flags & QS_NOINTR)) continue;
          return -1;
        }
        if(!r) {
          ctx->b[ctx->bpos] = 0;
          return 0;
        }
        break;
      }
      ctx->bufpos = 0;
      ctx->buflen = (fcl_size)r;
    }
    assert(ctx->bufpos<ctx->buflen && ctx->bpos<ctx->bsz);
    while(ctx->bufpos<ctx->buflen && ctx->bpos<ctx->bsz) {
      c = ctx->buf[ctx->bufpos++];
      if(c!='\n') { ctx->b[ctx->bpos++] = c; continue; }
      if((ctx->flags & QS_DOSNL) && ctx->bpos && ctx->b[ctx->bpos-1]=='\r')
        ctx->bpos--;
      ctx->b[ctx->bpos] = 0;
      return 1;
    }
  }
  return -1;
}

extern int qs_gets(qscan_ctx * ctx, char * buf, fcl_size maxlen,
                   fcl_size * rlen) {
  int r;
  ctx->b = buf;
  ctx->bsz = maxlen;
  ctx->bpos = 0;
  ctx->flags = QS_STATIC|QS_NOINTR;
  r = qscan(ctx);
  *rlen = ctx->bpos;
  return r;
}


