/*
 * This file is part of firk's Common C Library
 *  strerror.c - safe implementation for strerror()
 *
 * (c) firk 2022; firk@cantconnect.ru
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. The origin of this software must not be misrepresented; you must
 *    not claim that you wrote the original software.
 * 4. Altered versions in any form must be plainly marked as such, and
 *    must not be misinterpreted as being the original software.
 *
 * This software is provided by the author and contributors `as is'
 * without any express or implied warranty.
 */

#include <errno.h>
#include "types.h"
#include "string.h"
#include "malloc.h"
#include "qprint.h"
#include "strerror.h"

#include "config/work/strerror.inc"

#ifdef FCL_NO_STRERROR
extern int fcl_strerror_reinit(void) { return 0; }
extern int fcl_strerror_set(int e, char const * text) {
  errno = EINVAL;
  return -1;
}
extern char const * fcl_strerror(int e) {
  return strerror(e);
}
#else
/*
  #define FCL_ERRNO_NUM xxx
  static char const * fcl_errno_cache[FCL_ERRNO_NUM] = { "No error", ... };
*/

#ifndef FCL_STRERROR_STATIC
static char const * fcl_errno_cache_rt[FCL_ERRNO_NUM];
static char const errstr_override[1];

extern int fcl_strerror_reinit(void) {
  int e, se, ne;
  char const *s, *e0, *e1, *p;
  char t[10];
  fcl_size e0l, tl, e1l;
  
  /* fill fcl_errno_cache_rt[e] with strdup(strerror(e)),
     free existing data if any */
  for(e=0; e<FCL_ERRNO_NUM; e++) {
    if(fcl_errno_cache_rt[e]) {
      if(fcl_errno_cache_rt[e]!=errstr_override)
        fcl_free_c_check(fcl_errno_cache_rt[e]);
      fcl_errno_cache_rt[e] = NULL;
    }
    if(s = strerror(e)) {
      if(!(s = fcl_strdup(s))) return -1;
      fcl_errno_cache_rt[e] = s;
    }
  }
  
  /* remove dummy entries */
  qp_snprintf(t, sizeof(t), "%d", (int)(FCL_ERRNO_NUM-2));
  tl = strlen(t);
  if((e0=fcl_errno_cache_rt[FCL_ERRNO_NUM-2]) && (p=strstr(e0,t))) {
    e0l = p-e0;
    e1 = p+tl;
    e1l = strlen(e1);
    for(ne=FCL_ERRNO_NUM-1; ne>=0; ne--) {
      s = fcl_errno_cache_rt[ne];
      if(!s) continue;
      qp_snprintf(t, sizeof(t), "%d", ne);
      tl = strlen(t);
      if(strlen(s)!=e0l+tl+e1l || memcmp(s,e0,e0l) ||
        memcmp(s+e0l,t,tl) || memcmp(s+e0l+tl,e1,e1l)) break;
    }
    ne++;
    se = 0;
    if(ne>1 && (s=fcl_errno_cache_rt[0]) && strlen(s)==e0l+1+e1l &&
      !memcmp(s,e0,e0l) && s[e0l]=='0' && !memcmp(s+e0l+1,e1,e1l)) se++;
    for(e=ne; e<FCL_ERRNO_NUM; e++) if(fcl_errno_cache_rt[e]) {
      fcl_free_c_check(fcl_errno_cache_rt[e]);
      fcl_errno_cache_rt[e] = NULL;
    }
    if(se && fcl_errno_cache_rt[0]) {
      fcl_free_c_check(fcl_errno_cache_rt[0]);
      fcl_errno_cache_rt[0] = NULL;
    }
  }
  
  /* replace NULL's with errstr_override */
  for(e=0; e<FCL_ERRNO_NUM; e++) {
    if(!fcl_errno_cache_rt[e]) fcl_errno_cache_rt[e] = errstr_override;
  }
  return 0;
}

extern int fcl_strerror_set(int e, char const * text) {
  char const * s;
  if(e<1 || e>=FCL_ERRNO_NUM) { errno = EINVAL; return -1; }
  if(!text) s = errstr_override;
  else if(!(s = fcl_strdup(text))) return -1;
  if(fcl_errno_cache_rt[e]) {
    if(fcl_errno_cache_rt[e]!=errstr_override)
      fcl_free_c_check(fcl_errno_cache_rt[e]);
    fcl_errno_cache_rt[e] = NULL;
  }
  fcl_errno_cache_rt[e] = s;
  return -1;
}

#else
extern int fcl_strerror_reinit(void) { return 0; }
extern int fcl_strerror_set(int e, char const * text) {
  errno = EINVAL;
  return -1;
}
#endif

extern char const * fcl_strerror(int e) {
  char const * r;
  if(e<0) return "Invalid error number";
  if(e>=FCL_ERRNO_NUM) return "Unknown error number";
#ifndef FCL_STRERROR_STATIC
  if(r = fcl_errno_cache_rt[e]) {
    if(r!=errstr_override) return r;
  } else
#endif
  {
    if(r = fcl_errno_cache[e]) return r;
  }
  if(!e) return "No error";
  return "Unknown error number";
}
#endif

