/*
 * This file is part of firk's Common C Library
 *  strfmt.inc - string formatter static include
 *
 * (c) firk 2019-2022; firk@cantconnect.ru
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. The origin of this software must not be misrepresented; you must
 *    not claim that you wrote the original software.
 * 4. Altered versions in any form must be plainly marked as such, and
 *    must not be misinterpreted as being the original software.
 *
 * This software is provided by the author and contributors `as is'
 * without any express or implied warranty.
 */

#include "strerror.h"
/* derived from qlog.c ~ derived from farc archive/log_default.c */
/* how to use:
for standalone:
#define PA
#define CA
#define CHECK_STOP
static size_t remlen, maxlen, maxlen_present;
static void E_write(PA char const * s, size_t len);
static void E_repeat(PA char c, size_t len);
static void E_print_char(PA char c);
see (*)

for bundled into some stream wrapper:
#define PA cookie_t * cookie,
#define CA cookie,
#define CHECK_STOP if(cookie->err) return 0; (to fast exit from E_vprint)
#define remlen cookie->_remlen
#define maxlen cookie->_maxlen
#define maxlen_present cookie->_maxlen_present
static void E_write(PA char const * s, size_t len);
static void E_repeat(PA char c, size_t len);
static void E_print_char(PA char c);
see (*)

(*) in both cases:
1) printing functions should decrease remlen by number of printed bytes until
it becomes zero
2) if(maxlen_present) => same about maxlen, but bytes should be ignored when
maxlen becomes zero
example code to do all this:
E_write, E_repeat:
  if(remlen>len) remlen -= len; else remlen = 0;
  if(maxlen_present) { if(len>maxlen) len=maxlen; maxlen -= len; }
E_print_char:
  if(remlen) remlen--;
  if(maxlen_present) { if(maxlen) maxlen--; else return; }
 */
static void E_print_str(PA char const * s) {
  E_write(CA s, strlen(s));
}

static void E_print_escaped_char(PA char c) {
  char buf[4];
  if(c<0 || c>=32 && c!='"' && c!='\\' && c!=127) E_print_char(CA c);
  else switch(c) {
  case '\a': E_write(CA "\\a", 2); return;
  case '\b': E_write(CA "\\b", 2); return;
  case '\t': E_write(CA "\\t", 2); return;
  case '\n': E_write(CA "\\n", 2); return;
  case '\r': E_write(CA "\\r", 2); return;
  case '"': E_write(CA "\\\"", 2); return;
  case '\\': E_write(CA "\\\\", 2); return;
  case 127: E_write(CA "\\177", 4); return;
  default:
    buf[0] = '\\';
    buf[1] = '0';
    buf[2] = '0'+(c/8);
    buf[3] = '0'+(c&7);
    E_write(CA buf, 4);
  }
}

static void E_print_escaped_str(PA char const * s) {
  size_t p;
  char c;
  p = 0;
  while(c = s[p]) {
    if(c<0 || c>=32 && c!='"' && c!='\\' && c!=127) { p++; continue; }
    if(p) E_write(CA s, p);
    E_print_escaped_char(CA c);
    s += p+1; p = 0;
  }
  if(p) E_write(CA s, p);
}

static void E_print_escaped_bytes(PA char const * s, size_t len) {
  size_t p;
  char c;
  p = 0;
  while(p<len && (c=s[p])) {
    if(c<0 || c>=32 && c!='"' && c!='\\' && c!=127) { p++; continue; }
    if(p) E_write(CA s, p);
    E_print_escaped_char(CA c);
    s += p+1; len -= p+1; p = 0;
  }
  if(p) E_write(CA s, p);
}

#define F_DQ         256
#define F_0          64
#define F_0X         32
#define F_SPACE      16
#define F_SIGN       8
#define F_MINUS      4
#define F_L_ALIGN    2
#define F_ZERO_PAD   1

#define S_CHAR       1
#define S_SHORT      2
#define S_LONG       3
#ifdef HAVE_LONGLONG
#define S_LLONG      4
#endif
#define S_SIZET      5
#define S_8          6
#define S_16         7
#define S_32         8
#ifdef HAVE_INT64
#define S_64         9
#endif

/*
   up to 1 character added suffix (")
   up to 4 characters added prefix ("+0x)
 */
static void E_print_number(PA char * st, size_t len, size_t w, ushort flags) {
  size_t pw;
  pw = 0;
  if(flags & F_DQ) st[len++] = '"';
  if(flags & F_0X) { *(--st) = 'x'; *(--st) = '0'; pw += 2; }
  else if(flags & F_0) { *(--st) = '0'; pw++; }
  if(flags & F_MINUS) { *(--st) = '-'; pw++; }
  else if(flags & F_SIGN) { *(--st) = '+'; pw++; }
  else if(flags & F_SPACE) { *(--st) = ' '; pw++; }
  if(flags & F_DQ) { *(--st) = '"'; pw++; }
  len += pw;
  if(len>=w) {                       /* longer than width */
    E_write(CA st, len);
  } else if(flags & F_L_ALIGN) {     /* pad with right-spaces */
    E_write(CA st, len);
    E_repeat(CA ' ', w-len);
  } else if(!(flags & F_ZERO_PAD)) { /* pad with left-spaces */
    E_repeat(CA ' ', w-len);
    E_write(CA st, len);
  } else {                           /* pad with zeros */
    if(pw) E_write(CA st, pw);
    E_repeat(CA '0', w-len);
    E_write(CA st+pw, len-pw);
  }
}

static char const xx[17]="0123456789abcdef";
static char const XX[17]="0123456789ABCDEF";
#define E_PRINT(uname,utype,sname,stype) \
  static void E_print_##uname(PA utype v, size_t w, ushort flags) { \
    char dest[sizeof(v)*4+8]; size_t p; p = sizeof(dest)-1; \
    do { dest[--p] = (v%10)+'0'; v /= 10; } while(v && p>6); \
    E_print_number(CA dest+p, sizeof(dest)-1-p, w, flags); } \
  static void E_print_##uname##_o(PA utype v, size_t w, ushort flags) { \
    char dest[sizeof(v)*4+8]; size_t p; p = sizeof(dest)-1; \
    do { dest[--p] = (v%8)+'0'; v /= 8; } while(v && p>6); \
    E_print_number(CA dest+p, sizeof(dest)-1-p, w, flags); } \
  static void E_print_##uname##_x(PA utype v, size_t w, ushort flags) { \
    char dest[sizeof(v)*4+8]; size_t p; p = sizeof(dest)-1; \
    do { dest[--p] = xx[v%16]; v /= 16; } while(v && p>6); \
    E_print_number(CA dest+p, sizeof(dest)-1-p, w, flags); } \
  static void E_print_##uname##_X(PA utype v, size_t w, ushort flags) { \
    char dest[sizeof(v)*4+8]; size_t p; p = sizeof(dest)-1; \
    do { dest[--p] = XX[v%16]; v /= 16; } while(v && p>6); \
    E_print_number(CA dest+p, sizeof(dest)-1-p, w, flags); } \
  static void E_print_##sname(PA stype v, size_t w, ushort flags) { \
    if(v>=0) E_print_##uname(CA v, w, flags); \
    else E_print_##uname(CA -v, w, flags|F_MINUS); }
E_PRINT(uint, uint, int, int)
E_PRINT(ulong, ulong, long, long)
#ifdef HAVE_LONGLONG
E_PRINT(ull, ullong, ll, llong)
#endif
E_PRINT(uint32, uint32, int32, int32)
#ifdef HAVE_INT64
E_PRINT(uint64, uint64, int64, int64)
#endif
E_PRINT(uss, size_t, ss, ssize_t)

static int E_vprint(PA char const * fmt, va_list arg) {
  size_t p;
  char c;
  char const * s;
  size_t fw, fp, have_fp, fm, fwstr, dc, dp, sl;
  ushort flags;
  uchar bits;
  uint8 * tp; uint32 t32; int ii;
  p = 0;
  maxlen_present = 0;
  while(c = fmt[p]) {
    if(c!='%') { p++; continue; }
    if(p) E_write(CA fmt, p);
    CHECK_STOP
    fmt += p+1; p = 0;
    fw = 1; fwstr = 0; fp = 1; have_fp = 0; fm = 0; flags = 0; bits = 0;
    /*
    %[flags][width][.precision][:maxwidth][mods]{type}
      fw (field width) = [width] value (default: 1)
      fwstr (field width for strings %cCsS) = [width] value (default: 0)
      fm (field maxwidth) = [maxwidth] value (default: 0)
     */

again:
    c = *(fmt++);
    if(c=='"' && !(flags&F_DQ)) { flags |= F_DQ; goto again; }
    if(c=='-' && !(flags&F_L_ALIGN)) { flags |= F_L_ALIGN; goto again; }
    if(c=='+' && !(flags&F_SIGN)) { flags |= F_SIGN; goto again; }
    if(c==' ' && !(flags&F_SPACE)) { flags |= F_SPACE; goto again; }
    if(c>='0' && c<='9') {
      if(!(fw=c-'0')) flags |= F_ZERO_PAD;
      c = *(fmt++);
      if(!fw && c=='*') {
        fwstr = fw = (fcl_size)va_arg(arg,fcl_va_size);
        c = *(fmt++);
      } else {
        while(c>='0' && c<='9') {
          fw = fw*10+(size_t)(c-'0'); /* overflow check? */
          c = *(fmt++);
        }
        fwstr = fw;
      }
    } else if(c=='*') {
      fwstr = fw = (fcl_size)va_arg(arg,fcl_va_size);
      c = *(fmt++);
    }
    if(c=='.') {
      fp = 0;
      c = *(fmt++);
      if(c=='*') {
        fp = (fcl_size)va_arg(arg,fcl_va_size);
        c = *(fmt++);
      } else {
        while(c>='0' && c<='9') {
          fp = fp*10+(size_t)(c-'0'); /* overflow check? */
          c = *(fmt++);
        }
      }
      have_fp = 1;
    }
    if(c==':') {
      c = *(fmt++);
      if(c=='*') {
        fm = (fcl_size)va_arg(arg,fcl_va_size);
        c = *(fmt++);
      } else {
        while(c>='0' && c<='9') {
          fm = fm*10+(size_t)(c-'0'); /* overflow check? */
          c = *(fmt++);
        }
      }
    }
    if(c=='h') { c = *(fmt++); bits = S_SHORT;
      if(c=='h') { c = *(fmt++); bits = S_CHAR; }
    } else if(c=='l') { c = *(fmt++); bits = S_LONG;
#ifdef HAVE_LONGLONG
      if(c=='l') { c = *(fmt++); bits = S_LLONG; }
#endif
    } else if(c=='H') { c = *(fmt++); bits = S_16;
      if(c=='H') { c = *(fmt++); bits = S_8; }
    } else if(c=='L') { c = *(fmt++); bits = S_32;
#ifdef HAVE_INT64
      if(c=='L') { c = *(fmt++); bits = S_64; }
#endif
    } else if(c=='z') { c = *(fmt++); bits = S_SIZET; }
    remlen = fwstr; if(fm) { maxlen = fm; maxlen_present = 1; }
    switch(c) {
    case 'c':
      c = va_arg(arg, int);
      if(flags&F_DQ) E_print_char(CA '"');
      E_print_char(CA c);
      if(flags&F_DQ) E_print_char(CA '"');
      if(remlen) E_repeat(CA ' ', remlen);
      break;
    case 'C':
      c = va_arg(arg, int);
      if(flags&F_DQ) E_print_char(CA '"');
      E_print_escaped_char(CA c);
      if(flags&F_DQ) E_print_char(CA '"');
      if(remlen) E_repeat(CA ' ', remlen);
      break;
    case 's':
      if(!(s = va_arg(arg, char const*))) E_print_str(CA "(null)");
      else {
        if(flags&F_DQ) E_print_char(CA '"');
        if(!have_fp) E_print_str(CA s);
        else { if((sl=strlen(s))>fp) sl=fp; E_write(CA s, sl); }
        if(flags&F_DQ) E_print_char(CA '"');
      }
      if(remlen) E_repeat(CA ' ', remlen);
      break;
    case 'S':
      if(!(s = va_arg(arg, char const*))) E_print_str(CA "(null)");
      else {
        if(flags&F_DQ) E_print_char(CA '"');
        if(!have_fp) E_print_escaped_str(CA s);
        else { if((sl=strlen(s))>fp) sl=fp; E_print_escaped_bytes(CA s, sl); }
        if(flags&F_DQ) E_print_char(CA '"');
      }
      if(remlen) E_repeat(CA ' ', remlen);
      break;
    case 'b':
      if(!(s = va_arg(arg, char const*))) E_print_str(CA "(null)");
      else {
        if(flags&F_DQ) E_print_char(CA '"');
        E_write(CA s, fp);
        if(flags&F_DQ) E_print_char(CA '"');
      }
      if(remlen) E_repeat(CA ' ', remlen);
      break;
    case 'B':
      if(!(s = va_arg(arg, char const*))) E_print_str(CA "(null)");
      else {
        if(flags&F_DQ) E_print_char(CA '"');
        E_print_escaped_bytes(CA s, fp);
        if(flags&F_DQ) E_print_char(CA '"');
      }
      if(remlen) E_repeat(CA ' ', remlen);
      break;
    case 'd':
      switch(bits) { /* default: char, short, int, int8, int16 */ 
      default: E_print_int(CA va_arg(arg, int), fw, flags); break;
      case S_LONG: E_print_long(CA va_arg(arg, long), fw, flags); break;
#ifdef HAVE_LONGLONG
      case S_LLONG: E_print_ll(CA va_arg(arg, llong), fw, flags); break;
#endif
      case S_32: E_print_int32(CA va_arg(arg, int32), fw, flags); break;
#ifdef HAVE_INT64
      case S_64: E_print_int64(CA va_arg(arg, int64), fw, flags); break;
#endif
      case S_SIZET:
        E_print_ss(CA (fcl_ssize)va_arg(arg, fcl_va_ssize), fw, flags); break;
      }
      break;
    case 'u':
      switch(bits) {
      /* short unsigned types will be promoted to signed int, but:
         1) we have a problem with ushort on 16bit platform
         2) no harm to use uint instead of int
         */
      default: E_print_uint(CA va_arg(arg, uint), fw, flags); break;
      case S_LONG: E_print_ulong(CA va_arg(arg, ulong), fw, flags); break;
#ifdef HAVE_LONGLONG
      case S_LLONG: E_print_ull(CA va_arg(arg, ullong), fw, flags); break;
#endif
      case S_32: E_print_uint32(CA va_arg(arg, uint32), fw, flags); break;
#ifdef HAVE_INT64
      case S_64: E_print_uint64(CA va_arg(arg, uint64), fw, flags); break;
#endif
      case S_SIZET:
        E_print_uss(CA (fcl_size)va_arg(arg, fcl_va_size), fw, flags); break;
      }
      break;
    case 'o':
      switch(bits) {
      default: E_print_uint_o(CA va_arg(arg, uint), fw, flags); break;
      case S_LONG: E_print_ulong_o(CA va_arg(arg, ulong), fw, flags); break;
#ifdef HAVE_LONGLONG
      case S_LLONG: E_print_ull_o(CA va_arg(arg, ullong), fw, flags); break;
#endif
      case S_32: E_print_uint32_o(CA va_arg(arg, uint32), fw, flags); break;
#ifdef HAVE_INT64
      case S_64: E_print_uint64_o(CA va_arg(arg, uint64), fw, flags); break;
#endif
      case S_SIZET:
        E_print_uss_o(CA (fcl_size)va_arg(arg, fcl_va_size), fw, flags); break;
      }
      break;
    case 'x':
      switch(bits) {
      default: E_print_uint_x(CA va_arg(arg, uint), fw, flags); break;
      case S_LONG: E_print_ulong_x(CA va_arg(arg, ulong), fw, flags); break;
#ifdef HAVE_LONGLONG
      case S_LLONG: E_print_ull_x(CA va_arg(arg, ullong), fw, flags); break;
#endif
      case S_32: E_print_uint32_x(CA va_arg(arg, uint32), fw, flags); break;
#ifdef HAVE_INT64
      case S_64: E_print_uint64_x(CA va_arg(arg, uint64), fw, flags); break;
#endif
      case S_SIZET:
        E_print_uss_x(CA (fcl_size)va_arg(arg, fcl_va_size), fw, flags); break;
      }
      break;
    case 'X':
      switch(bits) {
      default: E_print_uint_X(CA va_arg(arg, uint), fw, flags); break;
      case S_LONG: E_print_ulong_X(CA va_arg(arg, ulong), fw, flags); break;
#ifdef HAVE_LONGLONG
      case S_LLONG: E_print_ull_X(CA va_arg(arg, ullong), fw, flags); break;
#endif
      case S_32: E_print_uint32_X(CA va_arg(arg, uint32), fw, flags); break;
#ifdef HAVE_INT64
      case S_64: E_print_uint64_X(CA va_arg(arg, uint64), fw, flags); break;
#endif
      case S_SIZET:
        E_print_uss_X(CA (fcl_size)va_arg(arg, fcl_va_size), fw, flags); break;
      }
      break;
    case 'p':
      E_print_ulong_x(CA (ulong)(va_arg(arg, void*)), fw, flags|F_0X);
      break;
    case '%':
      E_print_char(CA c);
      break;
    case '{':
      for(p=0; fmt[p] && fmt[p]!='}'; p++);
      if(fmt[p]!='}') { maxlen_present = 0; return -1; }
      if(p==3 && fmt[0]=='I' && fmt[1]=='P' && fmt[2]=='4') {
        tp = va_arg(arg, void*);
        E_print_uint(CA tp[0], 1, 0);
        E_print_char(CA '.');
        E_print_uint(CA tp[1], 1, 0);
        E_print_char(CA '.');
        E_print_uint(CA tp[2], 1, 0);
        E_print_char(CA '.');
        E_print_uint(CA tp[3], 1, 0);
      } else if(p==4 && fmt[0]=='I' && fmt[1]=='P'
        && fmt[2]=='3' && fmt[3]=='2') {
        t32 = va_arg(arg, uint32);
        E_print_uint(CA (t32>>24)&255, 1, 0);
        E_print_char(CA '.');
        E_print_uint(CA (t32>>16)&255, 1, 0);
        E_print_char(CA '.');
        E_print_uint(CA (t32>>8)&255, 1, 0);
        E_print_char(CA '.');
        E_print_uint(CA t32&255, 1, 0);
      } else if(p==3 && fmt[0]=='M' && fmt[1]=='A' && fmt[2]=='C') {
        tp = va_arg(arg, void*);
        E_print_uint_X(CA tp[0], 2, F_ZERO_PAD);
        E_print_char(CA ':');
        E_print_uint_X(CA tp[1], 2, F_ZERO_PAD);
        E_print_char(CA ':');
        E_print_uint_X(CA tp[2], 2, F_ZERO_PAD);
        E_print_char(CA ':');
        E_print_uint_X(CA tp[3], 2, F_ZERO_PAD);
        E_print_char(CA ':');
        E_print_uint_X(CA tp[4], 2, F_ZERO_PAD);
        E_print_char(CA ':');
        E_print_uint_X(CA tp[5], 2, F_ZERO_PAD);
      } else if(p==3 && fmt[0]=='E' && fmt[1]=='R' && fmt[2]=='R') {
        ii = va_arg(arg, int);
        E_print_int(CA ii, 1, 0); 
        E_print_str(CA " (");
        E_print_str(CA fcl_strerror(ii));
        E_print_char(CA ')');
      } else if(p>=4 && fmt[0]=='X' && fmt[1]=='D' && fmt[2]==':') {
        if(p==4 && fmt[3]=='*') {
          dc = (fcl_size)va_arg(arg,fcl_va_size);
        } else {
          dc = 0;
          for(dp=3; dp<p; dp++) {
            c = fmt[dp];
            if(c<'0' || c>'9') { maxlen_present = 0; return -1; }
            dc = dc*10+(size_t)(c-'0'); /* overflow check? */
          }
        }
        tp = va_arg(arg, void*);
        while(dc) {
          switch(bits) {
          default:
            E_print_uint_X(CA (*(uint*)tp), fw, flags);
            tp+=sizeof(uint); break;
          case S_CHAR:
            E_print_uint_X(CA (*(uchar*)tp), fw, flags);
            tp+=sizeof(uchar); break;
          case S_SHORT:
            E_print_uint_X(CA (*(ushort*)tp), fw, flags);
            tp+=sizeof(ushort); break;
          case S_LONG:
            E_print_ulong_X(CA (*(ulong*)tp), fw, flags);
            tp+=sizeof(ulong); break;
#ifdef HAVE_LONGLONG
          case S_LLONG:
            E_print_ull_X(CA (*(ullong*)tp), fw, flags);
            tp+=sizeof(ullong); break;
#endif
          case S_8:
            E_print_uint32_X(CA (*(uint8*)tp), fw, flags);
            tp+=sizeof(uint8); break;
          case S_16:
            E_print_uint32_X(CA (*(uint16*)tp), fw, flags);
            tp+=sizeof(uint16); break;
          case S_32:
            E_print_uint32_X(CA (*(uint32*)tp), fw, flags);
            tp+=sizeof(uint32); break;
#ifdef HAVE_INT64
          case S_64:
            E_print_uint64_X(CA (*(uint64*)tp), fw, flags);
            tp+=sizeof(uint64); break;
#endif
          case S_SIZET:
            E_print_uss_X(CA (*(fcl_size*)tp), fw, flags);
            tp+=sizeof(fcl_size); break;
          }
          dc--;
          if(dc) E_print_char(CA ' ');
        }
      } else { maxlen_present = 0; return -1; }
      fmt += p+1;
      p = 0;
      break;
    default:
      maxlen_present = 0; return -1;
    }
    maxlen_present = 0;
  }
  if(p) E_write(CA fmt, p);
  return 0;
}

