/*
 * This file is part of firk's Common C Library
 *  strparse.c - string parsing functions
 *
 * (c) firk 2015, 2016, 2019, 2020, 2021; firk@cantconnect.ru
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. The origin of this software must not be misrepresented; you must
 *    not claim that you wrote the original software.
 * 4. Altered versions in any form must be plainly marked as such, and
 *    must not be misinterpreted as being the original software.
 *
 * This software is provided by the author and contributors `as is'
 * without any express or implied warranty.
 */

#include <string.h>
#include "types.h"
#include "strparse.h"

extern char const * str_skip_spaces(char const * st) {
  char c;
  if(st) {
    while(c=*st) {
      if(c==' ' || c=='\t') st++;
      else break;
    }
  }
  return st;
}

extern char const * str_skip_prefix(char const * st, char const * prefix) {
  char c;
  if(!st) return NULL;
  while(c=*(prefix++)) if(c!=*(st++)) return NULL;
  return st;
}

extern char str_to_args(char * st, char * * argv, size_t * argc, size_t max,
                        uint flags) {
  char c, q;
  char *arg, *out;
  size_t n;
  n = 0;
  out = st;
  q = 0;
  while(1) {
    while((c=*st)==' ' || c=='\t') st++;
    if(!c) break;
    arg = out;
    while(1) {
      c = *st;
      if(!c || !q && (c==' ' || c=='\t')) break;
      if(c==q) { st++; q=0; continue; }
      if(!q && (c=='"' || c=='\'')) { st++; q=c; continue; }
      if(c=='\\') {
        c = st[1];
        if(!c) { *argc = n; return q?q:'\\'; }
        if(flags & 1) switch(c) {
        case 'a': c=7; break;
        case 'b': c=8; break;
        case 't': c=9; break;
        case 'n': c=10; break;
        case 'v': c=11; break;
        case 'f': c=12; break;
        case 'r': c=13; break;
        case '"': case '\'': break;
        default: *argc = n; return '\\';
        }
        st++;
      }
      *(out++) = c;
      st++;
    }
    if(q) { *argc = n; return q; }
    *(out++) = 0;
    if(n<max) argv[n] = arg;
    n++;
    if(!c) break;
    st++;
  }
  *argc = n;
  return 0;
}

extern unsigned char chr_to_value(char c) {
  if(c>='0' && c<='9') return (unsigned char)(c-'0');
  if(c>='A' && c<='Z') return (unsigned char)(c-'A'+10);
  if(c>='a' && c<='z') return (unsigned char)(c-'a'+10);
  return 255;
}

extern int str_get_bool(char const * st) {
  if(!st) return -1;
  if(!strcasecmp(st, "0") ||
     !strcasecmp(st, "no") ||
     !strcasecmp(st, "false") ||
     !strcasecmp(st, "n") ||
     !strcasecmp(st, "off")) return 0;
  if(!strcasecmp(st, "1") ||
     !strcasecmp(st, "yes") ||
     !strcasecmp(st, "true") ||
     !strcasecmp(st, "y") ||
     !strcasecmp(st, "on")) return 1;
  return -1;
}

#define STR_TO_I(ftname, uname, sname) \
extern int str_to_u##ftname(char const * st, uname * res) {\
  uname r, t;\
  char c;\
  r = 0;\
  if(!st) return -1;\
  if(!*st) return -1;\
  while(c=*st) {\
    if(c<'0' || c>'9') return -1;\
    t = r*10;\
    if(t/10!=r) return -1;\
    r = t + (unsigned char)(c-'0');\
    if(r<t) return -1;\
    st++;\
  }\
  *res = r;\
  return 0;\
} \
extern int strp_to_u##ftname(char const * * st, uname * res) {\
  char const * s;\
  uname r, t;\
  char c;\
  r = 0;\
  if(!st) return -1;\
  if(!(s=*st)) return -1;\
  while(c=*s) {\
    if(c<'0' || c>'9') break;\
    t = r*10;\
    if(t/10!=r) return -1;\
    r = t + (unsigned char)(c-'0');\
    if(r<t) return -1;\
    s++;\
  }\
  if(s==*st) return -1;\
  *st = s;\
  *res = r;\
  return 0;\
}\
extern int str_hex_to_u##ftname(char const * st, uname * res) {\
  uname r, t, u;\
  char c;\
  r = 0;\
  if(!st) return -1;\
  if(!*st) return -1;\
  while(c=*st) {\
    if((u=chr_to_value(c))>15) return -1;\
    t = r*16;\
    if(t/16!=r) return -1;\
    r = t + u;\
    if(r<t) return -1;\
    st++;\
  }\
  *res = r;\
  return 0;\
}\
extern int str_base_to_u##ftname(char const * st, uname * res, uchar b) {\
  uname r, t, u;\
  char c;\
  r = 0;\
  if(!st) return -1;\
  if(!*st) return -1;\
  while(c=*st) {\
    if((u=chr_to_value(c))>=b) return -1;\
    t = r*b;\
    if(t/b!=r) return -1;\
    r = t + u;\
    if(r<t) return -1;\
    st++;\
  }\
  *res = r;\
  return 0;\
}\
extern int strp_base_to_u##ftname(char const * * st, uname * res, uchar b) {\
  char const * s;\
  uname r, t, u;\
  char c;\
  r = 0;\
  if(!st) return -1;\
  if(!(s=*st)) return -1;\
  while(c=*s) {\
    if((u=chr_to_value(c))>=b) break;\
    t = r*b;\
    if(t/b!=r) return -1;\
    r = t + u;\
    if(r<t) return -1;\
    s++;\
  }\
  if(s==*st) return -1;\
  *st = s;\
  *res = r;\
  return 0;\
}\
extern int str_to_s##ftname(char const * st, sname * res) {\
  uname r;\
  sname rr;\
  int minus;\
  if(!st || !*st) return -1;\
  if(*st=='-') { minus=1; st++; } else\
  if(*st=='+') { minus=0; st++; } else minus=0;\
  if(str_to_u##ftname(st, &r)<0) return -1;\
  if(!r) { *res=0; return 0; }\
  if(!minus) { \
    rr = (sname)r;\
    if(rr<=0 || ((uname)rr)!=r) return -1;\
  } else { \
    rr = ((sname)-1)-(sname)(r-1);\
    if(rr>=0 || ((uname)(-1-rr))!=r-1) return -1; \
  } \
  *res = rr;\
  return 0;\
}\
extern int strp_to_s##ftname(char const * * st, sname * res) {\
  char const * s;\
  uname r;\
  sname rr;\
  int minus;\
  if(!st || !(s=*st)) return -1;\
  if(*s=='-') { minus=1; s++; } else\
  if(*s=='+') { minus=0; s++; } else minus=0;\
  if(strp_to_u##ftname(&s, &r)<0) return -1;\
  if(!r) { *st = s; *res=0; return 0; }\
  if(!minus) { \
    rr = (sname)r;\
    if(rr<=0 || ((uname)rr)!=r) return -1;\
  } else { \
    rr = ((sname)-1)-(sname)(r-1);\
    if(rr>=0 || ((uname)(-1-rr))!=r-1) return -1; \
  } \
  *st = s;\
  *res = rr;\
  return 0;\
}\
extern int str_base_to_s##ftname(char const * st, sname * res, uchar b) {\
  uname r;\
  sname rr;\
  int minus;\
  if(!st || !*st) return -1;\
  if(*st=='-') { minus=1; st++; } else\
  if(*st=='+') { minus=0; st++; } else minus=0;\
  if(str_base_to_u##ftname(st, &r, b)<0) return -1;\
  if(!r) { *res=0; return 0; }\
  if(!minus) { \
    rr = (sname)r;\
    if(rr<=0 || ((uname)rr)!=r) return -1;\
  } else { \
    rr = ((sname)-1)-(sname)(r-1);\
    if(rr>=0 || ((uname)(-1-rr))!=r-1) return -1; \
  } \
  *res = rr;\
  return 0;\
}\
extern int strp_base_to_s##ftname(char const * * st, sname * res, uchar b) {\
  char const * s;\
  uname r;\
  sname rr;\
  int minus;\
  if(!st || !(s=*st)) return -1;\
  if(*s=='-') { minus=1; s++; } else\
  if(*s=='+') { minus=0; s++; } else minus=0;\
  if(strp_base_to_u##ftname(&s, &r, b)<0) return -1;\
  if(!r) { *st = s; *res=0; return 0; }\
  if(!minus) { \
    rr = (sname)r;\
    if(rr<=0 || ((uname)rr)!=r) return -1;\
  } else { \
    rr = ((sname)-1)-(sname)(r-1);\
    if(rr>=0 || ((uname)(-1-rr))!=r-1) return -1; \
  } \
  *st = s;\
  *res = rr;\
  return 0;\
}

STR_TO_I(char, unsigned char, signed char)
STR_TO_I(short, unsigned short, signed short)
STR_TO_I(int, unsigned int, signed int)
STR_TO_I(long, unsigned long, signed long)
#ifdef HAVE_LONGLONG
STR_TO_I(llong, unsigned long long, signed long long)
#endif
#ifdef HAVE_INT8
STR_TO_I(int8, uint8, int8)
#endif
#ifdef HAVE_INT16
STR_TO_I(int16, uint16, int16)
#endif
#ifdef HAVE_INT32
STR_TO_I(int32, uint32, int32)
#endif
#ifdef HAVE_INT64
STR_TO_I(int64, uint64, int64)
#endif
STR_TO_I(size, fcl_size, fcl_ssize)


