/*
 * This file is part of firk's Common C Library
 *  strparse_fp.c - string parsing functions for floating-point values
 *
 * (c) firk 2019, 2021; firk@cantconnect.ru
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. The origin of this software must not be misrepresented; you must
 *    not claim that you wrote the original software.
 * 4. Altered versions in any form must be plainly marked as such, and
 *    must not be misinterpreted as being the original software.
 *
 * This software is provided by the author and contributors `as is'
 * without any express or implied warranty.
 */

#include <sys/types.h>
#include "types.h"
#include "strparse.h"

/* extended precision minimum exponent is about -32768-63,
   so we will not fit int16 anyway */
typedef int32 etype; /* type for keeping exponent */
typedef uint32 uetype; /* type for keeping abs(exponent) */
 /* minimum exponent value that we will try to process
    w/o just returning zero result */
#define EXP_MIN -1000000
 /* maximum exponent value before surely overflowing */
#define EXP_MAX 1000000
typedef long double best_fp;

typedef struct {
  int sign;
  char const * vstart;
  size_t dotpos, vlen;
  etype e;
} ppfdata;

static int preparse_float(char const * st, size_t maxlen, uint flags,
                          ppfdata * r, size_t * parsed_len) {
  size_t i0, i, j;
  int neg_exp;
  char c;
  uetype ur, ut;
  uchar uc;
  *parsed_len = 0;
  if(*st=='-')
    { if(flags&4) return FP_SIGN_DISALLOWED; r->sign = -1; i0 = 1; }
  else if(*st=='+')
    { if(flags&4) return FP_SIGN_DISALLOWED; r->sign = 1; i0 = 1; }
  else
    { r->sign = 0; i0 = 0; }
  r->vstart = st+i0;
  for(i=i0; maxlen?i<maxlen:st[i]; i++) {
    c = st[i];
    if(c=='.') { r->dotpos = i-i0; goto dot; }
    if((flags&1) && (c=='e' || c=='E')) { r->dotpos = i-i0; goto exponent; }
    if(c<'0' || c>'9') break;
  }
  if(i==i0) { *parsed_len = i; return FP_MISSING_DIGITS; }
  r->dotpos = i-i0;
  r->vlen = i-i0;
  r->e = 0;
  *parsed_len = i;
  return 0;
dot:
  if(i==i0 && !(flags&2)) { *parsed_len = i; return FP_MISSING_BEFORE_DOT; }
  for(i++; maxlen?i<maxlen:st[i]; i++) {
    c = st[i];
    if((flags&1) && (c=='e' || c=='E')) goto exponent;
    if(c<'0' || c>'9') break;
  }
  if(i==i0+1) { *parsed_len = i; return FP_MISSING_DIGITS; }
  r->vlen = i-i0;
  r->e = 0;
  *parsed_len = i;
  return 0;
exponent:
  if(i==i0 || !r->dotpos && i==i0+1)
    { *parsed_len = i; return FP_MISSING_BEFORE_EXP; }
  r->vlen = i-i0;
  i++;
  if(!(maxlen?i<maxlen:st[i])) { *parsed_len = i; return FP_MISSING_EXP; }
  c = st[i];
  neg_exp = 0;
  if(c=='+') i++; else if(c=='-') { neg_exp=1; i++; }
  if(maxlen) {
    if(i>=maxlen) { *parsed_len = i; return FP_MISSING_EXP; }
    st += i; maxlen -= i;
  } else {
    st += i;
    if(!st[i]) { *parsed_len = i; return FP_MISSING_EXP; }
  }
  ur = 0;
  for(j=0; maxlen?j<maxlen:st[j]; j++) {
    uc = (uchar)st[j]-'0';
    if(uc>9) break;
    ut = ur*10;
    if(ut/10!=ur) { *parsed_len = i+j; return FP_OVERFLOW_EXP; }
    ur = ut+uc;
    if(ur<ut) { *parsed_len = i+j; return FP_OVERFLOW_EXP; }
  }
  *parsed_len = i+j;
  if(!j) return FP_MISSING_EXP;
  if(!ur) r->e = 0;
  else if(!neg_exp) {
    r->e = (etype)ur;
    if(r->e<=0 || ur!=(uetype)r->e) return FP_OVERFLOW_EXP;
  } else {
    r->e = -1;
    r->e -= (etype)(ur-1);
    if(r->e>=0 || ur-1!=(uetype)(-1-r->e)) return FP_OVERFLOW_EXP;
  }
  return 0;
}

/*
  preprocess the string:
    align r->vstart to point to first significant (non-zero) digit
    reduce r->vlen to strip any tralining zeros too
    r->vstart[r->dotpos] is just to be ignored if r->dotpos<r->vlen
    r->e is normalized exponent: it is the number of digits before dot
  ret. 0 = got just zero; 1 = there are some non-zero digits; -1 = overflow */
#define CHECK(A) if(!(A)) { return -1; } /* TODO: fatal */
static int fix_ppfdata(ppfdata * r) {
  char const * st;
  size_t i, j, k;
  etype e;
  
  st = r->vstart;
  i = r->dotpos;
  j = r->vlen;
  e = r->e;
  CHECK(j==i || j>i && st[i]=='.');
  while(i && *st=='0') { st++; i--; j--; }
  while(j-i>1 && st[j-1]=='0') j--;
  if(j==i+1) j--;
  if(!j) return 0; /* no significant digits at all */
  if(j==i) { /* all significant digits are before dot */
    /* TODO: sign-typecast-safety */
    if(e>=EXP_MAX || ((uetype)(EXP_MAX-e))<=i) return -1;
    while(1) {
      CHECK(j);
      if(st[j-1]!='0') break;
      j--;
    }
    r->vstart = st;
    r->dotpos = i;
    r->vlen = j;
    r->e = e+i;
    return 1;
  }
  if(i) { /* there are significant digits both before and after dot */
    /* TODO: sign-typecast-safety */
    if(e>=EXP_MAX || ((uetype)(EXP_MAX-e))<=i) return -1;
    r->vstart = st;
    r->dotpos = i;
    r->vlen = j;
    r->e = e+i;
    return 1;
  }
  /* all significant digits are after dot */
  while(1) {
    CHECK(j>=2);
    if(st[1]!='0') break;
    st++;
    j--;
    if(e<=EXP_MIN) return 0;
    e--;
  }
  r->vstart = st;
  r->dotpos = 0;
  r->vlen = j;
  r->e = e;
  return 1;
}

static int parse_float(ppfdata * p, float * res);
static int parse_double(ppfdata * p, double * res);
static int parse_ldouble(ppfdata * p, long double * res);

extern int str_to_float(char const * st, float * res, uint flags) {
  ppfdata r;
  size_t len;
  int err;
  if(!st) return FP_NULL_INPUT;
  if(err = preparse_float(st, 0, flags, &r, &len)) return err;
  if(st[len]) return FP_EXTRA_SYMBOLS;
  switch(fix_ppfdata(&r)) {
  case 0: *res = 0; return 0;
  case -1: return FP_OVERFLOW;
  }
  if(parse_float(&r, res)<0) return FP_OVERFLOW;
  return 0;
}

extern int str_to_double(char const * st, double * res, uint flags) {
  ppfdata r;
  size_t len;
  int err;
  if(!st) return FP_NULL_INPUT;
  if(err = preparse_float(st, 0, flags, &r, &len)) return err;
  if(st[len]) return FP_EXTRA_SYMBOLS;
  switch(fix_ppfdata(&r)) {
  case 0: *res = 0; return 0;
  case -1: return FP_OVERFLOW;
  }
  if(parse_double(&r, res)<0) return FP_OVERFLOW;
  return 0;
}

extern int str_to_ldouble(char const * st, long double * res, uint flags) {
  ppfdata r;
  size_t len;
  int err;
  if(!st) return FP_NULL_INPUT;
  if(err = preparse_float(st, 0, flags, &r, &len)) return err;
  if(st[len]) return FP_EXTRA_SYMBOLS;
  switch(fix_ppfdata(&r)) {
  case 0: *res = 0; return 0;
  case -1: return FP_OVERFLOW;
  }
  if(parse_ldouble(&r, res)<0) return FP_OVERFLOW;
  return 0;
}

extern int strp_to_float(char const * * st, float * res, uint flags) {
  ppfdata r;
  size_t len;
  int err;
  char const * s;
  if(!st || !(s=*st)) return FP_NULL_INPUT;
  if(err = preparse_float(s, 0, flags, &r, &len)) return err;
  switch(fix_ppfdata(&r)) {
  case 0: *res = 0; *st=s+len; return 0;
  case -1: return FP_OVERFLOW;
  }
  if(parse_float(&r, res)<0) return FP_OVERFLOW;
  *st = s+len;
  return 0;
}

extern int strp_to_double(char const * * st, double * res, uint flags) {
  ppfdata r;
  size_t len;
  int err;
  char const * s;
  if(!st || !(s=*st)) return FP_NULL_INPUT;
  if(err = preparse_float(s, 0, flags, &r, &len)) return err;
  switch(fix_ppfdata(&r)) {
  case 0: *res = 0; *st=s+len; return 0;
  case -1: return FP_OVERFLOW;
  }
  if(parse_double(&r, res)<0) return FP_OVERFLOW;
  *st = s+len;
  return 0;
}

extern int strp_to_ldouble(char const * * st, long double * res, uint flags) {
  ppfdata r;
  size_t len;
  int err;
  char const * s;
  if(!st || !(s=*st)) return FP_NULL_INPUT;
  if(err = preparse_float(s, 0, flags, &r, &len)) return err;
  switch(fix_ppfdata(&r)) {
  case 0: *res = 0; *st=s+len; return 0;
  case -1: return FP_OVERFLOW;
  }
  if(parse_ldouble(&r, res)<0) return FP_OVERFLOW;
  *st = s+len;
  return 0;
}



static void parse_best_fp(ppfdata * p, best_fp * res) {
  /* TODO: looks like this may lose precision */
  best_fp sum, cexp, nexp;
  size_t i, dp;
  char const * st;
  uchar c;
  etype e;
  st = p->vstart; dp = p->dotpos; e = p->e;
  /* make VALUE = sum * pow(10,e) */
  sum = 0; cexp = 1;
  for(i=p->vlen; i--; ) if(i!=dp) {
    c = st[i]-'0';
    sum += cexp*c;
    if(e>EXP_MIN) {
      cexp *= 10;
      e--;
    } else {
      sum /= 10;
    }
  }
  /* now calc pow(10,e) */
  nexp = 1;
  if(e>0) {
    cexp = 10;
    while(e) {
      if(e & 1) nexp *= cexp;
      e >>= 1;
      cexp *= cexp;
    }
    sum *= nexp;
  } else if(e<0) {
    e = -e;
    cexp = 10;
    while(e) {
      if(e & 1) nexp *= cexp;
      e >>= 1;
      cexp *= cexp;
    }
    sum /= nexp;
  }
  *res = (p->sign>=0)?sum:-sum;
}

static int parse_float(ppfdata * p, float * res) {
  float val;
  best_fp r;
  parse_best_fp(p, &r);
  val = r;
  /* detect unwanted INF/NAN */
  if(val+1>val || val*2!=val) { *res = val; return 0; }
  return -1;
}

static int parse_double(ppfdata * p, double * res) {
  double val;
  best_fp r;
  parse_best_fp(p, &r);
  val = r;
  /* detect unwanted INF/NAN */
  if(val+1>val || val*2!=val) { *res = val; return 0; }
  return -1;
}

static int parse_ldouble(ppfdata * p, long double * res) {
  long double val;
  best_fp r;
  parse_best_fp(p, &r);
  val = r;
  /* detect unwanted INF/NAN */
  if(val+1>val || val*2!=val) { *res = val; return 0; }
  return -1;
}



