/*
 * This file is part of firk's Common C Library
 *  strparse_sys.c - string parsing functions for OS-related values
 *
 * (c) firk 2021; firk@cantconnect.ru
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. The origin of this software must not be misrepresented; you must
 *    not claim that you wrote the original software.
 * 4. Altered versions in any form must be plainly marked as such, and
 *    must not be misinterpreted as being the original software.
 *
 * This software is provided by the author and contributors `as is'
 * without any express or implied warranty.
 */

#include <sys/types.h>
#include <pwd.h>
#include <grp.h>
#include <errno.h>
#include "config/work/config.h"
#include "types.h"
#include "string.h"
#include "strparse.h"

extern int str_to_ip4_addr(char const * st, uint32 * ip) {
  uint8 t[4];
  if(strp_to_uint8(&st,t)<0 || *(st++)!='.' ||
     strp_to_uint8(&st,t+1)<0 || *(st++)!='.' ||
     strp_to_uint8(&st,t+2)<0 || *(st++)!='.' ||
     strp_to_uint8(&st,t+3)<0 || *st) return -1;
  *ip = t[0]*0x1000000 + t[1]*0x10000 + t[2]*0x100 + t[3];
  return 0;
}

extern int str_to_ip4_addr_port(char const * st, uint32 * ip, uint16 * port) {
  uint8 t[4];
  if(*st==':') {
    if(str_to_uint16(st+1, port)<0) return -1;
    return 2;
  }
  if(strp_to_uint8(&st,t)<0 || *(st++)!='.' ||
     strp_to_uint8(&st,t+1)<0 || *(st++)!='.' ||
     strp_to_uint8(&st,t+2)<0 || *(st++)!='.' ||
     strp_to_uint8(&st,t+3)<0) return -1;
  if(*st==':') {
    if(str_to_uint16(st+1, port)<0) return -1;
    *ip = t[0]*0x1000000 + t[1]*0x10000 + t[2]*0x100 + t[3];
    return 3;
  }
  if(*st) return -1;
  *ip = t[0]*0x1000000 + t[1]*0x10000 + t[2]*0x100 + t[3];
  return 1;
}

extern int str_to_eth_addr(char const * st, uint8 * mac) {
  uint8 t[6];
  uint i;
  if(strp_base_to_uint8(&st,t,16)<0 || *(st++)!=':' ||
     strp_base_to_uint8(&st,t+1,16)<0 || *(st++)!=':' ||
     strp_base_to_uint8(&st,t+2,16)<0 || *(st++)!=':' ||
     strp_base_to_uint8(&st,t+3,16)<0 || *(st++)!=':' ||
     strp_base_to_uint8(&st,t+4,16)<0 || *(st++)!=':' ||
     strp_base_to_uint8(&st,t+5,16)<0 || *st) return -1;
  for(i=0;i<6;i++) mac[i]=t[i];
  return 0;
}

extern int str_to_uid(char const * st, uint32 * uid) {
  struct passwd * r;
  if(str_to_uint32(st, uid)>=0) return 0;
#ifdef HAVE_GETPWNAM_R
  {
    struct passwd p;
    char buf[2048];
    int e;
    while((e = getpwnam_r(st, &p, buf, sizeof(buf), &r))==EINTR);
    if(r) { *uid = r->pw_uid; return 0; }
    return -1; /* we do not maintain errno, so any problems
      will be just "can't parse this uid-spec" */
  }
#else
#if HAVE_GETPWNAM
  {
    while(!(errno=0,r=getpwnam(st))) if(errno!=EINTR) return -1;
    *uid = r->pw_uid;
    return 0;
  }
#else
  return -1;
#endif
#endif
}

extern int str_to_gid(char const * st, uint32 * gid) {
  struct group * r;
  if(str_to_uint32(st, gid)>=0) return 0;
#ifdef HAVE_GETGRNAM_R
  {
    struct group p;
    char buf[2048];
    int e;
    while((e = getgrnam_r(st, &p, buf, sizeof(buf), &r))==EINTR);
    if(r) { *gid = r->gr_gid; return 0; }
    return -1; /* we do not maintain errno, so any problems
      will be just "can't parse this gid-spec" */
  }
#else
#if HAVE_GETGRNAM
  {
    while(!(errno=0,r=getgrnam(st))) if(errno!=EINTR) return -1;
    *gid = r->gr_gid;
    return 0;
  }
#else
  return -1;
#endif
#endif
}

extern int str_to_uid_gid(char const * st, uint32 * uid, uint32 * gid) {
  char tmp[1024];
  uint p;
  uint32 u,g;
  for(p=0; p<1024; p++) {
    if(!st[p]) {
      if(str_to_uid(st, uid)<0) return -1;
      return 1;
    }
    if(st[p]==':') {
      bcopy(st,tmp,p);
      tmp[p] = 0;
      if(str_to_uid(tmp, &u)<0 || str_to_gid(st+p+1, &g)<0) return -1;
      *uid = u; *gid = g;
      return 3;
    }
  }
  return -1;
}

extern int strp_to_ip4_addr(char const * * st, uint32 * ip) {
  char const * s;
  uint8 t[4];
  s  = *st;
  if(strp_to_uint8(&s,t)<0 || *(s++)!='.' ||
     strp_to_uint8(&s,t+1)<0 || *(s++)!='.' ||
     strp_to_uint8(&s,t+2)<0 || *(s++)!='.' ||
     strp_to_uint8(&s,t+3)<0) return -1;
  *ip = t[0]*0x1000000 + t[1]*0x10000 + t[2]*0x100 + t[3];
  *st = s;
  return 0;
}

extern int strp_to_ip4_addr_port(char const ** st, uint32 * ip, uint16 * port){
  char const * s;
  uint8 t[4];
  s  = *st;
  if(*s==':') {
    s++;
    if(strp_to_uint16(&s, port)<0) return -1;
    *st = s;
    return 2;
  }
  if(strp_to_uint8(&s,t)<0 || *(s++)!='.' ||
     strp_to_uint8(&s,t+1)<0 || *(s++)!='.' ||
     strp_to_uint8(&s,t+2)<0 || *(s++)!='.' ||
     strp_to_uint8(&s,t+3)<0) return -1;
  if(*s==':') {
    s++;
    if(strp_to_uint16(&s, port)<0) return -1;
    *ip = t[0]*0x1000000 + t[1]*0x10000 + t[2]*0x100 + t[3];
    *st = s;
    return 3;
  }
  *ip = t[0]*0x1000000 + t[1]*0x10000 + t[2]*0x100 + t[3];
  *st = s;
  return 1;
}

extern int strp_to_eth_addr(char const * * st, uint8 * mac) {
  char const * s;
  uint8 t[6];
  uint i;
  s  = *st;
  if(strp_base_to_uint8(&s,t,16)<0 || *(s++)!=':' ||
     strp_base_to_uint8(&s,t+1,16)<0 || *(s++)!=':' ||
     strp_base_to_uint8(&s,t+2,16)<0 || *(s++)!=':' ||
     strp_base_to_uint8(&s,t+3,16)<0 || *(s++)!=':' ||
     strp_base_to_uint8(&s,t+4,16)<0 || *(s++)!=':' ||
     strp_base_to_uint8(&s,t+5,16)<0) return -1;
  for(i=0;i<6;i++) mac[i]=t[i];
  *st = s;
  return 0;
}



