/*
 * This file is part of firk's Common C Library
 *  threads.c - threads
 *
 * (c) firk 2020; firk@cantconnect.ru
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. The origin of this software must not be misrepresented; you must
 *    not claim that you wrote the original software.
 * 4. Altered versions in any form must be plainly marked as such, and
 *    must not be misinterpreted as being the original software.
 *
 * This software is provided by the author and contributors `as is'
 * without any express or implied warranty.
 */

#include <stdio.h>
#include <stdlib.h>
#include <errno.h>
#include <assert.h>
#include <pthread.h>
#include <semaphore.h>
#include "types.h"
#include "malloc.h"
#define FCL_SIMPLE_NAMES
#include "threads.h"

struct fcl_thread {
  thread *      nextfree;
  int           created;
  pthread_t     pth;
  volatile int  state;
  sem_t         start;
  sem_t         finish;
  thread_func   tf;
  void *        tfp;
  int           tfr;
};


#define STATE_IDLE    0
#define STATE_INIT    1
#define STATE_RUNNING 2
#define STATE_STOPPED 3

static unsigned int maxn, usedn;
static thread * pool;
static thread * freelist;
static pthread_mutex_t lock;

extern void init_threads(unsigned int n) {
  if(!pool) {
    if(n<1) n = 1;
    maxn = n;
    pool = fcl_mallocz_a_check(sizeof(thread), n);
    if(pthread_mutex_init(&lock, NULL)<0) abort(); /* TODO: error handle */
  }
}
static void LOCK_BEGIN(void) { if(pthread_mutex_lock(&lock)<0) abort(); }
static void LOCK_END(void) { if(pthread_mutex_unlock(&lock)<0) abort(); }

static void * thread_manager(void * p);
static thread * alloc_thread(void) {
  thread * th;
  int err;
  
  assert(pool);
  LOCK_BEGIN();
  if(th=freelist) freelist = th->nextfree;
  else if(usedn<maxn) { th=pool+usedn; usedn++; }
  else { LOCK_END(); errno = EAGAIN; return NULL; }
  LOCK_END();
  assert(!th->state);
  if(!th->created) {
    if(sem_init(&th->start, 0, 0)<0) { err = errno; goto fail; }
    if(sem_init(&th->finish, 0, 0)<0) {
      err = errno; sem_destroy(&th->start); goto fail;
    }
    err = pthread_create(&th->pth, NULL/*attr*/, &thread_manager, (void*)th);
    if(err) { sem_destroy(&th->finish); sem_destroy(&th->start); goto fail; }
    th->created = 1;
  }
  return th;
fail:
  LOCK_BEGIN(); th->nextfree = freelist; freelist = th; LOCK_END();
  errno = err;
  return NULL;
}

static void free_thread(thread * th) {
  assert(th->created);
  assert(!th->state);
  LOCK_BEGIN();
  th->nextfree = freelist;
  freelist = th;
  LOCK_END();
}

/* initial state: state=STATE_IDLE, start=0, finish=0 */
static void * thread_manager(void * p) {
  thread * th;
  th = p;
  while(1) {
    while(sem_wait(&th->start)<0) if(errno!=EINTR) abort(); /* TODO */
    assert(th->state==STATE_INIT);
    th->state = STATE_RUNNING;
    th->tfr = th->tf(th->tfp);
    th->state = STATE_STOPPED;
    while(sem_post(&th->finish)<0) if(errno!=EINTR) abort(); /* TODO */
  }
}

extern thread * thread_create(thread_func tf, void * p) {
  thread * th;
  if(!(th = alloc_thread())) return NULL;
  th->tf = tf;
  th->tfp = p;
  th->state = STATE_INIT;
  while(sem_post(&th->start)<0) if(errno!=EINTR) abort(); /* TODO */
  return th;
}

extern int thread_is_finished(thread * th) {
  int v;
  assert(th->state!=STATE_IDLE);
  while(sem_getvalue(&th->finish, &v)<0) if(errno!=EINTR) abort(); /* TODO */
  return (v>0);
}

extern int thread_wait(thread * th) {
  int r;
  assert(th->state!=STATE_IDLE);
  while(sem_wait(&th->finish)<0) if(errno!=EINTR) abort(); /* TODO */
  assert(th->state==STATE_STOPPED);
  r = th->tfr;
  th->state = STATE_IDLE;
  free_thread(th);
  return r;
}

/* TODO: extern void thread_nowait(thread * th) { } */

