#include <stdio.h>
#include <string.h>


typedef int (*fn_opthandle)(char const * v);

typedef struct {
  char const * name;
  fn_opthandle handler;
  int allow_dup;
  char const * desc;
  int present;
} fcl_option;

static int opt_arch(char const * v);
static int opt_sse2(char const * v);
static int opt_SALSA20(char const * v);
static int opt_KUZ(char const * v);
static int opt_AES(char const * v);
static fcl_option opts[] = {
  { "arch", &opt_arch, 1, "={i386|i486|i586|i686|amd64|x86_64}\n" },
  { "sse2", &opt_sse2, 0, "={yes|no} Use SSE2 (default: yes)\n" },
  { "SALSA20", &opt_SALSA20, 0, "= force specified SALSA20 implementation\n" },
  { "KUZ", &opt_KUZ, 0, "= force specified Kuznyechik implementation\n" },
  { "AES", &opt_AES, 0, "= force specified AES implementation\n" },
  { NULL }
};

static void print_help(void) {
  fcl_option * o;
  char const * d;
  unsigned int j, f;
  fprintf(stderr, "Available configuration options:\n");
  for(o=opts; o->name; o++) {
    fprintf(stderr, "  --%s", o->name);
    if(!(d = o->desc)) { fputc('\n', stderr); continue; }
    if(*d=='=') f = 1; else { f = 0; fputc('\n', stderr); }
    while(*d) {
      for(j=0; d[j] && d[j]!='\n'; j++);
      if(f) f=0; else fputs("        ",stderr);
      fwrite(d, 1, j, stderr);
      fputc('\n',stderr);
      d+=j;
      if(*d) d++;
    }
  }
}

static char const * sp(char const * a, char const * p) {
  char c;
  while(c=*(p++)) {
    if(c!=*a) return NULL;
    a++;
  }
  return a;
}

static int have_arch, sse2 = 1;
static char const *SALSA20;
static char const *KUZ;
static char const *AES;
static int opt_SALSA20(char const * v) { SALSA20 = v; return 0; }
static int opt_KUZ(char const * v) { KUZ = v; return 0; }
static int opt_AES(char const * v) { AES = v; return 0; }

static int opt_arch(char const * v) {
  if(!strcmp(v,"i386") || !strcmp(v,"i486") ||
     !strcmp(v,"i586") || !strcmp(v,"i686")) {
    fprintf(stderr, "%s x86 32-bit arch\n", have_arch?"force":"have");
    have_arch = 32;
  } else if(!strcmp(v,"amd64") || !strcmp(v,"x86_64")) {
    fprintf(stderr, "%s x86 64-bit arch\n", have_arch?"force":"have");
    have_arch = 64;
  } else {
    fprintf(stderr, "%s non-x86 unknown arch\n", have_arch?"force":"have");
    have_arch = -1;
  }
  return 0;
}

static int opt_sse2(char const * v) {
  if(!strcmp(v,"yes")) sse2 = 1;
  else if(!strcmp(v,"no")) sse2 = 0;
  else return -2;
  return 0;
}


static void print_all(void) {
  if(!SALSA20) {
    if(have_arch==32 && sse2) SALSA20 = "asm_sse2_a";
    else if(have_arch==32 && !sse2) SALSA20 = "asm_ia";
    else SALSA20 = "";
  }
  if(!KUZ) {
    if(have_arch==32 && sse2) KUZ = "v3x2";
    else if(have_arch==64) KUZ = "v3x2_64";
    else KUZ = "";
  }
  if(!AES) {
    if(have_arch==32 && sse2) AES = "v2";
    else AES="";
  }
  printf("ARCH=%d\n"
         "SALSA20=\"%s\"\n"
         "KUZ=\"%s\"\n"
         "AES=\"%s\"\n",
         have_arch, SALSA20, KUZ, AES);
}

int main(int argc, char * * argv) {
  int j, r;
  char *a, *v, *vv, c;
  fcl_option * o;
  
  for(j=1; j<argc; j++) {
    a = argv[j];
    if(!strcmp(a,"--help")) {
      print_help();
      return -1;
    }
    if(a[0]!='-' || a[1]!='-') {
      fprintf(stderr, "[opts-config] bad argument: %s\n", a);
      return -1;
    }
    a += 2;
    if(v = strchr(a,'=')) *(v++) = 0;
    for(o=opts; o->name; o++) if(!strcmp(a, o->name)) break;
    if(!o->name) {
      fprintf(stderr, "[opts-config] unknown option: --%s\n", a);
      return -1;
    }
    if(!o->allow_dup && o->present) {
      fprintf(stderr, "[opts-config] duplicate option: --%s\n", a);
      return -1;
    }
    if(o->desc && *o->desc=='=' && !v) {
      fprintf(stderr, "[opts-config] option needs parameter: --%s\n", a);
      return -1;
    }
    if((!o->desc || *o->desc!='=') && v) {
      fprintf(stderr, "[opts-config] option doesn't need parameter: --%s\n", a);
      return -1;
    }
    if(v) for(vv=v; c=*vv; vv++) {
      if(!(c>='a' && c<='z' || c>='A'&& c<='Z' || c>='0' && c<='9'
        || c=='_' || c=='-' || c=='.')) {
        fprintf(stderr, "[opts-config] option parameter bad: --%s\n", a);
        return -1;
      }
    }
    o->present = 1;
    r = o->handler(v);
    if(r<0) {
      fprintf(stderr, "Error processing option: --%s\n", a);
      return -1;
    }
  }
  print_all();
  return 0;
}

