/*
 * This file is part of firk's Cryptography Library
 * (c) firk 2015-2022; firk@cantconnect.ru
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. The origin of this software must not be misrepresented; you must
 *    not claim that you wrote the original software.
 * 4. Altered versions in any form must be plainly marked as such, and
 *    must not be misinterpreted as being the original software.
 *
 * This software is provided by the author and contributors `as is'
 * without any express or implied warranty.
 */

#include <stdarg.h>
#include <stdlib.h>
#include <errno.h>
#include "../fcl/string.h"
#include <sys/types.h>
#include <unistd.h>
#include <fcntl.h>
#include "../fcl/types.h"
#include "rc4.h"
#include "sha1.h"
#include "sha256.h"
#include "sha512.h"
#include "aes/aes.h"
#include "aes/aes256.h"
#include "md5.h"
#include "hmac.h"
#include "kuznyechik/kuznyechik.h"
#include "salsa20/salsa20.h"
#include <stdio.h>

static void err_exit(char const * fmt, ...) {
  int errn;
  va_list arg;
  errn = errno;
  fprintf(stderr, "fcrypt: ");
  va_start(arg, fmt); vfprintf(stderr, fmt, arg); va_end(arg);
  fprintf(stderr, " error %d (%s)\n", errn, strerror(errn));
  exit(-1);
}

static void err2_exit(char const * fmt, ...) {
  int errn;
  va_list arg;
  errn = errno;
  fprintf(stderr, "fcrypt: ");
  va_start(arg, fmt); vfprintf(stderr, fmt, arg); va_end(arg);
  fprintf(stderr, "\n");
  exit(-1);
}

static void usage(void) {
  fprintf(stderr,
      "Usage: fcrypt [-b] <action> [<keyfile>] <filename>\n"
      "  -b means binary output (default is hex-text)\n"
      "  <action> is:\n"
      "    md5, sha1, sha256, sha512 - calculate hash of one file\n"
      "    hmac-md5, hmac-sha* - calculate HMAC with specified key\n"
      "    aes_enc, kuz_enc - encrypt second file with specified key\n"
      "    aes_dec, kuz_dec - decrypt second file with specified key\n");
  exit(-1);
}

static int binmode = 0;
static sha1_ctx  ctx1;
static sha256_ctx ctx256;
static sha512_ctx ctx512;
static md5_ctx   ctx5;
static hmac_sha1_key hkey1;
static hmac_sha256_key hkey256;
static hmac_sha512_key hkey512;
static hmac_md5_key hkey5;
static aes_key   aeskey;
static aes256_key aes256key;
static kuz_key   kuzkey;
static uint32    salsakey[8];
static uchar     hash[64]; /* max size of sha1, sha256, sha512 */
static uint      hashlen;
static uchar     buf[4096];
static uchar     buf2[8193];
static uchar h[17]="0123456789ABCDEF";

static void print_raw(uchar * b, uint sz) {
  if(fwrite(b, 1, sz, stdout)!=sz) err_exit("fwrite(stdout)");
}

static void print_bytes(uchar * b, uint sz) {
  uint j;
  uchar c;
  if(sz>4096) err2_exit("BUG");
  if(!binmode) {
    for(j=0;j<sz;j++) { c=b[j]; buf2[j*2]=h[c/16]; buf2[j*2+1]=h[c%16]; }
    buf2[j*2] = '\n';
    print_raw(buf2, j*2+1);
  } else {
    print_raw(b, sz);
  }
}

static void do_sha(int ac, char const * fn, char const * keyfn) {
  int   fd, ispipe;
  off_t fpos, fsz;
  size_t csz;
  ssize_t len;
  uchar keydata[1024];

  if(keyfn) {
    if((fd = open(keyfn, O_RDONLY))<0) err_exit("open(%s)", keyfn);
    fsz = lseek(fd, 0, SEEK_END);
    if(fsz<0) err_exit("lseek(%s, SEEK_END)", keyfn);
    lseek(fd, 0, SEEK_SET);
    if(fsz>(int)sizeof(keydata)) {
      err2_exit("HMAC key file too long (max 1024 bytes) (%s)", keyfn);
    }
    if(read(fd, keydata, fsz)!=fsz) {
      explicit_bzero(keydata, sizeof(keydata));
      err2_exit("HMAC key read error");
    }
    close(fd);
    switch(ac) {
    case 8:
      hmac_md5_initkey(&hkey5, keydata, fsz);
      hmac_md5_init(&ctx5, &hkey5); hashlen=16; break;
    case 1:
      hmac_sha1_initkey(&hkey1, keydata, fsz);
      hmac_sha1_init(&ctx1, &hkey1); hashlen=20; break;
    case 2:
      hmac_sha256_initkey(&hkey256, keydata, fsz);
      hmac_sha256_init(&ctx256, &hkey256); hashlen=32; break;
    case 3:
      hmac_sha512_initkey(&hkey512, keydata, fsz);
      hmac_sha512_init(&ctx512, &hkey512); hashlen=64; break;
    }
  } else {
    switch(ac) {
    case 8: md5_init(&ctx5); hashlen=16; break;
    case 1: sha1_init(&ctx1); hashlen=20; break;
    case 2: sha256_init(&ctx256); hashlen=32; break;
    case 3: sha512_init(&ctx512); hashlen=64; break;
    }
  }
  if((fd = open(fn, O_RDONLY))<0) err_exit("open(%s)", fn);
  ispipe = 0;
  if((fsz = lseek(fd, 0, SEEK_END))<0) {
    if(errno=ESPIPE) ispipe=1;
    else err_exit("lseek(%s, SEEK_END)", fn);
  }
  if(!ispipe) lseek(fd, 0, SEEK_SET);
  fpos = 0;
  for(fpos=0; ispipe || fpos<fsz; fpos+=len) {
    csz = sizeof(buf);
    if(!ispipe && (ssize_t)csz>fsz-fpos) csz=fsz-fpos;
    len = read(fd, buf, csz);
    if(len<0) err_exit("read(%s)", fn);
    if(!len) { if(ispipe) break; err2_exit("read(%s) unexpeced EOF", fn); }
    switch(ac) {
    case 8: md5_update(&ctx5, buf, (size_t)len); break;
    case 1: sha1_update(&ctx1, buf, (size_t)len); break;
    case 2: sha256_update(&ctx256, buf, (size_t)len); break;
    case 3: sha512_update(&ctx512, buf, (size_t)len); break;
    }
  }
  close(fd);
  if(keyfn) {
    switch(ac) {
    case 8: hmac_md5_final(&ctx5, &hkey5, hash); break;
    case 1: hmac_sha1_final(&ctx1, &hkey1, hash); break;
    case 2: hmac_sha256_final(&ctx256, &hkey256, hash); break;
    case 3: hmac_sha512_final(&ctx512, &hkey512, hash); break;
    }
    explicit_bzero(keydata, sizeof(keydata));
    explicit_bzero(&hkey5, sizeof(hkey5));
    explicit_bzero(&hkey1, sizeof(hkey1));
    explicit_bzero(&hkey256, sizeof(hkey256));
    explicit_bzero(&hkey512, sizeof(hkey512));
  } else {
    switch(ac) {
    case 8: md5_final(&ctx5, hash); break;
    case 1: sha1_final(&ctx1, hash); break;
    case 2: sha256_final(&ctx256, hash); break;
    case 3: sha512_final(&ctx512, hash); break;
    }
  }
}

static void blk_action(int ac, uchar * b) {
  uint32 d[4], i;
  switch(ac) {
  case 9: kuz_encrypt(&kuzkey, b); goto fin;
  case 10: kuz_decrypt(&kuzkey, b); goto fin;
  }
  for(i=0; i<4; i++) {
    d[i] = (((uint32)(b[i*4])) ) |
           (((uint32)(b[i*4+1]))<<8) |
           (((uint32)(b[i*4+2]))<<16) |
           (((uint32)(b[i*4+3]))<<24);
  }
  switch(ac) {
  case 4: aes_encrypt(&aeskey, d); break;
  case 5: aes_decrypt(&aeskey, d); break;
  case 6: aes256_encrypt(&aes256key, d); break;
  case 7: aes256_decrypt(&aes256key, d); break;
  }
  for(i=0; i<4; i++) {
    b[i*4] = (uchar)(d[i]);
    b[i*4+1] = (uchar)(d[i]>>8);
    b[i*4+2] = (uchar)(d[i]>>16);
    b[i*4+3] = (uchar)(d[i]>>24);
  }
  explicit_bzero(d, sizeof(d));
fin:
  ;
}

static void blk2_action(int ac, uchar * b, uint32 s) {
  uint32 tmp[16], nonce[2], seq[2];
  uint32 i;
  nonce[0] = nonce[1] = 0;
  seq[0] = s; seq[1] = 0;
  if(ac==11) salsa20_genrandom512(tmp, salsakey, nonce, seq);
  else chacha20_genrandom512(tmp, salsakey, nonce, seq);
  for(i=0; i<16; i++,b+=4) *(uint32*)b ^= tmp[i];
  explicit_bzero(tmp, sizeof(tmp));
}

static void do_blk(int ac, char const * keyfn, char const * fn) {
  int   fd, ispipe;
  off_t fpos, fsz;
  uchar keysz, ki;
  uchar key1[32];
  uint32 key2[8];
  int res;
  size_t csz;
  ssize_t len;
  size_t blen, bpos;
  uint32 count;

  if((fd = open(keyfn, O_RDONLY))<0) err_exit("open(%s)", keyfn);
  fsz = lseek(fd, 0, SEEK_END);
  if(fsz<0) err_exit("lseek(%s, SEEK_END)", keyfn);
  lseek(fd, 0, SEEK_SET);
  switch(ac) {
  case 11: case 12:
    if(fsz!=32) err2_exit("bad Salsa20 keyfile (%s)", keyfn);
    break;
  case 9: case 10:
    if(fsz!=32) err2_exit("bad Kuznyechik keyfile (%s)", keyfn);
    break;
  case 6: case 7:
    if(fsz!=32) err2_exit("bad AES256 keyfile (%s)", keyfn);
    break;
  case 4: case 5:
    if(fsz!=16 && fsz!=24 && fsz!=32) err2_exit("bad AES keyfile (%s)", keyfn);
    break;
  default:
    err2_exit("bad blk mode");
  }
  keysz = (uchar)fsz/4;
  if(read(fd, key1, keysz*4)!=keysz*4) {
    explicit_bzero(key1, sizeof(key1));
    err2_exit("key read error");
  }
  close(fd);
  if(ac>=11) {
    for(ki=0; ki<8; ki++) {
      salsakey[ki] = (((uint32)(key1[ki*4])) ) |
                     (((uint32)(key1[ki*4+1]))<<8) |
                     (((uint32)(key1[ki*4+2]))<<16) |
                     (((uint32)(key1[ki*4+3]))<<24);
    }
    explicit_bzero(key1, sizeof(key1));
  } else if(ac>=9) {
    kuz_prepare_key(&kuzkey, key1);
    explicit_bzero(key1, sizeof(key1));
  } else {
    for(ki=0; ki<keysz; ki++) {
      key2[ki] = (((uint32)(key1[ki*4])) ) |
                 (((uint32)(key1[ki*4+1]))<<8) |
                 (((uint32)(key1[ki*4+2]))<<16) |
                 (((uint32)(key1[ki*4+3]))<<24);
    }
    if(ac<6) res = aes_prepare_key(&aeskey, keysz, 4, key2);
    else res = aes256_prepare_key(&aes256key, key2);
    explicit_bzero(key1, sizeof(key1));
    explicit_bzero(key2, sizeof(key2));
    if(res<0) err2_exit("bad AES key (BUG)");
  }
  /* now process data */
  if((fd = open(fn, O_RDONLY))<0) err_exit("open(%s)", fn);
  ispipe = 0;
  if((fsz = lseek(fd, 0, SEEK_END))<0) {
    if(errno=ESPIPE) ispipe=1;
    else {
      explicit_bzero(&aeskey, sizeof(aeskey));
      explicit_bzero(&aes256key, sizeof(aes256key));
      explicit_bzero(&kuzkey, sizeof(kuzkey));
      explicit_bzero(&salsakey, sizeof(salsakey));
      err_exit("lseek(%s, SEEK_END)", fn);
    }
  }
  if(!ispipe) lseek(fd, 0, SEEK_SET);
  fpos = 0;
  blen = 0;
  count = 0;
  for(fpos=0; ispipe || fpos<fsz; fpos+=len) {
    csz = sizeof(buf)-blen;
    if(!ispipe && (ssize_t)csz>fsz-fpos) csz=fsz-fpos;
    len = read(fd, buf+blen, csz);
    if(!len && ispipe) break;
    if(len<=0) {
      explicit_bzero(&aeskey, sizeof(aeskey));
      explicit_bzero(&aes256key, sizeof(aes256key));
      explicit_bzero(&kuzkey, sizeof(kuzkey));
      explicit_bzero(&salsakey, sizeof(salsakey));
      explicit_bzero(buf, sizeof(buf));
      if(!len) err2_exit("read(%s) EOF", fn);
      err_exit("read(%s)", fn);
    }
    blen += len;
    if(ac>=11) {
      for(bpos=0; bpos+64<=blen; bpos+=64) blk2_action(ac, buf+bpos, ++count);
    } else {
      for(bpos=0; bpos+16<=blen; bpos+=16) blk_action(ac, buf+bpos);
    }
    if(bpos) {
      print_bytes(buf, bpos);
      explicit_bzero(buf, bpos);
      blen -= bpos;
      if(blen) bcopy(buf+bpos, buf, blen);
    }
  }
  close(fd);
  if(blen) {
    if(ac>=11) {
      bzero(buf+blen, 16-blen);
      blk_action(ac, buf);
      print_bytes(buf, 16);
      explicit_bzero(buf, 16);
    } else {
      bzero(buf+blen, 64-blen);
      blk2_action(ac, buf, ++count);
      print_bytes(buf, 64);
      explicit_bzero(buf, 64);
    }
  }
  explicit_bzero(&aeskey, sizeof(aeskey));
  explicit_bzero(&aes256key, sizeof(aes256key));
  explicit_bzero(&kuzkey, sizeof(kuzkey));
  explicit_bzero(&salsakey, sizeof(salsakey));
  explicit_bzero(buf, sizeof(buf));
}

int main(int argc, char const * const * argv) {
  int action;
  char const * actionstr;
  if(argc<3) usage();
  argc--; argv++;
  if(!strcmp(argv[0], "-b")) { binmode=1; argc--; argv++; }
  actionstr = argv[0]; argc--; argv++;
  if(!strcmp(actionstr, "sha1")) action = 1;
  else if(!strcmp(actionstr, "sha256")) action = 2;
  else if(!strcmp(actionstr, "sha512")) action = 3;
  else if(!strcmp(actionstr, "aes_enc")) action = 4;
  else if(!strcmp(actionstr, "aes_dec")) action = 5;
  else if(!strcmp(actionstr, "aes256_enc")) action = 6;
  else if(!strcmp(actionstr, "aes256_dec")) action = 7;
  else if(!strcmp(actionstr, "md5")) action = 8;
  else if(!strcmp(actionstr, "kuz_enc")) action = 9;
  else if(!strcmp(actionstr, "kuz_dec")) action = 10;
  else if(!strcmp(actionstr, "salsa20")) action = 11;
  else if(!strcmp(actionstr, "chacha20")) action = 12;
  else if(!strcmp(actionstr, "hmac-sha1")) action = 101;
  else if(!strcmp(actionstr, "hmac-sha256")) action = 102;
  else if(!strcmp(actionstr, "hmac-sha512")) action = 103;
  else if(!strcmp(actionstr, "hmac-md5")) action = 108;
  else usage();
  switch(action) {
  case 1: case 2: case 3: case 8:
    if(argc!=1) usage();
    do_sha(action, argv[0], NULL);
    print_bytes(hash, hashlen);
    break;
  case 101: case 102: case 103: case 108:
    if(argc!=2) usage();
    do_sha(action-100, argv[1], argv[0]);
    print_bytes(hash, hashlen);
    break;
  case 4: case 5: case 6: case 7: case 9: case 10: case 11: case 12:
    if(argc!=2) usage();
    do_blk(action, argv[0], argv[1]);
    break;
  default: usage();
  }
  return 0;
}
