/*
 * This file is part of firk's Cryptography Library
 * hmac.c - rfc2104 HMAC implementation
 * (c) firk 2020; firk@cantconnect.ru
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. The origin of this software must not be misrepresented; you must
 *    not claim that you wrote the original software.
 * 4. Altered versions in any form must be plainly marked as such, and
 *    must not be misinterpreted as being the original software.
 *
 * This software is provided by the author and contributors `as is'
 * without any express or implied warranty.
 */

#include <sys/types.h>
#include "../fcl/types.h"
#include "../fcl/string.h"
#include "md5.h"
#include "sha1.h"
#include "sha256.h"
#ifdef HAVE_INT64
#include "sha512.h"
#endif
#include "hmac.h"

#define PREPARE_KEY(hash_func, hash_len, pad, key, keylen) \
  if(keylen<=sizeof(pad)) { if(keylen) bcopy(key, pad, keylen); } \
  else { hash_func(pad, (char const*)key, keylen); keylen = hash_len; } \
  if(keylen<sizeof(pad)) bzero(pad+keylen, sizeof(pad)-keylen);

#define md5_len 16
#define sha1_len 20
#define sha256_len 32
#define sha512_len 64
#define md5_b 64
#define sha1_b 64
#define sha256_b 64
#define sha512_b 128

extern void hmac_md5_initkey(hmac_md5_key * k,
                              uint8 const * key, size_t keylen) {
  md5_ctx ctx;
  size_t n;
  uint8 pad[md5_b];
  PREPARE_KEY(md5_hash, md5_len, pad, key, keylen)
  for(n=0; n<sizeof(pad); n++) pad[n] ^= 0x36;
  md5_init(&ctx); md5_update(&ctx, pad, sizeof(pad));
  bcopy(ctx.state, k->istate, sizeof(ctx.state));
  for(n=0; n<sizeof(pad); n++) pad[n] ^= 0x6A; /* 0x36 ^ 0x5C */
  md5_init(&ctx); md5_update(&ctx, pad, sizeof(pad));
  bcopy(ctx.state, k->ostate, sizeof(ctx.state));
}

extern void hmac_md5_init(md5_ctx * ctx, hmac_md5_key const * k) {
  bcopy(k->istate, ctx->state, sizeof(ctx->state));
  ctx->len = md5_b;
}

extern void hmac_md5_final(md5_ctx * ctx, hmac_md5_key const * k,
                            uint8 * hash) {
  md5_final(ctx, hash);
  bcopy(k->ostate, ctx->state, sizeof(ctx->state));
  ctx->len = md5_b;
  md5_update(ctx, hash, md5_len);
  md5_final(ctx, hash);
}

extern void hmac_md5_hash(uint8 * hash, hmac_md5_key const * k,
                           char const * data, size_t len) {
  md5_ctx ctx;

  bcopy(k->istate, ctx.state, sizeof(ctx.state));
  ctx.len = md5_b;
  md5_update(&ctx, (uint8 const*)data, len);
  md5_final(&ctx, hash);
  bcopy(k->ostate, ctx.state, sizeof(ctx.state));
  ctx.len = md5_b;
  md5_update(&ctx, hash, md5_len);
  md5_final(&ctx, hash);
}


extern void hmac_sha1_initkey(hmac_sha1_key * k,
                              uint8 const * key, size_t keylen) {
  sha1_ctx ctx;
  size_t n;
  uint8 pad[sha1_b];
  PREPARE_KEY(sha1_hash, sha1_len, pad, key, keylen)
  for(n=0; n<sizeof(pad); n++) pad[n] ^= 0x36;
  sha1_init(&ctx); sha1_update(&ctx, pad, sizeof(pad));
  bcopy(ctx.state, k->istate, sizeof(ctx.state));
  for(n=0; n<sizeof(pad); n++) pad[n] ^= 0x6A; /* 0x36 ^ 0x5C */
  sha1_init(&ctx); sha1_update(&ctx, pad, sizeof(pad));
  bcopy(ctx.state, k->ostate, sizeof(ctx.state));
}

extern void hmac_sha1_init(sha1_ctx * ctx, hmac_sha1_key const * k) {
  bcopy(k->istate, ctx->state, sizeof(ctx->state));
  ctx->len = sha1_b;
}

extern void hmac_sha1_final(sha1_ctx * ctx, hmac_sha1_key const * k,
                            uint8 * hash) {
  sha1_final(ctx, hash);
  bcopy(k->ostate, ctx->state, sizeof(ctx->state));
  ctx->len = sha1_b;
  sha1_update(ctx, hash, sha1_len);
  sha1_final(ctx, hash);
}

extern void hmac_sha1_hash(uint8 * hash, hmac_sha1_key const * k,
                           char const * data, size_t len) {
  sha1_ctx ctx;

  bcopy(k->istate, ctx.state, sizeof(ctx.state));
  ctx.len = sha1_b;
  sha1_update(&ctx, (uint8 const*)data, len);
  sha1_final(&ctx, hash);
  bcopy(k->ostate, ctx.state, sizeof(ctx.state));
  ctx.len = sha1_b;
  sha1_update(&ctx, hash, sha1_len);
  sha1_final(&ctx, hash);
}


extern void hmac_sha256_initkey(hmac_sha256_key * k,
                              uint8 const * key, size_t keylen) {
  sha256_ctx ctx;
  size_t n;
  uint8 pad[sha256_b];
  PREPARE_KEY(sha256_hash, sha256_len, pad, key, keylen)
  for(n=0; n<sizeof(pad); n++) pad[n] ^= 0x36;
  sha256_init(&ctx); sha256_update(&ctx, pad, sizeof(pad));
  bcopy(ctx.state, k->istate, sizeof(ctx.state));
  for(n=0; n<sizeof(pad); n++) pad[n] ^= 0x6A; /* 0x36 ^ 0x5C */
  sha256_init(&ctx); sha256_update(&ctx, pad, sizeof(pad));
  bcopy(ctx.state, k->ostate, sizeof(ctx.state));
}

extern void hmac_sha256_init(sha256_ctx * ctx, hmac_sha256_key const * k) {
  bcopy(k->istate, ctx->state, sizeof(ctx->state));
  ctx->len = sha256_b;
}

extern void hmac_sha256_final(sha256_ctx * ctx, hmac_sha256_key const * k,
                            uint8 * hash) {
  sha256_final(ctx, hash);
  bcopy(k->ostate, ctx->state, sizeof(ctx->state));
  ctx->len = sha256_b;
  sha256_update(ctx, hash, sha256_len);
  sha256_final(ctx, hash);
}

extern void hmac_sha256_hash(uint8 * hash, hmac_sha256_key const * k,
                           char const * data, size_t len) {
  sha256_ctx ctx;

  bcopy(k->istate, ctx.state, sizeof(ctx.state));
  ctx.len = sha256_b;
  sha256_update(&ctx, (uint8 const*)data, len);
  sha256_final(&ctx, hash);
  bcopy(k->ostate, ctx.state, sizeof(ctx.state));
  ctx.len = sha256_b;
  sha256_update(&ctx, hash, sha256_len);
  sha256_final(&ctx, hash);
}


#ifdef HAVE_INT64
extern void hmac_sha512_initkey(hmac_sha512_key * k,
                              uint8 const * key, size_t keylen) {
  sha512_ctx ctx;
  size_t n;
  uint8 pad[sha512_b];
  PREPARE_KEY(sha512_hash, sha512_len, pad, key, keylen)
  for(n=0; n<sizeof(pad); n++) pad[n] ^= 0x36;
  sha512_init(&ctx); sha512_update(&ctx, pad, sizeof(pad));
  bcopy(ctx.state, k->istate, sizeof(ctx.state));
  for(n=0; n<sizeof(pad); n++) pad[n] ^= 0x6A; /* 0x36 ^ 0x5C */
  sha512_init(&ctx); sha512_update(&ctx, pad, sizeof(pad));
  bcopy(ctx.state, k->ostate, sizeof(ctx.state));
}

extern void hmac_sha512_init(sha512_ctx * ctx, hmac_sha512_key const * k) {
  bcopy(k->istate, ctx->state, sizeof(ctx->state));
  ctx->len = sha512_b; ctx->lenh = 0;
}

extern void hmac_sha512_final(sha512_ctx * ctx, hmac_sha512_key const * k,
                            uint8 * hash) {
  sha512_final(ctx, hash);
  bcopy(k->ostate, ctx->state, sizeof(ctx->state));
  ctx->len = sha512_b; ctx->lenh = 0;
  sha512_update(ctx, hash, sha512_len);
  sha512_final(ctx, hash);
}

extern void hmac_sha512_hash(uint8 * hash, hmac_sha512_key const * k,
                           char const * data, size_t len) {
  sha512_ctx ctx;

  bcopy(k->istate, ctx.state, sizeof(ctx.state));
  ctx.len = sha512_b; ctx.lenh = 0;
  sha512_update(&ctx, (uint8 const*)data, len);
  sha512_final(&ctx, hash);
  bcopy(k->ostate, ctx.state, sizeof(ctx.state));
  ctx.len = sha512_b; ctx.lenh = 0;
  sha512_update(&ctx, hash, sha512_len);
  sha512_final(&ctx, hash);
}
#endif


