/*
 * This file is part of firk's Cryptography Library
 * (c) firk 2020; firk@cantconnect.ru
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. The origin of this software must not be misrepresented; you must
 *    not claim that you wrote the original software.
 * 4. Altered versions in any form must be plainly marked as such, and
 *    must not be misinterpreted as being the original software.
 *
 * This software is provided by the author and contributors `as is'
 * without any express or implied warranty.
 */

#include <sys/types.h>
#include "../fcl/types.h"
#include "md5.h"

extern void md5_init(md5_ctx * ctx) {
    ctx->state[0] = 0x67452301;
    ctx->state[1] = 0xEFCDAB89;
    ctx->state[2] = 0x98BADCFE;
    ctx->state[3] = 0x10325476;
    ctx->len = 0;
}

static void uint32_store(uint8 * dest, uint32 src) {
    dest[0] = (uint8)(src);
    dest[1] = (uint8)(src>>8);
    dest[2] = (uint8)(src>>16);
    dest[3] = (uint8)(src>>24);
}

static uint32 ROR32(uint32 v, uint32 s) {
    /* we don't need to handle something outside 1<=s<=31
       here because all shift factors are hardcoded */
    return (v>>s)|(v<<(32-s));
}

extern void md5_transform(uint8 const * data,  uint32 * state) {
    uint32  buf[16], a, b, c, d;
    int     i;
    for(i=0; i<16; i++) {
        buf[i] = (((uint32)(data[i*4+0]))    ) +
                 (((uint32)(data[i*4+1]))<< 8) +
                 (((uint32)(data[i*4+2]))<<16) +
                 (((uint32)(data[i*4+3]))<<24);
    }
    a = state[0];    b = state[1];    c = state[2];    d = state[3];

#define f1(x,y,z) ( ((x)&(y)) | ((~(x))&(z)) )
#define f2(x,y,z) ( ((z)&(x)) | ((~(z))&(y)) )
#define f3(x,y,z) ( (x) ^ (y) ^ (z) )
#define f4(x,y,z) ( (y) ^ ((~(z))|(x)) )
#define F(f,a,b,c,d,i,s,C) a = b + ROR32(a + f(b,c,d) + buf[i] + C, s)
    F(f1,a,b,c,d, 0,25, 0xD76AA478);    F(f1,d,a,b,c, 1,20, 0xE8C7B756);
    F(f1,c,d,a,b, 2,15, 0x242070DB);    F(f1,b,c,d,a, 3,10, 0xC1BDCEEE);
    F(f1,a,b,c,d, 4,25, 0xF57C0FAF);    F(f1,d,a,b,c, 5,20, 0x4787C62A);
    F(f1,c,d,a,b, 6,15, 0xA8304613);    F(f1,b,c,d,a, 7,10, 0xFD469501);
    F(f1,a,b,c,d, 8,25, 0x698098D8);    F(f1,d,a,b,c, 9,20, 0x8B44F7AF);
    F(f1,c,d,a,b,10,15, 0xFFFF5BB1);    F(f1,b,c,d,a,11,10, 0x895CD7BE);
    F(f1,a,b,c,d,12,25, 0x6B901122);    F(f1,d,a,b,c,13,20, 0xFD987193);
    F(f1,c,d,a,b,14,15, 0xA679438E);    F(f1,b,c,d,a,15,10, 0x49B40821);

    F(f2,a,b,c,d, 1,27, 0xF61E2562);    F(f2,d,a,b,c, 6,23, 0xC040B340);
    F(f2,c,d,a,b,11,18, 0x265E5A51);    F(f2,b,c,d,a, 0,12, 0xE9B6C7AA);
    F(f2,a,b,c,d, 5,27, 0xD62F105D);    F(f2,d,a,b,c,10,23, 0x02441453);
    F(f2,c,d,a,b,15,18, 0xD8A1E681);    F(f2,b,c,d,a, 4,12, 0xE7D3FBC8);
    F(f2,a,b,c,d, 9,27, 0x21E1CDE6);    F(f2,d,a,b,c,14,23, 0xC33707D6);
    F(f2,c,d,a,b, 3,18, 0xF4D50D87);    F(f2,b,c,d,a, 8,12, 0x455A14ED);
    F(f2,a,b,c,d,13,27, 0xA9E3E905);    F(f2,d,a,b,c, 2,23, 0xFCEFA3F8);
    F(f2,c,d,a,b, 7,18, 0x676F02D9);    F(f2,b,c,d,a,12,12, 0x8D2A4C8A);

    F(f3,a,b,c,d, 5,28, 0xFFFA3942);    F(f3,d,a,b,c, 8,21, 0x8771F681);
    F(f3,c,d,a,b,11,16, 0x6D9D6122);    F(f3,b,c,d,a,14, 9, 0xFDE5380C);
    F(f3,a,b,c,d, 1,28, 0xA4BEEA44);    F(f3,d,a,b,c, 4,21, 0x4BDECFA9);
    F(f3,c,d,a,b, 7,16, 0xF6BB4B60);    F(f3,b,c,d,a,10, 9, 0xBEBFBC70);
    F(f3,a,b,c,d,13,28, 0x289B7EC6);    F(f3,d,a,b,c, 0,21, 0xEAA127FA);
    F(f3,c,d,a,b, 3,16, 0xD4EF3085);    F(f3,b,c,d,a, 6, 9, 0x04881D05);
    F(f3,a,b,c,d, 9,28, 0xD9D4D039);    F(f3,d,a,b,c,12,21, 0xE6DB99E5);
    F(f3,c,d,a,b,15,16, 0x1FA27CF8);    F(f3,b,c,d,a, 2, 9, 0xC4AC5665);

    F(f4,a,b,c,d, 0,26, 0xF4292244);    F(f4,d,a,b,c, 7,22, 0x432AFF97);
    F(f4,c,d,a,b,14,17, 0xAB9423A7);    F(f4,b,c,d,a, 5,11, 0xFC93A039);
    F(f4,a,b,c,d,12,26, 0x655B59C3);    F(f4,d,a,b,c, 3,22, 0x8F0CCC92);
    F(f4,c,d,a,b,10,17, 0xFFEFF47D);    F(f4,b,c,d,a, 1,11, 0x85845DD1);
    F(f4,a,b,c,d, 8,26, 0x6FA87E4F);    F(f4,d,a,b,c,15,22, 0xFE2CE6E0);
    F(f4,c,d,a,b, 6,17, 0xA3014314);    F(f4,b,c,d,a,13,11, 0x4E0811A1);
    F(f4,a,b,c,d, 4,26, 0xF7537E82);    F(f4,d,a,b,c,11,22, 0xBD3AF235);
    F(f4,c,d,a,b, 2,17, 0x2AD7D2BB);    F(f4,b,c,d,a, 9,11, 0xEB86D391);

    state[0] += a;    state[1] += b;    state[2] += c;    state[3] += d;
#undef F
#undef f1
#undef f2
#undef f3
#undef f4
}

extern void md5_update(md5_ctx * ctx, uint8 const * data, size_t len) {
    uint8 * cdata;
    uint32  a;

    cdata = ctx->data;
    a = (uint32)(ctx->len & 63);
     /* we can overflow 64bit byte-counter here;
        even easier is to overflow 64bit bitcounter (len*8) and
          formally it will be too-long-to-be-hashed sequence, but it will
          still produce some hash and we don't care about these overflows */
    ctx->len += (uint64)len;
    if(a) {
        while(a<64 && len>0) { cdata[a++] = *(data++); len--; }
        if(a<64) return;
        md5_transform(cdata,ctx->state);
        a = 0;
    }
    while(len>=64) {
        md5_transform(data,ctx->state);
        data += 64;
        len -= 64;
    }
    while(a<len) { cdata[a]=data[a]; a++; }
}

extern void md5_final(md5_ctx * ctx, uint8 * hash) {
    uint8 * cdata;
    uint    a;
    uint64 len;

    cdata = ctx->data;
    len = ctx->len;
    a = (uint)(len & 63);

    cdata[a++] = 0x80;
    if(a>56) {
        while(a<64) cdata[a++] = 0;
        md5_transform(cdata,ctx->state);
        a = 0;
    }
    while(a<56) cdata[a++] = 0;
    len<<=3;
    cdata[56] = (uint8)len; len>>=8;
    cdata[57] = (uint8)len; len>>=8;
    cdata[58] = (uint8)len; len>>=8;
    cdata[59] = (uint8)len; len>>=8;
    cdata[60] = (uint8)len; len>>=8;
    cdata[61] = (uint8)len; len>>=8;
    cdata[62] = (uint8)len; len>>=8;
    cdata[63] = (uint8)len;
    md5_transform(cdata,ctx->state);
    for(a=0;a<4;a++) uint32_store(hash+a*4, ctx->state[a]);
}

extern void md5_hash(uint8 * hash, char const * data, size_t len) {
    md5_ctx ctx;
    md5_init(&ctx);
    md5_update(&ctx, (uint8 const*)data, len);
    md5_final(&ctx, hash);
}


