/*
 * This file is part of firk's Cryptography Library
 * salsa20.c - Salsa20 and ChaCha20 stream cipher implementation
 * (c) firk 2015; firk@cantconnect.ru
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. The origin of this software must not be misrepresented; you must
 *    not claim that you wrote the original software.
 * 4. Altered versions in any form must be plainly marked as such, and
 *    must not be misinterpreted as being the original software.
 *
 * This software is provided by the author and contributors `as is'
 * without any express or implied warranty.
 */

#include <sys/types.h>
#include "../fcl/types.h"
#include "salsa20.h"

static uint32 ROL32(uint32 x, uchar s) {
  return ((x<<s) | (x>>(32-s)));
}

static void round4_salsa(uint32 * d0, uint32 * d1, uint32 * d2, uint32 * d3) {
  uint32 a, b, c, d;
  a=*d0;b=*d1;c=*d2;d=*d3;
  b ^= ROL32(a+d, 7);
  c ^= ROL32(b+a, 9);
  d ^= ROL32(c+b, 13);
  a ^= ROL32(d+c, 18);
  *d0=a;*d1=b;*d2=c;*d3=d;
}

static void round4_chacha(uint32 * d0, uint32 * d1, uint32 * d2, uint32 * d3) {
  uint32 a, b, c, d;
  a=*d0;b=*d1;c=*d2;d=*d3;
  a+=b; d=ROL32(d^a,16);
  c+=d; b=ROL32(b^c,12);
  a+=b; d=ROL32(d^a,8);
  c+=d; b=ROL32(b^c,7);
  *d0=a;*d1=b;*d2=c;*d3=d;
}

static void round16_salsa(uint32 * d) {
  round4_salsa(d+0, d+4, d+8, d+12);
  round4_salsa(d+5, d+9, d+13, d+1);
  round4_salsa(d+10, d+14, d+2, d+6);
  round4_salsa(d+15, d+3, d+7, d+11);
  round4_salsa(d+0, d+1, d+2, d+3);
  round4_salsa(d+5, d+6, d+7, d+4);
  round4_salsa(d+10, d+11, d+8, d+9);
  round4_salsa(d+15, d+12, d+13, d+14);
}

static void round16_chacha(uint32 * d) {
  round4_chacha(d+0, d+4, d+8, d+12);
  round4_chacha(d+1, d+5, d+9, d+13);
  round4_chacha(d+2, d+6, d+10, d+14);
  round4_chacha(d+3, d+7, d+11, d+15);
  round4_chacha(d+0, d+5, d+10, d+15);
  round4_chacha(d+1, d+6, d+11, d+12);
  round4_chacha(d+2, d+7, d+8, d+13);
  round4_chacha(d+3, d+4, d+9, d+14);
}

extern void salsa20_step(uint32 * d) {
  uint32 tmp[16];
  uint32 i;
  for(i=0;i<16;i++) tmp[i]=d[i];
  for(i=0;i<10;i++) round16_salsa(tmp);
  for(i=0;i<16;i++) d[i]+=tmp[i];
}

extern void chacha20_step(uint32 * d) {
  uint32 tmp[16];
  uint32 i;
  for(i=0;i<16;i++) tmp[i]=d[i];
  for(i=0;i<10;i++) round16_chacha(tmp);
  for(i=0;i<16;i++) d[i]+=tmp[i];
}

/* TODO: this is slightly incomplete (but works ok) */
extern void salsa20_genrandom512(uint32 * tmp, uint32 * key,
                                 uint32 * nonce, uint32 * seq) {
  tmp[0] = 0x61707865;
  tmp[5] = 0x3320646E;
  tmp[10] = 0x79622D32;
  tmp[15] = 0x6B206574;
  tmp[1] = key[0];
  tmp[2] = key[1];
  tmp[3] = key[2];
  tmp[4] = key[3];
  tmp[11] = key[4];
  tmp[12] = key[5];
  tmp[13] = key[6];
  tmp[14] = key[7];
  tmp[6] = nonce[0];
  tmp[7] = nonce[1];
  tmp[8] = seq[0];
  tmp[9] = seq[1];
  salsa20_step(tmp);
}

extern void chacha20_genrandom512(uint32 * tmp, uint32 * key,
                                  uint32 * nonce, uint32 * seq) {
  tmp[0] = 0x61707865;
  tmp[1] = 0x3320646E;
  tmp[2] = 0x79622D32;
  tmp[3] = 0x6B206574;
  tmp[4] = key[0];
  tmp[5] = key[1];
  tmp[6] = key[2];
  tmp[7] = key[3];
  tmp[8] = key[4];
  tmp[9] = key[5];
  tmp[10] = key[6];
  tmp[11] = key[7];
  tmp[14] = nonce[0];
  tmp[15] = nonce[1];
  tmp[12] = seq[0];
  tmp[13] = seq[1];
  chacha20_step(tmp);
}

#ifdef SALSA20_DEBUG
#include <stdio.h>
uchar ex0[64] = {
  0,0,0,0,0,0,0,0, 0,0,0,0,0,0,0,0,
  0,0,0,0,0,0,0,0, 0,0,0,0,0,0,0,0,
  0,0,0,0,0,0,0,0, 0,0,0,0,0,0,0,0,
  0,0,0,0,0,0,0,0, 0,0,0,0,0,0,0,0
};
uchar ex1[64] = {
  211,159,13,115,76,55,82,183,3,117,222,37,191,187,234,136,
  49,237,179,48,1,106,178,219,175,199,166,48,86,16,179,207,
  31,240,32,63,15,83,93,161,116,147,48,113,238,55,204,36,
  79,201,235,79,3,81,156,47,203,26,244,243,88,118,104,54
};
uchar ex2[64] = {
  88,118,104,54,79,201,235,79,3,81,156,47,203,26,244,243,
  191,187,234,136,211,159,13,115,76,55,82,183,3,117,222,37,
  86,16,179,207,49,237,179,48,1,106,178,219,175,199,166,48,
  238,55,204,36,31,240,32,63,15,83,93,161,116,147,48,113
};
uchar ex3[64] = {
  6,124,83,146,38,191,9,50,4,161,47,222,122,182,223,185,
  75,27,0,216,16,122,7,89,162,104,101,147,213,21,54,95,
  225,253,139,176,105,132,23,116,76,41,176,207,221,34,157,108,
  94,94,99,52,90,117,91,220,146,190,239,143,196,176,130,186
};

static void printb(char const * name, uchar * d, size_t len) {
  printf("%s =", name);
  while(len--) printf(" %3d", *(d++));
  printf("\n");
}

static void printh(char const * name, uchar * d, size_t len) {
  printf("%s = ", name);
  while(len--) printf("%02X", *(d++));
  printf("\n");
}

int main(void) {
  int i;
/*  printb("ex0", ex0, 64);
  salsa20_step((uint32*)ex0);
  printb("en0", ex0, 64);
  printb("ex1", ex1, 64);
  salsa20_step((uint32*)ex1);
  printb("en1", ex1, 64);
  printb("ex2", ex2, 64);
  salsa20_step((uint32*)ex2);
  printb("en2", ex2, 64);
  printb("ex3", ex3, 64);
  for(i=0;i<1000000;i++)salsa20_step((uint32*)ex3);
  printb("en3", ex3, 64);*/

  {
    uint8 tmp[64];
ex0[32]=1;
    chacha20_genrandom512((uint32*)tmp, (uint32*)ex0, (uint32*)(ex0+32), (uint32*)ex0);
 //   printb("stream", tmp, 64);
    printh("stream", tmp, 64);
  }
}
#endif
