/*
 * This file is part of firk's Cryptography Library
 * (c) firk 2014, 2015, 2020; firk@cantconnect.ru
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. The origin of this software must not be misrepresented; you must
 *    not claim that you wrote the original software.
 * 4. Altered versions in any form must be plainly marked as such, and
 *    must not be misinterpreted as being the original software.
 *
 * This software is provided by the author and contributors `as is'
 * without any express or implied warranty.
 */

#include <sys/types.h>
#include "../fcl/types.h"
#include "sha1.h"

extern void sha1_init(sha1_ctx * ctx) {
    ctx->state[0] = 0x67452301;
    ctx->state[1] = 0xEFCDAB89;
    ctx->state[2] = 0x98BADCFE;
    ctx->state[3] = 0x10325476;
    ctx->state[4] = 0xC3D2E1F0;
    ctx->len = 0;
}

static void uint32_store(uint8 * dest, uint32 src) {
    dest[0] = (uint8)(src>>24);
    dest[1] = (uint8)(src>>16);
    dest[2] = (uint8)(src>>8);
    dest[3] = (uint8)(src);
}

static uint32 ROR32(uint32 v, uint32 s) {
    /* we don't need to handle something outside 1<=s<=31
       here because all shift factors are hardcoded */
    return (v>>s)|(v<<(32-s));
}

extern void sha1_transform(uint8 const * data,  uint32 * state) {
    uint32  buf[80], t0, a, b, c, d, e, f, k;
    int     i;
    for(i=0; i<16; i++) {
        buf[i] = (((uint32)(data[i*4+0]))<<24) +
                 (((uint32)(data[i*4+1]))<<16) +
                 (((uint32)(data[i*4+2]))<< 8) +
                 (((uint32)(data[i*4+3]))    );
    }
    for(i=16;i<80;i++) {
        buf[i] = ROR32(buf[i-3] ^ buf[i-8] ^ buf[i-14] ^ buf[i-16], 31);
    }
    a = state[0];
    b = state[1];
    c = state[2];
    d = state[3];
    e = state[4];

    for(i=0; i<80; i++) {
        if(i<20) {
            f = (b & c) | ((~b) & d);
            k = 0x5A827999;
        } else if(i<40) {
            f = b ^ c ^ d;
            k = 0x6ED9EBA1;
        } else if(i<60) {
            f = (b & c) | (b & d) | (c & d);
            k = 0x8F1BBCDC;
        } else {
            f = b ^ c ^ d;
            k = 0xCA62C1D6;
        }
        t0 = ROR32(a, 27) + f + e + k + buf[i];
        e = d;
        d = c;
        c = ROR32(b, 2);
        b = a;
        a = t0;
    }

    state[0] += a;
    state[1] += b;
    state[2] += c;
    state[3] += d;
    state[4] += e;
}

extern void sha1_update(sha1_ctx * ctx, uint8 const * data, size_t len) {
    uint8 * cdata;
    uint32  a;

    cdata = ctx->data;
    a = (uint32)(ctx->len & 63);
     /* we can overflow 64bit byte-counter here;
        even easier is to overflow 64bit bitcounter (len*8) and
          formally it will be too-long-to-be-hashed sequence, but it will
          still produce some hash and we don't care about these overflows */
    ctx->len += (uint64)len;
    if(a) {
        while(a<64 && len>0) { cdata[a++] = *(data++); len--; }
        if(a<64) return;
        sha1_transform(cdata,ctx->state);
        a = 0;
    }
    while(len>=64) {
        sha1_transform(data,ctx->state);
        data += 64;
        len -= 64;
    }
    while(a<len) { cdata[a]=data[a]; a++; }
}

extern void sha1_final(sha1_ctx * ctx, uint8 * hash) {
    uint8 * cdata;
    uint    a;
    uint64 len;

    cdata = ctx->data;
    len = ctx->len;
    a = (uint)(len & 63);

    cdata[a++] = 0x80;
    if(a>56) {
        while(a<64) cdata[a++] = 0;
        sha1_transform(cdata,ctx->state);
        a = 0;
    }
    while(a<56) cdata[a++] = 0;
    len<<=3;
    cdata[63] = (uint8)len; len>>=8;
    cdata[62] = (uint8)len; len>>=8;
    cdata[61] = (uint8)len; len>>=8;
    cdata[60] = (uint8)len; len>>=8;
    cdata[59] = (uint8)len; len>>=8;
    cdata[58] = (uint8)len; len>>=8;
    cdata[57] = (uint8)len; len>>=8;
    cdata[56] = (uint8)len;
    sha1_transform(cdata,ctx->state);
    for(a=0;a<5;a++) uint32_store(hash+a*4, ctx->state[a]);
}

extern void sha1_hash(uint8 * hash, char const * data, size_t len) {
    sha1_ctx ctx;
    sha1_init(&ctx);
    sha1_update(&ctx, (uint8 const*)data, len);
    sha1_final(&ctx, hash);
}


