/*
 * This file is part of firk's Cryptography Library
 * (c) firk 2014, 2015; firk@cantconnect.ru
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. The origin of this software must not be misrepresented; you must
 *    not claim that you wrote the original software.
 * 4. Altered versions in any form must be plainly marked as such, and
 *    must not be misinterpreted as being the original software.
 *
 * This software is provided by the author and contributors `as is'
 * without any express or implied warranty.
 */

#include <sys/types.h>
#include "../fcl/types.h"
#include "sha256.h"

extern void sha256_init(sha256_ctx * ctx) {
    ctx->state[0] = 0x6A09E667;
    ctx->state[1] = 0xBB67AE85;
    ctx->state[2] = 0x3C6EF372;
    ctx->state[3] = 0xA54FF53A;
    ctx->state[4] = 0x510E527F;
    ctx->state[5] = 0x9B05688C;
    ctx->state[6] = 0x1F83D9AB;
    ctx->state[7] = 0x5BE0CD19;
    ctx->len = 0;
}

static void uint32_store(uint8 * dest, uint32 src) {
    dest[0] = (uint8)(src>>24);
    dest[1] = (uint8)(src>>16);
    dest[2] = (uint8)(src>>8);
    dest[3] = (uint8)(src);
}

static uint32 ROR32(uint32 v, uint32 s) {
    /* we don't need to handle something outside 1<=s<=31
       here because all shift factors are hardcoded */
    return (v>>s)|(v<<(32-s));
}

static const uint32 k[64] = {
   0x428A2F98, 0x71374491, 0xB5C0FBCF, 0xE9B5DBA5,
   0x3956C25B, 0x59F111F1, 0x923F82A4, 0xAB1C5ED5,
   0xD807AA98, 0x12835B01, 0x243185BE, 0x550C7DC3,
   0x72BE5D74, 0x80DEB1FE, 0x9BDC06A7, 0xC19BF174,
   0xE49B69C1, 0xEFBE4786, 0x0FC19DC6, 0x240CA1CC,
   0x2DE92C6F, 0x4A7484AA, 0x5CB0A9DC, 0x76F988DA,
   0x983E5152, 0xA831C66D, 0xB00327C8, 0xBF597FC7,
   0xC6E00BF3, 0xD5A79147, 0x06CA6351, 0x14292967,
   0x27B70A85, 0x2E1B2138, 0x4D2C6DFC, 0x53380D13,
   0x650A7354, 0x766A0ABB, 0x81C2C92E, 0x92722C85,
   0xA2BFE8A1, 0xA81A664B, 0xC24B8B70, 0xC76C51A3,
   0xD192E819, 0xD6990624, 0xF40E3585, 0x106AA070,
   0x19A4C116, 0x1E376C08, 0x2748774C, 0x34B0BCB5,
   0x391C0CB3, 0x4ED8AA4A, 0x5B9CCA4F, 0x682E6FF3,
   0x748F82EE, 0x78A5636F, 0x84C87814, 0x8CC70208,
   0x90BEFFFA, 0xA4506CEB, 0xBEF9A3F7, 0xC67178F2 };


extern void sha256_transform(uint8 const * data,  uint32 * state) {
    uint32  buf[64], t0, t1, a, b, c, d, e, f, g, h;
    int     i;
    for(i=0; i<16; i++) {
        buf[i] = (((uint32)(data[i*4+0]))<<24) +
                 (((uint32)(data[i*4+1]))<<16) +
                 (((uint32)(data[i*4+2]))<< 8) +
                 (((uint32)(data[i*4+3]))    );
    }
    for(i=16;i<64;i++) {
        t0 = ROR32(buf[i-15],7) ^ ROR32(buf[i-15],18) ^ (buf[i-15] >> 3);
        t1 = ROR32(buf[i-2],17) ^ ROR32(buf[i-2],19) ^ (buf[i-2] >> 10);
        buf[i] = buf[i-16] + buf[i-7] + t0 + t1;
    }
    a = state[0];
    b = state[1];
    c = state[2];
    d = state[3];
    e = state[4];
    f = state[5];
    g = state[6];
    h = state[7];

    for(i=0; i<64; i++) {
        t0 = (ROR32(a,2)^ROR32(a,13)^ROR32(a,22)) + ((a&b)^(a&c)^(b&c));
        t1 = (ROR32(e,6)^ROR32(e,11)^ROR32(e,25)) +
             ((e&f)^((~e)&g)) + h + k[i] + buf[i];
        h = g;
        g = f;
        f = e;
        e = d + t1;
        d = c;
        c = b;
        b = a;
        a = t1 + t0;
    }

    state[0] += a;
    state[1] += b;
    state[2] += c;
    state[3] += d;
    state[4] += e;
    state[5] += f;
    state[6] += g;
    state[7] += h;
}

extern void sha256_update(sha256_ctx * ctx, uint8 const * data, size_t len) {
    uint8 * cdata;
    uint32  a;

    cdata = ctx->data;
    a = (uint32)(ctx->len & 63);
     /* we can overflow 64bit byte-counter here;
        even easier is to overflow 64bit bitcounter (len*8) and
          formally it will be too-long-to-be-hashed sequence, but it will
          still produce some hash and we don't care about these overflows */
    ctx->len += (uint64)len;
    if(a) {
        while(a<64 && len>0) { cdata[a++] = *(data++); len--; }
        if(a<64) return;
        sha256_transform(cdata,ctx->state);
        a = 0;
    }
    while(len>=64) {
        sha256_transform(data,ctx->state);
        data += 64;
        len -= 64;
    }
    while(a<len) { cdata[a]=data[a]; a++; }
}

extern void sha256_final(sha256_ctx * ctx, uint8 * hash) {
    uint8 * cdata;
    uint   a;
    uint64 len;

    cdata = ctx->data;
    len = ctx->len;
    a = (uint)(len & 63);

    cdata[a++] = 0x80;
    if(a>56) {
        while(a<64) cdata[a++] = 0;
        sha256_transform(cdata,ctx->state);
        a = 0;
    }
    while(a<56) cdata[a++] = 0;
    len<<=3;
    cdata[63] = (uint8)len; len>>=8;
    cdata[62] = (uint8)len; len>>=8;
    cdata[61] = (uint8)len; len>>=8;
    cdata[60] = (uint8)len; len>>=8;
    cdata[59] = (uint8)len; len>>=8;
    cdata[58] = (uint8)len; len>>=8;
    cdata[57] = (uint8)len; len>>=8;
    cdata[56] = (uint8)len;
    sha256_transform(cdata,ctx->state);
    for(a=0;a<8;a++) uint32_store(hash+a*4, ctx->state[a]);
}

extern void sha256_hash(uint8 * hash, char const * data, size_t len) {
    sha256_ctx ctx;
    sha256_init(&ctx);
    sha256_update(&ctx, (uint8 const*)data, len);
    sha256_final(&ctx, hash);
}


