/*
 * This file is part of fproxy - multipurpose http(s) proxy server
 * config.c - config file reader
 * Copyright (c) 2020-2021 firk (firk@cantconnect.ru)
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. The origin of this software must not be misrepresented; you must
 *    not claim that you wrote the original software.
 * 4. Altered versions in any form must be plainly marked as such, and
 *    must not be misinterpreted as being the original software.
 *
 * This software is provided by the author and contributors `as is'
 * without any express or implied warranty.
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>
#include <fcl/types.h>
#define FCL_SIMPLE_NAMES
#include <fcl/strparse.h>
#include <fcl/malloc.h>
#include <fcl/qprint.h>
#include <libsimpletcp/libsimpletcp.h>
#include "eventlog.h"
#include "config.h"

typedef struct {
  char const * name;
  int          type;
  union {
    uchar        * _V_uchar;
    uint         * _V_uint;
    ulong        * _V_ulong;
    uint16       * _V_uint16;
    uint32       * _V_uint32;
    uint64       * _V_uint64;
    size_t       * _V_size_t;
    char const * * _V_string;
  };
  union {
    uchar        _D_uchar;
    uint         _D_uint;
    ulong        _D_ulong;
    uint16       _D_uint16;
    uint32       _D_uint32;
    uint64       _D_uint64;
    size_t       _D_size_t;
    char const * _D_string;
  };
} t_config_item;

#define _T_uchar      1
#define _T_uint       2
#define _T_ulong      3
#define _T_uint16     4
#define _T_uint32     5
#define _T_uint64     6
#define _T_size_t     7
#define _T_string     8

#define PAR(_type, _varname, _default)  { .name = #_varname, .type = _T_##_type, ._V_##_type = &prefs._varname, ._D_##_type = _default }
#define PARH(_type, _varname, _default)  { .name = #_varname, .type = _T_##_type | 256, ._V_##_type = &prefs._varname, ._D_##_type = _default }

t_prefs prefs;

static t_config_item config[] = {
  /* type           varname                         default    */
  PAR(string,       base_dir,                       "."                                  ),
  PAR(string,       listen.ip,                      "127.0.0.10"                         ),
  PAR(uint16,       listen.port,                    3128                                 ),
  PAR(uchar,        listen.raw_https,               0                                    ),
  PAR(uint16,       num_processes,                  40                                   ),
  PAR(uint,         connection_timeout.keepalive,   30                                   ),
  PAR(uint,         connection_timeout.idle,        36000                                ),
  PARH(uint16,      local_ssl.minver,               0x0303                               ),
  PAR(string,       local_ssl.cacert,               "./pem/cacert.pem"                   ),
  PAR(string,       local_ssl.cakey,                "./pem/cakey.pem"                    ),
  PAR(string,       local_ssl.proxykey,             "./pem/proxykey.pem"                 ),
  PAR(uchar,        local_ssl.wildcard_certs,       1                                    ),
  PAR(uchar,        local_ssl.no_SAN,               0                                    ),
  PARH(uint16,      inet_ssl.minver,                0x0301                               ),
  PAR(string,       inet_ssl.calist,                "/etc/ssl/certs/ca-certificates.crt" ),
  PARH(uint32,      rproxy.host_dns,                0x7F00000A                           ),
  PAR(uint16,       rproxy.port_dns,                1115                                 ),
  PARH(uint32,      rproxy.host_data,               0x7F00000A                           ),
  PAR(uint16,       rproxy.port_data,               1116                                 ),
  /* make room for easy setproctitle() */
  PAR(string,       helpers.ssl_client_wrapper,     "./././././././././././././././././././././././././././ssl-client-wrapper" ),
  PAR(string,       helpers.ssl_server_wrapper,     "./ssl-server-wrapper" ),
  { NULL }
};

static void config_check(void);
int config_load(char const * fn) {
  FILE * fp;
  t_config_item * c;
  char buf[1000];
  char ch, *p, *name;
  unsigned int ln;
  uint32 tmp32;
  
  bzero(&prefs, sizeof(prefs));
  for(c=config; c->name; c++) switch(c->type & 255) {
  default: goto unkt;
#define D(type) case _T_##type: *c->_V_##type = c->_D_##type; break;
  D(uchar)
  D(uint)
  D(ulong)
  D(uint16)
  D(uint32)
  D(uint64)
  D(string)
  D(size_t)
#undef D
  }
  
  if(!fn) return 0;
  ln = 0;
  if(!(fp=fopen(fn,"rb"))) { eventlog(eventlog_level_error, __FUNCTION__, "open config file \"%s\" error %d (%s)", fn, errno, strerror(errno)); exit(-1); }
  while(fgets(buf, sizeof(buf), fp)) {
    ln++;
    if(!(p=strchr(buf,'\n')) || p[1]) goto nt;
    *p = 0;
    p = (char*)skip_spaces(buf);
    if(!*p || *p=='#') continue;
    for(name=p; (ch=*p) && ch!=' ' && ch!='\t' && ch!='='; p++);
    if(!ch) goto bad;
    *(p++) = 0;
    if(ch!='=') {
      p = (char*)skip_spaces(p);
      if(*p!='=') goto bad;
      p++;
    }
    p = (char*)skip_spaces(p);
    for(c=config; c->name && strcmp(c->name,name); c++);
    if(!c->name) goto unk;
    switch(c->type) {
    default: goto unkt;
    case _T_uchar: if(str_to_uchar(p,c->_V_uchar)<0) goto badv; break;
    case _T_uint: if(str_to_uint(p,c->_V_uint)<0) goto badv; break;
    case _T_ulong: if(str_to_ulong(p,c->_V_ulong)<0) goto badv; break;
    case _T_uint16: if(str_to_uint16(p,c->_V_uint16)<0) goto badv; break;
    case _T_uint16|256: if(str_hex_to_uint16(p,c->_V_uint16)<0) goto badv; break;
    case _T_uint32: if(str_to_uint32(p,c->_V_uint32)<0) goto badv; break;
    case _T_uint32|256: if((tmp32=lst_parse_ipstr(p))==LST_INADDR_NONE) goto badv; *c->_V_uint32 = tmp32; break;
    case _T_uint64: if(str_to_uint64(p,c->_V_uint64)<0) goto badv; break;
    case _T_string: *c->_V_string = fcl_strdup_check(p); break;
    }
  }
  config_check();
  return 0;
nt: eventlog(eventlog_level_error, __FUNCTION__, "config file \"%s\":%u non-terminated line or line too long", fn, ln); exit(-1);
bad: eventlog(eventlog_level_error, __FUNCTION__, "config file \"%s\":%u syntax error", fn, ln); exit(-1);
unk: eventlog(eventlog_level_error, __FUNCTION__, "config file \"%s\":%u unknown option %s", fn, ln, name); exit(-1);
unkt: eventlog(eventlog_level_error, __FUNCTION__, "config file \"%s\":%u unknown type for option %s", fn, ln, c->name); exit(-1);
badv: eventlog(eventlog_level_error, __FUNCTION__, "config file \"%s\":%u bad value for option %s", fn, ln, c->name); exit(-1);
}

static void config_check(void) {
}

void config_dump(void) {
  t_config_item * c;
  uint32 n, *d;
  char buf[50];
  
  for(c=config; c->name; c++) switch(c->type) {
#define D(type, t2, fmt) case _T_##type: qp_printf(&qp_out, " %10s %-20s = "fmt"\n", #type, c->name, t2 (*c->_V_##type)); break;
#define DH(type, t2, fmt) case _T_##type|256: qp_printf(&qp_out, " %10s %-20s = "fmt"\n", #type, c->name, t2 (*c->_V_##type)); break;
  D(uchar, (uint), "%u")
  D(uint, , "%u")
  D(ulong, , "%lu")
  D(uint16, (uint), "%u")
  DH(uint16, (uint), "%04X")
  D(uint32, (ulong), "%lu")
  DH(uint32, (uint32), "%{IP32}")
  D(uint64, (ullong), "%llu")
  D(string, , "%s")
  D(size_t, (ullong), "%llu")
  }
  qp_flush(&qp_out);
}



extern int str_to_size_t(char const * st, size_t * res) {
  size_t r, t;
  char c;
  r = 0;
  if(!st) return -1;
  if(!*st) return -1;
  while(c=*st) {
    if(c<'0' || c>'9') return -1;
    t = r*10;
    if(t/10!=r) return -1;
    r = t + (unsigned char)(c-'0');
    if(r<t) return -1;
    st++;
  }
  *res = r;
  return 0;
}

