/*
 * This file is part of fproxy - multipurpose http(s) proxy server
 * dashboard.c - dashboard implementation
 * Copyright (c) 2018-2019 firk (firk@cantconnect.ru)
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. The origin of this software must not be misrepresented; you must
 *    not claim that you wrote the original software.
 * 4. Altered versions in any form must be plainly marked as such, and
 *    must not be misinterpreted as being the original software.
 *
 * This software is provided by the author and contributors `as is'
 * without any express or implied warranty.
 */

#include "config.h"
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <fcntl.h>
#include <unistd.h>
#include <errno.h>
#include <time.h>
#include "dashboard.h"
#include "eventlog.h"

static int fd = -1;
static unsigned int my_num = 0;

extern void dashboard_init(void) {
  unlink(DASHBOARD);
  fd = open(DASHBOARD, O_CREAT|O_RDWR|O_EXCL, 0644); /* RDWR needed for parent */
  if(fd<0) {
    eventlog(eventlog_level_fatal,__FUNCTION__,"open() errno %d (%s)", errno, strerror(errno));
    exit(-1);
  }
}

extern void dashboard_set_num(unsigned int n) {
  my_num = n;
}

/* this may be called after fork() if pwrite() has race conditions problem */
extern void dashboard_reopen(void) {
  if(fd>=0) close(fd);
  fd = open(DASHBOARD, O_WRONLY);
  if(fd<0) {
    eventlog(eventlog_level_fatal,__FUNCTION__,"open() errno %d (%s)", errno, strerror(errno));
    exit(-1);
  }
}

typedef struct {
  unsigned int is_active;
  unsigned int method;
  unsigned int is_ssl;
  char  host[256];
  char  url[512];
  char  rctype[64];
  char  ctype[64];
  time_t itime, rtime, etime, ktime;
  unsigned long long rclen, rdlen;
  unsigned long long clen, dlen;
} db_row;

static db_row row;

#ifdef _LINUX_
static void strlcpy(char * dst, char const * src, size_t sz) {
  if(!sz) return;
  strncpy(dst, src, sz);
  dst[sz-1] = 0;
}
#endif

static void dashboard_update() {
  if(pwrite(fd, &row, sizeof(row), sizeof(row)*(off_t)my_num)<0) {
  /* TODO: that to do? */
  }
}

extern void dashboard_begin_request(void) {
  row.is_active = 1;
  row.itime = time(NULL);
  dashboard_update();
}

extern void dashboard_set_request(unsigned int method, unsigned int is_ssl, char const * host, char const * url) {
  row.is_active = 2;
  row.method = method;
  row.is_ssl = is_ssl;
  strlcpy(row.host, host, sizeof(row.host));
  strlcpy(row.url, url, sizeof(row.url));
  row.ctype[0] = 0;
  row.clen = 0;
  row.dlen = 0;
  row.rtime = time(NULL);
  dashboard_update();
}

extern void dashboard_set_host(char const * host) {
  strlcpy(row.host, host, sizeof(row.host));
  dashboard_update();
}

extern void dashboard_set_rctype(char const * ctype) {
  strlcpy(row.rctype, ctype, sizeof(row.rctype));
  dashboard_update();
}

extern void dashboard_set_rclen(unsigned long long clen) {
  row.rclen = clen;
  dashboard_update();
}

extern void dashboard_set_rdlen(unsigned long long dlen) {
  row.rdlen = dlen;
  dashboard_update();
}

extern void dashboard_set_ctype(char const * ctype) {
  strlcpy(row.ctype, ctype, sizeof(row.ctype));
  dashboard_update();
}

extern void dashboard_set_clen(unsigned long long clen) {
  row.clen = clen;
  dashboard_update();
}

extern void dashboard_set_dlen(unsigned long long dlen) {
  row.dlen = dlen;
  dashboard_update();
}

extern void dashboard_received_request(void) {
  row.is_active = 3;
  dashboard_update();
}

extern void dashboard_resolved_domain(void) {
  row.is_active = 4;
  dashboard_update();
}

extern void dashboard_connected_target(void) {
  row.is_active = 5;
  dashboard_update();
}

extern void dashboard_keepalive(void) {
  row.is_active = 6;
  row.ktime = time(NULL);
  dashboard_update();
}

extern void dashboard_end_request(void) {
  row.is_active = 0;
  row.etime = time(NULL);
  dashboard_update();
}

extern int  dashboard_longest_keepalive(unsigned int numchilds, unsigned int minfree) {
  unsigned n, numfree, nmax;
  unsigned long t, maxt;
  time_t now;
  time(&now);
  maxt = 0; nmax = 0; numfree = 0;
  for(n=1; n<=numchilds; n++) {
    if(pread(fd, &row, sizeof(row), sizeof(row)*n)!=(ssize_t)sizeof(row) || !row.is_active) numfree++;
    else if(row.is_active==6) {
      t = (unsigned long)(now-row.ktime);
      if(t>maxt) { nmax=n; maxt=t; }
    }
  }
  if(numfree<minfree) return nmax;
  return 0;
}

#ifdef DASHBOARD_MAIN
static time_t now;
static int highlight = 0;
static unsigned long tdiff(time_t t) { return (unsigned long)(now-t); }
static void print_row(void) {
  char const *sstr, *mstr, *pstr;
  unsigned long td;
  if(highlight) {
    if(row.is_active>=3 && row.is_active<=5) fputs("\x1B[0;1;44m",stdout);
    else fputs("\x1B[0m",stdout);
  }
  printf("[%3u] ", my_num);
  switch(row.is_active) {
  case 0:
    if(!row.etime) { printf("NEVER USED"); break; }
    td = tdiff(row.etime);
    if(td<120) printf("IDLE %lus", tdiff(row.etime)); else printf("IDLE");
    break;
  case 1:
    printf("INIT %lus", tdiff(row.itime));
    break;
  case 2:
    sstr = "CWAIT"; goto go;
  case 3:
    sstr = " DNS "; goto go;
  case 4:
    sstr = "SWAIT"; goto go;
  case 5:
    sstr = "<--->"; goto go;
  case 6:
    sstr = "KEEPA"; row.rtime=row.ktime;
go: switch(row.method) {
    case M_GET:mstr="GET";break;
    case M_POST:mstr="POST";break;
    case M_HEAD:mstr="HEAD";break;
    default:mstr="???";
    }
    pstr = (row.is_ssl?"https":"http");
    if(strlen(row.host)>39) strcpy(row.host+36,"...");
    if(strlen(row.url)>63) strcpy(row.url+60,"...");
    if(strlen(row.ctype)>31) strcpy(row.ctype+28,"...");
    printf("%s %4lus %s %s://%s%s", sstr, tdiff(row.rtime), mstr, pstr, row.host, row.url);
    if(row.is_active>=4) printf(" %s %llu/%lluK", row.ctype, row.dlen/1024, row.clen/1024);
    else if(row.is_active==2) printf(" REQ:%llu/%llu", row.rdlen, row.rclen);
    break;
  }
  printf("\n");
  if(highlight) fputs("\x1B[0m",stdout);
}

static void print_state(void) {
  unsigned int n;

  fd = open(DASHBOARD, O_RDONLY);
  if(fd<0) {
    eventlog(eventlog_level_fatal,__FUNCTION__,"open() errno %d (%s)", errno, strerror(errno));
    exit(-1);
  }
  my_num = 0;
  time(&now);
  while(read(fd, &row, sizeof(row))==(ssize_t)sizeof(row)) { print_row(); my_num++; }
  close(fd);
}

int main(int argc, char * * argv) {
  int loop, n;
  struct tm * tm;
  char tmstr[64];
  char * arg;
  loop = 0;
  for(n=1; n<argc; n++) {
    arg = argv[n];
    if(!strcmp(arg,"--loop")) loop = 1;
    else if(!strcmp(arg,"--highlight")) highlight = 1;
    else if(!strncmp(arg,"--basedir=",10)) {
      if(chdir(arg+10)<0) {
        fprintf(stderr, "can't chdir(%s)\n", arg+10);
        return -1;
      }
    } else {
      fprintf(stderr, "unknown arg(%s)\n", arg);
      return -1;
    }
  }
  
  if(!loop) print_state();
  else {
    while(1) {
      time(&now);
      tm = localtime(&now);
      if(!strftime(tmstr, sizeof(tmstr), "%c", tm)) tmstr[0]=0;
      printf("\033[3J\033[H\033[2J       %s\n", tmstr);
      print_state();
      sleep(1);
    }
  }
  
  return 0;
}
#endif
