/*
 * This file is part of fproxy - multipurpose http(s) proxy server
 * gzip.c - zlib wrapper to deal with gzipped http
 * Copyright (c) 2018, 2021 firk (firk@cantconnect.ru)
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. The origin of this software must not be misrepresented; you must
 *    not claim that you wrote the original software.
 * 4. Altered versions in any form must be plainly marked as such, and
 *    must not be misinterpreted as being the original software.
 *
 * This software is provided by the author and contributors `as is'
 * without any express or implied warranty.
 */

#include <sys/types.h>
#include <string.h>
#include <strings.h>
#include <unistd.h>
#include <errno.h>
#include <assert.h>
#include <math.h>
#include <zlib.h>
#include <fcl/types.h>
#include <fcl/dynbufs.h>
#include "gzip.h"

/* Z_OK - ok
   Z_STREAM_ERROR - "invalid compression level"
   Z_DATA_ERROR - "invalid or incomplete defaltae data"
   Z_MEM_ERROR - "out of memory"
   Z_VERSION_ERROR - "zlib version mismatch" */

#define CSZ 16384

extern int gzip_compress(t_dynbuf * destb, t_dynbuf * srcb, int level, int * z_err) {
  int ret;
  z_stream strm;
  size_t csz;
  unsigned char * src;
  size_t srcsz;
  unsigned char out[CSZ];
  int read_error, write_error;
  
  src = srcb->buf;
  srcsz = srcb->len;
  
  strm.zalloc = Z_NULL;
  strm.zfree = Z_NULL;
  strm.opaque = Z_NULL;
  if(level<1 || level>9) level = Z_DEFAULT_COMPRESSION;
  ret = deflateInit2(&strm, level, Z_DEFLATED, MAX_WBITS|16, (MAX_MEM_LEVEL<8)?MAX_MEM_LEVEL:8, Z_DEFAULT_STRATEGY);
  if(ret!=Z_OK) { *z_err=ret; return -1; }
  
  strm.avail_in = 0; /* it is uInt not size_t - can't place entire src to it */
  strm.avail_out = sizeof(out);
  strm.next_out = out;
  
  write_error = read_error = 0;
  do {
    if(!strm.avail_in && srcsz) {
      csz=CSZ; if(CSZ>srcsz) csz=srcsz;
      strm.next_in=src; strm.avail_in=csz; src+=csz; srcsz-=csz;
    }
    if(!strm.avail_out) {
      if(dynbuf_write(destb, out, sizeof(out))<0) { write_error=1; break; }
      strm.avail_out = sizeof(out);
      strm.next_out = out;
    }
    ret = deflate(&strm, srcsz?Z_NO_FLUSH:Z_FINISH);
    assert(ret != Z_STREAM_ERROR); /* strm structure is corrupt */
    if(ret==Z_STREAM_END) {
      assert(!srcsz && !strm.avail_in);
      if(strm.avail_out<sizeof(out) && dynbuf_write(destb, out, sizeof(out)-strm.avail_out)<0) { write_error=1; break; }
      deflateEnd(&strm); *z_err = Z_OK; return 0;
    }
    assert(ret==Z_OK || ret==Z_BUF_ERROR);
    assert(!strm.avail_in && srcsz || !strm.avail_out);
  } while(1);
  deflateEnd(&strm); *z_err = Z_DATA_ERROR; return -2;
}

/* output (all values always stored, even in error cases):
  retvalue: 0=ok; -1=generic fail; -2=truncated input or not enough space in output (may be distinguished by destb->overflowed);
  *z_err:   error code from zlib (Z_DATA_ERROR for retvals -2 or -3); */
extern int gzip_decompress(t_dynbuf * destb, t_dynbuf * srcb, int * z_err) {
  int ret;
  z_stream strm;
  size_t csz;
  unsigned char * src;
  size_t srcsz;
  unsigned char out[CSZ];
  int read_error, write_error;
  
  src = srcb->buf;
  srcsz = srcb->len;
  
  strm.zalloc = Z_NULL;
  strm.zfree = Z_NULL;
  strm.opaque = Z_NULL;
  strm.avail_in = 0;
  strm.next_in = Z_NULL;
  strm.avail_out = 0;
  strm.next_out = Z_NULL;
  
  ret = inflateInit2(&strm, MAX_WBITS+16); /* +16 means gzip instead of zlib format */
  if(ret!=Z_OK) { *z_err=ret; return -1; }
  
  strm.avail_in = 0; /* it is uInt not size_t - can't place entire src to it */
  strm.avail_out = sizeof(out);
  strm.next_out = out;
  
  write_error = read_error = 0;
  do {
    if(!strm.avail_in) {
      csz=CSZ; if(CSZ>srcsz) csz=srcsz;
      if(!csz && strm.avail_out) { read_error = 1; break; }
      strm.next_in=src; strm.avail_in=csz; src+=csz; srcsz-=csz;
    }
    if(!strm.avail_out) {
      if(dynbuf_write(destb, out, sizeof(out))<0) { write_error=1; break; }
      strm.avail_out = sizeof(out);
      strm.next_out = out;
    }
    ret = inflate(&strm, Z_NO_FLUSH);
    assert(ret != Z_STREAM_ERROR); /* strm structure is corrupt */
    switch(ret) {
    case Z_NEED_DICT:
      ret = Z_DATA_ERROR;
    case Z_DATA_ERROR:
    case Z_MEM_ERROR:
    default:
      inflateEnd(&strm);
      *z_err = ret;
      return -1;
    case Z_OK: /* not a error conditions */
    case Z_BUF_ERROR:
    case Z_STREAM_END:
      break;
    }
  } while(ret!=Z_STREAM_END);
  if(write_error || strm.avail_out<sizeof(out) && dynbuf_write(destb, out, sizeof(out)-strm.avail_out)<0 || read_error) { inflateEnd(&strm); *z_err = Z_DATA_ERROR; return -2; }
  inflateEnd(&strm); *z_err = Z_OK; return 0;
}

extern int gzip_decompressfd(t_dynbuf * destb, int srcfd, int * z_err) {
  int ret;
  z_stream strm;
  unsigned char out[CSZ];
  unsigned char in[CSZ];
  int read_error, write_error;
  ssize_t rsz;
  
  strm.zalloc = Z_NULL;
  strm.zfree = Z_NULL;
  strm.opaque = Z_NULL;
  strm.avail_in = 0;
  strm.next_in = Z_NULL;
  strm.avail_out = 0;
  strm.next_out = Z_NULL;
  
  ret = inflateInit2(&strm, MAX_WBITS+16); /* +16 means gzip instead of zlib format */
  if(ret!=Z_OK) { *z_err=ret; return -1; }
  
  strm.avail_in = 0; /* it is uInt not size_t - can't place entire src to it */
  strm.avail_out = sizeof(out);
  strm.next_out = out;
  
  write_error = read_error = 0;
  do {
    if(!strm.avail_in) {
      while((rsz=read(srcfd, in, CSZ))<0) if(errno!=EINTR) { read_error=1; break; }
      if(!rsz && strm.avail_out) { read_error=1; break; }
      strm.next_in = in; strm.avail_in = rsz;
    }
    if(!strm.avail_out) {
      if(dynbuf_write(destb, out, sizeof(out))<0) { write_error=1; break; }
      strm.avail_out = sizeof(out);
      strm.next_out = out;
    }
    ret = inflate(&strm, Z_NO_FLUSH);
    assert(ret != Z_STREAM_ERROR); /* strm structure is corrupt */
    switch(ret) {
    case Z_NEED_DICT:
      ret = Z_DATA_ERROR;
    case Z_DATA_ERROR:
    case Z_MEM_ERROR:
    default:
      inflateEnd(&strm);
      *z_err = ret;
      return -1;
    case Z_OK: /* not a error conditions */
    case Z_BUF_ERROR:
    case Z_STREAM_END:
      break;
    }
  } while(ret!=Z_STREAM_END);
  if(write_error || strm.avail_out<sizeof(out) && dynbuf_write(destb, out, sizeof(out)-strm.avail_out)<0 || read_error) { inflateEnd(&strm); *z_err = Z_DATA_ERROR; return -2; }
  inflateEnd(&strm); *z_err = Z_OK; return 0;
}

/* -3 = mismatch */
extern int gzip_decompressfd_check(t_dynbuf * destb, int srcfd, int * z_err) {
  int ret;
  z_stream strm;
  size_t csz;
  unsigned char out[CSZ];
  unsigned char in[CSZ];
  int read_error, write_error;
  ssize_t rsz;
  size_t outpos;
  
  strm.zalloc = Z_NULL;
  strm.zfree = Z_NULL;
  strm.opaque = Z_NULL;
  strm.avail_in = 0;
  strm.next_in = Z_NULL;
  strm.avail_out = 0;
  strm.next_out = Z_NULL;
  
  ret = inflateInit2(&strm, MAX_WBITS+16); /* +16 means gzip instead of zlib format */
  if(ret!=Z_OK) { *z_err=ret; return -1; }
  
  strm.avail_in = 0; /* it is uInt not size_t - can't place entire src to it */
  strm.avail_out = sizeof(out);
  strm.next_out = out;
  
  write_error = read_error = 0; outpos = 0;
  do {
    if(!strm.avail_in) {
      while((rsz=read(srcfd, in, CSZ))<0) if(errno!=EINTR) { read_error=1; break; }
      if(!rsz && strm.avail_out) { read_error=1; break; }
      strm.next_in = in; strm.avail_in = rsz;
    }
    if(!strm.avail_out) {
      if(outpos>=destb->len || sizeof(out)>destb->len-outpos || bcmp(out, destb->buf+outpos, sizeof(out))) { write_error = 1; break; }
      strm.avail_out = sizeof(out);
      strm.next_out = out;
      outpos += sizeof(out);
    }
    ret = inflate(&strm, Z_NO_FLUSH);
    assert(ret != Z_STREAM_ERROR); /* strm structure is corrupt */
    switch(ret) {
    case Z_NEED_DICT:
      ret = Z_DATA_ERROR;
    case Z_DATA_ERROR:
    case Z_MEM_ERROR:
    default:
      inflateEnd(&strm);
      *z_err = ret;
      return -1;
    case Z_OK: /* not a error conditions */
    case Z_BUF_ERROR:
    case Z_STREAM_END:
      break;
    }
  } while(ret!=Z_STREAM_END);
  if(write_error) { inflateEnd(&strm); *z_err = Z_DATA_ERROR; return -3; }
  csz = sizeof(out)-strm.avail_out;
  if(outpos>destb->len || csz!=destb->len-outpos || csz && bcmp(out, destb->buf+outpos, csz)) { inflateEnd(&strm); *z_err = Z_DATA_ERROR; return -3; }
  if(read_error) { inflateEnd(&strm); *z_err = Z_DATA_ERROR; return -2; }
  inflateEnd(&strm); *z_err = Z_OK; return 0;
}

extern int calc_randomless(unsigned char const * b, size_t n) {
  size_t c[256], j;
  double avg, dd, sum;
  if(n<1) return 0;
  for(j=0;j<256;j++) c[j]=0;
  avg = ((double)n)/256;
  for(j=n; j; b++,j--) c[*b]++;
  sum = 0;
  for(j=0;j<256;j++) { dd = c[j]-avg; sum += dd*dd; }
       /* 0 <= sum < n*n */
  /* sum = 256*D[c];
     true random: D[c] = avg;    sum = n
     non-random: D[c] > avg
     very non-random: D[c] > avg*avg;    sum > 256*avg*avg = n*n/256
      */
  if(sum<n) sum /= n;
  else sum = log(sum)/log(n);
  /* <1 = random; 2 = maximum non-random */
  return (int)(sum*100);
  /* 140+ may try gzip 150+ surely needs gzip */
}
