/*
 * This file is part of fproxy - multipurpose http(s) proxy server
 * handle/cache.c - cache handler
 * Copyright (c) 2018-2021 firk (firk@cantconnect.ru)
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. The origin of this software must not be misrepresented; you must
 *    not claim that you wrote the original software.
 * 4. Altered versions in any form must be plainly marked as such, and
 *    must not be misinterpreted as being the original software.
 *
 * This software is provided by the author and contributors `as is'
 * without any express or implied warranty.
 */

#include <stdio.h>
#include <stdlib.h>
#include <fcntl.h>
#include <errno.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/file.h>
#include <fcl/types.h>
#include <fcl/malloc.h>
#include <fcl/string.h>
#include <fcl/dynbufs.h>
#include <fcrypt/sha256.h>
#include "../config.h"
#include "../eventlog.h"
#ifndef CACHE_WRITE_ONLY
#include "../util.h"
#else
extern int write_all(int fd, void const * buf, size_t sz);
#endif
#include "../gzip.h"
#include "cache.h"

extern void error_logf(char const * fmt, ...);

#ifndef CACHE_WRITE_ONLY

extern void free_cached_data(t_cached_data * cd) {
  if(cd->ctype) fcl_free_c_check(cd->ctype);
  if(cd->redir) fcl_free_c_check(cd->redir);
  dynbuf_clear(&cd->data);
  fcl_free_check(cd);
}

static int valid_b64(char c) { return (c>='A' && c<='Z' || c>='a' && c<='z' || c>='0' && c<='9' || c=='-' || c=='_'); }
/* it eats *src in any case
   returns NULL on malformed *src
  */
extern t_cached_data * parse_cached_data(t_dynbuf * src, char const * host, char const * url, int cache) {
  char const *ctype, *redir;
  t_cached_data * cd;
  char * x;
  size_t xlen, xpos, pp, jj;
  unsigned int status;
  int acao = 0;
  char hashp[80];
  t_dynbuf decomp;
  int zerr, errc;

  x = (char*)src->buf; xlen = src->len; errc = 1;
  if(xlen<4 || bcmp(x,"V78\n",4)) goto lfail;
  pp = 4;
/* (x+xpos) = get_next_line(); */ errc++; xpos = pp; while(pp<xlen && x[pp]!='\n') pp++; if(pp>=xlen) goto lfail; x[pp++] = 0;
  if(!strncmp(x+xpos,"HOST=",5)) { errc |= 256;
    if(strcmp(x+xpos+5,host)) { error_logf("parse_data: HOST mismatch"); goto fail; }
/* (x+xpos) = get_next_line(); */ xpos = pp; while(pp<xlen && x[pp]!='\n') pp++; if(pp>=xlen) goto lfail; x[pp++] = 0;
  }
  if(!strncmp(x+xpos,"URL=",4)) { errc |= 512;
    if(strcmp(x+xpos+4,url)) { error_logf("parse_data: URL mismatch"); goto fail; }
/* (x+xpos) = get_next_line(); */ xpos = pp; while(pp<xlen && x[pp]!='\n') pp++; if(pp>=xlen) goto lfail; x[pp++] = 0;
  }
  if(!strcmp(x+xpos,"ACAO")) { errc |= 1024;
    acao = 1;
/* (x+xpos) = get_next_line(); */ xpos = pp; while(pp<xlen && x[pp]!='\n') pp++; if(pp>=xlen) goto lfail; x[pp++] = 0;
  }
  if(!strncmp(x+xpos,"REDIR:",6)) goto handle_redir;
  status = atoi(x+xpos);
/* (x+xpos) = get_next_line(); */ errc++; xpos = pp; while(pp<xlen && x[pp]!='\n') pp++; if(pp>=xlen) goto lfail; x[pp++] = 0;
  errc++;
  if(xlen-pp<42 || xlen-pp>50 || x[pp+40]!='.') goto lfail;
  errc++;
  for(jj=0;jj<40;jj++) if(!valid_b64(x[pp+jj])) goto lfail;
  errc++;
  for(jj=41;jj<xlen-pp;jj++) if((x[pp+jj]<'0' || x[pp+jj]>'9') && x[pp+jj]!='.' && x[pp+jj]!='g' &&x[pp+jj]!='z') goto lfail;
  errc++;
  strcpy(hashp, cache?"./cache_data/":"./saved_data/");
  bcopy(x+pp, hashp+16, xlen-pp); hashp[16+(xlen-pp)] = 0;
  hashp[13] = hashp[16];
  hashp[14] = hashp[17];
  hashp[15] = '/';
  ctype = fcl_strdup_check(x+xpos);
  dynbuf_clear(src);
  
  cd = fcl_mallocz_check(sizeof(*cd));
  cd->status = status;
  cd->ctype = ctype;
  cd->allow_xdomain = acao;
  if(dynbuf_init_file(&cd->data,hashp)!=0) { dynbuf_clear(&cd->data); free((void*)ctype); free(cd); return NULL; }
  if(strstr(hashp,".gz")) {
    dynbuf_init(&decomp, 1000, 1000, 10000000);
    if(gzip_decompress(&decomp, &cd->data, &zerr)!=0) { error_logf("parse_data: gunzip error %d", zerr); dynbuf_clear(&decomp); dynbuf_clear(&cd->data); free((void*)ctype); free(cd); return NULL; }
    dynbuf_clear(&cd->data);
    cd->data = decomp;
  }
  return cd;
handle_redir:
  status = atoi(x+xpos+6);
  redir = fcl_mallocz_check(xlen-pp+1);
  bcopy(x+pp, (char*)redir, xlen-pp);
  dynbuf_clear(src);
  cd = fcl_mallocz_check(sizeof(*cd));
  cd->status = status;
  cd->redir = redir;
  return cd;
lfail:
  error_logf("parse_data: malformed file (err=%X)",errc);
fail:
  dynbuf_clear(src);
  return NULL;
}
#endif

static void b64hash(char * dest, uchar const * hash) {
  char const b64[]="ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz0123456789_-";
  size_t p;
  for(p=0; p<10; p++,dest+=4,hash+=3) {
    dest[0] = b64[hash[0] & 63];
    dest[1] = b64[(hash[0] >> 6) | ((hash[1] & 15) << 2)];
    dest[2] = b64[(hash[1] >> 4) | ((hash[2] & 3) << 4)];
    dest[3] = b64[(hash[2] >> 2) & 63];
  }
}

static int openat_nointr(int d, char const * name) {
  int fd;
  while((fd=openat(d,name,O_RDONLY))<0) if(errno!=EINTR) return -1;
  return fd;
}

#define IF_ERR_NOINTR(A) while((A)<0) if(errno!=EINTR)
/* always closes fd before return;
   ret. -1=i/o error; 0=file mismateches content; 1=file matches content */
static int compare_data(char const * fn, int fd, t_dynbuf * content, int gzip) {
  static char buf[65536];
  size_t p;
  ssize_t r;
  int rr, zerr;

  IF_ERR_NOINTR(flock(fd, LOCK_SH)) { error_logf("saving cache: flock(SH) file failed, error %d (%s), file (%s)", errno, strerror(errno), fn); close(fd); return -1; }
  if(!gzip) {
    for(p=0; p<=content->len; p+=r) {
      IF_ERR_NOINTR(r = read(fd,buf,sizeof(buf))) { error_logf("saving cache: file read error %d (%s), file (%s)", errno, strerror(errno), fn); close(fd); return -1; }
      if(p==content->len && !r) { close(fd); return 1; }
      if(p>=content->len || !r || (size_t)r>content->len-p || bcmp(buf, content->buf+p, r)) break;
    }
  } else {
    rr = gzip_decompressfd_check(content, fd, &zerr);
    if(rr == -3) { close(fd); return 0; }
    if(rr == 0) { close(fd); return 1; }
    close(fd); return -1;
  }
  close(fd);
  return 0;
}

extern int store_by_hash(int cache, t_dynbuf * content, char * hashp, int no_write) {
  uint16 hashi;
  int fd, dd, rr, gz;
  uchar hash[32];
  t_dynbuf comp;

  sha256_hash(hash, (char*)content->buf, content->len);
  strcpy(hashp, cache?"./cache_data/":"./saved_data/");
  b64hash(hashp+16, hash);
  hashp[13] = hashp[16];
  hashp[14] = hashp[17];
  hashp[15] = 0;
  hashp[56] = '.';
  hashi = 0;
  mkdir(cache?"./cache_data/":"./saved_data/", 0700);
  mkdir(hashp, 0700);
  IF_ERR_NOINTR(dd = open(hashp, O_DIRECTORY)) { error_logf("saving cache: open directory failed, error %d (%s), dir (%s)", errno, strerror(errno), hashp); return -1; }
  hashp[15] = '/';
  
  do {
    IF_ERR_NOINTR(flock(dd, LOCK_EX)) { hashp[15]=0; error_logf("saving cache: flock() directory failed, error %d (%s), dir (%s)", errno, strerror(errno), hashp); close(dd); return -1; }

    /* try raw */
    sprintf(hashp+57,"%u",(uint)hashi);
    if((fd = openat_nointr(dd, hashp+16))>=0) {
      flock(dd,LOCK_UN);
      if(rr = compare_data(hashp, fd, content, 0)) { close(dd); return (rr<0)?-1:2; }
      continue;
    }
    if(errno!=ENOENT) { error_logf("saving cache: open file for reading failed, error %d (%s), file (%s)", errno, strerror(errno), hashp); close(dd); return -1; }

    /* try .gz */
    sprintf(hashp+57,"%u.gz",(uint)hashi);
    if((fd = openat_nointr(dd, hashp+16))>=0) {
      flock(dd,LOCK_UN);
      if(rr = compare_data(hashp, fd, content, 1)) { close(dd); return (rr<0)?-1:2; }
      continue;
    }
    if(errno!=ENOENT) { error_logf("saving cache: open file for reading failed, error %d (%s), file (%s)", errno, strerror(errno), hashp); close(dd); return -1; }

    /* create new */
    gz = 0;
    if(calc_randomless(content->buf, content->len)>=140) {
    // TODO: check first 1024 bytes, and 1024 bytes from middle and 1024 from the end
      dynbuf_init(&comp, 1000, 1000, 1000000000);
      if(gzip_compress(&comp, content, 0, &rr)!=0) { dynbuf_clear(&comp); error_logf("saving cache: gzip error %d, file (%s)", rr, hashp); close(dd); return -1; }
      if(comp.len>content->len-content->len/3) dynbuf_clear(&comp);
      else gz = 1;
    }
    if(!gz) sprintf(hashp+57,"%u",(uint)hashi);
    if(no_write) { if(gz) dynbuf_clear(&comp); close(dd); return 1; }
    IF_ERR_NOINTR(fd = openat(dd, hashp+16, O_CREAT|O_EXCL|O_WRONLY, 0644)) { error_logf("saving cache: create file for writing failed, error %d (%s), file (%s)", errno, strerror(errno), hashp); close(dd); if(gz) dynbuf_clear(&comp); return -1; }
    IF_ERR_NOINTR(flock(fd, LOCK_EX)) { error_logf("saving cache: flock(EX) file failed, error %d (%s), file (%s)", errno, strerror(errno), hashp); close(fd); unlink(hashp); close(dd); if(gz) dynbuf_clear(&comp); return -1; }
    flock(dd,LOCK_UN);
    if(gz) content = &comp;
    if(write_all(fd,content->buf,content->len)<0 || fsync(fd)<0) {
      if(gz) dynbuf_clear(&comp);
      error_logf("saving cache: can't write file properly, file (%s)", hashp);
      ftruncate(fd, 0); /* do not unlink; this will lead to race problems */
      close(fd);
      close(dd);
      return -1;
    }
    if(gz) dynbuf_clear(&comp);
    close(fd); close(dd); return 1;
  } while(++hashi);

  error_logf("saving cache: collision counter overflow for file (%s/%s)", hashp, hashp+16);
  close(dd);
  return -1;
}

