/*
 * This file is part of fproxy - multipurpose http(s) proxy server
 * handle/handlereq.c - proxy http request handler
 * Copyright (c) 2018-2020 firk (firk@cantconnect.ru)
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. The origin of this software must not be misrepresented; you must
 *    not claim that you wrote the original software.
 * 4. Altered versions in any form must be plainly marked as such, and
 *    must not be misinterpreted as being the original software.
 *
 * This software is provided by the author and contributors `as is'
 * without any express or implied warranty.
 */

#define _GNU_SOURCE
#include "../config.h"
#include <sys/types.h>
#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>
#include <string.h>
#include <strings.h>
#include <errno.h>
#include <unistd.h>
#include <time.h>
#include <sys/types.h> // for cache.c
#include <sys/stat.h> // for cache.c
#include <dirent.h> // for cache.c
#include <fcntl.h> // for cache.c
#include <assert.h> // for cache.c
#include <sys/socket.h>
#include <sys/file.h>
#include <setjmp.h>
#include <fcl/types.h>
#include <fcl/strparse.h>
#include <libsimpletcp/libsimpletcp.h>
#include "../config.h"
#include "../eventlog.h"
#include "../helpers/lst_ssl.h"
#include "../helpers/remote.h"
#define UTIL_TIME
#include "../util.h"
#include "../http_packet.h"
#include "../gzip.h"
#include "../nslookup.h"
#include "handlereq.h"
#include "rules.h"
#include "rules.resolv.h"
#include "cache.h"
#include "../dashboard.h"

/* THIS IS STRONGLY SINGLE-THREAD */

extern int is_internal(void);
#define LOGFILE(X) (is_internal()?(LOG_BASE"/int_"X):(LOG_BASE"/"X))

/* input data */
static int in_FD;
static char in_METH;
static char const * in_RAW_URL;
static char http_v[2], in_HTTPS;
static t_http_headers const * in_HH;
static t_http_content const * in_HC;
static int is_keepalive;

/* runtime data */

static int conn_close;
static jmp_buf exitreq_save;
static void exitreq(void) { longjmp(exitreq_save, 1); }
static void exitreq_close(void) { conn_close=1; longjmp(exitreq_save, 1); }
static char const * HDR(char const * name) { return hh_get(in_HH,name); }

static unsigned int n_will_free;
static void * p_will_free[1000];

static void will_free(void * p) {
  if(n_will_free==1000) abort();
  p_will_free[n_will_free++] = p;
}


#define printf printf_to_socket
#define outstring outstring_to_socket

static void printf(char const * fmt, ...) {
  char * st;
  int r;
  va_list arg;
  va_start(arg, fmt);
  if(vasprintf(&st, fmt, arg)>=0) {
    lst_conn_send_all(in_FD, st, strlen(st)); // TODO: we CAN ignore errors here but better to stop handling of request after them
    free(st);
  } else abort(); // TODO
  va_end(arg);
}

static void outstring(char const * st) {
  lst_conn_send_all(in_FD, st, strlen(st)); // TODO: we CAN ignore errors here but better to stop handling of request after them
}

static void outbytes(void const * st, size_t sz) {
  lst_conn_send_all(in_FD, st, sz); // TODO: we CAN ignore errors here but better to stop handling of request after them
}

static void file_append_string(char const * fn, char const * st) {
  FILE * fp;
  if(fp=fopen(fn,"ab")) {
    fputs(st, fp);
    fclose(fp);
  }
}

static void file_append_printf(char const * fn, char const * fmt, ...) {
  FILE * fp;
  va_list arg;
  va_start(arg, fmt);
  if(fp=fopen(fn,"ab")) {
    vfprintf(fp, fmt, arg);
    fclose(fp);
  }
  va_end(arg);
}

static char const * get_timestr(void) {
  static char   timestr[64];
  time_t        now;
  struct tm *   tmnow;
  time(&now);
  tmnow = localtime(&now);
  if(!tmnow) sprintf(timestr, "%019lu", (unsigned long)now);
  else sprintf(timestr, "%04u/%02u/%02u %02u:%02u:%02u", tmnow->tm_year+1900, tmnow->tm_mon+1, tmnow->tm_mday, tmnow->tm_hour, tmnow->tm_min, tmnow->tm_sec);
  return timestr;
}

static void file_log_string(char const * fn, char const * st) {
  FILE * fp;
  if(fp=fopen(fn,"ab")) {
    fprintf(fp, "[%s] %s\n", get_timestr(), st);
    fclose(fp);
  }
}

static void file_log_printf(char const * fn, char const * fmt, ...) {
  FILE * fp;
  va_list arg;
  va_start(arg, fmt);
  if(fp=fopen(fn,"ab")) {
    fprintf(fp, "[%s] ", get_timestr());
    vfprintf(fp, fmt, arg);
    fputc('\n', fp);
    fclose(fp);
  }
  va_end(arg);
}

static void file_log_vprintf(char const * fn, char const * fmt, va_list arg) {
  FILE * fp;
  if(fp=fopen(fn,"ab")) {
    fprintf(fp, "[%s] ", get_timestr());
    vfprintf(fp, fmt, arg);
    fputc('\n', fp);
    fclose(fp);
  }
}

static char const * tmp_sprintf(char const * fmt, ...) {
  char * st;
  va_list arg;
  va_start(arg, fmt);
  if(vasprintf(&st, fmt, arg)>=0) {
    will_free(st);
  } else {
    st = "!!!ERROR vasprintf() failed ERROR!!!";
  }
  va_end(arg);
  return st;
}

static char const * tmp_vsprintf(char const * fmt, va_list arg) {
  char * st;
  if(vasprintf(&st, fmt, arg)>=0) {
    will_free(st);
  } else {
    st = "!!!ERROR vasprintf() failed ERROR!!!";
  }
  return st;
}

/* this is to avoid using unsupported statuses */
#define STATUS_200 200
#define STATUS_301 301
#define STATUS_302 302
#define STATUS_307 307
#define STATUS_400 400
#define STATUS_404 404
#define STATUS_500 500
#define STATUS_502 502
static void print_basic_headers(unsigned int status) {
  char const * fl;
  switch(status) {
  case STATUS_200: fl="200 OK"; break;
  case STATUS_301: fl="301 Moved Permanently"; break;
  case STATUS_302: fl="302 Found"; break;
  case STATUS_307: fl="307 Temporary Redirect"; break;
  case STATUS_400: fl="400 Bad Request"; break;
  case STATUS_404: fl="404 Not Found"; break;
  case STATUS_500: fl="500 Internal Server Error"; break;
  case STATUS_502: fl="502 Bad Gateway"; break;
  default: assert(!"unknown status");
  }
  printf("HTTP/%c.%c %s\r\nServer: proxy\r\nConnection: %s\r\n",
    http_v[0], http_v[1], fl, (is_keepalive?"keep-alive":"close"));
}

#include "inc.rewrite.c"
#include "inc.http_request.c"
#include "inc.http_sniff.c"
#include "inc.reply.c"
#include "inc.log.c"
#include "inc.cache.c"
#include "inc.proxy.c.sed"







/* what to do:
  handle incoming request, send http reply to fd, return */
extern int handle_request(int fd, char method, char const * raw_url, char v1, char v2, char ssl, t_http_headers const * hh, t_http_content const * hc) {
  char const * connmode;
  in_FD = fd;
  in_METH = method;
  in_RAW_URL = raw_url;
  http_v[0] = v1;
  http_v[1] = v2;
  in_HTTPS = ssl;
  in_HH = hh;
  in_HC = hc;
  is_keepalive = ((connmode=hh_get(hh,"Connection")) && !strcasecmp(connmode,"keep-alive"));
  
  conn_close = 0;
  n_will_free = 0;
  bzero(&_DSIGN_, sizeof(_DSIGN_));
  if(!setjmp(exitreq_save)) proxy_main();
  http_request_clear(&_DSIGN_);
  while(n_will_free--) free(p_will_free[n_will_free]);
  return conn_close;
}
