/*
 * This file is part of fproxy - multipurpose http(s) proxy server
 * handle/inc.cache.c - cache handler
 * Copyright (c) 2018-2020 firk (firk@cantconnect.ru)
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. The origin of this software must not be misrepresented; you must
 *    not claim that you wrote the original software.
 * 4. Altered versions in any form must be plainly marked as such, and
 *    must not be misinterpreted as being the original software.
 *
 * This software is provided by the author and contributors `as is'
 * without any express or implied warranty.
 */

#include "inc.rules.c"
// TODO: split this to separate .c file, not .inc

static int good_char(char c) {
  if(strchr("ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz0123456789-_()[]%!~@$&=+",c)) return 1;
  return 0;
}

static int good_host(char const * s) {
  size_t i;
  char c;
  for(i=0; c=s[i]; i++) {
    if(good_char(c) || c==':' || c=='.' && i!=0 && s[i+1] && s[i-1]!='.') continue;
    return 0;
  }
  return 1;
}

static int good_url(char const * s) {
  size_t i;
  char c;
  if(s[0]!='/') return 0;
  for(i=0; c=s[i]; i++) if(c<33 || c>126) return 0;
  return 1;
}

static char const * domain2dir(char const * domain) {
  size_t len, i, j, k;
  char c;
  char * r;
  if(!good_host(domain)) return NULL;
  len = strlen(domain);
  if(len>=280) return NULL; /* valid domain names are up to 255(256?) bytes, but we also have optional :port */
  /* TODO: dirname longer than 255 bytes will fail on most filesystems */
  for(i=j=k=0; c=domain[i]; i++) {
    if(c==':') break;
    if(c=='.') { if(i==j) return NULL; k=j; j=i+1; }
  }
  if(i==j) return NULL;
  assert(len<280 && i<=len && j<i && (k<j || k==j && j==0));
  i -= k;
  r = xmalloc(i+len+2);
  bcopy(domain+k, r, i);
  r[i++] = '/';
  bcopy(domain, r+i, len+1);
  return r;
}

static char const * url2file(char const * url) {
  char * r;
  size_t len, i;
  char c;
  if(!good_url(url)) return NULL;
  len = strlen(url);
  if(len>32760) return NULL; /* do not remove this check - other code relies on strlen(host)+strlen(url)+strlen(configured_cache_root) can't exceed or even be near any size_t limits */
  if(len>200) {
    r = xmalloc(100);
    r[0] = 'H';
    for(i=0; i<40; i++) {
      c = url[i];
      if(c=='/') c=' ';
      r[i+1] = c;
    }
    r[41] = '~';
    compact_string(r+42, url);
  } else {
    r = xmalloc(len+2);
    r[0] = 'U';
    for(i=0; i<len; i++) {
      c = url[i];
      if(c=='/') c=' ';
      r[i+1] = c;
    }
    r[len+1] = 0;
  }
  return r;
}

/* AB */
static char const * build_ddir(char const * base, char const * dstr) {
  char * dest;
  size_t l1, l2;
  l1 = strlen(base);
  l2 = strlen(dstr);
  dest = xmalloc(l1+l2+1);
  bcopy(base, dest, l1);
  bcopy(dstr, dest+l1, l2);
  dest[l1+l2] = 0;
  return dest;
}

/* AB/C.bin */
static char const * build_dupath(char const * base, char const * dstr, char const * ustr) {
  char * dest;
  size_t l1, l2, l3;
  l1 = strlen(base);
  l2 = strlen(dstr);
  l3 = strlen(ustr);
  dest = xmalloc(l1+l2+l3+6);
  bcopy(base, dest, l1);
  bcopy(dstr, dest+l1, l2);
  l1 += l2;
  dest[l1++] = '/';
  bcopy(ustr, dest+l1, l3);
  l1 += l3;
  dest[l1++] = '.';
  dest[l1++] = 'b';
  dest[l1++] = 'i';
  dest[l1++] = 'n';
  dest[l1++] = 0;
  return dest;
}

/* A/B.bin */
static char const * build_upath(char const * hpath, char const * ustr) {
  char * dest;
  size_t l1, l2;
  l1 = strlen(hpath);
  l2 = strlen(ustr);
  dest = xmalloc(l1+l2+6);
  bcopy(hpath, dest, l1);
  dest[l1++] = '/';
  bcopy(ustr, dest+l1, l2);
  l1+=l2;
  dest[l1++] = '.';
  dest[l1++] = 'b';
  dest[l1++] = 'i';
  dest[l1++] = 'n';
  dest[l1++] = 0;
  return dest;
}

/* A/B.tmp */
static char const * build_upath_tmp(char const * hpath, char const * ustr) {
  char * dest;
  size_t l1, l2;
  l1 = strlen(hpath);
  l2 = strlen(ustr);
  dest = xmalloc(l1+l2+6);
  bcopy(hpath, dest, l1);
  dest[l1++] = '/';
  bcopy(ustr, dest+l1, l2);
  l1+=l2;
  dest[l1++] = '.';
  dest[l1++] = 't';
  dest[l1++] = 'm';
  dest[l1++] = 'p';
  dest[l1++] = 0;
  return dest;
}

/* A/B.C.sav */
static char const * build_upath_ts(char const * hpath, char const * ustr, long ts) {
  char * dest;
  size_t l1, l2, l3;
  char tss[50];
  snprintf(tss,sizeof(tss),".%ld.sav",ts);
  l1 = strlen(hpath);
  l2 = strlen(ustr);
  l3 = strlen(tss);
  dest = xmalloc(l1+l2+l3+2);
  bcopy(hpath, dest, l1);
  dest[l1++] = '/';
  bcopy(ustr, dest+l1, l2);
  bcopy(tss, dest+l1+l2, l3);
  dest[l1+l2+l3] = 0;
  return dest;
}

static t_cached_data * get_offline_data(char const * host, char const * url, long offtime) {
  char const *dstr, *ustr, *hpath, *fpath;
  DIR * dp;
  struct dirent * pde;
  size_t ulen;
  long ft, ndiff, best, diff;
  t_dynbuf buf;
  t_cached_data * cd;
  char tmpt[50];
  size_t len1, len2, len3;
  
//  rewrite_url(host, &url); // already rewriten in get_cached_data() which calls this func
  if(!(dstr=domain2dir(host)) || !(ustr=url2file(url))) { if(dstr)free((void*)dstr); return NULL; }
  ulen = strlen(ustr);
  
  hpath = build_ddir("./saved_real/",dstr);
  if(!(dp=opendir(hpath))) goto fail;
  best = -1;
  while(pde = readdir(dp)) {
    if(!skip_prefix(pde->d_name, ustr) || pde->d_name[ulen]!='.' || pde->d_name[ulen+1]=='b') continue;
    if((ft=strtol(pde->d_name+ulen+1,NULL,10))<0) continue; /* prevent negative values */
    snprintf(tmpt,sizeof(tmpt),".%ld.sav",ft);
    if(strcmp(pde->d_name+ulen,tmpt)) continue;
    ndiff = labs(ft-offtime);
    if(best<0 || ndiff<diff) { best=ft; diff=ndiff; }
    if(!ndiff) break;
  }
  closedir(dp);
  if(best<0) goto fail;
  
  fpath = build_upath_ts(hpath,ustr,best);
  switch(dynbuf_init_file(&buf,fpath)) {
  default: // -2, -3 = i/o error or not enough memory //// TODO ////
    error_logf("file unreadable or too big: %s", fpath);
  case -1:
    dynbuf_clear(&buf);
    break; // no such file - ok, skip it
  case 0:
    if(!(cd=parse_cached_data(&buf,host,url,0))) { error_logf("error loading offline saved data from: %s",fpath); break; }
    cd->offline = 1;
    free((void*)fpath); free((void*)hpath); free((void*)dstr); free((void*)ustr);
    return cd;
  }
  free((void*)fpath);
fail: free((void*)hpath); free((void*)dstr); free((void*)ustr); return NULL;
}


static t_cached_data * get_cached_data(char const * host, char const * url) {
  char const *dstr, *ustr, *fpath;
  t_dynbuf buf;
  t_cached_data * cd;
  long off;
  
  rewrite_url(host, &url);
  if(!(dstr=domain2dir(host)) || !(ustr=url2file(url))) { if(dstr)free((void*)dstr); return NULL; }
  fpath = build_dupath("./cache_real/",dstr,ustr);
  free((void*)dstr); free((void*)ustr);
  switch(dynbuf_init_file_mintime(&buf,fpath,min_cache_date())) {
  default: // -2, -3 = i/o error or not enough memory //// TODO ////
    error_logf("file unreadable or too big: %s", fpath);
  case -1:
    dynbuf_clear(&buf);
    break; // no such file - ok, skip it
  case 0:
    if(!(cd=parse_cached_data(&buf,host,url,1))) { error_logf("error loading cached data from: %s", fpath); break; }
    if(!strcmp(host,"fonts.gstatic.com")) cd->allow_xdomain=1; // for youtube
    free((void*)fpath);
    return cd;
  }

  free((void*)fpath);
  if(off=is_offline()) return get_offline_data(host, url, off);
  return NULL;
}

static void b64hash(char * dest, uchar const * hash) {
  char const b64[]="ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz0123456789_-";
  size_t p;
  for(p=0; p<10; p++,dest+=4,hash+=3) {
    dest[0] = b64[hash[0] & 63];
    dest[1] = b64[(hash[0] >> 6) | ((hash[1] & 15) << 2)];
    dest[2] = b64[(hash[1] >> 4) | ((hash[2] & 3) << 4)];
    dest[3] = b64[(hash[2] >> 2) & 63];
  }
}

static void got_data(char const * host, char const * url, char const * ctype, t_dynbuf * content, int is_acao, int is_damaged) {
  char const *dstr, *hpath, *fpath, *fspath, *fspath_tmp, *ustr;
  char * hpp;
  size_t pp;
  int fd, nc;
  char hashp[80];
  
  if(is_damaged) return;
  rewrite_url(host, &url);
  if(!(nc=need_cache(host, url, ctype)) && !is_save(host,url)) return;
  if(!(dstr=domain2dir(host)) || !(ustr=url2file(url))) { if(dstr)free((void*)dstr); error_logf("saving cache: bad domain and/or url: %s %s", host, url); return; }
  if(store_by_hash(nc, content, hashp, 0)<0) return;
  
  if(nc) {
    hpath = build_ddir("./cache_real/",dstr);
    mkdir("./cache", 0700); mkdir("./cache_real", 0700); hpp = strrchr(hpath,'/'); *hpp = 0; mkdir(hpath, 0700); *hpp = '/'; mkdir(hpath, 0700);
    fpath = build_upath_ts(hpath,ustr,time(NULL));
    fspath = build_upath(hpath,ustr);
    fspath_tmp = build_upath_tmp(hpath,ustr);
    cache_logf("cached %s %s %lubytes",fpath,ctype,(unsigned long)content->len);
  } else {
    hpath = build_ddir("./saved_real/",dstr);
    mkdir("./saved", 0700); mkdir("./saved_real", 0700); hpp = strrchr(hpath,'/'); *hpp = 0; mkdir(hpath, 0700); *hpp = '/'; mkdir(hpath, 0700);
    fpath = build_upath_ts(hpath,ustr,time(NULL));
    fspath = fspath_tmp = NULL;
//    cache_logf("saved %s %s %lubytes",fpath,ctype,(unsigned long)content->len);
  }
  free((void*)hpath); free((void*)dstr); free((void*)ustr);

  if((fd=open(fpath,O_WRONLY|O_CREAT|O_EXCL, 0644))<0) {
    error_logf("saving cache: open file for writing failed, error %d (%s), file (%s)", errno, strerror(errno), fpath);
    free((void*)fpath);
    if(fspath) { free((void*)fspath); free((void*)fspath_tmp); }
    return;
  }
  if(write_all(fd,"V78\nHOST=",9)<0 ||
     write_all(fd,host,strlen(host))<0 ||
     write_all(fd,"\nURL=",5)<0 ||
     write_all(fd,url,strlen(url))<0 ||
     is_acao && write_all(fd,"\nACAO",5)<0 ||
     write_all(fd,"\n200\n",5)<0 ||
     write_all(fd,ctype,strlen(ctype))<0 ||
     write_all(fd,"\n",1)<0 ||
     write_all(fd,hashp+16,strlen(hashp+16))<0) {
    error_logf("saving cache: can't write file properly, file (%s)", fpath);
    close(fd);
    unlink(fpath);
    free((void*)fpath);
    if(fspath) { free((void*)fspath); free((void*)fspath_tmp); }
    return;
  }
  if(close(fd)<0) {
    error_logf("saving cache: can't write file properly, file (%s)", fpath);
    unlink(fpath);
    free((void*)fpath);
    if(fspath) { free((void*)fspath); free((void*)fspath_tmp); }
    return;
  }
  if(fspath) {
    unlink(fspath_tmp);
    if(link(fpath,fspath_tmp)<0) error_logf("saving cache: can't make hardlink from \"%s\" to \"%s\"", fpath, fspath_tmp);
    if(rename(fspath_tmp,fspath)<0) error_logf("saving cache: can't rename from \"%s\" to \"%s\"", fspath_tmp, fspath);
  }
  pp = strlen(fpath);
  assert(pp>1);
  free((void*)fpath);
  if(fspath) { free((void*)fspath); free((void*)fspath_tmp); }
}
