/*
 * This file is part of fproxy - multipurpose http(s) proxy server
 * handle/inc.http_request.c - http(s) proxied client
 * Copyright (c) 2018-2020,2022 firk (firk@cantconnect.ru)
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. The origin of this software must not be misrepresented; you must
 *    not claim that you wrote the original software.
 * 4. Altered versions in any form must be plainly marked as such, and
 *    must not be misinterpreted as being the original software.
 *
 * This software is provided by the author and contributors `as is'
 * without any express or implied warranty.
 */


#define TMPBUFSZ 32768 /* should fit signed int; so 16384 on 16bit */

extern uint32 http_proxy_addr;
extern uint16 http_proxy_port;

static int resolve_addr(char * https, char const * host, uint32 * r_addr, uint16 * r_port) {

  *r_port = ((*https)?443:80);
  {
    char *hh, *p;
    hh = xstrdup(host);
    if(p=strchr(hh,':')) {
      *p=0;
      if(str_to_uint16(p+1, r_port)<0) { free(hh); return -1; }
    }
    if(test_resolv_rule(hh, r_addr, https, r_port)) { free(hh); return 0; }
    free(hh);
  }
  
  if(nslookup(host, r_addr, r_port)<0) return -1;
  return 0;
}

#define err_req    1 /* error constructing request */
#define err_net    2 /* system or networking or ssl problem */
#define err_bad    3 /* bad http response (also truncated response here) */
#define err_badst  4 /* bad http status */
#define status_raw 5 /* raw mode ok */
#define status_ok  6 /* standard mode ok */
#define status_rcv 7 /* receiving response body */

typedef struct {
  int          R;
  char const * errmsg;
  char const * certs;
  t_dynbuf     warnlog; /* non-critical problems log */
  uchar *      tmpbuf; /* internally used temporary buffer; also used to hold dynamic errmsg */
  size_t       recvlen;
  t_dynbuf     req; /* request headers */
  uint32       addr;
  uint16       port;
  char const * addrstr;
  char const * sni_host;
  t_dynbuf     resp; /* response headers */
  unsigned     http; /* http status */
  char const * statusline;
  unsigned long long clen; /* value from header; 0=missing or zero */
  t_http_headers hdrs; /* parsed response headers */
  t_http_content content; /* response body */
  int          socket; /* only vaild for status_raw and status_rcv */
} t_http_request;

static void http_request_clear(t_http_request * $) {
  dynbuf_clear(&$->warnlog);
  if($->tmpbuf) free((void*)$->tmpbuf);
  dynbuf_clear(&$->req);
  if($->addrstr) free((void*)$->addrstr);
  if($->sni_host) free((void*)$->sni_host);
  dynbuf_clear(&$->resp);
  if($->statusline) free((void*)$->statusline);
  hh_clear(&$->hdrs);
  hc_clear(&$->content);
  // do not free errmsg - it is static string or marked as will_free()
  //if($->errmsg) free((void*)$->errmsg);
  // socket is to be closed by other code depending on context ...
  bzero($, sizeof(*$));
}

#define ERRS http_request_set_err_static
#define ERRF http_request_set_err_fmt
static void ERRS(t_http_request * $, int R, char const * st) {
  $->R = R;
  $->errmsg = st;
}
static void ERRF(t_http_request * $, int R, char const * fmt, ...) {
  va_list arg;
  va_start(arg,fmt);
  vsnprintf((char*)$->tmpbuf, TMPBUFSZ/2, fmt, arg);
  va_end(arg);
  $->R = R;
  $->errmsg = (char*)$->tmpbuf;
}


static int http_request_recv_all(t_http_request * $);

static int _prepare_http_request_raw(t_http_request * $, char meth, char const * host, char const * url, t_http_headers const * hdrs, t_http_content const * data, char is__raw) {
  t_dynbuf req;
  unsigned int n;
  char const *name, *value;
  
  dynbuf_init_check(&req, 4096, 4096, 65536);
  dynbuf_printf(&req, "%s %s HTTP/1.1\r\nHost: %s\r\nConnection: close\r\n", method_get_name(meth), url, host);
  if(!is__raw) dynbuf_printstr(&req, "Accept-Encoding: gzip\r\n");
  for(n=0; n<hdrs->n; n++) {
    name = hdrs->h[n*2];
    value = hdrs->h[n*2+1];
    if(!strcmp(name,"Connection") || !strcmp(name,"Host") || !strcmp(name,"Content-Length") || 
       !strcmp(name,"Accept-Encoding")) continue;
    dynbuf_printf(&req, "%s: %s\r\n", name, value);
  }
  if(data) dynbuf_printf(&req, "Content-Length: %lu\r\n", (unsigned long)data->total);
  dynbuf_printstr(&req, "\r\n");
  if(req.overflowed) { dynbuf_clear(&req); ERRS($, err_req, "request header too big"); return -1; }
  $->req = req;
  return 0;
}

static int http_request(t_http_request * $, char https, char meth, char const * host, char const * url, t_http_headers const * hdrs, t_http_content const * data, char is__raw, int use_rproxy) {
  lst_conn_open_ssl_ctx tmp;
  unsigned int n;
  char const *name, *value;
  uint32 addr;
  uint16 port;
  char addrstr[50];
  int fd, tmpr;
  unsigned long rbpos;
  unsigned char * tb;
  
  bzero($, sizeof(*$)); dynbuf_init(&$->warnlog, 256, 256, 65536);
  tb = $->tmpbuf = xmalloc(TMPBUFSZ);
  
  /* prepare request body */
  if(_prepare_http_request_raw($, meth, host, url, hdrs, data, is__raw)<0) return -1;
  
  /* get ip address */
  {
    if(resolve_addr(&https, host, &addr, &port)<0) { ERRF($, err_net, "failed resolve ip address for (%s)", host); return -1; }
    strcpy(addrstr, https?"ssl://":"tcp://");
    lst_print_ip_port(addrstr+6, addr, port);
    $->addr = addr;
    $->port = port;
    $->addrstr = xstrdup(addrstr);
  }
  dashboard_resolved_domain();

  /* connect */
  {
    char *sni_host, *tmpp;
    int attempt;
    char sslerr[1024];
    
    sni_host = xstrdup(host);
    $->sni_host = sni_host;
    if(tmpp=strchr(sni_host,':')) *tmpp=0;
    
    for(attempt=0; attempt<3; attempt++) {
    /* TODO: add proper timeouts of 20 sec */
      $->certs = NULL;
      dynbuf_printf(&$->warnlog, "trying %s/%s\n", addrstr, sni_host);
      if(use_rproxy)       fd = lst_remote_conn_open(prefs.rproxy.host_data, prefs.rproxy.port_data, addr, port, 0);
      else
      if(!http_proxy_port) fd = lst_conn_open(addr, port, 0);
      else                 fd = lst_remote_conn_open(http_proxy_addr, http_proxy_port, addr, port, 0);
      if(fd>=0 && https) { $->certs=(char*)tb+TMPBUFSZ/2; fd = lst_conn_open_ssl(fd, addr, port, sni_host, SSL_R_MINVER, SSL_R_CALIST, sslerr, sizeof(sslerr), &tmp, (char*)tb+TMPBUFSZ/2, TMPBUFSZ/2, 0); }
      if(fd>=0) break;
      dynbuf_printf(&$->warnlog, "fail, errno=%d errmsg=%s\n", errno, sslerr);
    }
    
    if(fd<0) {
      if(https && !errno) ERRF($, err_net, "connect(%s/%s) failed, errmsg: %s", addrstr, sni_host, sslerr);
      else                ERRF($, err_net, "connect(%s/%s) failed, errno %d (%s)", addrstr, sni_host, errno, strerror(errno));
      return -1;
    }
    dashboard_connected_target();
  }
  
  /* send request */
  if(lst_conn_s_send_all(fd, $->req.buf, $->req.len)<0) { ERRF($, err_net, "send request failed, errno %d (%s)", errno, strerror(errno)); close(fd); return -1; }
  if(data && data->total) {
    if(!data->use_fd) {
      if(lst_conn_s_send_all(fd, data->d, data->total)<0) { ERRF($, err_net, "send request body failed, errno %d (%s)", errno, strerror(errno)); close(fd); return -1; }
    } else {
      if(lseek(data->fd, 0, SEEK_SET)<0) { ERRF($, err_net, "reading back request body from tmpfile failed, lseek() errno %d (%s)", errno ,strerror(errno)); close(fd); return -1; }
      for(rbpos=0; rbpos<data->total; ) {
        tmpr = read_safe(data->fd, tb, TMPBUFSZ);
        if(tmpr<0) { ERRF($, err_net, "reading back request body from tmpfile failed, read() errno %d (%s), after %lu of %lu", errno, strerror(errno), (ulong)rbpos, (ulong)data->total); close(fd); return -1; }
        if(tmpr==0) { ERRF($, err_net, "reading back request body from tmpfile failed (EOF) after %lu of %lu", (ulong)rbpos, (ulong)data->total); close(fd); return -1; }
        if(lst_conn_s_send_all(fd, tb, tmpr)<0) { ERRF($, err_net, "send request body failed, errno %d (%s)", errno, strerror(errno)); close(fd); return -1; }
        rbpos += tmpr;
      }
    }
  }

  /* receive reply */
  {
    char *resp, *hdr, save;
    size_t csz, k;
    unsigned status;
    size_t hpos;
    int hhr, first;
    size_t recvlen;
    
    recvlen = 0;
    k = 0;
    do {
      if(recvlen>=TMPBUFSZ) { ERRS($, err_bad, "response header length exceeds 32767 bytes"); close(fd); return -1; }
      csz = TMPBUFSZ-recvlen;
      if(csz>10000) csz=10000;
      tmpr = lst_conn_recv(fd, tb+recvlen, csz);
      if(tmpr<0) { ERRS($, err_bad, "connection closed before response header fully received"); close(fd); return -1; }
      if(!tmpr) continue;
      recvlen += tmpr;
      for(csz=0; csz<recvlen; csz++) {
        if(recvlen-csz>=4 && tb[csz]=='\r' && tb[csz+1]=='\n' && tb[csz+2]=='\r' && tb[csz+3]=='\n') { csz+=2; k=2; break; }
        if(recvlen-csz>=3 && tb[csz]=='\n' && tb[csz+1]=='\r' && tb[csz+2]=='\n') { csz+=1; k=2; break; }
        if(recvlen-csz>=3 && tb[csz]=='\r' && tb[csz+1]=='\n' && tb[csz+2]=='\n') { csz+=2; k=1; break; }
        if(recvlen-csz>=2 && tb[csz]=='\n' && tb[csz+1]=='\n') { csz+=1; k=1; break; }
      }
    } while(!k);
    dynbuf_init_data(&$->resp, tb, csz+1); /* k>0 allows us to safely add 1 */
    resp = (char*)$->resp.buf;
    resp[csz] = 0;
    if(recvlen -= csz+k) bcopy(tb+csz+k, tb, recvlen);
    /* parse headers */
    if(strncmp(resp,"HTTP/1.0 ",9) && strncmp(resp,"HTTP/1.1 ",9)) { ERRS($, err_bad, "bad protocol in response header"); close(fd); return -1; }
    $->http = status = atoi(resp+9);
    if(status<100 || status>999)  { ERRF($, err_badst, "bad status (%u) in response header",status); close(fd); return -1; }
    if(status>=400 && status<=599)  dynbuf_printf(&$->warnlog, "error status (%u) in response header\n",status);
    first = 1;
    while(csz) {
      hdr = resp;
      for(hpos=0; hpos<csz && resp[hpos]!='\n'; hpos++);
      if(hpos>=csz) abort(); /* this should not happen due to code above */
      resp += hpos+1;
      csz -= hpos+1;
      if(hpos && hdr[hpos-1]=='\r') hpos--;
      save = hdr[hpos];
      hdr[hpos] = 0;
      if(first) { $->statusline = xstrdup(hdr); first=0; hhr=0; }
      else hhr = hh_add(&$->hdrs, hdr);
      hdr[hpos] = save;
      if(hhr<0) dynbuf_printf(&$->warnlog, "bad response header - errcode %d\n",hhr);
    }
    if(value=hh_get(&$->hdrs,"Content-Type")) {
      dashboard_set_ctype(value);
    }
    if(value=hh_get(&$->hdrs,"Content-Length")) {
      $->clen = strtoull(value,NULL,10);
      dashboard_set_clen($->clen);
      if(!$->clen) {
        hc_init(&$->content, 0);
        $->recvlen = recvlen;
        $->R = status_ok;
        close(fd);
        return 0;
      }
    }
    /* receive rest */
    if(is__raw && status>=200 && status<=299) {
      hc_init(&$->content, 0);
      hc_add(&$->content, tb, recvlen); // TODO: error check
      $->R = status_raw;
      $->socket = fd;
      return 0;
    }
    hc_init(&$->content, 0);
    $->recvlen = recvlen;
    $->R = status_rcv;
    $->socket = fd;
    return http_request_recv_all($);
  }
}

static int http_request_recv_all(t_http_request * $) {
  int is_chunked, fd;
  char const * value;
  unsigned char *tb;
  char *zz;
  size_t recvlen, pos, recvsz;
  t_http_content * hcres;
  unsigned long chunksize;

  assert($->R==status_rcv);
  fd = $->socket;
  tb = $->tmpbuf;
  recvlen = $->recvlen;
  hcres = &$->content;
  is_chunked = 0;
  if(value=hh_get(&$->hdrs,"Transfer-Encoding")) {
    if(!strcmp(value,"chunked")) is_chunked=1;
    else { ERRS($, err_bad, "unsupported Transfer-Encoding"); close(fd); return -1; }
  }

  if(is_chunked) {
    int state;

    state = 1; // 0=waiting for \r\n before size; 1=reading size; 2=reading remaining data; 3=waiting for last \r\n
    while(1) {
oh:   switch(state) {
      case 0:
        if(recvlen<2) break;
        if(tb[0]!='\r' || tb[1]!='\n') { ERRS($, err_bad, "malformed chunked encoding (1)"); close(fd); return -1; }
        tb += 2;
        recvlen -= 2;
        state = 1;
      case 1:
        for(pos=0; pos+2<=recvlen; pos++) if(tb[pos]=='\r' && tb[pos+1]=='\n') {
          tb[pos]=0;
          if(zz=strchr((char*)tb,';')) *zz=0;
          if(zz=strchr((char*)tb,' ')) *zz=0;
          if(str_hex_to_ulong((char*)tb,&chunksize)<0) { ERRS($, err_bad, "malformed chunked encoding (2)"); close(fd); return -1; }
          tb += pos+2;
          recvlen -= pos+2;
          if(chunksize) state=2; else state=3;
          goto oh;
        }
        break;
      case 2:
        pos = recvlen;
        if(pos>chunksize) pos=chunksize;
        if(hc_add(hcres, tb, pos)<0) { ERRS($, err_bad, "not enough storage space to save response body"); close(fd); return -1; }
        dashboard_set_dlen(hcres->cur);
        if(hcres->cur>10000000) { ERRS($, err_bad, "response body length exceeds 10Mbytes"); close(fd); return -1; }
        tb += pos;
        recvlen -= pos;
        chunksize -= pos;
        if(!chunksize) { state=0; continue; }
        assert(!recvlen);
        break;
      case 3:
        if(recvlen<2) break;
        if(tb[0]!='\r' || tb[1]!='\n') { ERRS($, err_bad, "malformed chunked encoding (3)"); close(fd); return -1; }
        tb += 2;
        recvlen -= 2;
        // TODO: deal with remaining data if any // will be needed for keep-alive
        close(fd);
        $->R = status_ok;
        return 0;
      }
      if(tb!=$->tmpbuf) { if(recvlen) bcopy(tb, $->tmpbuf, recvlen); tb = $->tmpbuf; }

      pos = TMPBUFSZ-recvlen;
      if(!pos) { ERRS($, err_bad, "malformed chunked encoding (4)"); close(fd); return -1; }
      if(lst_conn_s_recv(fd, tb+recvlen, pos, &recvsz)<0) { ERRS($, err_bad, "connection closed before chunked data fully received"); close(fd); return -1; }
      recvlen += recvsz;
    }
  }
  
  if(hc_add(hcres, tb, recvlen)<0) { ERRS($, err_bad, "not enough storage space to save response body"); close(fd); return -1; }
  while(1) {
    if(lst_conn_s_recv(fd, tb, TMPBUFSZ, &recvsz)<0) break; /* connections looks closed for some reason */
    if(!recvsz) continue;
    if(hc_add(hcres, tb, recvsz)<0) { ERRS($, err_bad, "not enough storage space to save response body"); close(fd); return -1; }
    dashboard_set_dlen(hcres->cur);
    if(hcres->cur>10000000) { ERRS($, err_bad, "response body length exceeds 10Mbytes"); close(fd); return -1; }
  }
  close(fd);
  $->R = status_ok;
  return 0;
}

#undef ERRS
#undef ERRF

