/*
 * This file is part of fproxy - multipurpose http(s) proxy server
 * handle/rules.c - rules-files parse and lookup
 * Copyright (c) 2018-2021 firk (firk@cantconnect.ru)
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. The origin of this software must not be misrepresented; you must
 *    not claim that you wrote the original software.
 * 4. Altered versions in any form must be plainly marked as such, and
 *    must not be misinterpreted as being the original software.
 *
 * This software is provided by the author and contributors `as is'
 * without any express or implied warranty.
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <strings.h>
#include <fcl/types.h>
#include "../util.h"
#include "../eventlog.h"
#include "rules.h"

/*
# format
# {w/sub|nosub|all}\t{all|spec}\thostname hostname hostname, ...
# w/sub means block both hostname and *.hostname and *.*.hostname and so on
# nosub means block only hostname and not sub-domains
# all means match any host name at all, hostnames list ignored
# all means block all url on this hostname(s)
# spec means block only urls that specified in next lines in format:
# \t{exact|pref|wpre|query}\turl
# exact means block only exact match
# pref means block all urls that starting with specified line
# wpre means block "url", "url/anything" and "url?anything"
# query means block "url" and "url?anything" or "url&anything" if url contains "?"
# ext means block by "...extension" or "...extension?..."
*/

#define RULES_DOMAIN_WITH_SUB 1
#define RULES_DOMAIN_ANY      2
#define RULES_DOMAIN_ANYURL   4
#define RULES_DOMAIN_ALLOW    16
#define RULES_URL_EXACT 0
#define RULES_URL_PREF  1
#define RULES_URL_WPRE  2
#define RULES_URL_QUERY 3
#define RULES_URL_CTYPE 4
#define RULES_URL_EXT   5

#define _MAX_DOMAINS 10000
#define _MAX_HOSTS 16
#define _MAX_URLS 1000

typedef struct {
  unsigned int type;
  char const * url;
} t_rules_url;
typedef struct {
  unsigned int flags;
  unsigned int n_hosts;
  char const * hosts[_MAX_HOSTS];
  unsigned int n_urls;
  t_rules_url * urls;
} t_rules_domain;
typedef struct {
  unsigned int n_domains;
  t_rules_domain * * domains;
} t_rules_list;

static t_rules_domain * rules_add_domain(t_rules_list * r) {
  t_rules_domain * d;
  unsigned int n;
  t_rules_domain * * domains;
  n = r->n_domains;
  if(n>=_MAX_DOMAINS) return NULL;
  domains = realloc(r->domains, (n+1)*sizeof(t_rules_domain*));
  if(!domains) return NULL;
  r->domains = domains;
  d = xmallocz(sizeof(*d));
  domains[n] = d;
  r->n_domains = n+1;
  return d;
}

static int domain_add_host(t_rules_domain * d, char const * host) {
  unsigned int n;
  n = d->n_hosts;
  if(n>=_MAX_HOSTS) return -1;
  d->hosts[n] = xstrdup(host);
  d->n_hosts = n+1;
  return 0;
}

static int domain_add_url(t_rules_domain * d, unsigned int type, char const * url) {
  unsigned int n;
  t_rules_url * urls;
  n = d->n_urls;
  if(n>=_MAX_URLS) return -1;
  urls = realloc(d->urls, (n+1)*sizeof(t_rules_url));
  if(!urls) return -1;
  d->urls = urls;
  urls[n].type = type;
  urls[n].url = xstrdup(url);
  d->n_urls = n+1;
  return 0;
}

static void rules_list_clear(t_rules_list * r) {
  unsigned int n, n1, n2, n3;
  t_rules_domain * d;
  t_rules_domain * * domains;
  t_rules_url * urls;
  if(domains=r->domains) {
    for(n=r->n_domains; n--; ) {
      d = domains[n];
      for(n1=d->n_hosts; n1--; ) free((void*)d->hosts[n1]);
      if(urls=d->urls) {
        for(n1=d->n_urls; n1--; ) free((void*)urls[n1].url);
        free(urls);
      }
      free(d);
    }
    free(domains);
  }
  bzero(r, sizeof(*r));
}

static int rules_list_load(t_rules_list * r, char const * fn) {
  FILE * fp;
  char buf[1024];
  char *url, *p, *p2;
  size_t len;
  unsigned int ln, type;
  t_rules_domain * d;
  
  d = NULL;
  bzero(r, sizeof(*r));
  if(!(fp=fopen(fn,"r"))) { eventlog(eventlog_level_error,__FUNCTION__,"error opening rules file \"%s\" for reading", fn); return -1; }
  ln = 0;
  while(fgets(buf, sizeof(buf), fp)) {
    ln++;
    len = strlen(buf);
    if(!len || buf[len-1]!='\n') { eventlog(eventlog_level_error,__FUNCTION__,"line too long or missing end-of-line at \"%s\":%u", fn, ln); goto fail; }
    buf[--len] = 0;
    for(len=0; buf[len] && buf[len]!='#'; len++);
    while(len && buf[len-1]==' ') len--;
    buf[len] = 0;
    if(!len) continue;
    if(buf[0]=='\t') {
      p = buf+1;
      if(*p=='\t') p++;
      if(!d) { eventlog(eventlog_level_error,__FUNCTION__,"url specification w/o domain at \"%s\":%u", fn, ln); goto fail; }
      if(url=skip_prefix(p,"exact\t")) type=RULES_URL_EXACT;
      else if(url=skip_prefix(p,"pref\t")) type=RULES_URL_PREF;
      else if(url=skip_prefix(p,"wpre\t")) type=RULES_URL_WPRE;
      else if(url=skip_prefix(p,"query\t")) type=RULES_URL_QUERY;
      else if(url=skip_prefix(p,"ctype\t")) type=RULES_URL_CTYPE;
      else if(url=skip_prefix(p,"ext\t")) type=RULES_URL_EXT;
      else { eventlog(eventlog_level_error,__FUNCTION__,"url type unknown at \"%s\":%u", fn, ln); goto fail; }
      if(domain_add_url(d, type, url)<0) { eventlog(eventlog_level_error,__FUNCTION__,"url - too many urls for domain at \"%s\":%u", fn, ln); goto fail; }
      continue;
    }
    if(p2=skip_prefix(buf,"allow\t")) type=RULES_DOMAIN_ALLOW;
    else if(p2=skip_prefix(buf,"deny\t")) type=0;
    else if(p2=skip_prefix(buf,"block\t")) type=0;
    else { eventlog(eventlog_level_error,__FUNCTION__,"rule type unknown at \"%s\":%u", fn, ln); goto fail; }
    if(p=skip_prefix(p2,"w/sub\t")) type |= RULES_DOMAIN_WITH_SUB;
    else if(p=skip_prefix(p2,"nosub\t")) ;
    else if(p=skip_prefix(p2,"all\t")) type |= RULES_DOMAIN_ANY;
    else { eventlog(eventlog_level_error,__FUNCTION__,"domain type unknown at \"%s\":%u", fn, ln); goto fail; }
    if(p2=skip_prefix(p,"all\t")) type |= RULES_DOMAIN_ANYURL;
    else if(p2=skip_prefix(p,"spec\t")) ;
    else { eventlog(eventlog_level_error,__FUNCTION__,"domain url-type unknown at \"%s\":%u", fn, ln); goto fail; }
    if(!(d=rules_add_domain(r))) { eventlog(eventlog_level_error,__FUNCTION__,"domain - too many domains in entire config at \"%s\":%u", fn, ln); goto fail; }
    d->flags = type;
    while(*p2) {
      p = strchr(p2,' ');
      if(p) *(p++) = 0;
      if(domain_add_host(d, p2)<0) { eventlog(eventlog_level_error,__FUNCTION__,"domain - too many hostnames in line at \"%s\":%u", fn, ln); goto fail; }
      if(!p) break;
      while(*p==' ') p++;
      p2 = p;
    }
  }
  fclose(fp);
  return 0;
fail:
  rules_list_clear(r);
  fclose(fp);
  return -1;
}

/* 1=matched to w/sub rule, 2=matched to exact rule, 3=match to all rule */
static int host_match(t_rules_domain * d, char const * host) {
  int sub;
  unsigned int nh, mh;
  char const * h;
  size_t hostlen, hlen;
  
  if(d->flags & RULES_DOMAIN_ANY) return 3;
  if(d->flags & RULES_DOMAIN_WITH_SUB) sub=1; else sub=0;
  nh = d->n_hosts;
  hostlen = strlen(host);
  for(mh=0; mh<nh; mh++) {
    h = d->hosts[mh];
    hlen = strlen(h);
    if(hostlen==hlen && !strcasecmp(h, host)) return (sub?1:2);
    if(sub && hostlen>hlen && host[hostlen-hlen-1]=='.' && !strcasecmp(h, host+(hostlen-hlen))) return 1;
  }
  return 0;
}

/* content types looks like: "CATEGORY/DETAIL; PARAMS" with "; PARAMS" part may be missing */
static int ctype_match(char const * pattern, char const * ctype) {
  char oldc, c;
  if(!ctype || !*ctype) return 0;
  oldc = 0;
  while(c=*(pattern++)) {
    oldc = c;
    if(c!=*ctype) return 0;
    ctype++;
  }
  c = *ctype;
  if(!c || c==';') return 1;
  if(oldc=='/') return 1;
  return 0;
}

static int url_match(t_rules_url * d, char const * url, char const * ctype) {
  size_t n, ul;
  char const *u1, *u2;
  u1 = d->url;
  if(d->type==RULES_URL_CTYPE) return ctype_match(u1, ctype); /* special case */
  if(d->type==RULES_URL_EXT) {
    ul = strlen(u1);
    for(n=0; url[n] && url[n]!='?'; n++);
    return (n>=ul && !memcmp(url+(n-ul), u1, ul));
  }
  u2 = skip_prefix((char*)url, u1);
  if(!u2) return 0; /* prefix mismatched */
  if(!*u2) return 1; /* exact match */
  switch(d->type) {
  default: /* RULES_URL_EXACT */
    return 0;
  case RULES_URL_PREF:
    return 1;
  case RULES_URL_WPRE:
    return (*u2=='?' || *u2=='/');
  case RULES_URL_QUERY:
    for(n=0; u1[n] && u1[n]!='?'; n++);
    if(u1[n]=='?') return (*u2=='&');
    else return (*u2=='?');
  }
}


/* 0=no match, 1-6=matched block rule, 11-16=matched allow rule
   1=matched to w/sub rule, 2=matched to exact rule, 3=match to all rule
   same+3 (4,5,6) means that url match was also used
  */
static int rules_list_match(t_rules_list * r, char const * host, char const * url, char const * ctype) {
  t_rules_domain * d;
  unsigned int nd, nu, md, mu;
  int res;
  nd = r->n_domains;
  for(md=0; md<nd; md++) {
    d = r->domains[md];
    if(!(res=host_match(d, host))) continue;
    if(d->flags & RULES_DOMAIN_ALLOW) res += 10;
    if(d->flags & RULES_DOMAIN_ANYURL) return res;
    nu = d->n_urls;
    for(mu=0; mu<nu; mu++) if(url_match(d->urls+mu, url, ctype)) return res+3;
  }
  return 0;
}

extern char const * rules_match_name(int r) {
  switch(r) {
  default: return "no match";
  case 1:  return "(sub)domain";
  case 2:  return "exact domain";
  case 3:  return "[all-matching rule]";
  case 4:  return "url on (sub)domain";
  case 5:  return "url on exact domain";
  case 6:  return "url on any domain";
  case 100:return "(default deny)";
  }
}

static t_rules_list list_block, list_cookie, list_cache, list_is_raw, list_save, list_sniff, list_rproxy;
extern void rules_load(void) {
  if(rules_list_load(&list_block, "rules_block.txt")<0) {
    eventlog(eventlog_level_fatal,__FUNCTION__,"can't load BLOCK rules");
    exit(-1);
  }
  if(rules_list_load(&list_cookie, "rules_cookie.txt")<0) {
    eventlog(eventlog_level_fatal,__FUNCTION__,"can't load COOKIE rules");
    exit(-1);
  }
  if(rules_list_load(&list_cache, "rules_cache.txt")<0) {
    eventlog(eventlog_level_fatal,__FUNCTION__,"can't load CACHE rules");
    exit(-1);
  }
  if(rules_list_load(&list_is_raw, "rules_raw.txt")<0) {
    eventlog(eventlog_level_fatal,__FUNCTION__,"can't load IS_RAW rules");
    exit(-1);
  }
  if(rules_list_load(&list_save, "rules_save.txt")<0) {
    eventlog(eventlog_level_fatal,__FUNCTION__,"can't load SAVE rules");
    exit(-1);
  }
  if(rules_list_load(&list_sniff, "rules_sniff.txt")<0) {
    eventlog(eventlog_level_fatal,__FUNCTION__,"can't load SNIFF rules");
    exit(-1);
  }
  if(rules_list_load(&list_rproxy, "rules_rproxy.txt")<0) {
    eventlog(eventlog_level_fatal,__FUNCTION__,"can't load RPROXY rules");
    exit(-1);
  }
}

extern int is_blocked(char const * host, char const * url) {
  int r;
  r = rules_list_match(&list_block, host, url, NULL);
  if(!r) return 100;
  if(r>=1 && r<=6) return r;
  return 0;
}

extern int block_cookie(char const * host, char const * url) {
  int r;
  r = rules_list_match(&list_cookie, host, url, NULL);
  if(!r) return 100;
  if(r>=1 && r<=6) return r;
  return 0;
}

extern int need_cache(char const * host, char const * url, char const * ctype) {
  int r;
  r = rules_list_match(&list_cache, host, url, ctype);
  if(r>=11 && r<=16) return r-10;
  return 0;
}

extern int is_raw(char const * host, char const * url) {
  int r;
  r = rules_list_match(&list_is_raw, host, url, NULL);
  if(r>=11 && r<=16) return r-10;
  return 0;
}

extern int is_save(char const * host, char const * url) {
  int r;
  r = rules_list_match(&list_save, host, url, NULL);
  if(r>=11 && r<=16) return r-10;
  return 0;
}

extern int is_sniff(char const * host, char const * url) {
  int r;
  r = rules_list_match(&list_sniff, host, url, NULL);
  if(r>=11 && r<=16) return r-10;
  return 0;
}

extern int is_rproxy(char const * host, char const * url) {
  int r;
  r = rules_list_match(&list_rproxy, host, url, NULL);
  if(r>=11 && r<=16) return r-10;
  return 0;
}

