/*
 * This file is part of fproxy - multipurpose http(s) proxy server
 * handle/rules.resolv.c - rules_resolv.txt parse and lookup
 * Copyright (c) 2018 firk (firk@cantconnect.ru)
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. The origin of this software must not be misrepresented; you must
 *    not claim that you wrote the original software.
 * 4. Altered versions in any form must be plainly marked as such, and
 *    must not be misinterpreted as being the original software.
 *
 * This software is provided by the author and contributors `as is'
 * without any express or implied warranty.
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <strings.h>
#include <fcl/types.h>
#include <fcl/strparse.h>
#include <libsimpletcp/libsimpletcp.h>
#include "../util.h"
#include "../eventlog.h"
#include "rules.resolv.h"

/*
# format
# hostspec hostspec hostspec
# =>\tconnspec
# hostspec is [http://|https://]hostname[:port][/prefix]
# connspec is [http://|https://]ip[:port][/prefix]
# unspecified protocol in hostspec = match both http/https
# unspecified protocol in connspec = keep original protocol
# unspecified port in hostspec = match any port
# unspecified port in connspec = (allowed only if protocol also unspecified) keep original port
# /prefix is unsupported for now
*/

#define _MAX_HOSTS 16
#define _MAX_DOMAINS 10000

typedef struct {
  char const * name;
  char const * prefix;
  uint16 port; /* 0=any */
  char proto; /* 0=any, 1=http, 2=https */
} t_rules_host;
typedef struct {
  unsigned int n_hosts;
  t_rules_host * hosts[_MAX_HOSTS];
  char const * new_prefix;
  uint32 new_addr;
  uint16 new_port;
  char   new_proto;
} t_rules_domain;
typedef struct {
  unsigned int n_domains;
  t_rules_domain * * domains;
} t_rules_list;

static t_rules_domain * rules_add_domain(t_rules_list * r) {
  t_rules_domain * d;
  unsigned int n;
  t_rules_domain * * domains;
  n = r->n_domains;
  if(n>=_MAX_DOMAINS) return NULL;
  domains = realloc(r->domains, (n+1)*sizeof(t_rules_domain*));
  if(!domains) return NULL;
  r->domains = domains;
  d = xmallocz(sizeof(*d));
  domains[n] = d;
  r->n_domains = n+1;
  return d;
}

/* -1=too many hosts; -2=syntax error */
static int domain_add_host(t_rules_domain * d, char const * hostspec) {
  unsigned int n;
  t_rules_host h;
  char *hs, *p0, *p;
  n = d->n_hosts;
  if(n>=_MAX_HOSTS) return -1;
  bzero(&h, sizeof(h));
  p0 = hs = xstrdup(hostspec);
  if(p=strstr(hs,"://")) {
    *p = 0;
    if(!strcmp(hs,"http")) h.proto = 1;
    else if(!strcmp(hs,"https")) h.proto = 2;
    else { free(p0); return -2; }
    hs = p+3;
  }
  if(p=strchr(hs,'/')) {
    h.prefix = xstrdup(p);
    *p = 0;
  }
  if(p=strchr(hs,':')) {
    *p = 0;
    if(str_to_uint16(p+1, &h.port)<0 || !h.port) { free(p0); if(h.prefix) free((void*)h.prefix); return -2; }
  }
  h.name = xstrdup(hs);
  free(p0);
  d->hosts[n] = xmemdup(&h, sizeof(h));
  d->n_hosts = n+1;
  return 0;
}

/* -2=syntax error */
static int domain_parse_connspec(t_rules_domain * d, char const * connspec) {
  char *cs, *p0, *p;
  char const * pref;
  
  pref = NULL;
  p0 = cs = xstrdup(connspec);
  if(p=strstr(cs,"://")) {
    *p = 0;
    if(!strcmp(cs,"http")) d->new_proto = 1;
    else if(!strcmp(cs,"https")) d->new_proto = 2;
    else { free(p0); return -2; }
    cs = p+3;
  }
  if(p=strchr(cs,'/')) {
    pref = xstrdup(p);
    *p = 0;
  }
  if(p=strchr(cs,':')) {
    *p = 0;
    if(str_to_uint16(p+1, &d->new_port)<0 || !d->new_port) { free(p0); if(pref) free((void*)pref); return -2; }
  } else if(d->new_proto) { /* unspecified port is allowed only with unspecified protocol */
    free(p0); if(pref) free((void*)pref); return -2;
  }
  if(cs[0] && !(d->new_addr = lst_parse_ipstr(cs))) { free(p0); if(pref) free((void*)pref); return -2; }
  free(p0);
  d->new_prefix = pref;
  return 0;
}

static void rules_list_clear(t_rules_list * r) {
  unsigned int n, n1, n2, n3;
  t_rules_domain * d;
  t_rules_domain * * domains;
  t_rules_host * host;
  if(domains=r->domains) {
    for(n=r->n_domains; n--; ) {
      d = domains[n];
      for(n1=d->n_hosts; n1--; ) {
        host = d->hosts[n1];
        if(host->name) free((void*)host->name);
        if(host->prefix) free((void*)host->prefix);
        free(host);
      }
      if(d->new_prefix) free((void*)d->new_prefix);
      free(d);
    }
    free(domains);
  }
  bzero(r, sizeof(*r));
}

static int resolv_rules_list_load(t_rules_list * r, char const * fn) {
  FILE * fp;
  char buf[1024];
  char *url, *p, *p2;
  size_t len;
  unsigned int ln, type;
  t_rules_domain * d;
  
  d = NULL;
  bzero(r, sizeof(*r));
  if(!(fp=fopen(fn,"r"))) { eventlog(eventlog_level_error,__FUNCTION__,"error opening resolv-rules file \"%s\" for reading", fn); return -1; }
  ln = 0;
  while(fgets(buf, sizeof(buf), fp)) {
    ln++;
    len = strlen(buf);
    if(!len || buf[len-1]!='\n') { eventlog(eventlog_level_error,__FUNCTION__,"line too long or missing end-of-line at \"%s\":%u", fn, ln); goto fail; }
    buf[--len] = 0;
    for(len=0; buf[len] && buf[len]!='#'; len++);
    while(len && buf[len-1]==' ') len--;
    buf[len] = 0;
    if(!len) continue;
    if(buf[0]=='=' && buf[1]=='>' && buf[2]=='\t') {
      if(!d) { eventlog(eventlog_level_error,__FUNCTION__,"connspec w/o hostspec in previous line at \"%s\":%u", fn, ln); goto fail; }
      if(domain_parse_connspec(d, buf+3)<0) { eventlog(eventlog_level_error,__FUNCTION__,"connspec syntax error at \"%s\":%u", fn, ln); goto fail; }
      d = NULL;
      continue;
    }
    if(d) { eventlog(eventlog_level_error,__FUNCTION__,"hostspec after another hostspec w/o connspec at \"%s\":%u", fn, ln); goto fail; }
    if(!(d=rules_add_domain(r))) { eventlog(eventlog_level_error,__FUNCTION__,"domain - too many domains in entire config at \"%s\":%u", fn, ln); goto fail; }
    p2 = buf;
    while(*p2) {
      p = strchr(p2,' ');
      if(p) *(p++) = 0;
      if(domain_add_host(d, p2)<0) { eventlog(eventlog_level_error,__FUNCTION__,"domain - too many hostnames or syntax error in line at \"%s\":%u", fn, ln); goto fail; }
      if(!p) break;
      while(*p==' ') p++;
      p2 = p;
    }
  }
  fclose(fp);
  return 0;
fail:
  rules_list_clear(r);
  fclose(fp);
  return -1;
}

/* 1=matched */
static int host_match(t_rules_domain * d, char const * host, int is_https, uint16 port) {
  int sub;
  unsigned int nh, mh;
  t_rules_host * h0;
  char const * h;
  size_t hostlen, hlen;
  
  nh = d->n_hosts;
  hostlen = strlen(host);
  for(mh=0; mh<nh; mh++) {
    h0 = d->hosts[mh];
    h = h0->name;
    hlen = strlen(h);
    if((!h0->proto || h0->proto==1 && !is_https || h0->proto==2 && is_https) && (!h0->port || h0->port==port) && !h0->prefix){
      if(h[0]=='*' && h[1]==0) return 1;
      if(hostlen==hlen && !strcasecmp(h, host)) return 1;
      if(h[0]=='*' && h[1]=='.' && hostlen>=hlen && !strcasecmp(h+1, host+(hostlen-hlen+1))) return 1;
    }
  }
  return 0;
}

static t_rules_domain * search_resolv_rule(t_rules_list * r, char const * host, int is_https, uint16 port) {
  t_rules_domain * d;
  unsigned int nd, md;
  nd = r->n_domains;
  for(md=0; md<nd; md++) {
    d = r->domains[md];
    if(host_match(d, host, is_https, port)) return d;
  }
  return NULL;
}



static t_rules_list list_resolv;
extern void resolv_rules_load(void) {
  if(resolv_rules_list_load(&list_resolv, "rules_resolv.txt")<0) {
    eventlog(eventlog_level_fatal,__FUNCTION__,"can't load RESOLV rules");
    exit(-1);
  }
}

extern int test_resolv_rule(char const * host, uint32 * addr, char * is_https, uint16 * port) {
  t_rules_domain * d;
  d = search_resolv_rule(&list_resolv, host, *is_https, *port);
  if(!d) return 0;
  if(d->new_port) *port = d->new_port;
  if(d->new_proto==1) *is_https = 0;
  else if(d->new_proto==2) *is_https = 1;
  *addr = d->new_addr;
  return d->new_addr?1:0; /* allow silent changing ofr https-flag and port w/o affecting name resolving */
}
