/*
 * This file is part of fproxy - multipurpose http(s) proxy server
 * helpers/cert_gen.c - server certificate generator
 * Copyright (c) 2018-2020 firk (firk@cantconnect.ru)
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. The origin of this software must not be misrepresented; you must
 *    not claim that you wrote the original software.
 * 4. Altered versions in any form must be plainly marked as such, and
 *    must not be misinterpreted as being the original software.
 *
 * This software is provided by the author and contributors `as is'
 * without any express or implied warranty.
 */

#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>
#include <string.h>
#include <errno.h>
#include <unistd.h>
#include <fcntl.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/file.h>
#include <openssl/evp.h>
#include <openssl/ssl.h>
#include <openssl/err.h>
#include <openssl/x509v3.h>
#include "cert_gen.h"

#define CERTS_DIR "./dyn-certs/"
#define CERTS_SERIAL "./dyn-cert-serial"

static void ssl_fail(char const * fmt, ...) {
  va_list arg;
  va_start(arg, fmt);
  fprintf(stderr, "ERROR: ");
  vfprintf(stderr, fmt, arg);
  fprintf(stderr, "\n");
  ERR_print_errors_fp(stderr);
  va_end(arg);
  // may be SSL_CTX_free() here?
  exit(-1);
}

static void raw_fail(char const * fmt, ...) {
  va_list arg;
  va_start(arg, fmt);
  fprintf(stderr, "ERROR: ");
  vfprintf(stderr, fmt, arg);
  fprintf(stderr, "\n");
  va_end(arg);
  // may be SSL_CTX_free() here?
  exit(-1);
}

/************************************************************************************************/

static long alloc_serial(void) {
  int fd, fd2, err;
  char buf[50];
  long serial;
  if((fd=open(CERTS_SERIAL,O_RDWR))<0) raw_fail("open serial file errno %d (%s)", errno, strerror(errno));
  do { err=flock(fd,LOCK_EX); } while(err<0 && errno==EINTR);
  if(err<0) raw_fail("lock serial file errno %d (%s)", errno, strerror(errno));
  err = read(fd, buf, sizeof(buf));
  if(err<0) raw_fail("read serial file errno %d (%s)", errno, strerror(errno));
  if(err<1 || err>20 || buf[err-1]!='\n') raw_fail("serial file malformed");
  buf[err-1] = 0;
  serial = atol(buf)+1;
  snprintf(buf, sizeof(buf), "%ld\n", serial);
  if(lseek(fd, 0, SEEK_SET)<0) raw_fail("rewind serial file errno %d (%s)", errno, strerror(errno));
  if(ftruncate(fd, 0)<0) raw_fail("truncate serial file errno %d (%s) - serial file may be now corrupt", errno, strerror(errno));
  if(write(fd, buf, strlen(buf))!=(int)strlen(buf)) raw_fail("write serial file failed - serial file now corrupt"); // what to do?
  if(close(fd)<0) raw_fail("close serial file errno %d (%s) - serial file may be now corrupt", errno, strerror(errno));
  return serial;
}

#define MAX_HOSTNAME_LEN 255 /* rfc1123 2.1 recommendation; actually DNS protocol limits length to 253 bytes */
   /* NOTES:
    1) will reject FQDN in form like www.example.net. (with finalizing dot), so finalizing dot should be stripped before
    2) will reject any uppercase characters, so they should be lowercased before
    3) allows '_' symbols which is not allowed by standards but in fact allowed by browsers and DNS servers
    */
static int hostname_valid(char const * h) {
  int i, j;
  char c;
  j = 0;
  for(i=0; i<MAX_HOSTNAME_LEN; i++) {
    c = h[i];
    if(c=='.') {
      if(!i || h[i-1]=='.' || h[i+1]=='.' || !h[i+1]) return 0;
      j=1;
      continue;
    }
    if(!c) return (i>=3)?1:0; /* not checking j==1 because of localhost (and possibly some other private domains) w/o dots */
    if(c>='0' && c<='9' || c>='a' && c<='z' || c=='_' || c=='-') continue;
    if(c=='*' && !i && h[1]=='.' && strchr(h+2,'.')) continue;
    return 0;
  }
  if(!h[i]) return 1;
  return 0;
}

/************************************************************************************************/

extern EVP_PKEY * read_pkey(char const * fn) {
  EVP_PKEY * pkey;
  FILE * fp;
  if(!(fp=fopen(fn,"rb"))) raw_fail("can't open pkey file \"%s\"", fn);
  if(!(pkey = PEM_read_PrivateKey(fp, NULL, NULL, NULL))) ssl_fail("ssl can't read pkey file \"%s\"", fn);
  fclose(fp);
  return pkey;
}

extern X509 * read_cert(char const * fn) {
  X509 * cert;
  FILE * fp;
  if(!(fp=fopen(fn,"rb"))) raw_fail("can't open cert file \"%s\"", fn);
  if(!(cert = PEM_read_X509(fp, NULL, NULL, NULL))) ssl_fail("ssl can't read cert file \"%s\"", fn);
  fclose(fp);
  return cert;
}

static int add_ext(X509 * cacert, X509 *cert, int nid, char *value) {
  X509_EXTENSION *ext;
  X509V3_CTX ctx;
  int r;
  X509V3_set_ctx_nodb(&ctx);
  X509V3_set_ctx(&ctx, cacert, cert, NULL, NULL, 0);
  if(!(ext = X509V3_EXT_conf_nid(NULL, &ctx, nid, value))) return 0;
  r = X509_add_ext(cert, ext, -1);
  X509_EXTENSION_free(ext);
  return r;
}


extern X509 * gen_dyncert(EVP_PKEY * cakey, X509 * cacert, EVP_PKEY * serverkey, char const * servername, int use_SAN) {
//  X509_REQ * req;
  X509_NAME * name;
  X509_NAME * caname;
  X509 * cert;
  ASN1_INTEGER * serial;
  char tmp[MAX_HOSTNAME_LEN+100];
  char const * CN;
/*
  if(!(req=X509_REQ_new())) ssl_fail("X509_REQ_new() failed");
  if(X509_REQ_set_version(req,2)<=0) ssl_fail("X509_REQ_set_version() failed");
  if(!(name=X509_REQ_get_subject_name(req))) ssl_fail("X509_REQ_get_subject_name() failed");
  if(X509_NAME_add_entry_by_txt(name,"C",MBSTRING_ASC,(unsigned char*)"RU",-1,-1,0)<=0 ||
     X509_NAME_add_entry_by_txt(name,"O",MBSTRING_ASC,(unsigned char*)"proxy",-1,-1,0)<=0 ||
     X509_NAME_add_entry_by_txt(name,"CN",MBSTRING_ASC,(unsigned char*)servername,-1,-1,0)<=0) ssl_fail("X509_NAME failed");
  if(X509_REQ_set_pubkey(req, ............)<=0) ssl_fail("X509_REQ_set_pubkey() failed");
  if(X509_REQ_sign(req, serverkey, EVP_sha256())<=0) ssl_fail("X509_REQ_sign() failed");
*/

  if(!hostname_valid(servername)) return NULL;
  snprintf(tmp,sizeof(tmp),"DNS:%s",servername);
  CN = servername;
  if(use_SAN) CN = "proxycert";
  if(!(cert=X509_new())) ssl_fail("X509_new() failed");
  if(X509_set_version(cert,2)<=0) ssl_fail("X509_set_version() failed");
  if(!(name=X509_get_subject_name(cert))) ssl_fail("X509_get_subject_name() failed");
  if(X509_NAME_add_entry_by_txt(name,"C",MBSTRING_ASC,(unsigned char*)"RU",-1,-1,0)<=0 ||
     X509_NAME_add_entry_by_txt(name,"O",MBSTRING_ASC,(unsigned char*)"proxy",-1,-1,0)<=0 ||
     X509_NAME_add_entry_by_txt(name,"CN",MBSTRING_ASC,(unsigned char*)CN,-1,-1,0)<=0) ssl_fail("X509_NAME failed");
  if(!(caname=X509_get_subject_name(cacert))) ssl_fail("X509_get_subject_name(ca) failed");
  if(X509_set_issuer_name(cert,caname)<=0) ssl_fail("X509_set_issuer_name() failed");
  /* private key as second arg? looks like ok - browser shows only public key in that place after this */
  if(X509_set_pubkey(cert, serverkey)<=0) ssl_fail("X509_set_pubkey() failed");
  serial = ASN1_INTEGER_new();
  ASN1_INTEGER_set(serial, alloc_serial());
  if(X509_set_serialNumber(cert, serial)<=0) ssl_fail("X509_set_serialNumber() failed");
  ASN1_INTEGER_free(serial);
  if(X509_gmtime_adj(X509_get_notBefore(cert),0)<=0 ||
     X509_gmtime_adj(X509_get_notAfter(cert),1000000)<=0) ssl_fail("X509_gmtime_adj() failed");
  if(use_SAN && !add_ext(cacert, cert, NID_subject_alt_name, tmp)) ssl_fail("add_ext() failed");
  if(X509_sign(cert, cakey, EVP_sha256())<=0) ssl_fail("X509_sign() failed");
  
  return cert;
}

extern X509 * gen_dyncert_cache(EVP_PKEY * cakey, X509 * cacert, EVP_PKEY * serverkey, char const * servername, int use_SAN) {
  char fn[100+MAX_HOSTNAME_LEN];
  char fn2[100+MAX_HOSTNAME_LEN];
  char *pfn2;
  struct stat st;
  FILE * fp;
  X509 * cert;
  
  if(!hostname_valid(servername)) return NULL;
  snprintf(fn, sizeof(fn), "%s%s", CERTS_DIR, servername);
  if(strlen(fn)+1==sizeof(fn)) return NULL;
  snprintf(fn2, sizeof(fn2), "%s.%u.%s", CERTS_DIR, (unsigned)getpid(), servername);
  /* prevent problems with creating temporary file; it protected from collisions by getpid(), and servername is only for debugging here and may be truncated */
  if(pfn2=strrchr(fn2,'/')) pfn2++; else pfn2=fn2;
  if(strlen(pfn2)>255) pfn2[255]=0;
  
  if(lstat(fn, &st)>=0 && time(NULL)-st.st_mtime<500000 && (fp=fopen(fn,"rb"))) {
    if(!(cert = PEM_read_X509(fp, NULL, NULL, NULL))) ssl_fail("ssl can't read cert-cache file \"%s\"", fn);
    fclose(fp);
    return cert;
  }
  cert = gen_dyncert(cakey, cacert, serverkey, servername, use_SAN);
  if(!(fp=fopen(fn2,"wb"))) raw_fail("can't open cert-cache file \"%s\" for writing, errno %d (%s)", fn2, errno, strerror(errno));
  if(PEM_write_X509(fp, cert)<=0) ssl_fail("ssl can't write cert-cache file \"%s\"", fn2);
  if(fclose(fp)<0) raw_fail("can't close cert-cache file \"%s\" after writing, errno %d (%s)", fn2, errno, strerror(errno));
  if(rename(fn2,fn)<0) raw_fail("can't rename cert-cache file \"%s\" => \"%s\" after writing, errno %d (%s)", fn2, fn, errno, strerror(errno));
  return cert;
}

