/*
 * This file is part of fproxy - multipurpose http(s) proxy server
 * helpers/cert_match.c - check that certificate matches hostname
 * Copyright (c) 2018-2019 firk (firk@cantconnect.ru)
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. The origin of this software must not be misrepresented; you must
 *    not claim that you wrote the original software.
 * 4. Altered versions in any form must be plainly marked as such, and
 *    must not be misinterpreted as being the original software.
 *
 * This software is provided by the author and contributors `as is'
 * without any express or implied warranty.
 */

#include <string.h>
#include <openssl/asn1.h>
#include <openssl/ssl.h>
#include <openssl/x509v3.h>
#include "cert_match.h"

#if OPENSSL_VERSION_NUMBER >= 0x10100000L
#define ASN1_STRING_data ASN1_STRING_get0_data
#endif

static int wildcard_host_match(char const * wildcard, char const * needed_name) {
  size_t len;

#ifdef DEBUG_SSL_CLIENT
  fprintf(stderr, "DEBUG: check wildcard '%s' against needed name '%s'\n", wildcard, needed_name);
#endif

  len = strlen(wildcard);
  if(!len || wildcard[len-1]=='.') return 0;
  if(wildcard[0]=='*' && wildcard[1]=='.') {
    wildcard+=2;
    if(strchr(wildcard,'*')) return 0;
    if(!strchr(wildcard,'.')) return 0;
    needed_name = strchr(needed_name,'.');
    if(!needed_name) return 0;
    return !strcasecmp(wildcard, needed_name+1);
  } else {
    if(strchr(wildcard,'*')) return 0;
    return !strcasecmp(wildcard, needed_name);
  }
  return 0;
}

static int wildcard_asn1_host_match(ASN1_STRING * asn1name, char const * needed_name) {
  int len;
  char name[1024];

  if(!asn1name) return CERT_MATCH_INTERNAL_BUG;
  len = ASN1_STRING_length(asn1name);
  if(len<0 || (unsigned)len>=sizeof(name)) return CERT_MATCH_MALFORMED_CERT;
  memcpy(name, (char const*)ASN1_STRING_data(asn1name), len);
  name[len] = 0;
  if(strlen(name)!=(unsigned)len) return CERT_MATCH_MALFORMED_CERT;
  if(wildcard_host_match(name, needed_name)) return CERT_MATCH_OK;
  return CERT_MATCH_NO;
}

extern int cert_match(X509 const * cert, char const * hostname) {
  STACK_OF(GENERAL_NAME) * SAN; /* TODO: what if X509_get_ext_d2i() returns something of another type? */
  GENERAL_NAME const * SAN_name;
  int n, i, res;
  
  X509_NAME * subj;
  X509_NAME_ENTRY * CN;
  ASN1_STRING * CN_value;
  
  if(!cert) return CERT_MATCH_MISSING_CERT;
  if(!hostname) return CERT_MATCH_MISSING_HOSTNAME;
  SAN = X509_get_ext_d2i((X509*)cert, NID_subject_alt_name, NULL /*&crit*/, NULL);
  if(SAN/* && CRIT>=0 */) {
    /* we shouldn't fallback to CommonName is SAN exist */
    n = sk_GENERAL_NAME_num(SAN);
    res = CERT_MATCH_NO;
    for(i=0; i<n; i++) {
      SAN_name = sk_GENERAL_NAME_value(SAN, i);
      if(SAN_name->type==GEN_DNS) {
        res = wildcard_asn1_host_match(SAN_name->d.dNSName, hostname);
        if(res!=CERT_MATCH_NO) break;
      }
    }
    sk_GENERAL_NAME_pop_free(SAN, GENERAL_NAME_free);
    return res;
  }

  if(!(subj = X509_get_subject_name((X509*)cert))) return CERT_MATCH_MALFORMED_CERT;
  n = X509_NAME_get_index_by_NID(subj, NID_commonName, -1);
  if(n<0) return CERT_MATCH_MISSING_CN;
  CN = X509_NAME_get_entry(subj, n);
  if(!CN) return CERT_MATCH_INTERNAL_BUG; /* this shouldn't happen */
  CN_value = X509_NAME_ENTRY_get_data(CN);
  if(!CN_value) return CERT_MATCH_INTERNAL_BUG; /* this shouldn't happen */
  
  return wildcard_asn1_host_match(CN_value, hostname);
}
