/*
 * This file is part of fproxy - multipurpose http(s) proxy server
 * helpers/cert_verify.c - check if certificate trusted or not
 * Copyright (c) 2020, 2022 firk (firk@cantconnect.ru)
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. The origin of this software must not be misrepresented; you must
 *    not claim that you wrote the original software.
 * 4. Altered versions in any form must be plainly marked as such, and
 *    must not be misinterpreted as being the original software.
 *
 * This software is provided by the author and contributors `as is'
 * without any express or implied warranty.
 */

#include <stdio.h>
#include <string.h>
#include <errno.h>
#include <unistd.h>
#include <fcntl.h>
#include <sys/file.h>
#include <fcl/types.h>
#include <fcrypt/sha256.h>
#include <openssl/ssl.h>
#include <openssl/x509.h>
#include <openssl/x509v3.h>
#include <openssl/asn1.h>
#include <openssl/err.h>
#include "http_download.h"
#include "cert_verify.h"

#if OPENSSL_VERSION_NUMBER >= 0x10100000L
#define ASN1_STRING_data ASN1_STRING_get0_data
#endif

/*
  as for ctx->untrusted:
    looks like it is a private variable not planned to be accessed by app; but looks to work fine with openssl 1.0.1 and 1.1.1
    openssl 1.0 has X509_STORE_CTX_set_chain() to set it and nothing to get it
    openssl 1.1 has X509_STORE_CTX_set0_untrusted() to set it and X509_STORE_CTX_get0_untrusted() to get it, and struct become private
  
  also don't know if untrusted chain can be freed after trusted chain built by X509_verify_cert()
  for safety don't free it; anyway this is one-run app and memory leaks will not stack
   */

#if OPENSSL_VERSION_NUMBER >= 0x10100000L
#define my_get_untrusted(ctx) X509_STORE_CTX_get0_untrusted(ctx)
#define my_set_untrusted(ctx, u) X509_STORE_CTX_set0_untrusted(ctx, u)
#define my_get_cert(ctx) X509_STORE_CTX_get0_cert(ctx)
#define my_get_chain(ctx) X509_STORE_CTX_get0_chain(ctx)
#else
#define my_get_untrusted(ctx) ((ctx)->untrusted)
#define my_set_untrusted(ctx, u) ((ctx->untrusted) = u)    /* X509_STORE_CTX_set_chain(ctx, u); */
#define my_get_cert(ctx) ((ctx)->cert)
#define my_get_chain(ctx) X509_STORE_CTX_get_chain(ctx)
#endif

static int do_download_aia_issuers = 0;
static int do_save_chains = 0;
static int do_dump_chains = 0;
static STACK_OF(X509) *saved_untrusted;
static STACK_OF(X509) *saved_trusted;

static void dump_chain(STACK_OF(X509) *chain);
#ifdef SSL_CLIENT_ALLOW_CERT_CACHE

#define CACHE_PATH "./cert-pin/cache/"
#define PROXY_SOCKET_PATH "./internal/http.sock"

/* TODO: what X509_STORE_add_lookup() does? may be it is useful here */

#define is_issued(issuer, x) (X509_check_issued((issuer), (x))==X509_V_OK)

static char * name_to_fname(char * dest, char const * certname) {
  /* uses dest[0..32+strlen(CACHE_PATH)] */
  static char const h[37] = "0123456789abcdefghijklmnopqrstuvwxyz";
  uint8 hash[32], cc;
  unsigned int jj, pl;
  
  pl = strlen(CACHE_PATH);
  memcpy(dest, CACHE_PATH, pl); dest += pl;
  sha256_hash(hash, certname, strlen(certname)+1);
  for(jj=0; jj<32; jj++) *(dest++) = h[hash[jj]%36];
  *(dest++) = '.';
  return dest;
}

static int read_all(int fd, unsigned char * dest, int len) {
  int r;
  while(len>0) {
    r = read(fd, dest, len);
    if(r<0) { if(errno==EINTR) continue; return -1; }
    if(!r) return -1;
    dest += r; len -= r;
  }
  return 0;
}

static int read_max(int fd, unsigned char * dest, int len) {
  int r, rr;
  rr = 0;
  while(len>0) {
    r = read(fd, dest, len);
    if(r<0) { if(errno==EINTR) continue; return -1; }
    if(!r) return rr;
    dest += r; len -= r; rr += r;
  }
  return rr;
}

static int write_all(int fd, unsigned char * dest, int len) {
  int r;
  while(len>0) {
    r = write(fd, dest, len);
    if(r<0) { if(errno==EINTR) continue; return -1; }
    if(!r) return -1;
    dest += r; len -= r;
  }
  return 0;
}

#define IF_ERR(X) while((X)<0) if(errno!=EINTR)
static void save_to_cache(X509 * cert) {
  X509_NAME * xn;
  char * name;
  char fname[100+33+50], *pfn;
  unsigned int jj;
  unsigned char * bincert;
  unsigned char * bincert2;
  int bincertlen;
  int fd;
  off_t fs;
  
  if(!(xn = X509_get_subject_name(cert))) return; /* TODO: error log? */
  if(!(name = X509_NAME_oneline(xn, NULL, 0))) return; /* TODO: error log? */
  pfn = name_to_fname(fname, name);
  OPENSSL_free(name);
  
  bincert = NULL;
  bincertlen = i2d_X509(cert, &bincert);
  if(bincertlen<0) { while(ERR_get_error()); return; } /* TODO: error log? */
  if(!bincertlen) { OPENSSL_free(bincert); return; } /* TODO: error log? */
  if(!(bincert2=OPENSSL_malloc(bincertlen))) { OPENSSL_free(bincert); return; } /* TODO: error log? */
  
  for(jj=0; jj<1000000; jj++) {
    snprintf(pfn, 50, "%u", jj);
    IF_ERR(fd = open(fname, O_RDWR|O_CREAT, 0644)) goto fail;
    IF_ERR(flock(fd, LOCK_EX)) { close(fd); goto fail; }
    IF_ERR(fs=lseek(fd, 0, SEEK_END)) { close(fd); goto fail; }
    if(!fs) {
      if(write_all(fd, bincert, bincertlen)<0) { ftruncate(fd, 0); close(fd); goto fail; } /* truncate but not unlink because unlink may cause race */
      close(fd);
#ifdef DEBUG_SSL_CLIENT
      fprintf(stderr, "DEBUG: cache-saved cert '%s'\n", fname);
#endif
      goto ok;
    }
    if(fs==bincertlen) {
      IF_ERR(lseek(fd, 0, SEEK_SET)) { close(fd); goto fail; }
      if(read_all(fd, bincert2, bincertlen)<0) { close(fd); goto fail; }
      if(!memcmp(bincert, bincert2, bincertlen)) { close(fd); goto ok; }
    }
    close(fd);
  }
fail:
#ifdef DEBUG_SSL_CLIENT
  pfn[0]='*'; pfn[1]=0;
  fprintf(stderr, "DEBUG: system error while cache-save cert '%s'\n", fname);
#endif
  OPENSSL_free(bincert); OPENSSL_free(bincert2); return;
ok: OPENSSL_free(bincert); OPENSSL_free(bincert2); return;
}

  /*  search issuer for cert (cache, AIA);
    return referenced X509 struct if found, NULL if not */
static X509 * search_issuer_from_aia(X509 * cert);
static X509 * search_issuer(X509 * cert) {
  X509_NAME * xn;
  char * name;
  char fname[100+33+50], *pfn;
  unsigned int jj;
  unsigned char buf[10000]; /* never seen certs > 2kb */
  unsigned char const * bb;
  int blen;
  int fd;
  X509 * issuer;
  
  if(!(xn = X509_get_issuer_name(cert))) return NULL; /* TODO: error log? */
  if(!(name = X509_NAME_oneline(xn, NULL, 0))) return NULL; /* TODO: error log? */
  pfn = name_to_fname(fname, name);
  OPENSSL_free(name);
  
  for(jj=0; jj<1000000; jj++) {
    snprintf(pfn, 50, "%u", jj);
    IF_ERR(fd = open(fname, O_RDONLY)) { if(errno==ENOENT) goto not_found; goto fail; }
    IF_ERR(flock(fd, LOCK_EX)) { close(fd); goto fail; }
    if((blen=read_max(fd, buf, sizeof(buf)))<0) { close(fd); goto fail; }
    close(fd);
    if(!blen || blen==sizeof(buf)) continue;
    bb = buf;
    if(!(issuer = d2i_X509(NULL, &bb, blen))) { while(ERR_get_error()); goto fail; }
    if(is_issued(issuer, cert)) return issuer;
    X509_free(issuer);
  }
fail:
#ifdef DEBUG_SSL_CLIENT
  fprintf(stderr, "DEBUG: system error while cache-load cert '%s'\n", fname);
#endif
  return NULL;
not_found:
#ifdef DEBUG_SSL_CLIENT
  fprintf(stderr, "DEBUG: cache miss at %s\n", fname);
#endif
  return search_issuer_from_aia(cert);
}

  /* try to use AIA; TODO: should handle all AIA entries, there may be more than one */
static X509 * search_issuer_from_aia(X509 * cert) {
#ifdef SSL_CLIENT_ALLOW_AIA
  int pos;
  AUTHORITY_INFO_ACCESS * aia;
  int len, i;
  ACCESS_DESCRIPTION * ad;
  char url[200];
  uint8 * bincert;
  size_t bincertlen;
  unsigned int bincertstatus;
  unsigned char const * bb;
  X509 * issuer;
  int rr;

  url[0] = 0;
  pos = -1;
#if 0
  while((pos = X509_get_ext_by_NID(cert, NID_info_access, pos))>=0) {
    X509_EXTENSION * ext;
    ext = X509_get_ext(cert, pos);
    aia = X509V3_ext_d2i(ext);
#endif

  while(aia = X509_get_ext_d2i(cert, NID_info_access, NULL, &pos)) {
    len = sk_ACCESS_DESCRIPTION_num(aia);
    for(i=0; i<len; i++) {
      ad = sk_ACCESS_DESCRIPTION_value(aia, i);
      if(OBJ_obj2nid(ad->method)==NID_ad_ca_issuers && ad->location->type==GEN_URI) {
        strncpy(url, (char const*)(ASN1_STRING_data(ad->location->d.uniformResourceIdentifier)), sizeof(url));
        if(url[sizeof(url)-1]) url[0]=0; /* TODO: proper error handle */
        if(url[0]) goto fin;
      }
    }
  }
fin:
  AUTHORITY_INFO_ACCESS_free(aia);

  if(!do_download_aia_issuers) {
    fprintf(stderr, "DEBUG: possible cert location: %s\n", url);
    return NULL;
  }
#ifdef DEBUG_SSL_CLIENT
  fprintf(stderr, "DEBUG: downloading AIA issuer cert: %s\n", url);
#endif
  
  rr = http_download(PROXY_SOCKET_PATH, url, 10000, &bincert, &bincertlen, &bincertstatus);
  if(rr==0) {
    bb = bincert;
    if(!(issuer = d2i_X509(NULL, &bb, bincertlen))) { while(ERR_get_error()); } else {
      if(is_issued(issuer, cert)) {
        save_to_cache(issuer);
        return issuer;
      }
      X509_free(issuer);
    }
  }
  fprintf(stderr, "DEBUG: downloading AIA issuer cert failed, reason %d\n", rr);
#endif
  return NULL; /* TODO: error log? */
}


/* check if cert issued by something from trust store, so we should stop completing chain */
static int is_pre_final(X509_STORE_CTX * ctx, X509 * cert) {
  X509 * issuer;
#if OPENSSL_VERSION_NUMBER >= 0x10100000L
  if(X509_STORE_CTX_get_get_issuer(ctx)(&issuer, ctx, cert)<=0) return 0;
#else
  if(ctx->get_issuer(&issuer, ctx, cert)<=0) return 0;
#endif
  X509_free(issuer);
  return 1;
}

/* in fact the only practical purpose of this function is to handle misconfigured webservers sending only one cert w/o chain */
/* TODO: support alternative chains if first matched fails; */
static STACK_OF(X509) * fix_chain(X509_STORE_CTX * ctx, STACK_OF(X509) * chain, X509 * cert0) {
  X509 *cert, *issuer;
  int depth, len, i, j;
  char * name;
  STACK_OF(X509) * new_chain;
  
   /* TODO: check +-1 bugs here; is seems that X509_verify_cert from 1.0.1 allows trusted chain length <= depth+1 */
#if OPENSSL_VERSION_NUMBER >= 0x10100000L
  depth = X509_VERIFY_PARAM_get_depth(X509_STORE_CTX_get0_param(ctx));
#else
  depth = ctx->param->depth;
#endif

  len = sk_X509_num(chain);
#ifdef DEBUG_SSL_CLIENT
  for(i=-1; i<len; i++) {
    if(i<0) cert = cert0; else cert = sk_X509_value(chain, i);
    name = X509_NAME_oneline(X509_get_subject_name(cert), NULL, 0);
    fprintf(stderr, "DEBUG: cert[%d] subject = %s\n", i, name);
    if(name) OPENSSL_free(name);
    name = X509_NAME_oneline(X509_get_issuer_name(cert), NULL, 0);
    fprintf(stderr, "DEBUG: cert[%d] issuer = %s\n", i, name);
    if(name) OPENSSL_free(name);
  }
#endif
  
  cert = cert0; i = -1;
  while(1) {
    if(is_issued(cert, cert)) {
#ifdef DEBUG_SSL_CLIENT
      fprintf(stderr, "DEBUG: cert[%d] is self-signed, chain ended\n", i);
#endif
      return chain;
    }
    if(is_pre_final(ctx, cert)) {
#ifdef DEBUG_SSL_CLIENT
      fprintf(stderr, "DEBUG: cert[%d] was issued by something from trusted-cert-store, chain ended\n", i);
#endif
      return chain;
    }
    if(depth<1) {
#ifdef DEBUG_SSL_CLIENT
      fprintf(stderr, "DEBUG: cert[%d] is not final, but depth limit reached, chain ended, may refuse\n", i);
#endif
      return chain;
    }
/*
    if(!(xn = X509_get_issuer_name(cert))) {
#ifdef DEBUG_SSL_CLIENT
      fprintf(stderr, "DEBUG: cert[%d] hasn't issuer, will refuse\n", i);
#endif
      X509err(0, 0);
      return NULL;
    }
    if(!(name = X509_NAME_oneline(xn, NULL, 0))) {
#ifdef DEBUG_SSL_CLIENT
      fprintf(stderr, "DEBUG: cert[%d] X509_NAME_oneline(issuer) failed, will refuse\n", i);
#endif
      X509err(0, 0);
      return NULL;
    }
*/
    for(j=0; j<len; j++) if((issuer=sk_X509_value(chain, j)) && is_issued(issuer, cert)) break;
    if(j==len) break;
#ifdef DEBUG_SSL_CLIENT
    fprintf(stderr, "DEBUG: cert[%d] issued by cert[%d]\n", i, j);
#endif
    i = j; cert = issuer; depth--;
    save_to_cache(issuer);
  }
  /* we have incomplete chain, last cert is [i] (cert) */
  
  if(is_pre_final(ctx, cert)) {
#ifdef DEBUG_SSL_CLIENT
    fprintf(stderr, "DEBUG: cert[%d] was issued by something from trusted-cert-store, chain ended\n", i);
#endif
    return chain;
  }
  
#ifdef DEBUG_SSL_CLIENT
  fprintf(stderr, "DEBUG: cert[%d]=cert[*0] was not issued by something from trusted-cert-store - chain incomplete, trying to recover\n", i);
#endif
  new_chain = NULL; /* do not do sk_X509_dup() until we aren't sure we need it */
  i = 0;
  
  while(1) {
    if(!(issuer = search_issuer(cert))) {
#ifdef DEBUG_SSL_CLIENT
      fprintf(stderr, "DEBUG: unable to find cert[*%d] issuer in cache, chain ended\n", i);
#endif
      return new_chain?new_chain:chain;
    }
#ifdef DEBUG_SSL_CLIENT
    fprintf(stderr, "DEBUG: cert[*%d] issued by cert[*%d]\n", i, i+1);
#endif
    if(!new_chain && !(new_chain=sk_X509_dup(chain))) {
      X509err(X509_F_X509_VERIFY_CERT, ERR_R_MALLOC_FAILURE);
      X509_free(issuer);
      return NULL;
    }
    if(!sk_X509_push(new_chain, issuer)) {
      X509err(X509_F_X509_VERIFY_CERT, ERR_R_MALLOC_FAILURE);
      X509_free(issuer);
      sk_X509_free(new_chain);
      return NULL;
    }
    i++; cert = issuer; depth--;
    if(is_issued(cert, cert)) {
#ifdef DEBUG_SSL_CLIENT
      fprintf(stderr, "DEBUG: cert[*%d] is self-signed, chain ended\n", i);
#endif
      return new_chain;
    }
    if(is_pre_final(ctx, cert)) {
#ifdef DEBUG_SSL_CLIENT
      fprintf(stderr, "DEBUG: cert[*%d] was issued by something from trusted-cert-store, chain ended\n", i);
#endif
      return new_chain;
    }
    if(depth<1) {
#ifdef DEBUG_SSL_CLIENT
      fprintf(stderr, "DEBUG: cert[*%d] is not final, but depth limit reached, chain ended, may refuse\n", i);
#endif
      return new_chain;
    }
  }
}

/* TODO: openssl 1.1.1 source has (1.0.1 hasn't) ssl/ssl_cert.c/ssl_build_cert_chain() function which does some weird things; */

extern int cert_verify_with_cache(X509_STORE_CTX * ctx, void * udata) {
  int r;
  STACK_OF(X509) *orig_chain;
  STACK_OF(X509) *new_chain;
  STACK_OF(X509) *trusted_chain;
  orig_chain = my_get_untrusted(ctx);
  if(do_save_chains) saved_untrusted = sk_X509_dup(orig_chain);
  if(!(new_chain = fix_chain(ctx, orig_chain, my_get_cert(ctx)))) {
    /* TODO: check how X509err() works in practice */
    return -1;
  } else if(new_chain==orig_chain) {
    r = X509_verify_cert(ctx);
  } else {
    my_set_untrusted(ctx, new_chain);
    r = X509_verify_cert(ctx);
    my_set_untrusted(ctx, orig_chain);
    /* sk_X509_free(new_chain);  - see notes above */
  }
  trusted_chain = my_get_chain(ctx);
  /* as for openssl 1.1.1k:
      when verify successful: trusted_chain contains entire trusted chain, including certs from site, possible intermediates from cache and CA cert from local trust store
      when verify failed: trusted_chain is a copy of untrusted_chain */
  if(do_dump_chains>=2 || do_dump_chains && r<=0) dump_chain(trusted_chain);
  if(do_save_chains) saved_trusted = sk_X509_dup(trusted_chain);
  return r;
}

#endif

extern int cert_verify_without_cache(X509_STORE_CTX * ctx, void * udata) {
  int r;
  STACK_OF(X509) *trusted_chain;
  if(do_save_chains) saved_untrusted = sk_X509_dup(my_get_untrusted(ctx));
  r = X509_verify_cert(ctx);
  trusted_chain = my_get_chain(ctx);
  if(do_dump_chains>=2 || do_dump_chains && r<=0) dump_chain(trusted_chain);
  if(do_save_chains) saved_trusted = sk_X509_dup(trusted_chain);
  return r;
}

extern void cert_verify_set_flags(int download_aia_issuers) {
  do_download_aia_issuers = download_aia_issuers;
}

extern void cert_verify_set_save_chains(int save_chains) {
  do_save_chains = save_chains;
}

extern void cert_verify_set_dump_chains(int level) {
  do_dump_chains = level;
}

extern STACK_OF(X509) * cert_verify_get_saved_untrusted(void) { return saved_untrusted; }
extern STACK_OF(X509) * cert_verify_get_saved_trusted(void) { return saved_trusted; }

static void dump_cert(int j, X509 *cert);
static void dump_chain(STACK_OF(X509) *chain) {
  int i, j;
  X509 *cert;

  i = sk_X509_num(chain);
  for(j=0; j<i; j++) {
    cert = sk_X509_value(chain, j);
    dump_cert(j, cert);
  }
}

static size_t b64encode(char * dst, unsigned char const * src, size_t srclen) {
  static char const h[65]="ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz0123456789+/";
  size_t j,d;
  for(j=d=0; j+3<=srclen; j+=3,d+=4) {
    *(dst++) = h[(src[j]>>2)&63];
    *(dst++) = h[((src[j]<<4)|(src[j+1]>>4))&63];
    *(dst++) = h[((src[j+1]<<2)|(src[j+2]>>6))&63];
    *(dst++) = h[(src[j+2])&63];
  }
  if(j+1==srclen) {
    *(dst++) = h[(src[j]>>2)&63];
    *(dst++) = h[(src[j]<<4)&63];
    *(dst++) = '=';
    *(dst++) = '=';
    d+=4;
  } else if(j+2==srclen) {
    *(dst++) = h[(src[j]>>2)&63];
    *(dst++) = h[((src[j]<<4)|(src[j+1]>>4))&63];
    *(dst++) = h[(src[j+1]<<2)&63];
    *(dst++) = '=';
    d+=4;
  }
  return d;
}

static void dump_cert(int j, X509 *cert) {
  unsigned char * bincert;
  int bincertlen;
  char b64[8192];
  
  bincert = NULL;
  bincertlen = i2d_X509(cert, &bincert);
  if(bincertlen<0 || bincertlen>5000) { while(ERR_get_error()); fprintf(stderr,"CERT: %d-\n", j); return; }
  b64[b64encode(b64, bincert, bincertlen)] = 0;
  OPENSSL_free(bincert);
  fprintf(stderr, "CERT: %d=%s\n", j, b64);
  return;
}

