/*
 * This file is part of fproxy - multipurpose http(s) proxy server
 * helpers/certs_pin.c - maintain site-vs-certificate user-configured matches
 * Copyright (c) 2018,2020 firk (firk@cantconnect.ru)
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. The origin of this software must not be misrepresented; you must
 *    not claim that you wrote the original software.
 * 4. Altered versions in any form must be plainly marked as such, and
 *    must not be misinterpreted as being the original software.
 *
 * This software is provided by the author and contributors `as is'
 * without any express or implied warranty.
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <strings.h>
#include <errno.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <dirent.h>
#include <unistd.h>
#include <fcntl.h>
#include "certs_pin.h"

#define CERT_PIN_DIR "./cert-pin/certs/"
#define CERT_PIN_Q_DIR "./cert-pin/queue/"
/* CERT_PIN_DIR/hostname/ contents:
     A<...> - specially allowed cert in file contents
     D<...> - specially disallowed cert in file contents
     Q<...> - temporary file created for unknown cert for approving/rejecting
     P - disallow all certs except specified in A<...> files (file contents of "P" is ignored)
   letters after A or D is ignored, it is just random characters to make unique file name
 */

   
   
static int valid_hostname(char const * hostname) {
  size_t i;
  char c;
  
  if(hostname[0]=='.') return 0;
  for(i=0; c=hostname[i]; i++) {
    if(c<=32 || c>=127 || c=='/' || c=='\\') return 0;
  }
  if(i<3 || i>255) return 0; /* TODO: what is real upper limit for hostname len? what is limit of filename len? */
  return 1;
}

static struct dirent * readdir_ex(DIR * dp) {
  errno = 0;
  return readdir(dp);
}

/* return:
          system errors:
  -5 = malloc failed
  -4 = empty cert
  -3 = unexpected error accessing files about that hostname
  -2 = bad hostname
          normal conditions:
  -1 = discard this cert;            zfn[] contains "D..." for explicitly discarded cert -OR- "Q..." newly created question for pinned mode
   0 = skip to normal cert checking; zfn[] contains "Q..." for existing or newly created (for do_ask=1) question -OR- "" when question didn't exist and do_ask=0
   1 = forcefully accept this cert;  zfn[] contains "A..." for explicitly accepted cert
   
 */


extern int cert_pin_check(unsigned char const * cert, size_t certlen, char const * hostname, int do_ask, char * zfn) {
  char dn[400], qn[400], fn[500], c, xfn[32], afn[32], dfn[32], qfn[32];
  DIR * dp;
  struct dirent * de;
  int res;
  int dirnotexist, pinmode, discarded, approved, question;
  int fd;
  unsigned char * cert2;
  struct stat st;
  unsigned long mnum, cnum;

  /* prepare */
  if(certlen<1) return -4;
  if(!valid_hostname(hostname)) return -2;
  snprintf(dn, sizeof(dn), CERT_PIN_DIR"%s", hostname);
  snprintf(qn, sizeof(qn), CERT_PIN_Q_DIR"%s", hostname);
  mnum = 0; dirnotexist = pinmode = discarded = approved = question = 0;

  /* check existing state: mnum, dirnotexist, pinmode, discarded, approved, question */
  if(!(dp = opendir(dn))) {
    if(errno!=ENOENT) return -3;
    dirnotexist = 1;
  } else {
    if(!(cert2 = malloc(certlen))) { closedir(dp); return -5; }
    while(de = readdir_ex(dp)) {
      c = de->d_name[0];
      if(c=='.') continue; /* ".", ".." (special dirs) and various ".XXXXXXXXXX" (files to be ignored) */
      if(de->d_type!=DT_REG || strlen(de->d_name)>31) goto file_err;
      strncpy(xfn, de->d_name, 32);
      if(xfn[31]) exit(-1);
      if(c=='P' && !xfn[1]) { pinmode=1; continue; }
      if(c!='A' && c!='D' && c!='Q') goto file_err;
      cnum = strtoul(xfn+1, NULL, 10);
      if(cnum>mnum) mnum=cnum;
      snprintf(fn, sizeof(fn), "%s/%s", dn, xfn);
      if(lstat(fn, &st)<0) goto file_err;
      if(st.st_size!=(off_t)certlen) continue;
      if((fd = open(fn, O_RDONLY))<0) goto file_err;
      if(read(fd, cert2, certlen)!=(ssize_t)certlen) { close(fd); goto file_err; }
      close(fd);
      if(memcmp(cert, cert2, certlen)) continue;
      if(c=='D') { bcopy(xfn, dfn, 32); discarded = 1; }
      if(c=='A') { bcopy(xfn, afn, 32); approved = 1; }
      if(c=='Q') { bcopy(xfn, qfn, 32); question = 1; }
    }
    if(errno) { file_err: free(cert2); closedir(dp); return -3; }
    free(cert2);
    closedir(dp);
  }

  /* existing state fully read, now handle it */
  if(!discarded && !approved && !question && (do_ask || pinmode)) {
    if(dirnotexist && mkdir(dn,0755)<0) return -3;
    snprintf(qfn, 32, "Q%010lu", mnum+1);
    snprintf(fn, sizeof(fn), "%s/%s", dn, qfn);
    if((fd = open(fn, O_WRONLY|O_CREAT|O_EXCL, 0444))<0) return -3;
    if(write(fd, cert, certlen)!=(ssize_t)certlen) { close(fd); unlink(fn); return -3; }
    close(fd);
    close(open(qn,O_WRONLY|O_CREAT,0444));
    question = 1;
  }
  
  if(discarded) { bcopy(dfn,zfn,32); return -1; }
  if(approved) { bcopy(afn,zfn,32); return 1; }
  if(question) { bcopy(qfn,zfn,32); if(pinmode) return -1; return 0; }
  bzero(zfn,32); return 0;
}
