/*
 * This file is part of fproxy - multipurpose http(s) proxy server
 * helpers/certs_pin_mgr.c - maintain site-vs-certificate user-configuration
 * Copyright (c) 2019 firk (firk@cantconnect.ru)
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. The origin of this software must not be misrepresented; you must
 *    not claim that you wrote the original software.
 * 4. Altered versions in any form must be plainly marked as such, and
 *    must not be misinterpreted as being the original software.
 *
 * This software is provided by the author and contributors `as is'
 * without any express or implied warranty.
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <strings.h>
#include <errno.h>
#include <assert.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <dirent.h>
#include <unistd.h>
#include <fcntl.h>
#include "certs_pin.h"

#define CERT_PIN_DIR "./cert-pin/"
/* CERT_PIN_DIR/hostname/ contents:
     A<...> - specially allowed cert in file contents
     D<...> - specially disallowed cert in file contents
     Q<...> - temporary file created for unknown cert for approving/rejecting
     P - disallow all certs except specified in A<...> files (file contents of "P" is ignored)
   letters after A or D is ignored, it is just random characters to make unique file name
 */

   
   
static int valid_hostname(char const * hostname) {
  size_t i;
  char c;
  
  if(hostname[0]=='.') return 0;
  for(i=0; c=hostname[i]; i++) {
    if(c<=32 || c>=127 || c=='/' || c=='\\') return 0;
  }
  if(i<3 || i>255) return 0; /* TODO: what is real upper limit for hostname len? what is limit of filename len? */
  return 1;
}

static struct dirent * readdir_ex(DIR * dp) {
  errno = 0;
  return readdir(dp);
}

/* return:
          system errors:
  -5 = malloc failed
  -4 = empty cert
  -3 = unexpected error accessing files about that hostname
  -2 = bad hostname
          normal conditions:
  -1 = discard this cert;            zfn[] contains "D..." for explicitly discarded cert -OR- "Q..." newly created question for pinned mode
   0 = skip to normal cert checking; zfn[] contains "Q..." for existing or newly created (for do_ask=1) question -OR- "" when question didn't exist and do_ask=0
   1 = forcefully accept this cert;  zfn[] contains "A..." for explicitly accepted cert
   
 */

static void got_error(char const * fmt, ...);
static char ask_what_to_do(char const * hostname, char const * xfn, char const * fn);

extern void check_hostname(char const * hostname) {
  char dn[400], fn[500], fn2[500], c, xfn[32], afn[32], dfn[32], qfn[32];
  DIR * dp;
  struct dirent * de;
  int res;
  int pinmode, discarded, approved, question;
  int fd;

  if(!valid_hostname(hostname)) { got_error("invalid hostname directory found:\n\n%s", hostname); return; }
  snprintf(dn, sizeof(dn), CERT_PIN_DIR"%s", hostname);

  if(!(dp = opendir(dn))) { got_error("can't open hostname directory:\n\n%s", hostname); return; }
  /* check for pinmode */
  pinmode = 0;
  while(de = readdir_ex(dp)) {
    c = de->d_name[0];
    if(c=='.') continue; /* ".", ".." (special dirs) and various ".XXXXXXXXXX" (files to be ignored) */
    if(de->d_type!=DT_REG || strlen(de->d_name)>31)
      { bad_ent: got_error("got invalid entry inside hostname directory:\n\n%s", hostname); closedir(dp); return; }
    strncpy(xfn, de->d_name, 32);
    assert(!xfn[31]);
    if(c=='P' && !xfn[1]) { pinmode=1; continue; }
    if(c!='A' && c!='D' && c!='Q') goto bad_ent;
  }
  if(errno) { read_err: got_error("error getting filelist for hostname directory:\n\n%s", hostname); closedir(dp); return; }
again:
  rewinddir(dp);
  /* process questions */
  while(de = readdir_ex(dp)) {
    c = de->d_name[0];
    if(c=='.') continue; /* ".", ".." (special dirs) and various ".XXXXXXXXXX" (files to be ignored) */
    if(de->d_type!=DT_REG || strlen(de->d_name)>31) goto bad_ent;
    strncpy(xfn, de->d_name, 32);
    assert(!xfn[31]);
    if(c=='P' && !xfn[1]) continue;
    if(c=='Q') {
      snprintf(fn, sizeof(fn), "%s/%s", dn, xfn);
      c = ask_what_to_do(hostname, xfn, fn);
      if(c=='A' || c=='D') {
        snprintf(fn2, sizeof(fn2), "%s/%c%s", dn, c, xfn+1);
        if(link(fn, fn2)<0) got_error("error renaming (%s) to (%c%s)\nerrno %d (%s)\n\nhostname: %s", xfn, c, xfn+1, errno, strerror(errno), hostname);
        else if(unlink(fn)<0) got_error("error deleting (%s) after creating (%c%s)\nerrno %d (%s)\n\nhostname: %s", xfn, c, xfn+1, errno, strerror(errno), hostname);
      } else if(c==0) {
        if(unlink(fn)<0) got_error("error deleting (%s)\nerrno %d (%s)\n\nhostname: %s", xfn, errno, strerror(errno), hostname);
      }
      goto again;
    }
    if(c!='A' && c!='D') goto bad_ent;
  }
  if(errno) goto read_err;
  closedir(dp);
  return;
}

static void got_error(char const * fmt, ...) {
  // TODO: something interactive
}
// 0=delete 1=error 'D'=discard 'A'=accept
static char ask_what_to_do(char const * hostname, char const * xfn, char const * fn) {
  struct stat st;
  if(lstat(fn, &st)<0) { got_error("error reading (%s)\nerrno %d (%s)\n\nhostname: %s", xfn, errno, strerror(errno), hostname); return 1; }
  // display time: st.st_mtime
  // openssl x509 -inform DER -in {fn} -noout -text
  // ask: accept/discard/forget
  return 1;
}
