/*
 * This file is part of fproxy - multipurpose http(s) proxy server
 * helpers/client.c - main source file for ssl-client-wrapper
 * Copyright (c) 2018-2020, 2022 firk (firk@cantconnect.ru)
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. The origin of this software must not be misrepresented; you must
 *    not claim that you wrote the original software.
 * 4. Altered versions in any form must be plainly marked as such, and
 *    must not be misinterpreted as being the original software.
 *
 * This software is provided by the author and contributors `as is'
 * without any express or implied warranty.
 */

#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>
#include <string.h>
#include <errno.h>
#include <unistd.h>
#include <openssl/evp.h>
#include <openssl/ssl.h>
#include <openssl/err.h>
#include <libsimpletcp/libsimpletcp.h>
#include "cert_match.h"
#include "certs_pin.h"
#include "cert_verify.h"

static char * argv0;
static size_t argv0len;

static void ssl_fail(char const * fmt, ...) {
  va_list arg;
  va_start(arg, fmt);
  fprintf(stderr, "ERROR: ");
  vfprintf(stderr, fmt, arg);
  fprintf(stderr, "\n");
  ERR_print_errors_fp(stderr);
  va_end(arg);
  // may be SSL_CTX_free() here?
  exit(-1);
}

static void raw_fail(char const * fmt, ...) {
  va_list arg;
  va_start(arg, fmt);
  fprintf(stderr, "ERROR: ");
  vfprintf(stderr, fmt, arg);
  fprintf(stderr, "\n");
  va_end(arg);
  // may be SSL_CTX_free() here?
  exit(-1);
}

static void setup_minversion(SSL_CTX * ctx, unsigned short minver, int legacy_server) {
  if(minver>0x0200) SSL_CTX_set_options(ctx, SSL_OP_NO_SSLv2);
  if(minver>0x0300) SSL_CTX_set_options(ctx, SSL_OP_NO_SSLv3);
  if(minver>0x0301) SSL_CTX_set_options(ctx, SSL_OP_NO_TLSv1);
#ifdef SSL_OP_NO_TLSv1_1
  if(minver>0x0302) SSL_CTX_set_options(ctx, SSL_OP_NO_TLSv1_1);
#endif
#ifdef SSL_OP_NO_TLSv1_2
  if(minver>0x0303) SSL_CTX_set_options(ctx, SSL_OP_NO_TLSv1_2);
#endif
  /* allow "legacy insecure renegotiation with unpatched servers"; idk what is it; see man SSL_CTX_set_options */
  /* "legacy server" means server with openssl <0.9.8m which is vulnerable;
     enabling or disabling of this option does not help to avoid vulnerability, but may prevent from connecting to vulnerable server;
     client have no way to prevent attack if server is vulnerable */
  if(legacy_server) SSL_CTX_set_options(ctx, SSL_OP_LEGACY_SERVER_CONNECT);
  else SSL_CTX_clear_options(ctx, SSL_OP_LEGACY_SERVER_CONNECT);
}

/* optional, usually not used */
static void setup_client_cert(SSL_CTX * ctx, char const * certfile, char const * keyfile, int flag) {
  /* Load the client certificate into the SSL_CTX structure */
  if(SSL_CTX_use_certificate_file(ctx, certfile, SSL_FILETYPE_PEM)<=0) ssl_fail("failed loading client-cert file");
  /* Load the private-key corresponding to the client certificate */
  if(SSL_CTX_use_PrivateKey_file(ctx, keyfile, SSL_FILETYPE_PEM)<=0) ssl_fail("failed loading client-key file");
  /* Check if the client certificate and private-key matches */
  if(!SSL_CTX_check_private_key(ctx)) raw_fail("client-cert and client-key mismatch");
}

/* this is required to verify server cert */
static void setup_trusted_ca_list(SSL_CTX * ctx, char const * listfile, int enforce) {
  if(listfile && *listfile && !SSL_CTX_load_verify_locations(ctx, listfile, NULL)) ssl_fail("error loading trusted ca list");
  SSL_CTX_set_verify(ctx, enforce?SSL_VERIFY_PEER:SSL_VERIFY_NONE, NULL);
  /* in SSL_VERIFY_NONE mode, certificate still checked but not fatal;
     SSL_get_verify_result() can be used to get results */
  SSL_CTX_set_verify_depth(ctx, 9); /* allow max 9 ca-certs (10 total) in server cert-chain */
}


/*
  SSL_CTX_set_cert_verify_callback() may be used to replace default cert verify function X509_verify_cert() with custom one
 */

static void verify_cert(SSL * ssl, X509 * server_cert, char const * hostname) {
  int match;
  long vres;
  unsigned char * bincert;
  int bincertlen;
  int problems;
  char zfn[32];
  int rrr;
  
  bincert = NULL;
  bincertlen = i2d_X509(server_cert, &bincert);
  if(bincertlen<0) ssl_fail("server certificate looks like malformed - can't reencode it");
  if(!bincertlen) raw_fail("server certificate looks like malformed - zero-length encoding?");
  
  vres = SSL_get_verify_result(ssl);
  match = cert_match(server_cert, hostname);
  problems = (vres!=X509_V_OK || match<0);


  rrr = cert_pin_check(bincert, bincertlen, hostname, problems, zfn);
  switch(rrr) {
  case -1:
    if(zfn[0]=='Q') raw_fail("server certificate should be manually revised (%s) for hostname '%s'", zfn, hostname);
    raw_fail("server certificate was discarded due to local blacklist (%s) for hostname '%s'", zfn, hostname);
  case 1:
    return; /* just accept */
  case 0:
    break;
  case -2:
    raw_fail("server certificate for hostname '%s' can't be checked due to BAD HOSTNAME", hostname);
  case -3:
    raw_fail("server certificate for hostname '%s' can't be checked due to INTERNAL FILESYSTEM PROBLEM", hostname);
  case -4:
    raw_fail("server certificate for hostname '%s' can't be checked due to EMPTY CERT", hostname);
  case -5:
    raw_fail("server certificate for hostname '%s' can't be checked due to INTERNAL MALLOC PROBLEM", hostname);
  default:
    raw_fail("server certificate for hostname '%s' can't be checked due to internal error %d", hostname, rrr);
  }

  if(vres==X509_V_ERR_UNABLE_TO_GET_ISSUER_CERT_LOCALLY) raw_fail("server certificate (%s) is not valid or not trusted, verify code %ld (%s) - most likely due to missing certificate chain", zfn, vres, X509_verify_cert_error_string(vres));
  if(vres!=X509_V_OK) raw_fail("server certificate (%s) is not valid or not trusted, verify code %ld (%s)", zfn, vres, X509_verify_cert_error_string(vres));
  if(match==CERT_MATCH_NO) raw_fail("server certificate (%s) is not for hostname '%s'", zfn, hostname);
  if(match<0) raw_fail("server certificate (%s) from hostname '%s' has some problems, code %d", zfn, hostname, match);
}

static SSL * start_connection(SSL_CTX * ctx, int fd, uint32 addr, uint16 port, char const * hostname, int * r_fd, int do_dump_certs) {
  int ret;
  SSL * ssl;
  X509 * server_cert;
  char * tmp;
  char addrstr[50];
  size_t j;
  char c;
  
  for(j=0; c=hostname[j]; j++) if(c<33 || c>126 || c=='<' || c=='>' || c=='\\' || c=='/' || c=='?' || c=='*' || c=='|') break;
  if(!j || hostname[j]) raw_fail("malformed hostname");
  cert_verify_set_dump_chains(2);
  lst_print_ip_port(addrstr,addr,port);
  if(argv0) snprintf(argv0, argv0len, "SSL(%s/%s)",addrstr,hostname);
  if(fd<0) {
    fd = lst_conn_open(addr, port, LST_FLAG_NONBLOCK_AFTER);
    if(fd<0) raw_fail("error establishing TCP connection: %d (%s)", errno, strerror(errno));
  }
  ssl = SSL_new(ctx);
  if(!ssl) ssl_fail("SSL_new() failed");
  if(hostname && !SSL_set_tlsext_host_name(ssl, hostname)) ssl_fail("SSL_set_tlsext_host_name() failed");
  if(!SSL_set_fd(ssl, fd)) ssl_fail("SSL_set_fd() failed");
  while((ret=SSL_connect(ssl))<=0) { /* non-blocking side effects workaround */
    if(ret<0) switch(SSL_get_error(ssl,ret)) {
    case SSL_ERROR_WANT_READ:  lst_conn_wait_readable(fd, 1000); continue;
    case SSL_ERROR_WANT_WRITE: lst_conn_wait_writable(fd, 1000); continue;
    }
    ssl_fail("SSL_connect() returned %d", ret);
  }
  server_cert = SSL_get_peer_certificate(ssl);
  if(!server_cert) raw_fail("missing server certificate (how it can happen?)");

  verify_cert(ssl, server_cert, hostname);

#ifdef DEBUG_SSL_CLIENT
  fprintf(stderr, "DEBUG: protocol = %s\nDEBUG: cipher = %s\nDEBUG: verify_result = %ld\n", SSL_get_version(ssl), SSL_get_cipher(ssl), SSL_get_verify_result(ssl));
  if(!server_cert) fprintf(stderr, "DEBUG: server_cert is MISSING\n"); else {
    /* X509_NAME_oneline() properly handles NULL in first arg so no check for that */
    tmp = X509_NAME_oneline(X509_get_subject_name(server_cert),NULL,0);
    fprintf(stderr, "DEBUG: server_cert subject: %s\n", tmp);
    if(tmp) OPENSSL_free(tmp);
    tmp = X509_NAME_oneline(X509_get_issuer_name(server_cert),NULL,0);
    fprintf(stderr, "DEBUG: server_cert issuer: %s\n", tmp);
    if(tmp) OPENSSL_free(tmp);
  }
#endif
  if(server_cert) X509_free(server_cert); // needs to be called after SSL_get_peer_certificate()
  fprintf(stderr,"OK: connection to %s/%s established\n", addrstr, hostname);
  *r_fd = fd;
  return ssl;
}

/* wait_reply means "wait for remote shutdown too";
   this is not essential by protocol, but may be needed in some cases by app */
static void end_connection(SSL * ssl, int fd, int wait_reply) {
  int ret;
  while((ret=SSL_shutdown(ssl))<=0) { /* non-blocking side effects workaround */
    if(ret<0) switch(SSL_get_error(ssl,ret)) {
    case SSL_ERROR_WANT_READ:  lst_conn_wait_readable(fd, 1000); continue;
    case SSL_ERROR_WANT_WRITE: lst_conn_wait_writable(fd, 1000); continue;
    default: ssl_fail("SSL_shutdown() returned %d", ret);
    }
    if(!wait_reply) break;
  }
  if(close(fd)<0) raw_fail("close() errno %d (%s)", errno, strerror(errno));
  SSL_free(ssl);
}

static void dump_cert(unsigned int p, X509 * cert) {
  unsigned int i;
  X509_NAME * xn;
  char * name;

  /* TODO: proper dump */

  for(i=p; i--; ) fputc(' ', stdout);
  xn = X509_get_subject_name(cert);
  name = X509_NAME_oneline(xn,NULL,0);
  fprintf(stdout, "subject: %s\n", name);
  if(name) OPENSSL_free(name);

  for(i=p; i--; ) fputc(' ', stdout);
  xn = X509_get_issuer_name(cert);
  name = X509_NAME_oneline(xn,NULL,0);
  fprintf(stdout, "issuer: %s\n", name);
  if(name) OPENSSL_free(name);
}

static void get_cert(SSL_CTX * ctx, int fd, uint32 addr, uint16 port, char const * hostname, int do_dump_certs) {
  int ret;
  SSL * ssl;
  X509 * server_cert;
  STACK_OF(X509) * chain;
  char * tmp;
  char addrstr[50];
  int i, j;
  char c;
  
  for(j=0; c=hostname[j]; j++) if(c<33 || c>126 || c=='<' || c=='>' || c=='\\' || c=='/' || c=='?' || c=='*' || c=='|') break;
  if(!j || hostname[j]) raw_fail("malformed hostname");
  cert_verify_set_save_chains(1);
  cert_verify_set_dump_chains(2);
  lst_print_ip_port(addrstr,addr,port);
  if(argv0) snprintf(argv0, argv0len, "SSL(%s/%s)",addrstr,hostname);
  if(fd<0) {
    fd = lst_conn_open(addr, port, LST_FLAG_NONBLOCK_AFTER);
    if(fd<0) raw_fail("error establishing TCP connection: %d (%s)", errno, strerror(errno));
  }
  ssl = SSL_new(ctx);
  if(!ssl) ssl_fail("SSL_new() failed");
  if(hostname && !SSL_set_tlsext_host_name(ssl, hostname)) ssl_fail("SSL_set_tlsext_host_name() failed");
  if(!SSL_set_fd(ssl, fd)) ssl_fail("SSL_set_fd() failed");
  while((ret=SSL_connect(ssl))<=0) { /* non-blocking side effects workaround */
    if(ret<0) switch(SSL_get_error(ssl,ret)) {
    case SSL_ERROR_WANT_READ:  lst_conn_wait_readable(fd, 1000); continue;
    case SSL_ERROR_WANT_WRITE: lst_conn_wait_writable(fd, 1000); continue;
    }
    ssl_fail("SSL_connect() returned %d", ret);
  }
  cert_verify_set_save_chains(0);

  fprintf(stderr,"OK: pre-connection to %s/%s established\n", addrstr, hostname);
  /* TODO: use client bufs instead of printf */
  /* TODO: html? */
  fprintf(stdout, "protocol = %s\ncipher = %s\nverify_result = %ld\n", SSL_get_version(ssl), SSL_get_cipher(ssl), SSL_get_verify_result(ssl));

  if(!(server_cert = SSL_get_peer_certificate(ssl))) fprintf(stdout, "server_cert is MISSING\n");
  else {
    fprintf(stdout,"server_cert:\n");
    dump_cert(2, server_cert);
    X509_free(server_cert); // needs to be called after SSL_get_peer_certificate()
  }
  if(!(chain = cert_verify_get_saved_untrusted())) fprintf(stdout, "untrusted_chain is MISSING\n");
  else {
    fprintf(stdout,"untrusted_chain:\n");
    i = sk_X509_num(chain);
    for(j=0; j<i; j++) {
      server_cert = sk_X509_value(chain, j);
      fprintf(stdout,"  untrusted[%d]:\n", j);
      dump_cert(4, server_cert);
    }
    sk_X509_free(chain);
  }
  if(!(chain = cert_verify_get_saved_trusted())) fprintf(stdout, "trusted_chain is MISSING\n");
  else {
    fprintf(stdout,"trusted_chain:\n");
    i = sk_X509_num(chain);
    for(j=0; j<i; j++) {
      server_cert = sk_X509_value(chain, j);
      fprintf(stdout,"  trusted[%d]:\n", j);
      dump_cert(4, server_cert);
    }
    sk_X509_free(chain);
  }
  end_connection(ssl, fd, 0);
}

/* SSL_write() is not in-place replacement for write() in case of being deferred;
   it should be retries with exactly same args and same data in buffer;
   also it usually does success until all bytes sent;
   see SSL_set_mode SSL_MODE_ENABLE_PARTIAL_WRITE, SSL_MODE_ACCEPT_MOVING_WRITE_BUFFER for details */
/* this wrapper always sends all bytes until return or get fatal error */
static void send_ssl_bytes(SSL * ssl, int fd, unsigned char const * b, size_t n) {
  int ret;
  int csz;
  while(n) {
    csz = 16000;
    if(((unsigned)csz)>n) csz = (int)n;
    ret = SSL_write(ssl, b, csz);
    if(ret>0) {
      if(ret>csz) raw_fail("BUG! ret>csz for SSL_write()");
      n -= ret;
      b += ret;
      continue;
    }
    switch(SSL_get_error(ssl,ret)) {
    case SSL_ERROR_WANT_READ:  lst_conn_wait_readable(fd, 1000); continue;
    case SSL_ERROR_WANT_WRITE: lst_conn_wait_writable(fd, 1000); continue;
    case SSL_ERROR_ZERO_RETURN: raw_fail("remote peer closed SSL session while we are trying SSL_write()");
    case SSL_ERROR_SYSCALL: raw_fail("SSL_write() failed due to IO errno %d (%s)", errno, strerror(errno));
    case SSL_ERROR_SSL: ssl_fail("SSL_write() failed in protocol handling");
    default: ssl_fail("SSL_write() failed due to unknown reason");
    }
  }
}

/* may recv less bytes than requested */
static size_t recv_ssl_bytes(SSL * ssl, int fd, unsigned char * b, size_t n, int * p_eof) {
  int ret;
  int csz;
  while(n) {
    csz = 16000;
    if(((unsigned)csz)>n) csz = (int)n;
    ret = SSL_read(ssl, b, csz);
    if(ret>0) {
      if(ret>csz) raw_fail("BUG! ret>csz for SSL_read()");
      n -= ret;
      b += ret;
      return ret;
    }
    switch(SSL_get_error(ssl,ret)) {
    case SSL_ERROR_WANT_READ:  return 0; // lst_conn_wait_readable(fd, 1); continue;
    case SSL_ERROR_WANT_WRITE: lst_conn_wait_writable(fd, 1000); continue;
    case SSL_ERROR_ZERO_RETURN: *p_eof = 1; return 0; // do not fail, it is graceful EOF // raw_fail("remote peer closed SSL session while we are trying SSL_read()");
    case SSL_ERROR_SYSCALL: raw_fail("SSL_read() failed due to IO errno %d (%s)", errno, strerror(errno));
    case SSL_ERROR_SSL: ssl_fail("SSL_read() failed in protocol handling");
    default: ssl_fail("SSL_read() failed due to unknown reason");
    }
  }
  return 0;
}

/*
this is not correct; it only returns number of bytes that already pulled by previous actions but not
received by SSL_read(); it seems to ignore any pending data in socket buffer */
/*static int is_readable(SSL * ssl) {
  if(SSL_pending(ssl)) return 1;
  return 0;
}*/

/* CLIENT = parent process who interacts with us via fd0/fd1/fd2
   SERVER = remote server accessed via OpenSSL */
static unsigned char SCbuf[20000];
static unsigned char CSbuf[20000];
static size_t SCbytes, CSbytes;
static int input_closed;

static int update_client_bufs(void) {
  ssize_t r;
  int work = 0;
  if(CSbytes<sizeof(CSbuf)) {
    r = read(0, CSbuf+CSbytes, sizeof(CSbuf)-CSbytes);
    if(r<0 && errno!=EWOULDBLOCK && errno!=EAGAIN && errno!=EINTR && errno!=ENOBUFS && errno!=ENOMEM) exit(-1); /* client disconnected, nothing to do now */
    if(!r) input_closed = 1;
    if(r>0) { work=1; CSbytes += r; }
  }
  if(SCbytes) {
    r = write(1, SCbuf, SCbytes);
    if(!r || r>0 && ((size_t)r)>SCbytes) abort(); /* should not happen */
    if(r<0 && errno!=EWOULDBLOCK && errno!=EAGAIN && errno!=EINTR && errno!=ENOBUFS && errno!=ENOMEM) exit(-1); /* client disconnected, nothing to do now */
    if(r>0) {
      work = 1;
      SCbytes -= r;
      if(SCbytes) memmove(SCbuf, SCbuf+r, SCbytes);
    }
  }
  return work;
}

static void CS_drain(size_t n) {
  CSbytes -= n;
  if(CSbytes) memmove(CSbuf, CSbuf+n, CSbytes);
}

static void do_things(SSL_CTX * ctx, int base_fd) {
  SSL * ssl;
  int work, fd, eof_flag, do_dump_certs;
  size_t n, n0;
  uint32 c_addr;
  uint16 c_port;
  
  if(lst_conn_set_nonblock(0)<0 || lst_conn_set_nonblock(1)<0 || base_fd>=0 && lst_conn_set_nonblock(base_fd)<0) raw_fail("can't set nonblock");
  ssl = NULL;
  
  work = 0;
  input_closed = 0;
  do_dump_certs = 0;
#ifdef SSL_CLIENT_ALLOW_CERT_CACHE
  SSL_CTX_set_cert_verify_callback(ctx, &cert_verify_with_cache, NULL);
  cert_verify_set_flags(1);
#else
  SSL_CTX_set_cert_verify_callback(ctx, &cert_verify_without_cache, NULL);
#endif
  while(!ssl) { /* pre-connection loop */
    if(update_client_bufs()) work = 1;
    if(input_closed) exit(0);
    if(!work) { if(CSbytes==sizeof(CSbuf)) raw_fail("input data malformed"); usleep(10000); continue; }
    work = 0;
    if(!CSbytes) continue;
    switch(CSbuf[0]) {
    case 0:
      if(CSbytes<4) continue;
      setup_minversion(ctx, CSbuf[1]*0x100+CSbuf[2], CSbuf[3]);
      CS_drain(4);
      work = 1;
      continue;
    case 'V': /* humanfriendly alias for 0 */
      for(n=1; n<CSbytes && CSbuf[n]!='\n'; n++);
      if(CSbytes-n<1) continue;
      CSbuf[n] = 0;
      if(n!=3) raw_fail("syntax error");
      setup_minversion(ctx, (CSbuf[1]&15)*0x100+(CSbuf[2]&15), 0);
      CS_drain(4);
      work = 1;
      continue;
    case 1:
      for(n=1; n<CSbytes && CSbuf[n]; n++);
      if(CSbytes-n<2) continue;
      n0 = ++n;
      for(; n<CSbytes && CSbuf[n]; n++);
      if(CSbytes-n<2) continue;
      setup_client_cert(ctx, (char*)(CSbuf+1), (char*)(CSbuf+n0), CSbuf[n+1]);
      CS_drain(n+2);
      work = 1;
      continue;
    case 2:
      for(n=1; n<CSbytes && CSbuf[n]; n++);
      if(CSbytes-n<2) continue;
      setup_trusted_ca_list(ctx, (char*)(CSbuf+1), CSbuf[n+1]);
      CS_drain(n+2);
      work = 1;
      continue;
    case 'T': /* human-friendly alias for 2 */
      for(n=1; n<CSbytes && CSbuf[n]!='\n'; n++);
      if(CSbytes-n<1) continue;
      CSbuf[n] = 0;
      setup_trusted_ca_list(ctx, (char*)(CSbuf+1), 0);
      CS_drain(n+1);
      work = 1;
      continue;
    case 3:
    case 4:
      for(n=7; n<CSbytes && CSbuf[n]; n++);
      if(CSbytes-n<1) continue;
      c_addr = CSbuf[1]*0x1000000+CSbuf[2]*0x10000+CSbuf[3]*0x100+CSbuf[4];
      c_port = CSbuf[5]*0x100+CSbuf[6];
      if(CSbuf[0]==3) ssl = start_connection(ctx, base_fd, c_addr, c_port, (char*)(CSbuf+7), &fd, do_dump_certs);
      else get_cert(ctx, base_fd, c_addr, c_port, (char*)(CSbuf+7), do_dump_certs);
      CS_drain(n+1);
      work = 1;
      continue;
    case 5:
    case 'S':
    case 's':
      do_dump_certs = 1;
      CS_drain(1);
      work = 1;
      continue;
    case 'C': /* human-friendly alias for 3 */
    case 'c':
      for(n=1; n<CSbytes && CSbuf[n]!='/'; n++);
      if(CSbytes-n<1) continue;
      CSbuf[n] = 0;
      if(lst_parse_addrstr((char*)(CSbuf+1), &c_addr, &c_port)!=3) raw_fail("syntax error");
      CSbuf[n] = '/';
      n0 = ++n;
      for( ; n<CSbytes && CSbuf[n]!='\n'; n++);
      if(CSbytes-n<1) continue;
      CSbuf[n] = 0;
      if(CSbuf[0]=='C') ssl = start_connection(ctx, base_fd, c_addr, c_port, (char*)(CSbuf+n0), &fd, do_dump_certs);
      else get_cert(ctx, base_fd, c_addr, c_port, (char*)(CSbuf+n0), do_dump_certs);
      CS_drain(n+1);
      work = 1;
      continue;
    default: raw_fail("bad opcode 0x%02X", CSbuf[0]);
    }
  }
  
//  setup_minversion(ctx, 0x0300, 0);
//  // setup_client_cert()
//  setup_trusted_ca_list(ctx, "/etc/ssl/certs/ca-certificates.crt", 1);
//  ssl = start_connection(ctx, 0x);
  work = (CSbytes?1:0);
  eof_flag = 0;
  while(1) { /* connected loop */
    if(update_client_bufs()) work = 1;
    if(input_closed && !CSbytes) break;
    if(eof_flag && !SCbytes) break;
    if(!SCbytes) {
      SCbytes = recv_ssl_bytes(ssl, fd, SCbuf, sizeof(SCbuf), &eof_flag);
      if(SCbytes) work = 1;
    }
    if(!work) { usleep(10000); continue; }
    work = 0;
    if(CSbytes) {
      send_ssl_bytes(ssl, fd, CSbuf, CSbytes);
      CSbytes = 0;
    }
  }
  end_connection(ssl, fd, 0);
}

int main(int argc, char * * argv) {
  SSL_CTX * ctx;
  int base_fd;
  int n;
  char * arg;
  
  argv0 = argv[0];
  if(argv0) argv0len = strlen(argv0)+1;
  if(!argv0len) argv0 = NULL;
  base_fd = -1;
  for(n=1; n<argc; n++) {
    arg = argv[n];
    if(!strcmp(arg,"--fd")) base_fd = 3;
    if(!strcmp(arg,"--silent")) argv0 = NULL;
  }
  lst_init();
  
  /**** init[1] register error strings */
  /* ERR_load_BIO_strings() - what is it? */
  SSL_load_error_strings();

  /**** init[2] register algorythms */
  SSL_library_init();
  /* OpenSSL_add_all_algorithms(); - if ALL algorythms needed, not just SSL/TLS (idk what it means, see manpages) */

  /**** init[3] create ssl context */
  ctx = SSL_CTX_new(SSLv23_client_method()); /* this construction means: create context and allow all known protocols; see manpage if doubt */
  if(!ctx) ssl_fail("SSL_CTX_new() failed");

  /**** do all things */
  do_things(ctx, base_fd);

  /**** done[3] free ssl context */
  SSL_CTX_free(ctx);

  /**** done[2] unregoster algorythms */
  EVP_cleanup();

  /**** done[1] unregister error strings */
  ERR_free_strings();
  return 0;
}
