/*
 * This file is part of fproxy - multipurpose http(s) proxy server
 * helpers/http_download.c - download certificates or just small files via http
 * Copyright (c) 2020 firk (firk@cantconnect.ru)
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. The origin of this software must not be misrepresented; you must
 *    not claim that you wrote the original software.
 * 4. Altered versions in any form must be plainly marked as such, and
 *    must not be misinterpreted as being the original software.
 *
 * This software is provided by the author and contributors `as is'
 * without any express or implied warranty.
 */

#define FCL_SIMPLE_NAMES
#include <string.h>
#include <errno.h>
#include <sys/types.h>
#include <fcl/types.h>
#include <fcl/strparse.h>
#include <fcl/malloc.h>
#include <fcl/dynbufs.h>
#include <libsimpletcp/libsimpletcp.h>
#include "http_download.h"

static int send_str(int fd, char const * st) { return lst_conn_s_send_all(fd, st, strlen(st)); }

/* return:
  -1 = bad input data - url protocol
  -2 = bad input data - url host
  -3 = bad input data - url port
  -4 = bad input data - url path
  -5 = networking error (connecting), errno set
  -6 = networking error (after connecting), errno set
  -7 = bad reply (should not happen - it is reply from local proxy)
  0 = ok
  */
extern int http_download(char const * proxysock, char const * url, size_t size_limit, uint8 * * res, size_t * rbytes, unsigned int * status) {
  char const * proto;
  char * host;
  size_t p, pp;
  char c;
  uint16 port, stt;
  int fd, err;
  t_dynbuf resp;
  uint8 buf[16384];
  size_t rsize;
  char *h, *v;
  uint32 clen;
  int have_clen;
  
  if(!strncasecmp(url,"http://",7)) { proto="http://"; url+=7; }
  else if(!strncasecmp(url,"https://",8)) { proto="https://"; url+=8; }
  else return -1;
  for(p=0; url[p] && url[p]!='/'; p++);
  host = fcl_substrdup_check(url, p);
  url += p;
  if(*url!='/') url = "/";
  for(p=0; (c=host[p]) && (c!=':'); p++) {
    if(c>='0' && c<='9' || c>='a' && c<='z' || c=='_' || c=='-') continue;
    if(c>='A' && c<='Z') { host[p] = c-'A'+'a'; continue; }
    if(c=='.' && p && host[p-1]!='.' && host[p+1] && host[p+1]!='.') continue;
    return -2;
  }
  if(!p) return -2;
  if(c==':') if(str_to_uint16(host+p+1, &port)<0 || !port || port!=80 && port!=81 && port!=88 && port!=443 && port!=1443 && port!=8080) return -3;
  for(p=0; c=url[p]; p++) if(c<=32 || c>=127) return -4;
  
  if((fd=lst_conn_open_unix(proxysock, 0))<0) return -5;

  if(send_str(fd, "GET ")<0 || send_str(fd, proto)<0 || send_str(fd, host)<0 || send_str(fd, url)<0 || send_str(fd, " HTTP/1.0\r\n"
    "Connection: close\r\n"
    "User-Agent: FirkSslProxy\r\n"
    "Host: ")<0 || send_str(fd, host)<0 || send_str(fd,"\r\n"
    "\r\n")<0) { lst_conn_close(fd, LST_FLAG_IGNORE_ERRORS); return -6; }

  dynbuf_init_check(&resp, 8192, 8192, size_limit);
  while(lst_conn_s_recv(fd, buf, sizeof(buf), &rsize)>=0) {
    if(rsize) dynbuf_write(&resp, buf, rsize);
    if(resp.overflowed) { errno=ENOSPC; break; }
  }
  lst_conn_close(fd, LST_FLAG_IGNORE_ERRORS);
  if(errno!=ECONNRESET) { fd=errno; dynbuf_clear(&resp); errno=fd; return -6; }
#define  b  resp.buf
  clen = 0; have_clen = 0;
  for(pp=p=0; p<resp.len; p++) {
    if(b[p]=='\r' && p+1<resp.len && b[p+1]=='\n') continue;
    if(b[p]!='\n') {
      if(b[p]<=31 || b[p]==127) { badresp: dynbuf_clear(&resp); return -7; }
      continue;
    }
    if(!p || b[p-1]!='\r') goto badresp;
    b[p-1] = 0;
    h = (char*)(b+pp);
    if(!pp) {
      if(p<13 || strncmp(h,"HTTP/1.",7) || h[7]!='0' && h[7]!='1' || h[8]!=' ' || h[12] && h[12]!=' ') goto badresp;
      h += 9;
      if(strp_to_uint16((char const**)&h, &stt)<0 || stt<100 || stt>999 || *h && *h!=' ') goto badresp;
    } else {
      if(v=(char*)skip_prefix(h,"Content-Length:")) {
        if(have_clen || str_to_uint32(skip_spaces(v), &clen)<0) goto badresp;
        have_clen = 1;
      }
    }
    pp = p+1;
    if(!*h) break;
  }
  if(p>=resp.len) goto badresp;
  if(resp.len-pp!=clen) goto badresp;
  dynbuf_drain(&resp, pp);
  *res = b;
  *rbytes = clen;
  *status = stt;
  return 0;
}
