/*
 * This file is part of fproxy - multipurpose http(s) proxy server
 * helpers/lst_ssl.c - libsimpletcp-like functions for using ssl-wrappers
 * Copyright (c) 2018,2019,2022 firk (firk@cantconnect.ru)
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. The origin of this software must not be misrepresented; you must
 *    not claim that you wrote the original software.
 * 4. Altered versions in any form must be plainly marked as such, and
 *    must not be misinterpreted as being the original software.
 *
 * This software is provided by the author and contributors `as is'
 * without any express or implied warranty.
 */

#include <sys/types.h>
#include <stdio.h>
#include <string.h>
#include <unistd.h>
#include <errno.h>
#include <sys/socket.h>
#include <libsimpletcp/libsimpletcp.h>
#include "../config.h"
#include "lst_ssl.h"

extern char * * environ;

#define JUST_ERRNO(name) if(errmsgsize) snprintf(errmsg, errmsgsize, name"() errno %d (%s)", errno, strerror(errno))
#define JUST_ERRSTR(str) if(errmsgsize) snprintf(errmsg, errmsgsize, "%s", str)

/* TODO: LST_FLAG_NONBLOCK_BEFORE handled improperly */
/*
  when returned -1 it sets errno OR (errno=0 AND errmsg contains some message)
    errno=0 and errmsg[0]=0 means it should be message but it was not received for some reason

  raw_fd: -1 = connect to addr:port
          >=0 = use "addr:port" in log messages but do not connect, use raw_fd as already connected socket; it eats raw_fd in any case
*/
extern int lst_conn_open_ssl(int raw_fd, uint32 addr, uint16 port, char const * hostname, uint16 minver, char const * cacert, char * errmsg, size_t errmsgsize, lst_conn_open_ssl_ctx * ctx, char * cbuf, size_t cbufsz, uint32 flags) {
  int sv[4], err, i;
  char * argv[3];
  pid_t r;
  unsigned char buf0[256];

  ctx->cbuf = cbuf;
  ctx->cbufsz = cbuf?cbufsz:0;
  ctx->cbufpos = 0;
  ctx->cstr = 0;
  if(cbuf && cbufsz) *cbuf=0;
  if(!errmsg) errmsgsize = 0;
  if(errmsgsize) *errmsg = 0;
  if(socketpair(PF_UNIX, SOCK_STREAM, 0, sv)<0) { JUST_ERRNO("socketpair"); if(raw_fd>=0) close(raw_fd); errno=0; return -1; }
  if(socketpair(PF_UNIX, SOCK_STREAM, 0, sv+2)<0) { JUST_ERRNO("socketpair"); if(raw_fd>=0) close(raw_fd); close(sv[0]); close(sv[1]); errno=0; return -1; }
  if(sv[0]<=2 || sv[1]<=2 || sv[2]<=2 || sv[3]<=2) { JUST_ERRSTR("got fds<=2"); if(raw_fd>=0) close(raw_fd); close(sv[0]); close(sv[1]); close(sv[2]); close(sv[3]); errno=0; return -1; }
  if((r=fork())<0) { JUST_ERRNO("fork"); if(raw_fd>=0) close(raw_fd); close(sv[0]); close(sv[1]); close(sv[2]); close(sv[3]); errno=0; return -1; }
  if(!r) { /* child */
    if(close(sv[0])<0 || close(sv[2])<0 || dup2(sv[1],0)<0 || dup2(sv[1],1)<0 || dup2(sv[3],2)<0) _exit(-1);
    if(raw_fd>=0 && raw_fd!=3 && dup2(raw_fd,3)<0) _exit(-1);
    if(raw_fd<0) close(3);
    for(i=4; i<1000; i++) close(i);
    argv[0] = (char*)prefs.helpers.ssl_client_wrapper;
    argv[1] = ((raw_fd>=0)?"--fd":NULL);
    argv[2] = NULL;
    execve(argv[0], argv, environ);
    _exit(-1);
  }
  if(raw_fd>=0) close(raw_fd);
  close(sv[1]);
  close(sv[3]);
  buf0[0] = 0;
  buf0[1] = minver/256;
  buf0[2] = minver%256;
  buf0[3] = 0;
  buf0[4] = 2;
  buf0[5] = 0;
  buf0[6] = 0;
  buf0[7] = 3;
  buf0[8] = (addr >> 24) & 255;
  buf0[9] = (addr >> 16) & 255;
  buf0[10] = (addr >> 8) & 255;
  buf0[11] = (addr) & 255;
  buf0[12] = (port >> 8) & 255;
  buf0[13] = (port) & 255;
  buf0[14] = 0;
  if(lst_conn_send_all(sv[0],buf0,5)<0) { fail: err=errno; close(sv[0]); close(sv[2]); errno=err; return -1; }
  if(lst_conn_send_all(sv[0],cacert,strlen(cacert))<0) goto fail;
  if(lst_conn_send_all(sv[0],buf0+5,9)<0) goto fail;
  if(lst_conn_send_all(sv[0],hostname,strlen(hostname))<0) goto fail;
  if(lst_conn_send_all(sv[0],buf0+14,1)<0) goto fail;
  ctx->fd = sv[0];
  ctx->meta = sv[2];
  ctx->buf[0] = '\n'; ctx->pos = 1;
  if(flags & LST_FLAG_NONBLOCK_AFTER) ctx->nonblock = 1; else ctx->nonblock = 0;
  if(flags & LST_FLAG_NONBLOCK_BEFORE) {
    if(lst_conn_set_nonblock(sv[0])<0) goto fail;
    if(lst_conn_set_nonblock(sv[2])<0) goto fail;
    return lst_wait_conn_open_ssl(ctx, errmsg, errmsgsize);
  }
  while(1) {
    i = lst_wait_conn_open_ssl(ctx, errmsg, errmsgsize);
    if(i>=-1) return i;
    usleep(1000);
  }
}

/* returns:
  >=0 - returned fd is ready to work; *ctx is not used now
  -1  - error occured, errno/errmsg set properly, *ctx is not used now
  -2  - need wait more - call this function again with same args
  */
extern int lst_wait_conn_open_ssl(lst_conn_open_ssl_ctx * ctx, char * errmsg, size_t errmsgsize) {
  int err, i, meta;
  unsigned int pos, pp, lastpp;
  unsigned char * buf;
  meta = ctx->meta;
  buf = ctx->buf;
  pos = ctx->pos;
  while(1) {
    i = lst_conn_recv(meta, buf+pos, sizeof(ctx->buf)-pos);
    if(i<0) goto fail;
    if(!i) { ctx->pos=pos; return -2; }
    pos += i;
    if(pos<2) continue;
    for(lastpp=pp=0; pp<pos; pp++) {
      if(buf[pp]!='\n') {
        if(ctx->cstr) {
          if(ctx->cbufpos<ctx->cbufsz) ctx->cbuf[ctx->cbufpos++] = buf[pp];
          lastpp = pp+1;
        } else {
          if(pp-lastpp>16) lastpp=pp;
        }
        continue;
      }
      if(ctx->cstr) {
        if(ctx->cbufpos<ctx->cbufsz) ctx->cbuf[ctx->cbufpos++] = '\n';
        if(ctx->cbufpos<ctx->cbufsz) ctx->cbuf[ctx->cbufpos] = 0; else if(ctx->cbufsz) ctx->cbuf[ctx->cbufsz-1] = 0;
        ctx->cstr=0;
      }
      lastpp = pp;
      if(pos-pp>=7 && buf[pp+1]=='C' && buf[pp+2]=='E' && buf[pp+3]=='R' && buf[pp+4]=='T' && buf[pp+5]==':' &&
         buf[pp+6]==' ') { lastpp=pp+=7; ctx->cstr=1; pp--; continue; }
      if(pos-pp>=8 && buf[pp+1]=='E' && buf[pp+2]=='R' && buf[pp+3]=='R' && buf[pp+4]=='O' && buf[pp+5]=='R' &&
         buf[pp+6]==':' && buf[pp+7]==' ') { pp+=8; goto err_found; }
      if(pos-pp>=5 && buf[pp+1]=='O' && buf[pp+2]=='K' &&
         buf[pp+3]==':' && buf[pp+4]==' ') { pp+=5; goto ok_found; }
    }
    if(lastpp) { pos -= lastpp; memmove(buf, buf+lastpp, pos); }
  }
err_found:
  pos -= pp; if(pos) memmove(buf, buf+pp, pos);
  while(errmsgsize>=2) {
    if(pos>=errmsgsize-1) {
      memcpy(errmsg, buf, errmsgsize-1);
      errmsg[errmsgsize-1] = 0;
      errno = 0;
      goto fail;
    }
    if(pos) {
      memcpy(errmsg, buf, pos);
      errmsg += pos;
      errmsgsize -= pos;
      *errmsg = 0;
    }
    i = lst_conn_recv(meta, buf, sizeof(ctx->buf));
    if(i<0) { errno=0; goto fail; }
    if(!i) usleep(1000); /* this waiting is waiting for local unix socket and that side surely already has all data to send => no blocking possible */
    pos = i;
  }
  if(errmsgsize) *errmsg=0;
  errno = 0;
  goto fail;
ok_found:
  pos -= pp; if(pos) memmove(buf, buf+pp, pos);
  for(pp=0; pp<pos; pp++) if(buf[pp]=='\n') break;
  while(pp==pos) {
    i = lst_conn_recv(meta, buf, sizeof(ctx->buf));
    if(i<0) goto fail;
    if(!i) { usleep(1000); continue; }
    pos = i;
    for(pp=0; pp<pos; pp++) if(buf[pp]=='\n') break;
  }
  close(meta); /* TODO: child may do EPIPE; also this fd is useful - should send it to called somehow */
  if(ctx->nonblock) {
    if(lst_conn_set_nonblock(ctx->fd)<0) { err=errno; close(ctx->fd); errno=err; return -1; }
  }
  return ctx->fd;
fail:
  err=errno; close(ctx->fd); close(ctx->meta); errno=err; return -1;
}

/* it eats raw_fd in any case */
extern int lst_conn_accept_ssl_ex(int raw_fd, uint16 minver, char const * cacert, char const * cakey, char const * serverkey, char * errmsg, size_t errmsgsize, uint32 flags, char const * force_hostname) {
  int sv[4], err, i;
  char * argv[5];
  pid_t r;
  unsigned char buf0[256], buf[1024];
  unsigned int pos, pp, lastpp;
  char host_arg[512];
  
  if(!errmsg) errmsgsize = 0;
  if(errmsgsize) *errmsg = 0;
  if(socketpair(PF_UNIX, SOCK_STREAM, 0, sv)<0) { JUST_ERRNO("socketpair"); close(raw_fd); errno=0; return -1; }
  if(socketpair(PF_UNIX, SOCK_STREAM, 0, sv+2)<0) { JUST_ERRNO("socketpair"); close(raw_fd); close(sv[0]); close(sv[1]); errno=0; return -1; }
  if(sv[0]<=2 || sv[1]<=2 || sv[2]<=2 || sv[3]<=2 || raw_fd<=2) { JUST_ERRSTR("got fds<=2"); close(raw_fd); close(sv[0]); close(sv[1]); close(sv[2]); close(sv[3]); errno=0; return -1; }
  if((r=fork())<0) { JUST_ERRNO("fork"); close(raw_fd); close(sv[0]); close(sv[1]); close(sv[2]); close(sv[3]); errno=0; return -1; }
  if(!r) { /* child */
    if(close(sv[0])<0 || close(sv[2])<0 || dup2(sv[1],0)<0 || dup2(sv[1],1)<0 || dup2(sv[3],2)<0 || raw_fd!=3 && dup2(raw_fd,3)<0) _exit(-1);
    for(i=4; i<1000; i++) close(i);
    i = 0;
    argv[i++] = (char*)prefs.helpers.ssl_server_wrapper;
    if(prefs.local_ssl.wildcard_certs) argv[i++] = "--gen-wildcard";
    if(prefs.local_ssl.no_SAN) argv[i++] = "--no-SAN";
    if(force_hostname) {
      snprintf(host_arg, sizeof(host_arg), "--hostname=%s", force_hostname);
      argv[i++] = host_arg;
    }
    argv[i++] = NULL;
    execve(argv[0], argv, environ);
    _exit(-1);
  }
  close(raw_fd);
  close(sv[1]);
  close(sv[3]);
  buf0[0] = 0;
  buf0[1] = minver/256;
  buf0[2] = minver%256;
  buf0[3] = 0;
  buf0[4] = 1;
  /* cacer,t cakey, serverkey */
  buf0[5] = 1;
  buf0[6] = 2;
  buf0[7] = 0;
  buf0[8] = 0;
  buf0[9] = 3;
  if(lst_conn_send_all(sv[0],buf0,5)<0) { fail: JUST_ERRNO("internal send");  close(sv[0]); close(sv[2]); errno=0; return -1; }
  if(lst_conn_send_all(sv[0],cacert,strlen(cacert)+1)<0) goto fail;
  if(lst_conn_send_all(sv[0],cakey,strlen(cakey)+1)<0) goto fail;
  if(lst_conn_send_all(sv[0],serverkey,strlen(serverkey)+1)<0) goto fail;
  if(lst_conn_send_all(sv[0],buf0+5,5)<0) goto fail;
  buf[0] = '\n'; pos = 1;
  while(1) {
    i = lst_conn_recv(sv[2], buf+pos, sizeof(buf)-pos);
    if(i<0) { fail2: JUST_ERRNO("internal recv"); fail3: close(sv[0]); close(sv[2]); errno=0; return -1; }
    pos += i;
    if(!i) { usleep(1000); continue; }
    if(pos<2) continue;
    for(lastpp=pp=0; pp<pos; pp++) {
      if(buf[pp]!='\n') continue;
      lastpp = pp;
      if(pos-pp>=8 && buf[pp+1]=='E' && buf[pp+2]=='R' && buf[pp+3]=='R' && buf[pp+4]=='O' && buf[pp+5]=='R' &&
         buf[pp+6]==':' && buf[pp+7]==' ') { pp+=8; goto err_found; }
      if(pos-pp>=5 && buf[pp+1]=='O' && buf[pp+2]=='K' &&
         buf[pp+3]==':' && buf[pp+4]==' ') { pp+=5; goto ok_found; }
    }
    if(lastpp) { pos -= lastpp; if(pos) memmove(buf, buf+lastpp, pos); }
  }
err_found:
  pos -= pp; if(pos) memmove(buf, buf+pp, pos);
  while(errmsgsize>=2) {
    if(pos>=errmsgsize-1) {
      memcpy(errmsg, buf, errmsgsize-1);
      errmsg[errmsgsize-1] = 0;
      goto fail3;
    }
    if(pos) {
      memcpy(errmsg, buf, pos);
      errmsg += pos;
      errmsgsize -= pos;
      *errmsg = 0;
    }
    i = lst_conn_recv(sv[2], buf, sizeof(buf));
    if(i<0) goto fail3;
    if(!i) usleep(1000);
    pos = i;
  }
  if(errmsgsize) *errmsg=0;
  goto fail3;
ok_found:
  pos -= pp; if(pos) memmove(buf, buf+pp, pos);
  for(pp=0; pp<pos; pp++) if(buf[pp]=='\n') break;
  while(pp==pos) {
    i = lst_conn_recv(sv[2], buf, sizeof(buf));
    if(i<0) goto fail2;
    if(!i) { usleep(1000); continue; }
    pos = i;
    for(pp=0; pp<pos; pp++) if(buf[pp]=='\n') break;
  }
  close(sv[2]); /* TODO: child may do EPIPE; also thi fd is useful - should send it to called somehow */
  if(flags & (LST_FLAG_NONBLOCK_BEFORE | LST_FLAG_NONBLOCK_AFTER)) {
    if(lst_conn_set_nonblock(sv[0])<0) { JUST_ERRNO("set_nonblock"); close(sv[0]); errno=0; return -1; }
  }
  return sv[0];
}
