/*
 * This file is part of fproxy - multipurpose http(s) proxy server
 * helpers/remote.c - tunneling of proxied connection support
 * Copyright (c) 2018-2019, 2021 firk (firk@cantconnect.ru)
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. The origin of this software must not be misrepresented; you must
 *    not claim that you wrote the original software.
 * 4. Altered versions in any form must be plainly marked as such, and
 *    must not be misinterpreted as being the original software.
 *
 * This software is provided by the author and contributors `as is'
 * without any express or implied warranty.
 */


#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <strings.h>
#include <assert.h>
#include <errno.h>
#include <libsimpletcp/libsimpletcp.h>

#define MAX_NSLOOKUP_LEN 255
#ifdef BUILD_REMOTE_HANDLER
#include <netdb.h>
#include <sys/socket.h>
#include <arpa/inet.h>
#include <unistd.h>

#define SC_BUF_SIZE 262144
#define CS_BUF_SIZE 16384
#define MAX_CONN 100
typedef struct {
  int state;
  int fd;
  int rfd;
  char sc_buf[SC_BUF_SIZE];
  char cs_buf[CS_BUF_SIZE];
  size_t sc_head, sc_tail, cs_bytes;
  size_t remaining;
} t_conn;

static t_conn conns[MAX_CONN];

static uint32 do_nslookup(char const * hostname) {
  struct hostent * h;
  
  if(strlen(hostname)>255) return (uint32)-1;
  h = gethostbyname(hostname);
  if(!h || h->h_addrtype!=AF_INET || !h->h_addr_list || !h->h_addr_list[0]) return (uint32)-1;
  return ntohl(*(uint32*)(h->h_addr_list[0]));
}

// return <0 = will close conn; >0 = did something, not idle
static int handle_conn(t_conn * c) {
  int r, work;
  size_t sch, sct, scs, p, b;
  
  work = 0;
  assert(c->cs_bytes<CS_BUF_SIZE || c->state==4 && c->cs_bytes==CS_BUF_SIZE);
  assert(c->state>0);
  assert(c->fd>=0);
  if(c->cs_bytes<CS_BUF_SIZE) {
    r = lst_conn_recv(c->fd, c->cs_buf+c->cs_bytes, CS_BUF_SIZE-c->cs_bytes);
    if(r<0) return -1;
    if(r>0) { work=2; c->cs_bytes += r; }
  }
  sch = c->sc_head;
  sct = c->sc_tail;
  if(sch<sct) {
    r = lst_conn_send(c->fd, c->sc_buf+sct, SC_BUF_SIZE-sct);
    if(r<0) return -1;
    if(r>0) { work=2; sct = (sct+r)%SC_BUF_SIZE; }
  }
  if(sct<sch) {
    r = lst_conn_send(c->fd, c->sc_buf+sct, sch-sct);
    if(r<0) return -1;
    if(r>0) { work=2; sct = (sct+r)%SC_BUF_SIZE; }
  }
  c->sc_tail = sct;
  
  b = c->cs_bytes;
  switch(c->state) {
  case 1:
    assert(!sch && !sct);
    for(p=0; p<b; p++) if(!c->cs_buf[p]) break;
    if(p>MAX_NSLOOKUP_LEN) return -1;
    if(p>=b) return work;
    if(p>0) { // dns request
      if(b!=p+1) return -1;
      *(uint32*)(c->sc_buf) = do_nslookup((char*)c->cs_buf);
      c->sc_head = 4;
      c->state = 12;
      c->cs_bytes = 0;
//      printf("domain %s is 0x%08X\n", (char*)c->cs_buf, *(uint32*)(c->sc_buf));
      return work;
    }
    // proxy request
    if(b<7) return work;
    if(b>7) return -1;
//printf("connecting %u.%u.%u.%u:%u\n",(uchar)c->cs_buf[4],(uchar)c->cs_buf[3],(uchar)c->cs_buf[2],(uchar)c->cs_buf[1],*(uint16*)(c->cs_buf+5));
    c->rfd = lst_conn_open(*(uint32*)(c->cs_buf+1), *(uint16*)(c->cs_buf+5), LST_FLAG_NONBLOCK_BEFORE|LST_FLAG_NONBLOCK_AFTER);
    if(c->rfd<0) return -1;
    if(c->remaining) lst_conn_set_tcp_nodelay(c->rfd);
    c->state = 3;
    c->cs_bytes = 0;
    return work;
  case 12:
    if(b) return -1;
    if(sct==sch) { c->state = 1; c->sc_head = 0; c->sc_tail = 0; return 0; }
    return work;
  case 2:
    if(sct==sch) return -1;
    return work;
  case 3:
    if(b) return -1;
    r = lst_conn_wait_connection(c->rfd, 0);
    work = 1;
    if(r<0) return -1;
    if(r>0) {
      work = 2;
      c->state = 4;
      assert(!sch && !sct);
      c->sc_tail = 0;
      c->sc_head = 1;
      c->sc_buf[0] = 1;
    }
    return work;
  case 4:
    if(b) {
      if(!c->remaining) r = lst_conn_send(c->rfd, c->cs_buf, b);
      else { r = lst_conn_send(c->rfd, c->cs_buf, (b>3)?3:b); usleep(10000); }
      if(r<0) return -1;
      if(r>0) {
        if(c->remaining>(size_t)r) c->remaining-=r; else c->remaining=0;
        work = 2;
        assert((unsigned)r<=b);
        b -= r;
        if(b) bcopy(c->cs_buf+r, c->cs_buf, b);
        c->cs_bytes = b;
      }
    }
    if((sch+1)%SC_BUF_SIZE!=sct) {
      if(!sct) scs = SC_BUF_SIZE-1-sch; else if(sch>=sct) scs = SC_BUF_SIZE-sch; else scs = sct-1-sch;
      r = lst_conn_recv(c->rfd, c->sc_buf+sch, scs);
      if(r<0) { lst_conn_close(c->rfd,0); c->rfd=-1; c->state=2; return work; }
      if(r>0) { work=2; c->sc_head = (sch+r)%SC_BUF_SIZE; }
    }
    return work;
  default: assert(!"invalid state");
  }
  return 0;
}

int main(int argc, char const * const * argv) {
  uint32 addr, raddr;
  uint16 port, rport;
  int fd, accepted_fd, i, r, rr, chunk;
  if(argc==3 && !strcmp(argv[2],"chunk")) { chunk=1; argc=2; } else chunk=0;
  if(argc!=2 || lst_parse_addrstr(argv[1],&addr,&port)!=3) {
    fprintf(stderr,"Usage: proxy_r <listen-ip>:<listen-port>\n");
    return -1;
  }
  fd = lst_conn_listen(addr, port, 100, LST_FLAG_NONBLOCK_AFTER|LST_FLAG_REUSE_ADDR);
  if(fd<0) {
    fprintf(stderr,"proxy_r: failed to open listen socket at %s errno %d (%s)\n", argv[1], errno, strerror(errno));
    return -1;
  }
  fprintf(stderr,"proxy_r: started listening at %s\n", argv[1]);
  bzero(conns, sizeof(conns));
  accepted_fd = -1;
  while(1) {
try_accept:
    if(accepted_fd<0) accepted_fd = lst_conn_accept(fd, &raddr, &rport, LST_FLAG_NONBLOCK_AFTER);
    if(accepted_fd>=0) for(i=0; i<MAX_CONN; i++) if(!conns[i].state) {
      conns[i].state = 1;
      conns[i].fd = accepted_fd;
      conns[i].rfd = -1;
      conns[i].sc_head = 0;
      conns[i].sc_tail = 0;
      conns[i].cs_bytes = 0;
      if(chunk) conns[i].remaining = 200;
      accepted_fd = -1;
      goto try_accept;
    }
    rr = 0;
    for(i=0; i<MAX_CONN; i++) if(conns[i].state) {
      r = handle_conn(conns+i);
      if(r<0) {
        conns[i].state = 0;
        if(conns[i].fd>=0) lst_conn_close(conns[i].fd,0);
        if(conns[i].rfd>=0) lst_conn_close(conns[i].rfd,0);
      }
      if(r && !rr) rr=1;
      if(r==2) rr=2;
    }
    if(rr==0) usleep(50000);
    if(rr==1) usleep(1000);
  }
  lst_conn_close(fd,0);
}
#else
#include "remote.h"
/* client */
extern uint32 lst_remote_nslookup(int * proxy_fd, uint32 proxy_addr, uint16 proxy_port, char const * d) {
  int fd, r;
  size_t len;
  char buf[4];
  fd = *proxy_fd;
  if(fd<0) *proxy_fd = fd = lst_conn_open(proxy_addr, proxy_port, 0);
  if(fd<0) return (uint32)-1;
  len = strlen(d);
  if(len<1 || len>MAX_NSLOOKUP_LEN) return (uint32)-1;
  if(lst_conn_send_all(fd, d, len+1)<0) { lst_conn_close(fd,0); *proxy_fd = -1; return (uint32)-1; }
  len = 0;
  while(len<4) {
    r = lst_conn_recv(fd, buf+len, 4-len);
    if(r<0) { lst_conn_close(fd,0); *proxy_fd = -1; return (uint32)-1; }
    assert((unsigned)r<=4-len);
    len += r;
  }
  return *(uint32*)buf;
}
extern int lst_remote_conn_open(uint32 proxy_addr, uint16 proxy_port, uint32 addr, uint16 port, uint32 flags) {
  int fd, r;
  char buf[7];
  fd = lst_conn_open(proxy_addr, proxy_port, 0);
  if(fd<0) return -1;
  buf[0] = 0;
  *(uint32*)(buf+1) = addr;
  *(uint16*)(buf+5) = port;
  if(lst_conn_send_all(fd, buf, 7)<0) { lst_conn_close(fd,0); return -1; }
  while(1) {
    r = lst_conn_recv(fd, buf, 1);
    if(r<0) { lst_conn_close(fd,0); return -1; }
    if(r==1) break;
    assert(!r);
  }
  if(flags & LST_FLAG_NONBLOCK_AFTER) {
    if(lst_conn_set_nonblock(fd)<0) { lst_conn_close(fd,0); return -1; }
  }
  return fd;
}
#endif
