/*
 * This file is part of fproxy - multipurpose http(s) proxy server
 * helpers/server.c - main source file for ssl-server-wrapper
 * Copyright (c) 2018-2019,2022 firk (firk@cantconnect.ru)
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. The origin of this software must not be misrepresented; you must
 *    not claim that you wrote the original software.
 * 4. Altered versions in any form must be plainly marked as such, and
 *    must not be misinterpreted as being the original software.
 *
 * This software is provided by the author and contributors `as is'
 * without any express or implied warranty.
 */

#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>
#include <string.h>
#include <errno.h>
#include <unistd.h>
#include <openssl/evp.h>
#include <openssl/ssl.h>
#include <openssl/err.h>
#include <libsimpletcp/libsimpletcp.h>
#include "cert_gen.h"

static void ssl_fail(char const * fmt, ...) {
  va_list arg;
  va_start(arg, fmt);
  fprintf(stderr, "ERROR: ");
  vfprintf(stderr, fmt, arg);
  fprintf(stderr, "\n");
  ERR_print_errors_fp(stderr);
  va_end(arg);
  // may be SSL_CTX_free() here?
  exit(-1);
}

static void raw_fail(char const * fmt, ...) {
  va_list arg;
  va_start(arg, fmt);
  fprintf(stderr, "ERROR: ");
  vfprintf(stderr, fmt, arg);
  fprintf(stderr, "\n");
  va_end(arg);
  // may be SSL_CTX_free() here?
  exit(-1);
}

static void setup_minversion(SSL_CTX * ctx, unsigned short minver, int legacy_server) {
  if(minver>0x0200) SSL_CTX_set_options(ctx, SSL_OP_NO_SSLv2);
  if(minver>0x0300) SSL_CTX_set_options(ctx, SSL_OP_NO_SSLv3);
  if(minver>0x0301) SSL_CTX_set_options(ctx, SSL_OP_NO_TLSv1);
#ifdef SSL_OP_NO_TLSv1_1
  if(minver>0x0302) SSL_CTX_set_options(ctx, SSL_OP_NO_TLSv1_1);
#endif
#ifdef SSL_OP_NO_TLSv1_2
  if(minver>0x0303) SSL_CTX_set_options(ctx, SSL_OP_NO_TLSv1_2);
#endif
  /* allow "legacy insecure renegotiation with unpatched servers"; idk what is it; see man SSL_CTX_set_options */
  if(legacy_server) SSL_CTX_set_options(ctx, SSL_OP_LEGACY_SERVER_CONNECT);
  else SSL_CTX_clear_options(ctx, SSL_OP_LEGACY_SERVER_CONNECT);
  SSL_CTX_set_options(ctx, SSL_OP_CIPHER_SERVER_PREFERENCE);
}

static EVP_PKEY *dyn_ca_key, *dyn_server_key;
static X509 *dyn_ca_cert;
static int gen_wildcard = 0;
static int use_SAN = 1;
static int ignore_SNI = 0;
static char const *default_hostname;


static int servername_callback(SSL * ssl, int * dummy, void * arg) {
  char const * sni;
  char const *tt;
  char *sni2;
  X509 * newcert;
  sni = SSL_get_servername(ssl, TLSEXT_NAMETYPE_host_name);
  if(!sni || !*sni) return SSL_TLSEXT_ERR_NOACK;
  sni2 = NULL;
  if(gen_wildcard && (tt=strchr(sni,'.')) && tt>sni && strchr(tt+1,'.')) {
    sni2 = strdup(tt-1); /* just fallback to non-wildcard if fail */
    if(sni2) *sni2='*';
  }
  newcert = gen_dyncert_cache(dyn_ca_key, dyn_ca_cert, dyn_server_key, sni2?sni2:sni, use_SAN);
  if(sni2) free(sni2);
  if(!newcert) raw_fail("bad hostname");
  if(SSL_use_certificate(ssl, newcert)<=0) ssl_fail("failed loading dynamic server-cert");
  if(!SSL_check_private_key(ssl)) raw_fail("dynamic server-cert and server-key mismatch");
  /* key already loaded and not changed */
  return SSL_TLSEXT_ERR_OK;
}

static void setup_server_cert(SSL_CTX * ctx, char const * certfile, char const * keyfile, char const * genkeyfile, int autogen_SNI) {
  X509 * newcert;
  if(autogen_SNI) {
    /* TODO: what of these should be freed some time? /// don't care, this program will handle and exit anyway */
    dyn_ca_key = read_pkey(keyfile);
    dyn_server_key = read_pkey(genkeyfile);
    dyn_ca_cert = read_cert(certfile);
    if(!ignore_SNI) SSL_CTX_set_tlsext_servername_callback(ctx, servername_callback);
    if(!default_hostname) {
      newcert = gen_dyncert_cache(dyn_ca_key, dyn_ca_cert, dyn_server_key, "localhost", 0);
      if(!newcert) raw_fail("can't generate cert for localhost - dyncert dir too long?");
      if(SSL_CTX_use_certificate(ctx, newcert)<=0) ssl_fail("failed loading default server-cert");
      if(SSL_CTX_use_PrivateKey(ctx, dyn_server_key)<=0) ssl_fail("failed loading server-key");
      if(!SSL_CTX_check_private_key(ctx)) raw_fail("default server-cert and server-key mismatch");
    } else {
      newcert = gen_dyncert_cache(dyn_ca_key, dyn_ca_cert, dyn_server_key, default_hostname, 0);
      if(!newcert) raw_fail("can't generate cert for specified hostname - bad hostname? dyncert dir too long?");
      if(SSL_CTX_use_certificate(ctx, newcert)<=0) ssl_fail("failed loading generated server-cert");
      if(SSL_CTX_use_PrivateKey(ctx, dyn_server_key)<=0) ssl_fail("failed loading server-key");
      if(!SSL_CTX_check_private_key(ctx)) raw_fail("generated server-cert and server-key mismatch");
    }
  } else {
    /* Load the client certificate into the SSL_CTX structure */
    if(SSL_CTX_use_certificate_file(ctx, certfile, SSL_FILETYPE_PEM)<=0) ssl_fail("failed loading server-cert file");
    /* Load the private-key corresponding to the client certificate */
    if(SSL_CTX_use_PrivateKey_file(ctx, keyfile, SSL_FILETYPE_PEM)<=0) ssl_fail("failed loading server-key file");
    /* Check if the client certificate and private-key matches */
    if(!SSL_CTX_check_private_key(ctx)) raw_fail("server-cert and server-key mismatch");
  }
}

/* this is required to verify server cert */
static void setup_trusted_ca_list(SSL_CTX * ctx, char const * listfile, int enforce) {
  if(listfile && *listfile && !SSL_CTX_load_verify_locations(ctx, listfile, NULL)) ssl_fail("error loading trusted ca list");
  SSL_CTX_set_verify(ctx, enforce?SSL_VERIFY_PEER:SSL_VERIFY_NONE, NULL);
  /* in SSL_VERIFY_NONE mode, certificate still checked but not fatal;
     SSL_get_verify_result() can be used to get results */
  SSL_CTX_set_verify_depth(ctx, 9); /* allow max 9 ca-certs (10 total) in server cert-chain */
}

static SSL * accept_connection_by_fd(SSL_CTX * ctx, int fd) {
  int ret;
  SSL * ssl;
  char * tmp;
  
  ssl = SSL_new(ctx);
  if(!ssl) ssl_fail("SSL_new() failed");
  if(!SSL_set_fd(ssl, fd)) ssl_fail("SSL_set_fd() failed");
  while((ret=SSL_accept(ssl))<=0) { /* non-blocking side effects workaround */
    if(ret<0) switch(SSL_get_error(ssl,ret)) {
    case SSL_ERROR_WANT_READ:  lst_conn_wait_readable(fd, 1000); continue;
    case SSL_ERROR_WANT_WRITE: lst_conn_wait_writable(fd, 1000); continue;
    }
    ssl_fail("SSL_accept() returned %d", ret);
  }

#ifdef DEBUG_SSL_SERVER
  fprintf(stderr, "DEBUG: protocol = %s\nDEBUG: cipher = %s\n", SSL_get_version(ssl), SSL_get_cipher(ssl));
#endif
  fprintf(stderr,"OK: accepted connection\n");
  return ssl;
}

/* wait_reply means "wait for remote shutdown too";
   this is not essential by protocol, but may be needed in some cases by app */
static void end_connection(SSL * ssl, int fd, int wait_reply) {
  int ret;
  while((ret=SSL_shutdown(ssl))<=0) { /* non-blocking side effects workaround */
    if(ret<0) switch(SSL_get_error(ssl,ret)) {
    case SSL_ERROR_WANT_READ:  lst_conn_wait_readable(fd, 1000); continue;
    case SSL_ERROR_WANT_WRITE: lst_conn_wait_writable(fd, 1000); continue;
    default: ssl_fail("SSL_shutdown() returned %d", ret);
    }
    if(!wait_reply) break;
  }
  if(close(fd)<0) raw_fail("close() errno %d (%s)", errno, strerror(errno));
  SSL_free(ssl);
}

/* SSL_write() is not in-place replacement for write() in case of being deferred;
   it should be retries with exactly same args and same data in buffer;
   also it usually does success until all bytes sent;
   see SSL_set_mode SSL_MODE_ENABLE_PARTIAL_WRITE, SSL_MODE_ACCEPT_MOVING_WRITE_BUFFER for details */
/* this wrapper always sends all bytes until return or get fatal error */
static void send_ssl_bytes(SSL * ssl, int fd, unsigned char const * b, size_t n) {
  int ret;
  int csz;
  while(n) {
    csz = 16000;
    if(((unsigned)csz)>n) csz = (int)n;
    ret = SSL_write(ssl, b, csz);
    if(ret>0) {
      if(ret>csz) raw_fail("BUG! ret>csz for SSL_write()");
      n -= ret;
      b += ret;
      continue;
    }
    switch(SSL_get_error(ssl,ret)) {
    case SSL_ERROR_WANT_READ:  lst_conn_wait_readable(fd, 1000); continue;
    case SSL_ERROR_WANT_WRITE: lst_conn_wait_writable(fd, 1000); continue;
    case SSL_ERROR_ZERO_RETURN: raw_fail("remote peer closed SSL session while we are trying SSL_write()");
    case SSL_ERROR_SYSCALL: raw_fail("SSL_write() failed due to IO errno %d (%s)", errno, strerror(errno));
    case SSL_ERROR_SSL: ssl_fail("SSL_write() failed in protocol handling");
    default: ssl_fail("SSL_write() failed due to unknown reason");
    }
  }
}

/* may recv less bytes than requested */
static size_t recv_ssl_bytes(SSL * ssl, int fd, unsigned char * b, size_t n) {
  int ret;
  int csz;
  while(n) {
    csz = 16000;
    if(((unsigned)csz)>n) csz = (int)n;
    ret = SSL_read(ssl, b, csz);
    if(ret>0) {
      if(ret>csz) raw_fail("BUG! ret>csz for SSL_read()");
      n -= ret;
      b += ret;
      return ret;
    }
    switch(SSL_get_error(ssl,ret)) {
    case SSL_ERROR_WANT_READ:  return 0; // lst_conn_wait_readable(fd, 1); continue;
    case SSL_ERROR_WANT_WRITE: lst_conn_wait_writable(fd, 1000); continue;
    case SSL_ERROR_ZERO_RETURN: raw_fail("remote peer closed SSL session while we are trying SSL_read()");
    case SSL_ERROR_SYSCALL: raw_fail("SSL_read() failed due to IO errno %d (%s)", errno, strerror(errno));
    case SSL_ERROR_SSL: ssl_fail("SSL_read() failed in protocol handling");
    default: ssl_fail("SSL_read() failed due to unknown reason");
    }
  }
  return 0;
}

/*
this is not correct; it only returns number of bytes that already pulled by previous actions but not
received by SSL_read(); it seems to ignore any pending data in socket buffer */
/*static int is_readable(SSL * ssl) {
  if(SSL_pending(ssl)) return 1;
  return 0;
}*/

/* CLIENT = parent process who interacts with us via fd0/fd1/fd2
   SERVER = remote server accessed via OpenSSL */
static unsigned char SCbuf[20000];
static unsigned char CSbuf[20000];
static size_t SCbytes, CSbytes;
static int input_closed;

static int update_client_bufs(void) {
  ssize_t r;
  int work = 0;
  if(CSbytes<sizeof(CSbuf)) {
    r = read(0, CSbuf+CSbytes, sizeof(CSbuf)-CSbytes);
    if(r<0 && (errno==ECONNABORTED || errno==ECONNRESET || errno==ECONNREFUSED)) input_closed = 1; else /* we should wait while SSL side receive all data */
    if(r<0 && errno!=EWOULDBLOCK && errno!=EAGAIN && errno!=EINTR && errno!=ENOBUFS && errno!=ENOMEM) exit(-1); /* client disconnected, nothing to do now */
    if(!r) input_closed = 1;
    if(r>0) { work=1; CSbytes += r; }
  }
  if(SCbytes) {
    r = write(1, SCbuf, SCbytes);
    if(!r || ((size_t)r)>SCbytes || r<0 && errno!=EWOULDBLOCK && errno!=EAGAIN && errno!=EINTR && errno!=ENOBUFS && errno!=ENOMEM) exit(-1); /* client disconnected, nothing to do now */
    if(r>0) {
      work = 1;
      SCbytes -= r;
      if(SCbytes) memmove(SCbuf, SCbuf+r, SCbytes);
    }
  }
  return work;
}

static void CS_drain(size_t n) {
  CSbytes -= n;
  if(CSbytes) memmove(CSbuf, CSbuf+n, CSbytes);
}

static void do_things(SSL_CTX * ctx) {
  SSL * ssl;
  int work, fd;
  size_t n, n0, n1;
  
  if(lst_conn_set_nonblock(0)<0 || lst_conn_set_nonblock(1)<0 || lst_conn_set_nonblock(3)) raw_fail("can't set nonblock");
  ssl = NULL;
  
  work = 0;
  input_closed = 0;
  while(!ssl) { /* pre-connection loop */
    if(update_client_bufs()) work = 1;
    if(input_closed) exit(0);
    if(!work) { if(CSbytes==sizeof(CSbuf)) raw_fail("input data malformed"); usleep(10000); continue; }
    work = 0;
    if(!CSbytes) continue;
    switch(CSbuf[0]) {
    case 0:
      if(CSbytes<4) continue;
      setup_minversion(ctx, CSbuf[1]*0x100+CSbuf[2], CSbuf[3]);
      CS_drain(4);
      work = 1;
      continue;
    case 1:
      for(n=1; n<CSbytes && CSbuf[n]; n++);
      if(CSbytes-n<2) continue;
      n0 = ++n;
      for(; n<CSbytes && CSbuf[n]; n++);
      if(CSbytes-n<2) continue;
      n1 = ++n;
      for(; n<CSbytes && CSbuf[n]; n++);
      if(CSbytes-n<2) continue;
      setup_server_cert(ctx, (char*)(CSbuf+1), (char*)(CSbuf+n0), (char*)(CSbuf+n1), CSbuf[n+1]);
      CS_drain(n+2);
      work = 1;
      continue;
    case 2:
      for(n=1; n<CSbytes && CSbuf[n]; n++);
      if(CSbytes-n<2) continue;
      setup_trusted_ca_list(ctx, (char*)(CSbuf+1), CSbuf[n+1]);
      CS_drain(n+2);
      work = 1;
      continue;
    case 3:
      fd = 3;
      ssl = accept_connection_by_fd(ctx, fd);
      CS_drain(1);
      work = 1;
      continue;
    default: raw_fail("bad opcode 0x%02X", CSbuf[0]);
    }
  }
  
  work = (CSbytes?1:0);
  while(1) { /* connected loop */
    if(update_client_bufs()) work = 1;
    if(input_closed && !CSbytes) break;
    if(!SCbytes) {
      SCbytes = recv_ssl_bytes(ssl, fd, SCbuf, sizeof(SCbuf));
      if(SCbytes) work = 1;
    }
    if(!work) { usleep(10000); continue; }
    work = 0;
    if(CSbytes) {
      send_ssl_bytes(ssl, fd, CSbuf, CSbytes);
      CSbytes = 0;
    }
  }
  end_connection(ssl, fd, 0);
}

/* fd3 is accepted socket (connected to remote client) */
int main(int argc, char * * argv) {
  SSL_CTX * ctx;
  int i;
  
  for(i=1; i<argc; i++) {
    if(!strcmp(argv[i],"--gen-wildcard")) gen_wildcard=1;
    if(!strcmp(argv[i],"--no-SAN")) use_SAN=0;
    if(!strncmp(argv[i],"--hostname=",11)) default_hostname = argv[i]+11;
    if(!strncmp(argv[i],"--force-hostname=",17)) { default_hostname = argv[i]+17; ignore_SNI=1; }
  }
  lst_init();
  
  /**** init[1] register error strings */
  /* ERR_load_BIO_strings() - what is it? */
  SSL_load_error_strings();

  /**** init[2] register algorythms */
  SSL_library_init();
  /* OpenSSL_add_all_algorithms(); - if ALL algorythms needed, not just SSL/TLS (idk what it means, see manpages) */

  /**** init[3] create ssl context */
  ctx = SSL_CTX_new(SSLv23_server_method()); /* this construction means: create context and allow all known protocols; see manpage if doubt */
  if(!ctx) ssl_fail("SSL_CTX_new() failed");

  /**** do all things */
  do_things(ctx);

  /**** done[3] free ssl context */
  SSL_CTX_free(ctx);

  /**** done[2] unregoster algorythms */
  EVP_cleanup();

  /**** done[1] unregister error strings */
  ERR_free_strings();
  return 0;
}
