/*
 * This file is part of fproxy - multipurpose http(s) proxy server
 * http_packet.c - http protocol handler
 * Copyright (c) 2018-2020 firk (firk@cantconnect.ru)
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. The origin of this software must not be misrepresented; you must
 *    not claim that you wrote the original software.
 * 4. Altered versions in any form must be plainly marked as such, and
 *    must not be misinterpreted as being the original software.
 *
 * This software is provided by the author and contributors `as is'
 * without any express or implied warranty.
 */

#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>
#include <string.h>
#include <strings.h>
#include <errno.h>
#include <fcntl.h>
#include <unistd.h>
#include <time.h>
#include <assert.h>
#include <sys/stat.h>
#include <fcl/types.h>
#include <fcl/malloc.h>
#include "eventlog.h"
#include "util.h"
#include "http_packet.h"


/**********************************************************************************************************************/
/**********************************************************************************************************************/
/**********************************************************************************************************************/

/* ret. 0 = ok; -1 = too many headers; -2 = too long header name; -3 = bad header name; -4 = too long header value; -5 = missing ':' */
extern int hh_add(t_http_headers * hh, char const * text) {
  int r;
  unsigned int n;
  char *hname, *hvalue;
  
  if(hh->n>MAX_HEADERS) abort();
  if(hh->n==MAX_HEADERS) return -1;
  r = hh_parse(text, &hname, &hvalue);
  if(r<0) return r;
  n = hh->n++;
  hh->h[n*2] = hname;
  hh->h[n*2+1] = hvalue;
  return 0;
}

/* ret. 0 = ok; -2 = too long header name; -3 = bad header name; -4 = too long header value; -5 = missing ':' */
extern int hh_parse(char const * text, char * * r_name, char * * r_value) {
  unsigned int n, ow;
  char c;
  char name[MAX_HEADER_NAME];
  char *hname, *hvalue;
  size_t vlen;
  
  while(*text==' ' || *text=='\t') text++;
  for(n=0; n<MAX_HEADER_NAME; n++) {
    c = text[n];
    if(!c) return -5;
    if(c==':') break;
    name[n] = c;
  }
  if(n==MAX_HEADER_NAME) return -2;
  text += n+1;
  while(n && (name[n-1]==' ' || name[n-1]=='\t')) n--;
  name[n] = 0;
  if(!n) return -3;
  for(n=ow=0; c=name[n]; n++) {
    if(c=='_' || c=='-') { name[n]='-'; ow=0; continue; }
    if(c>='A' && c<='Z') { if(ow) name[n]=c-'A'+'a'; ow=1; continue; }
    if(c>='a' && c<='z') { if(!ow) name[n]=c-'a'+'A'; ow=1; continue; }
    if(c>='0' && c<='9') { ow=1; continue; }
    return -3;
  }
  while(*text==' ' || *text=='\t') text++;
  if(strlen(text)>=MAX_HEADER_VALUE) return -4;
  hname = fcl_strdup_check(name);
  hvalue = fcl_strdup_check(text);
  vlen = strlen(hvalue);
  while(vlen && (hvalue[vlen-1]==' ' || hvalue[vlen-1]=='\t')) hvalue[--vlen] = 0;
  *r_name = hname;
  *r_value = hvalue;
  return 0;
}

extern void hh_clear(t_http_headers * hh) {
  unsigned int n;
  for(n=hh->n; n; ) {
    n--;
    free(hh->h[n*2]);
    free(hh->h[n*2+1]);
  }
  hh->n = 0;
}

extern char const * hh_get(t_http_headers const * hh, char const * name) {
  unsigned int n;
  for(n=hh->n; n; ) {
    n--;
    if(!strcmp(name, hh->h[n*2])) return hh->h[n*2+1];
  }
  return NULL;
}


/**********************************************************************************************************************/
/**********************************************************************************************************************/
/**********************************************************************************************************************/

static int hc_openfile(void) {
  char fn[250];
  int fd;
  snprintf(fn, sizeof(fn), "./proxy_temp/%u", (unsigned)getpid());
  if(strlen(fn)+1==sizeof(fn)) {
    eventlog(eventlog_level_fatal, __FUNCTION__, "[BUG] tmpfile name too long");
    exit(-1);
  }
  unlink(fn);
  while((fd = open(fn, O_CREAT|O_EXCL|O_RDWR, 0))<0) if(errno!=EINTR) return -1;
  unlink(fn);
  return fd;
}

extern int hc_init(t_http_content * hc, unsigned long max_size) {
  bzero(hc, sizeof(*hc));
  hc->total = max_size;
  if(max_size<=MAX_MEM_CONTENT) {
    hc->alloc = max_size?max_size:INIT_MEM_CONTENT;
    hc->d = fcl_malloc_check(hc->alloc);
    return 0;
  }
  if(max_size<=MAX_FILE_CONTENT) {
    if((hc->fd = hc_openfile())<0) return -1;
    hc->use_fd = 1;
    return 0;
  }
  errno=EFBIG; return -1;
}

extern void hc_clear(t_http_content * hc) {
  if(hc->d) { assert(hc->alloc); free(hc->d); hc->d=NULL; }
  if(hc->use_fd) { assert(hc->fd>=0); assert(hc->cur>MAX_MEM_CONTENT || hc->total>MAX_MEM_CONTENT); close(hc->fd); hc->use_fd=0; }
  bzero(hc, sizeof(*hc));
}

extern int hc_add(t_http_content * hc, unsigned char const * buf, size_t len) {
  unsigned long cur, max, alloc, newalloc;
  int errs;
  unsigned char * b;
  
  if(!len) return 0;
  cur = hc->cur;
  alloc = hc->alloc;
  if(max=hc->total) {
    assert(cur<=max);
    if(len>max-cur) {
      eventlog(eventlog_level_fatal, __FUNCTION__, "[BUG] try to add data more than declared initially");
      exit(-1);
    }
    if(!hc->use_fd) {
      assert(alloc==max);
      assert(max<=MAX_MEM_CONTENT);
      memcpy(hc->d+cur, buf, len);
      hc->cur = cur+len;
      return 0;
    }
  } else { /* dynamic sized */
    if(!hc->use_fd) {
      assert(cur<=MAX_MEM_CONTENT);
      if(len<=MAX_MEM_CONTENT-cur) {
        assert(cur<=alloc);
        if(cur+len>alloc) {
          newalloc = cur+len;
          if(newalloc/2>MAX_MEM_CONTENT-newalloc) newalloc=MAX_MEM_CONTENT; else newalloc+=newalloc/2;
          b = realloc(hc->d, newalloc);
          if(!b) goto go_file;
          hc->d = b;
          hc->alloc = alloc = newalloc;
        }
        assert(cur+len<=alloc);
        memcpy(hc->d+cur, buf, len);
        hc->cur = cur+len;
        return 0;
      }
go_file:
      if((hc->fd = hc_openfile())<0) return -1;
      hc->use_fd = 1;
      if(write_all(hc->fd, hc->d, cur)<0) { errs=errno; close(hc->fd); errno=errs; hc->fd=-1; hc->use_fd=0; return -1; }
      free(hc->d); hc->d=NULL; hc->alloc=0;
    }
  }
  assert(!hc->d && !hc->alloc);
  assert(hc->use_fd && hc->fd>=0 && cur<=MAX_FILE_CONTENT);
  if(len>MAX_FILE_CONTENT-cur) { errno=EFBIG; return -1; }
  if(write_all_2(hc->fd, buf, len, &hc->cur)<0) return -1;
  return 0;
}


/**********************************************************************************************************************/
/**********************************************************************************************************************/
/**********************************************************************************************************************/

static int spc(char c) { return (c==' ' || c=='\t'); }
/* return:
    0 = ok
    -1 = syntax error while parsing http-method
    -2 = bad method (hp->mstr contains method name)
    -3 = missing version
    -4 = bad url
    -5 = url too long
    -6 = bad version
 */
extern int hp_parse_firstline(t_http_packet * hp, char * fl) {
  char ch, *url;
  size_t n, pos;
  
  n = 0;
  while(1) {
    if(n>=sizeof(hp->mstr) || !(ch=fl[n])) return -1;
    if(spc(ch)) { hp->mstr[n]=0; fl+=n+1; break; }
    if(ch>='a' && ch<='z') ch=ch-'a'+'A';
    if(ch<'A' || ch>'Z') return -1;
    hp->mstr[n] = ch;
    n++;
  }
  if(!strcmp(hp->mstr,"GET")) hp->method=M_GET;
  else if(!strcmp(hp->mstr,"HEAD")) hp->method=M_HEAD;
  else if(!strcmp(hp->mstr,"POST")) hp->method=M_POST;
  else if(!strcmp(hp->mstr,"CONNECT")) hp->method=M_CONNECT;
  else if(!strcmp(hp->mstr,"OPTIONS")) hp->method=M_OPTIONS;
  else return -2;
  while(spc(*fl)) fl++;
  
  for(url=fl; !spc(ch=*fl); fl++) if(!ch) return -3;
  *(fl++) = 0;
  while(spc(*fl)) fl++;
  if(hp->method!=M_CONNECT) {
    if(!strncmp(url, "http://", 7)) { url = strchr(url+7, '/'); if(!url) return -4; }
    else if(*url!='/') return -4;
  }
  n = strlen(url);
  if(n>=sizeof(hp->raw_url)) return -5;
  bcopy(url, hp->raw_url, n);
  hp->raw_url[n] = 0;
  
  if(fl[0]!='H' || fl[1]!='T' || fl[2]!='T' || fl[3]!='P' || fl[4]!='/' || fl[5]<'1' || fl[5]>'9' || fl[6]!='.' || fl[7]<'0' || fl[7]>'9' || fl[8]) return -6;
  hp->v1 = fl[5];
  hp->v2 = fl[7];
  return 0;
}

