/*
 * This file is part of fproxy - multipurpose http(s) proxy server
 *
 * Copyright (c) 2017-2018 firk (firk@cantconnect.ru)
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. The origin of this software must not be misrepresented; you must
 *    not claim that you wrote the original software.
 * 4. Altered versions in any form must be plainly marked as such, and
 *    must not be misinterpreted as being the original software.
 *
 * This software is provided by the author and contributors `as is'
 * without any express or implied warranty.
 */

/* (c) firk 2017-2018 */
/* this is old proxy
1) it is single-process
2) it only tunnels all connections to real web-server, removing SSL wrap if needed
*/
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <strings.h>
#include <errno.h>
#include <unistd.h>
#include <sys/wait.h>
#include "eventlog.h"
#include "helpers/libsimpletcp.h"
#include "helpers/lst_ssl.h"

#define IP "127.0.0.10"
#define PORT 3128
#define TARGET_IP "127.0.0.11"
#define MAXCONN 100
#define FIRSTLINELEN 4096
#define INBSIZE 4096
#define OUTBSIZE 4096

#define state_closed 1
#define state_connected 2
#define state_work 3
#define state_https_pending 4
#define state_tunnel 5

typedef struct {
  int   state;
  int   fd;
  int   rfd;
  int   method;
  char  v1,v2,https;
  uint32 addr;
  uint16 port;
  uint32 inbytes;
  uint32 outbytes;
  uint32 maxrecv;
  uint8 inb[INBSIZE+40]; /* room for https marker */
  uint8 outb[OUTBSIZE];
} t_connection;

static int get_line(char const * data, unsigned int len) {
  unsigned int n;
  for(n=0; n<len; n++) {
    if(!data[n]) return -1;
    if(data[n]=='\n') return n+1;
  }
  return 0;
}

#define M_GET 0
#define M_HEAD 1
#define M_POST 2
#define M_CONNECT 3

/* return NULL on error, return url else */
static char * parse_firstline(t_connection * conn, char * fl) {
  char mstr[16], ch, *url;
  unsigned int n, p;
  int method;
  
  n = 0;
  while(1) {
    if(n>=sizeof(mstr) || !(ch=fl[n])) return NULL;
    if(ch==' ' || ch=='\t') { mstr[n]=0; fl+=n+1; break; }
    if(ch>='a' && ch<='z') ch=ch-'a'+'A';
    if(ch<'A' || ch>'Z') return NULL;
    mstr[n] = ch;
    n++;
  }

  if(!strcmp(mstr,"GET")) method=M_GET;
  else if(!strcmp(mstr,"HEAD")) method=M_HEAD;
  else if(!strcmp(mstr,"POST")) method=M_POST;
  else if(!strcmp(mstr,"CONNECT")) method=M_CONNECT;
  else { eventlog(eventlog_level_warn,__FUNCTION__,"[%d] bad method '%s'",conn->fd,mstr); return NULL; }
  conn->method=method;
  while(ch=*fl) { if(ch!=' ' && ch!='\t') break; fl++; }
  
  url = fl;
  while(ch=*fl) { if(ch==' ' || ch=='\t') break; fl++; }
  if(!ch) { eventlog(eventlog_level_warn,__FUNCTION__,"[%d] missing version",conn->fd); return NULL; }
  *fl=0;
  fl++;
  
  while(ch=*fl) { if(ch!=' ' && ch!='\t') break; fl++; }
  if(fl[0]!='H' || fl[1]!='T' || fl[2]!='T' || fl[3]!='P' || fl[4]!='/' || fl[5]<'1' || fl[5]>'9' || fl[6]!='.' || fl[7]<'0' || fl[7]>'9' || fl[8]) {
    eventlog(eventlog_level_warn,__FUNCTION__,"[%d] bad version",conn->fd); return NULL;
  }
  conn->v1 = fl[5];
  conn->v2 = fl[7];
  return url;
}

/* return: 0=nothing; 1=some work done; -1=need close */
static int conn_process(t_connection * c) {
  int r, work;
  unsigned int n, nn;
  uint8 * buf;
  char fl[FIRSTLINELEN], nl[FIRSTLINELEN+40], *mstr;
  char * url; /* points somewhere inside fl[] */
  uint32 rtry;
  char errmsg[1024];
  
  work = 0;
  if(c->state!=state_connected && c->state!=state_work && c->state!=state_https_pending && c->state!=state_tunnel) {
    eventlog(eventlog_level_fatal,__FUNCTION__,"[%d] bad state %d - internal bug, exiting for safety", c->fd, c->state);
    abort();
  }
  if(c->outbytes) {
    r = lst_conn_send(c->fd, c->outb, c->outbytes);
    if(r<0) { eventlog(eventlog_level_warn,__FUNCTION__,"[%d] write error %d, closing", c->fd, lst_get_errcode()); return -1; }
    if(r>0) { c->outbytes -= r; if(c->outbytes) bcopy(c->outb+r, c->outb, c->outbytes); work = 1; }
  }
  if(c->inbytes<INBSIZE && c->maxrecv) {
    rtry = INBSIZE-c->inbytes;
    if(rtry>c->maxrecv) rtry = c->maxrecv;
    r = lst_conn_recv(c->fd, c->inb+c->inbytes, rtry);
    if(r<0) { eventlog(eventlog_level_warn,__FUNCTION__,"[%d] read error %d, closing", c->fd, lst_get_errcode()); return -1; }
    if(r>0) { c->inbytes += r; work = 1; }
  }
  if(c->state==state_tunnel) {
    if(c->inbytes) {
      r = lst_conn_send(c->rfd, c->inb, c->inbytes);
      if(r<0) { eventlog(eventlog_level_warn,__FUNCTION__,"[%d] r-write error %d, closing", c->fd, lst_get_errcode()); return -1; }
      if(r>0) { c->inbytes -= r; if(c->inbytes) bcopy(c->inb+r, c->inb, c->inbytes); work = 1; }
    }
    if(c->outbytes<OUTBSIZE) {
      r = lst_conn_recv(c->rfd, c->outb+c->outbytes, OUTBSIZE-c->outbytes);
      if(r<0) { eventlog(eventlog_level_warn,__FUNCTION__,"[%d] r-read error %d, closing", c->fd, lst_get_errcode()); return -1; }
      if(r>0) { c->outbytes += r; work = 1; }
    }
  }
  if(!work) return 0;
  /* parse received */
  while(1) switch(c->state) {
  default:
    eventlog(eventlog_level_fatal,__FUNCTION__,"[%d] bad state %d - internal bug#2, exiting for safety", c->fd, c->state);
    abort();
  case state_connected:
    buf = c->inb;
    if(!(r=get_line((char*)buf, c->inbytes))) return 1;
    if(r<0) { eventlog(eventlog_level_warn,__FUNCTION__,"[%d] bad input, closing", c->fd); return -1; }
    n = r-1;
    if(n && buf[n-1]=='\r') n--;

    if(n>=FIRSTLINELEN) { eventlog(eventlog_level_warn,__FUNCTION__,"[%d] firstline too long, closing", c->fd); return -1; }
    bcopy(buf,fl,n); fl[n] = 0;
    c->inbytes -= r;
    if(c->inbytes) bcopy(buf+r, buf, c->inbytes);
    if(!(url=parse_firstline(c, fl))) { eventlog(eventlog_level_warn,__FUNCTION__,"[%d] bad firstline", c->fd); return -1; }
    if(c->method==M_CONNECT) {
      if(c->https) { eventlog(eventlog_level_warn,__FUNCTION__,"[%d] bad firstline - CONNECT twice", c->fd); return -1; }
      /* ignore all, just skip all headers and make proxy */
      c->state=state_https_pending;
      break;
    }
    if(!strncmp(url,"http://",7)) {
      url = strchr(url+7,'/');
      if(!url) { eventlog(eventlog_level_warn,__FUNCTION__,"[%d] bad url", c->fd); return -1; }
    }
    if(c->method==M_HEAD) mstr="HEAD"; else if(c->method==M_POST) mstr="POST"; else mstr="GET";
    snprintf(nl,sizeof(nl),"%s %s HTTP/%c.%c\r\n",mstr,url,c->v1,c->v2); /* it should fit because it shorter than original firstline */
    eventlog(eventlog_level_trace,__FUNCTION__,"[%d] %s",c->fd,nl);
    if(c->https) strcat(nl,"Proxy-Https: yes\r\n");
    /* place firstline back */
    n=strlen(nl);
    if(sizeof(c->inbytes)/*INBSIZE*/-c->inbytes<n) { eventlog(eventlog_level_fatal,__FUNCTION__,"err"); abort(); }
    bcopy(buf,buf+n,c->inbytes);
    c->inbytes+=n;
    bcopy(nl,buf,n);
    c->state=state_work;
    /* and connect to target */
    c->rfd = lst_conn_open(lst_parse_ipstr(TARGET_IP),80,LST_FLAG_NONBLOCK_AFTER);
    if(c->rfd<0) { eventlog(eventlog_level_error,__FUNCTION__,"[%d] failed connect to target, errcode %d",c->fd,lst_get_errcode()); return -1; }
    eventlog(eventlog_level_debug,__FUNCTION__,"connected to target");
    c->state = state_tunnel;
  case state_tunnel:
    return 1;
  case state_https_pending:
    buf = c->inb;
    if(!(r=get_line((char*)buf, c->inbytes))) return 1;
    if(r<0) { eventlog(eventlog_level_warn,__FUNCTION__,"[%d] bad input, closing", c->fd); return -1; }
    n = r-1;
    if(n && buf[n-1]=='\r') n--;
    c->inbytes -= r;
    if(c->inbytes) bcopy(buf+r, buf, c->inbytes);
    if(!n) {
      strcpy((char*)c->outb, "HTTP/1.1 200 Connection Established\r\nServer: proxy\r\n\r\n");
      c->outbytes = strlen((char*)c->outb);
      c->state = state_work;
      c->maxrecv = 0;
    }
    break;
  case state_work:
    if(c->outbytes) return 1;
    c->maxrecv = INBSIZE;
    c->fd = lst_conn_accept_ssl(c->fd, 0x0303, "./cacert.pem", "./cakey.pem", "./proxykey.pem", errmsg, sizeof(errmsg), LST_FLAG_NONBLOCK_AFTER);
    if(c->fd<0) {
      if(errno) eventlog(eventlog_level_error,__FUNCTION__,"[%d] failed init SSL, errno %d (%s)",c->fd,errno,strerror(errno));
      else eventlog(eventlog_level_error,__FUNCTION__,"[%d] failed init SSL, errmsg: %s",c->fd,errmsg);
      return -1;
    }
    eventlog(eventlog_level_debug,__FUNCTION__,"connected to target");
    c->state = state_connected;
    c->https = 1;
    return 1;
  }
}

static t_connection conns[MAXCONN];
static unsigned int nconns = 0;

static void wipe_zombies(void) { int status; while(waitpid(-1, &status, WNOHANG)>0); }

int main(unsigned int argc, char const * const * argv) {
  uint32 addr;
  uint16 port;
  int fd, afd, err, work, r;
  char tmp[32];
  t_connection * c;
  unsigned int n;
  char errmsg[1024];
  
  fd = lst_conn_listen(lst_parse_ipstr(IP),PORT,100,LST_FLAG_NONBLOCK_AFTER|LST_FLAG_REUSE_ADDR);
  if(fd<0) { eventlog(eventlog_level_fatal,__FUNCTION__,"listen errcode %d", lst_get_errcode()); return -1; }
  while(1) {
    work=0;
    wipe_zombies();
    /* accept */
    afd = lst_conn_accept(fd, &addr, &port, LST_FLAG_NONBLOCK_AFTER);
    if(afd<0) {
      err = lst_get_errcode();
      if(err!=LST_EWOULDBLOCK && err!=LST_EINTR) eventlog(eventlog_level_error,__FUNCTION__,"accept() errcode %d",err);
    } else {
      work=1;
      eventlog(eventlog_level_debug,__FUNCTION__,"[%d] accepted from %s",afd,lst_print_ip_port(tmp,addr,port));
      if(nconns>=MAXCONN) {
        eventlog(eventlog_level_warn,__FUNCTION__,"[%d] too many connections, dropping connection from %s",afd,lst_print_ip_port(tmp,addr,port));
        lst_conn_close(afd,0);
      } else {
        c = conns+nconns;
        c->state = state_connected;
        c->https = 0;
        c->addr = addr;
        c->port = port;
        c->fd = afd;
        c->rfd = -1;
        c->inbytes = 0;
        c->outbytes = 0;
        c->maxrecv = INBSIZE; // 20; // "CONNECT x HTTP/1.0\n\n"
//        c->rfd = -1;
        nconns++;
        
        
#ifdef RAW_HTTPS
    c->fd = lst_conn_accept_ssl(c->fd, 0x0303, "./cacert.pem", "./cakey.pem", "./proxykey.pem", errmsg, sizeof(errmsg), LST_FLAG_NONBLOCK_AFTER);
    if(c->fd<0) {
      if(errno) eventlog(eventlog_level_error,__FUNCTION__,"[%d] failed init SSL, errno %d (%s)",c->fd,errno,strerror(errno));
      else eventlog(eventlog_level_error,__FUNCTION__,"[%d] failed init SSL, errmsg: %s",c->fd,errmsg);
      c->state = state_closed;
    } else {
      eventlog(eventlog_level_debug,__FUNCTION__,"connected to target");
      c->state = state_connected;
    }
#endif
        
        
        
      }
    }
    /* process */
    for(n=0; n<nconns; n++)
#ifdef RAW_HTTPS
    if(conns[n].state!=state_closed)
#endif
    {
      if(r=conn_process(conns+n)) work=1;
      if(r<0) conns[n].state=state_closed;
    }
    /* delete closed */
    for(n=0; n<nconns; ) {
      c = conns+n;
      if(c->state!=state_closed) { n++; continue; }
      eventlog(eventlog_level_debug,__FUNCTION__,"[%d] closed from %s",c->fd,lst_print_ip_port(tmp,c->addr,c->port));
      lst_conn_close(c->fd,0);
      if(c->rfd>=0) lst_conn_close(c->rfd,0);
      nconns--;
      if(n<nconns) bcopy(conns+nconns, c, sizeof(*c));
      conns[nconns].state = 0; /* for safety */
    }
    /* idle */
    if(!work) usleep(10000);
  }
  return 0;
}

