/*
 * This file is part of fproxy - multipurpose http(s) proxy server
 * proxy.c - main source file
 * Copyright (c) 2017-2020,2022 firk (firk@cantconnect.ru)
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. The origin of this software must not be misrepresented; you must
 *    not claim that you wrote the original software.
 * 4. Altered versions in any form must be plainly marked as such, and
 *    must not be misinterpreted as being the original software.
 *
 * This software is provided by the author and contributors `as is'
 * without any express or implied warranty.
 */

#include "config.h"
#include <sys/types.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <strings.h>
#include <errno.h>
#include <signal.h>
#include <unistd.h>
#include <time.h>
#include <assert.h>
#include <sys/wait.h>
#include <sys/socket.h>
#include <fcl/types.h>
#include <fcl/strparse.h>
#include <libsimpletcp/libsimpletcp.h>
#include "eventlog.h"
#include "helpers/lst_ssl.h"
#include "util.h"
#include "http_packet.h"
#include "handle/handlereq.h"
#include "handle/rules.h"
#include "handle/rules.resolv.h"
#include "nslookup.h"
#include "dashboard.h"

static int watch_fd, my_num, internal_mode;
extern int is_internal(void) { return internal_mode; }

static volatile int do_exit = 0;
static void int_sig_handle(int signum) { do_exit = 1; }
static volatile int do_usr1 = 0;
static void usr1_sig_handle(int signum) { do_usr1 = 1; }

static void wipe_zombies(void) { int status; while(waitpid(-1, &status, WNOHANG)>0); }

static int parent_alive(void) {
  char c;
  int r;
  do { r = read(watch_fd, &c, 1); } while(r<0 && errno==EINTR);
  assert(r<=0); /* nobody will send data to here */
  if(r<0 && (errno==EWOULDBLOCK || errno==EAGAIN)) return 1;
  return 0;
}

static int child_main(int lfd);
       uint32 http_proxy_addr;
       uint16 http_proxy_port;

int main(int argc, char const * const * argv) {
  static pid_t childs[MAX_NUM_PROCESSES];
  int fd, ifd, sv[2], jj, is_debug, is_dump, is_rproxy;
  pid_t r;
  unsigned int n, minfree, num_processes;
  struct sigaction sa;
  char const * arg;
  uint16 listen_port;
  char const * confpath;
  
  eventlog(eventlog_level_info,__FUNCTION__,"Starting fproxy (v"FPROXY_VERSION")");
  eventlog(eventlog_level_info,__FUNCTION__,"Copyright (c) 2017-2021 firk (firk@cantconnect.ru)");
  eventlog(eventlog_level_info,__FUNCTION__,"Homepage: https://firk.cantconnect.ru/projects/fproxy/");
  
  listen_port = 0; num_processes = 0; is_debug = is_dump = is_rproxy = 0; confpath = NULL;
  for(jj=1; jj<argc; jj++) {
    arg = argv[jj];
    if(!strncmp(arg,"--conf=",7)) {
      if(confpath) { eventlog(eventlog_level_fatal,__FUNCTION__,"duplicate --conf= option"); return -1; }
      confpath = arg+7;
      continue;
    }
    if(!strcmp(arg,"--debug")) {
      is_debug = 1;
      continue;
    }
    if(!strcmp(arg,"--confdebug")) {
      is_dump = 1;
      continue;
    }
    if(!strcmp(arg,"--proxy")) {
      is_rproxy = 1;
      continue;
    }
    if(!strncmp(arg,"--port=",7)) {
      if(listen_port) { eventlog(eventlog_level_fatal,__FUNCTION__,"duplicate --port= option"); return -1; }
      if(str_to_uint16(arg+7, &listen_port)<0 || !listen_port) { eventlog(eventlog_level_fatal,__FUNCTION__,"bad --port= option"); return -1; }
      continue;
    }
    if(!strncmp(arg,"--numproc=",10)) {
      if(num_processes) { eventlog(eventlog_level_fatal,__FUNCTION__,"duplicate --numproc= option"); return -1; }
      if(str_to_uint(arg+10, &num_processes)<0 || !num_processes) { eventlog(eventlog_level_fatal,__FUNCTION__,"bad --numproc= option"); return -1; }
      continue;
    }
    
    eventlog(eventlog_level_fatal,__FUNCTION__,"unknown argument: %s", arg); return -1;
  }
  
  config_load(confpath);
  if(is_dump) { config_dump(); return 0; }
  if(is_rproxy) {
    http_proxy_addr = prefs.rproxy.host_data;
    http_proxy_port = prefs.rproxy.port_data;
    nslookup_set_proxy(prefs.rproxy.host_dns, prefs.rproxy.port_dns);
  }
  if(chdir(prefs.base_dir)<0) { eventlog(eventlog_level_fatal,__FUNCTION__,"chdir(%s) error %d (%s)", prefs.base_dir, errno, strerror(errno)); return -1; }
  if(!listen_port) listen_port = prefs.listen.port;
  if(!num_processes) num_processes = prefs.num_processes;
  if(num_processes>MAX_NUM_PROCESSES) {
    eventlog(eventlog_level_error,__FUNCTION__,"reducing number of processes from %u to %u", num_processes, (unsigned)MAX_NUM_PROCESSES);
    num_processes = MAX_NUM_PROCESSES;
  }
  
  rules_load(); // TODO:make it more user-friendly?
  resolv_rules_load(); // TODO:make it more user-friendly?
  
  /* master-slave watching socket */
  if(socketpair(PF_UNIX, SOCK_STREAM, 0, sv)<0) { eventlog(eventlog_level_fatal,__FUNCTION__,"socketpair() error %d (%s)", errno, strerror(errno)); return -1; }
  if(lst_conn_set_nonblock(sv[0])<0) { eventlog(eventlog_level_fatal,__FUNCTION__,"set_nonblock() error %d (%s)", errno, strerror(errno)); close(sv[0]); close(sv[1]); return -1; }
  
  /* listening socket */
  fd = lst_conn_listen(lst_parse_ipstr(prefs.listen.ip), listen_port, num_processes, LST_FLAG_REUSE_ADDR);
  if(fd<0) { eventlog(eventlog_level_fatal,__FUNCTION__,"listen errcode %d", lst_get_errcode()); close(sv[0]); close(sv[1]); return -1; }
  
  /* internal listening socket for AIA issuer certs downloads */
  if(is_debug) {
    ifd = -1;
    eventlog(eventlog_level_debug,__FUNCTION__,"skipping AIA listening due to debug mode");
  } else {
    unlink("./internal/http.sock");
    ifd = lst_conn_listen_unix("./internal/http.sock", num_processes, LST_FLAG_REUSE_ADDR);
    if(ifd<0) eventlog(eventlog_level_error,__FUNCTION__,"internal listen errcode %d - ignoring", lst_get_errcode());
  }
  
  /* this should be done after listening socket to avoid damaging of existing dashboard */
  dashboard_init();
  
  /* pre-fork */
  if(is_debug) goto single_process;
  for(n=0; n<num_processes; n++) {
    r = fork();
    if(r<0) {
      eventlog(eventlog_level_fatal,__FUNCTION__, "prefork failed - fork() #%u errno %d (%s)", n+1, errno, strerror(errno));
      while(n--) kill(childs[n], 9);
      return -1;
    }
    if(!r) goto child_process;
    childs[n] = r;
  }
  
  /* parent process */
  close(sv[0]);
  sa.sa_flags = 0; sigemptyset(&sa.sa_mask); sa.sa_handler = int_sig_handle;
  sigaction(SIGINT, &sa, NULL);
  sigaction(SIGQUIT, &sa, NULL);
  sigaction(SIGTERM, &sa, NULL);
  minfree = num_processes/6;
  if(minfree<2) minfree=2;
  while(!do_exit) {
    if(n=dashboard_longest_keepalive(num_processes, minfree)) if(n>=1 && n<=num_processes) kill(childs[n-1], SIGUSR1);
    usleep(100000);
  }
  for(n=0; n<num_processes; n++) kill(childs[n], 9);
  return 0;

child_process:
  close(sv[1]);
  watch_fd = sv[0];
  my_num = n+1;
  sa.sa_flags = SA_RESTART; sigemptyset(&sa.sa_mask); sa.sa_handler = usr1_sig_handle;
  sigaction(SIGUSR1, &sa, NULL);
  if(!n && ifd>=0) {
    internal_mode = 1;
    return child_main(ifd);
  }
  return child_main(fd);

single_process:
  if(ifd>=0) close(ifd);
  watch_fd = sv[0];
  my_num = 1;
  return child_main(fd);
}

static int start_ssl(int fd, char const * force_hostname) {
  int newfd;
  char errmsg[1024];
  newfd = lst_conn_accept_ssl_ex(fd, SSL_MINVER, SSL_CACERT, SSL_CAKEY, SSL_PROXYKEY, errmsg, sizeof(errmsg), LST_FLAG_NONBLOCK_AFTER, force_hostname);
  if(newfd<0) {
    if(errno) eventlog(eventlog_level_error,__FUNCTION__,"(%d) failed init SSL, errno %d (%s)",my_num,errno,strerror(errno));
    else eventlog(eventlog_level_error,__FUNCTION__,"(%d) failed init SSL, errmsg: %s",my_num,errmsg);
  } else {
    eventlog(eventlog_level_debug,__FUNCTION__,"(%d) enabled SSL",my_num);
  }
  return newfd;
}

#define INBSIZE 16384
#define OUTBSIZE 16384


typedef struct {
  int state;
  int fd;
  int ssl;
  uint32 addr;
  uint16 port;
  uint32 inbytes;
  uint32 outbytes;
  uint8 inb[INBSIZE];
  uint8 outb[OUTBSIZE];
  t_http_packet p;
  time_t start_time;
  time_t active_time;
} t_connection;

static int conn_process(t_connection * c);
static int child_main(int lfd) {
  int r, err;
  unsigned int n;
  char addrstr[50];
  static t_connection c;
  time_t now;

  dashboard_set_num(my_num);
again:
  c.p.hh.n = 0;
  c.p.hc.d = NULL;
  c.p.hc.use_fd = 0;
  while(1) {
    if(!parent_alive()) { lst_conn_close(lfd,0); return 0; }
    wipe_zombies();
    if(!internal_mode) c.fd = lst_conn_accept(lfd, &c.addr, &c.port, LST_FLAG_NONBLOCK_AFTER);
    else { c.addr=c.port=0; c.fd = lst_conn_accept_unix(lfd, LST_FLAG_NONBLOCK_AFTER); }
    if(c.fd>=0) {
      if(!internal_mode) lst_print_ip_port(addrstr,c.addr,c.port);
      else strcpy(addrstr,"INTERNAL");
      eventlog(eventlog_level_debug,__FUNCTION__,"(%d) accepted from %s",my_num,addrstr);
      break;
    }
    err = lst_get_errcode();
    if(err!=LST_EWOULDBLOCK && err!=LST_EINTR) eventlog(eventlog_level_error,__FUNCTION__,"(%d) accept() errcode %d",my_num,err);
  }

  dashboard_begin_request();
  c.state = state_connected;
  c.active_time = c.start_time = time(NULL);
  c.ssl = 0;
  c.inbytes = 0;
  c.outbytes = 0;
  if(prefs.listen.raw_https) {
    c.fd = start_ssl(c.fd, NULL);
    if(c.fd<0) { dashboard_end_request(); goto again; }
    c.ssl = 1; // comment this to allow CONNECT inside THIS ssl layer
  }

  /* initialized values: c.state=connected, c.fd>=0, c.ssl, c.addr, c.port, c.inbytes=0, c.outbytes=0, c.hh.n=0, c.hc.fd=-1 */
  while(1) {
    if(!parent_alive()) { lst_conn_close(lfd,0); lst_conn_close(c.fd,0); return 0; }
    r = conn_process(&c);
    time(&now);
    if(r>0) { c.active_time = now; continue; }
    if(r<0 || c.state==state_closed && !c.outbytes || prefs.connection_timeout.idle && now-c.active_time>(int)prefs.connection_timeout.idle) { /* TODO: signed overflow */
      eventlog(eventlog_level_debug,__FUNCTION__,"(%d) closed from %s",my_num,addrstr);
      lst_conn_close(c.fd,0);
      hh_clear(&c.p.hh);
      hc_clear(&c.p.hc);
      c.fd = -1;
      dashboard_end_request();
      goto again;
    }
    if(!r) usleep(10000);
  }
}
  
static int get_line(char * data, unsigned int len) {
  unsigned int n;
  for(n=0; n<len; n++) {
    if(!data[n]) return -1;
    if(data[n]=='\n') {
      if(n && data[n-1]=='\r') data[n-1]=0; else data[n]=0;
      return n+1;
    }
  }
  return 0;
}

static int get_raw_ip_from_CONNECT(char * dest, char const * src) {
  uint32 ip;
  uint16 port;
  if(lst_parse_addrstr(src,&ip,&port)!=3) return 0;
  lst_print_ip(dest, ip);
  return 1;
}

static int conn_process_request(t_connection * c);

static void drain_in(t_connection * c, uint32 count) {
  if(!count) return;
  assert(count<=c->inbytes);
  c->inbytes -= count;
  if(c->inbytes) bcopy(c->inb+count, c->inb, c->inbytes);
}

/* return: 0=nothing; 1=some work done; -1=need close */
static int conn_process(t_connection * c) {
  int r, work, rrr;
  unsigned int nn;
  uint8 * buf;
  char const * tmp;
  unsigned long clen, avail;
  char conn_host[4096];
  size_t j;
  
  work = 0;
  if(c->state!=state_closed &&
     c->state!=state_connected &&
     c->state!=state_https_pending2 &&
     c->state!=state_request_pending &&
     c->state!=state_content_pending &&
     c->state!=state_process_request) {
    eventlog(eventlog_level_fatal,__FUNCTION__,"(%d) bad state %d - internal bug, exiting", my_num, c->state);
    abort();
  }
  if(c->outbytes) {
    r = lst_conn_send(c->fd, c->outb, c->outbytes);
    if(r<0) { eventlog(eventlog_level_warn,__FUNCTION__,"(%d) write error %d, closing", my_num, lst_get_errcode()); return -1; }
    if(r>0) { c->outbytes -= r; if(c->outbytes) bcopy(c->outb+r, c->outb, c->outbytes); work = 1; }
  }
  if(c->state==state_closed) return work;
  if(c->inbytes<INBSIZE && c->state!=state_https_pending2 && c->state!=state_process_request) {
    r = lst_conn_recv(c->fd, c->inb+c->inbytes, INBSIZE-c->inbytes);
    if(r<0) { if(c->inbytes || c->state!=state_connected) eventlog(eventlog_level_warn,__FUNCTION__,"(%d) read error %d, closing", my_num, lst_get_errcode()); return -1; }
    if(r>0) { c->inbytes += r; work = 1; }
  }
  if(c->state==state_connected && !c->inbytes && !c->outbytes) {
     avail = (unsigned long)(time(NULL)-c->start_time);
     if(avail>prefs.connection_timeout.keepalive || avail>1 && do_usr1) { do_usr1 = 0; c->state=state_closed; return 0; }
  }
  do_usr1 = 0;
  if(!work) return 0;
  /* parse received */
  while(1) switch(c->state) {
  default:
    eventlog(eventlog_level_fatal,__FUNCTION__,"(%d) bad state %d - internal bug#2, exiting for safety", my_num, c->state);
    abort();
  case state_connected:
    if(c->outbytes) { c->start_time=time(NULL); return 1; } /* needed for keep-alive case of this state after some request processed */
    buf = c->inb;
    if(!(r=get_line((char*)buf, c->inbytes))) return 1;
    if(r<0) { eventlog(eventlog_level_warn,__FUNCTION__,"(%d) bad input, closing", my_num); return -1; }
    
    switch(hp_parse_firstline(&c->p, (char*)buf)) {
    case 0: break;
    case -2: eventlog(eventlog_level_warn,__FUNCTION__,"(%d) bad method '%s'", my_num, c->p.mstr); return -1;
    case -3: eventlog(eventlog_level_warn,__FUNCTION__,"(%d) missing version", my_num); return -1;
    case -4: eventlog(eventlog_level_warn,__FUNCTION__,"(%d) bad url", my_num); return -1;
    case -5: eventlog(eventlog_level_warn,__FUNCTION__,"(%d) url too long", my_num); return -1;
    case -6: eventlog(eventlog_level_warn,__FUNCTION__,"(%d) bad version", my_num); return -1;
    default: eventlog(eventlog_level_warn,__FUNCTION__,"(%d) bad firstline", my_num); return -1;
    }
    if(c->p.method==M_CONNECT && c->ssl) { eventlog(eventlog_level_warn,__FUNCTION__,"(%d) bad firstline - CONNECT inside SSL", my_num); return -1; }
    drain_in(c, r);
    c->state = state_request_pending;
    dashboard_set_request(c->p.method, c->ssl, "", c->p.raw_url);
    break;
  case state_https_pending2:
    assert(!c->inbytes); // should not happen
    if(c->outbytes) return 1;
    if(get_raw_ip_from_CONNECT(conn_host, c->p.raw_url)) c->fd = start_ssl(c->fd, conn_host);
    else c->fd = start_ssl(c->fd, NULL);
    if(c->fd<0) return -1; // error already reported
    c->state = state_connected;
    c->start_time = time(NULL);
    c->ssl = 1;
    return 1;
  case state_request_pending:
    assert(!c->outbytes); // should not happen
    buf = c->inb;
    if(!(r=get_line((char*)buf, c->inbytes))) return 1;
    if(r<0) { eventlog(eventlog_level_warn,__FUNCTION__,"(%d) bad input, closing", my_num); return -1; }
    if(*buf) {
      if(c->p.method!=M_CONNECT) if((rrr=hh_add(&c->p.hh, (char*)buf))<0) { eventlog(eventlog_level_warn,__FUNCTION__,"(%d) too many or too bad headers (%d), closing", my_num, rrr); return -1; }
      drain_in(c, r);
      break;
    }
    drain_in(c, r);
    if(c->p.method==M_CONNECT) {
      if(c->inbytes) { eventlog(eventlog_level_warn,__FUNCTION__,"(%d) extra input after CONNECT request, closing", my_num); return -1; }
      strcpy((char*)c->outb, "HTTP/1.1 200 Connection Established\r\nServer: proxy\r\n\r\n");
      c->outbytes = strlen((char*)c->outb);
      c->state = state_https_pending2;
      break;
    }
    if(tmp=hh_get(&c->p.hh,"Host")) dashboard_set_host(tmp);
    clen = 0;
    if((tmp=hh_get(&c->p.hh,"Content-Length")) && str_to_ulong(tmp,&clen)<0) { eventlog(eventlog_level_warn,__FUNCTION__,"(%d) bad Content-Length, closing", my_num); return -1; }
    dashboard_set_rclen(clen);
    c->p.clen = clen;
    if(clen) {
      if(c->p.method!=M_POST) { eventlog(eventlog_level_warn,__FUNCTION__,"(%d) non-zero Content-Length with non-POST method, closing", my_num); return -1; }
      if(hc_init(&c->p.hc, clen)<0) { eventlog(eventlog_level_error,__FUNCTION__,"(%d) failed to init content storage for clen=%lu, closing", my_num, clen); return -1; }
      c->state = state_content_pending;
    } else {
      if(c->inbytes) { eventlog(eventlog_level_warn,__FUNCTION__,"(%d) extra input after request headers, closing",my_num); return -1; }
      c->state = state_process_request;
      conn_process_request(c);
      return 1;
    }
    break;
  case state_content_pending:
    if(c->outbytes) return 1; /* prevent problems with request handler */
    if(c->inbytes) {
      avail = c->p.hc.total-c->p.hc.cur;
      if(c->inbytes>avail) { eventlog(eventlog_level_warn,__FUNCTION__,"(%d) extra input after request body, closing",my_num); return -1; }
      if(hc_add(&c->p.hc, c->inb, c->inbytes)<0) { eventlog(eventlog_level_error,__FUNCTION__,"(%d) error storing request body",my_num); return -1; }
      c->inbytes = 0;
      avail = c->p.hc.total-c->p.hc.cur;
      dashboard_set_rdlen(c->p.hc.cur);
      if(!avail) { c->state = state_process_request; conn_process_request(c); return 1; }
    }
    return 1;
  case state_process_request:
    assert(!c->inbytes); // should not happen
    return 1;
  }
}

#ifdef PROXY_DEBUG_DUMP_REQUEST
static void debug_dump_request(t_connection * c) {
  unsigned int n;
  
  eventlog(eventlog_level_trace,__FUNCTION__,"(%d) %s %s HTTP/%c.%c",my_num,method_get_name(c->p.method),c->p.raw_url,c->p.v1,c->p.v2);
  for(n=0; n<c->p.hh.n; n++) eventlog(eventlog_level_trace,__FUNCTION__,"(%d) %s: %s",my_num,c->p.hh.h[n*2],c->p.hh.h[n*2+1]);
}
#else
#define debug_dump_request(c) ((void)0)
#endif

static int conn_process_request(t_connection * c) {
  char const *mstr, *connmode;
  int conn_close;
  
  dashboard_received_request();
  assert(c->inbytes==0);
  assert(c->outbytes==0);
  debug_dump_request(c);
  conn_close = handle_request(c->fd, c->p.method, c->p.raw_url, c->p.v1, c->p.v2, c->ssl, &c->p.hh, &c->p.hc);

  if(!conn_close && (connmode=hh_get(&c->p.hh,"Connection")) && !strcasecmp(connmode,"keep-alive")) {
    c->state = state_connected;
    c->start_time = time(NULL);
    dashboard_keepalive();
  } else {
    c->state = state_closed;
  }
  hh_clear(&c->p.hh);
  hc_clear(&c->p.hc);

  return 0;
}




extern char const * method_get_name(char m) {
  switch(m) {
  case M_GET: return "GET";
  case M_HEAD: return "HEAD";
  case M_POST: return "POST";
  case M_CONNECT: return "CONNECT";
  case M_OPTIONS: return "OPTIONS";
  }
  abort();
  return NULL;
}
