/*
 * This file is part of fproxy - multipurpose http(s) proxy server
 * util.c - various utility functions
 * Copyright (c) 2018-2021 firk (firk@cantconnect.ru)
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. The origin of this software must not be misrepresented; you must
 *    not claim that you wrote the original software.
 * 4. Altered versions in any form must be plainly marked as such, and
 *    must not be misinterpreted as being the original software.
 *
 * This software is provided by the author and contributors `as is'
 * without any express or implied warranty.
 */

#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>
#include <string.h>
#include <strings.h>
#include <errno.h>
#include <fcntl.h>
#include <unistd.h>
#include <time.h>
#include <assert.h>
#include <sys/stat.h>
#include <fcl/types.h>
#include <fcl/malloc.h>
#include <fcrypt/sha256.h>
#include "eventlog.h"
#define UTIL_TIME
#include "util.h"

extern void * xmalloc(size_t sz) { return fcl_malloc_check(sz); }
extern void * xmallocz(size_t sz) { return fcl_mallocz_check(sz); }
extern char * xstrdup(char const * st) { return fcl_strdup_check(st); /* TODO: prevent NULL input -> NULL output? */ }
extern void * xmemdup(void const * p0, size_t sz) { return fcl_memdup_check(p0, sz); /* TODO: prevent NULL input -> NULL output? */ }

extern void strtolower(char * st) {
  char c;
  while(c=*st) {
    if(c>='A' && c<='Z') *st = c-'A'+'a';
    st++;
  }
}

extern char * skip_prefix(char * st, char const * prefix) {
  char c;
  while(c=*(prefix++)) {
    if((*st)!=c) return NULL;
    st++;
  }
  return st;
}

extern int write_all_2(int fd, void const * buf, size_t sz, unsigned long * inc) {
  char const * b;
  ssize_t wsz;
  size_t csz;
  b = buf;
  while(sz) {
    csz = sz;
    while(((ssize_t)csz)<=0) csz/=2;
    while((wsz = write(fd, b, csz))<0) if(errno!=EINTR && errno!=EWOULDBLOCK && errno!=EAGAIN) return -1;
    assert(wsz>0 && wsz<=(ssize_t)csz && csz<=sz);
    sz -= wsz;
    b += wsz;
    if(inc) (*inc) += wsz;
  }
  return 0;
}

extern ssize_t read_safe(int fd, void * buf, size_t sz) {
  ssize_t r;
  if(!sz) return 0;
  while(((ssize_t)sz)<=0) sz/=2;
  while((r = read(fd, buf, sz))<0) if(errno!=EINTR) return -1;
  return r;
}

static char b64(unsigned char c) { /* this is not standard base64 but just some 64 chars */
  if(c<26) return c+'A';
  if(c<52) return c-26+'a';
  if(c<62) return c-52+'0';
  if(c==62) return '-';
  return '_';
}

extern void compact_string(char * dest, char const * src) {
  uint8 hash[33];
  unsigned int i;
  sha256_hash(hash, src, strlen(src));
  hash[32] = 0;
  for(i=0; i<11; i++) {
    dest[i*4+0] = b64(hash[i*3] & 0x3F);
    dest[i*4+1] = b64((hash[i*3]>>6) | ((hash[i*3+1]<<2) & 0x3C));
    dest[i*4+2] = b64((hash[i*3+1]>>4) | ((hash[i*3+2]<<4) & 0x30));
    dest[i*4+3] = b64((hash[i*3+2]>>2));
  }
  dest[43] = 0;
}





extern void dynbuf_init(t_dynbuf * dynbuf, size_t init, size_t step, size_t max) {
  if(init<64) init=64;
  if(step<64) step=64;
  if(max<init) max=init;
  if(max>10000000) max=10000000;
  if(init>max) init=max;
  dynbuf->len = 0;
  dynbuf->alloc = init;
  dynbuf->step = step;
  dynbuf->max = max;
  dynbuf->overflowed = 0;
  dynbuf->buf = fcl_malloc_check(init);
}

/* ret. -1=file can't be opened; -2=file exists but can't read; -3=not enough space in dynbuf */
extern int  dynbuf_init_file(t_dynbuf * dynbuf, char const * fn) {
  return dynbuf_init_file_mintime(dynbuf, fn, 0);
}

extern int  dynbuf_init_file_mintime(t_dynbuf * dynbuf, char const * fn, time_t mintime) {
  int fd;
  off_t sz;
  ssize_t sz2, rsz;
  size_t sz3, pos;
  unsigned char * buf;
  dynbuf_init(dynbuf, 0, 0, 0); /* prevent garbage on return -1 */
  if((fd=open(fn, O_RDONLY))<0) return -1;
  if(mintime) {
    struct stat st;
    if(fstat(fd, &st)<0) { close(fd); return -2; }
    if(st.st_mtime<mintime) { close(fd); errno=ENOENT; return -1; }
  }
  if((sz=lseek(fd,0,SEEK_END))<0) { close(fd); return -2; }
  if(lseek(fd,0,SEEK_SET)<0) { close(fd); return -2; }
  sz2 = sz;
  if(sz2!=sz) { close(fd); return -3; }
  sz3 = sz2;
  if(!(buf=fcl_malloc(sz3))) { close(fd); return -3; }
  free(dynbuf->buf);
  dynbuf->buf = buf;
  dynbuf->alloc = sz3;
  if(dynbuf->max<sz3) dynbuf->max=sz3;
  
  for(pos=0; pos<sz3; ) {
    while((rsz = read(fd, buf+pos, sz3-pos))<0) if(errno!=EINTR) { close(fd); return -2; }
    if(rsz<=0) { close(fd); return -2; }
    pos += rsz;
    dynbuf->len = pos;
  }
  close(fd);
  return 0;
}


