/*
 * This file is part of fproxy - multipurpose http(s) proxy server
 * util/cache_add.c - insert datafile into hashed storage
 * Copyright (c) 2021 firk (firk@cantconnect.ru)
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. The origin of this software must not be misrepresented; you must
 *    not claim that you wrote the original software.
 * 4. Altered versions in any form must be plainly marked as such, and
 *    must not be misinterpreted as being the original software.
 *
 * This software is provided by the author and contributors `as is'
 * without any express or implied warranty.
 */

#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>
#include <assert.h>
#include <errno.h>
#include <dirent.h>
#include <fcntl.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/file.h>
#include <fcl/types.h>
#include <fcl/string.h>
#include <fcl/malloc.h>
#include <fcl/dynbufs.h>
#include "../gzip.h"
#include "../handle/cache.h"

extern void error_logf(char const * fmt, ...) {
  va_list arg;
  va_start(arg, fmt);
  vprintf(fmt, arg);
  fputc('\n',stdout);
  va_end(arg);
}

extern int write_all(int fd, void const * buf, size_t sz) {
  char const * b;
  ssize_t wsz;
  size_t csz;
  b = buf;
  while(sz) {
    csz = sz;
    while(((ssize_t)csz)<=0) csz/=2;
    while((wsz = write(fd, b, csz))<0) if(errno!=EINTR && errno!=EWOULDBLOCK && errno!=EAGAIN) return -1;
    assert(wsz>0 && wsz<=(ssize_t)csz && csz<=sz);
    sz -= wsz;
    b += wsz;
  }
  return 0;
}

static int is_cache, demo, v;
static char buf[65536];

int main(int argc, char * * argv) {
  char const *cmd, *fn;
  int fd;
  t_dynbuf data;
  ssize_t r;
  int rr;
  char hashp[80];
  
  if(argc!=2 && argc!=3) {
usage:
    fprintf(stderr,"Usage: %s <mode>[d][v] [<filename>]\n"
                   "  <mode> is c (cache_data) or s (saved_data)\n"
                   "  d is demo-mode (do not actually write anything)\n"
                   "  v is verbose-mode (display not only hash but also new/exising file)\n"
                   "  <filename> is file to insert (stdin used by default)\n",argv[0]);
    return -1;
  }
  cmd = argv[1];
  fn = (argc==3)?argv[2]:NULL;
  if(cmd[0]=='c') is_cache = 1; else if(cmd[0]=='s') is_cache = 0; else goto usage;
  if(cmd[1]=='d' && cmd[2]=='v' && !cmd[3]) demo = v = 1;
  else if(cmd[1]=='v' && cmd[2]=='d' && !cmd[3]) demo = v = 1;
  else if(cmd[1]=='d' && !cmd[2]) demo = 1;
  else if(cmd[1]=='v' && !cmd[2]) v = 1;
  else if(cmd[1]) goto usage;
  if(!fn) fd = 0; else {
    while((fd=open(fn,O_RDONLY|O_NOCTTY))<0) if(errno!=EINTR) {
      fprintf(stderr, "open(%s) error %d (%s)\n", fn, errno, strerror(errno));
      return -1;
    }
  }
  dynbuf_init(&data, 1000, 1000, (size_t)-1);
  if(data.overflowed) {
    fprintf(stderr, "out of memory or file too big (%s)\n", fn);
    return -1;
  }
  while(r=read(fd,buf,sizeof(buf))) {
    if(r<0) {
      if(errno==EINTR) continue;
      fprintf(stderr, "read(%s) error %d (%s)\n", fn, errno, strerror(errno));
      return -1;
    }
    dynbuf_write(&data, buf, r);
    if(data.overflowed) {
      fprintf(stderr, "out of memory or file too big (%s)\n", fn);
      return -1;
    }
  }
  if(fd>2) close(fd);

  rr = store_by_hash(is_cache, &data, hashp, demo);
  if(rr<0) { 
    fprintf(stderr, "data-write error for (%s)\n", fn);
    return -1;
  }
  /* rr=1 new file; rr=2 existing */
  if(v) fputs((rr==2)?"existing ":"new ",stdout);
  fputs(hashp,stdout);
  return 0;
}



