/*
 * This file is part of fproxy - multipurpose http(s) proxy server
 * util/cacheconv-78.c - cache db format converter to V78
 * Copyright (c) 2018-2021 firk (firk@cantconnect.ru)
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. The origin of this software must not be misrepresented; you must
 *    not claim that you wrote the original software.
 * 4. Altered versions in any form must be plainly marked as such, and
 *    must not be misinterpreted as being the original software.
 *
 * This software is provided by the author and contributors `as is'
 * without any express or implied warranty.
 */

#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>
#include <assert.h>
#include <errno.h>
#include <dirent.h>
#include <fcntl.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/file.h>
#include <fcl/types.h>
#include <fcl/string.h>
#include <fcl/malloc.h>
#include <fcl/dynbufs.h>
#include "../gzip.h"


extern int store_by_hash(int cache, t_dynbuf * content, char * hashp);
extern void error_logf(char const * fmt, ...) {
  va_list arg;
  va_start(arg, fmt);
  vprintf(fmt, arg);
  fputc('\n',stdout);
  va_end(arg);
}


static DIR * opendir_v(char const * dir) {
  DIR * dp;
  if(!(dp = opendir(dir))) printf("ERROR: opendir(%s) errno %d (%s)\n", dir, errno, strerror(errno));
  return dp;
}

static struct dirent * readdir_v(DIR * dp, char const * dir) {
  struct dirent * de;
  errno = 0;
  de = readdir(dp);
  if(!de && errno) printf("ERROR: readdir(%s) errno %d (%s)\n", dir, errno, strerror(errno));
  return de;
}

static void new_subdir(char * dest, size_t destsz, char const * base, char const * sub) {
  size_t s1, s2;
  s1 = strlen(base);
  s2 = strlen(sub);
  if(s1>=destsz || s2>=destsz-s1 || destsz-s1-s2<2) { printf("FATAL: too long name!\n"); abort(); }
  memcpy(dest, base, s1);
  dest[s1++] = '/';
  memcpy(dest+s1, sub, s2);
  dest[s1+s2] = 0;
}

static int is_cache, demo;

typedef struct { unsigned tot, new, hit, redir, alr, err; } counters;
/*
   tot = total processed (except errors or already)
   new = converted to new hash
   hit = converted to duplicated hash
   redir = converted redirect (no data at all)
   alr = already V78
   err = fail
 */
static void update_dir(char const * dir);
static void update_dir2(char const * dir, counters * cc);
static void update_dir3(char const * dir, counters * cc);
static int v = 0;
static int demo = 0;
int main(int argc, char * * argv) {
  if(argc==2 && strchr(argv[1],'v')) v = 1;
  if(argc==2 && strchr(argv[1],'d')) demo = 1;
  is_cache = 1;
  update_dir("./cache_real");
  is_cache = 0;
  update_dir("./saved_real");
  return 0;
}


static void update_dir(char const * dir) { /* /cache_real */
  DIR * dp;
  struct dirent *de;
  char sub[10000];
  counters tot, cur;
  
  tot = (counters) { };
  if(!(dp = opendir_v(dir))) return;
  printf("updating %s\n", dir);
  while(de=readdir_v(dp,dir)) {
    if(de->d_type!=DT_DIR) continue;
    if(de->d_name[0]=='.') continue;
    new_subdir(sub, sizeof(sub), dir, de->d_name);
    cur = (counters) { };
    update_dir2(sub, &cur);
    printf("[%s] converted %u (uniq=%u;dedup=%u;redir=%u), already (or hardlinked) %u, failed %u files\n",sub,cur.tot,cur.new,cur.hit,cur.redir,cur.alr,cur.err);
    tot.tot += cur.tot;
    tot.new += cur.new;
    tot.hit += cur.hit;
    tot.redir += cur.redir;
    tot.alr += cur.alr;
    tot.err += cur.err;
  }
  closedir(dp);
  printf("converted %u (uniq=%u;dedup=%u;redir=%u), already (or hardlinked) %u, failed %u files\n",tot.tot,tot.new,tot.hit,tot.redir,tot.alr,tot.err);
}

static void update_dir2(char const * dir, counters * cc) { /* cache_real/level2host */
  DIR * dp;
  struct dirent *de;
  char sub[10000];
  counters cur;
  
  if(!(dp = opendir_v(dir))) return;
  while(de=readdir_v(dp,dir)) {
    if(de->d_type!=DT_DIR) continue;
    if(de->d_name[0]=='.') continue;
    new_subdir(sub, sizeof(sub), dir, de->d_name);
    cur = (counters) { };
    update_dir3(sub, &cur);
    printf("[%s] converted %u (uniq=%u;dedup=%u;redir=%u), already (or hardlinked) %u, failed %u files\n",sub,cur.tot,cur.new,cur.hit,cur.redir,cur.alr,cur.err);
    cc->tot += cur.tot;
    cc->new += cur.new;
    cc->hit += cur.hit;
    cc->redir += cur.redir;
    cc->alr += cur.alr;
    cc->err += cur.err;
  }
  closedir(dp);
}

static void update_file(char const * dir, int dd, char const * fn, counters * cc);
static void update_dir3(char const * dir, counters * cc) { /* cache_real/level2host/fullhostname */
  DIR * dp;
  struct dirent *de;
  char const * oname;
  size_t olen;
  int dd;
  char sub[10000];
  int r;
  
  if(!(dp = opendir_v(dir))) return;
  printf("host %s\n", dir);
  if((dd = open(dir, O_DIRECTORY))<0) { printf("FATAL: open(%s) error %d (%s)\n", dir, errno, strerror(errno)); abort(); }
  while(de=readdir_v(dp,dir)) {
    if(de->d_type!=DT_REG) continue;
    if(de->d_name[0]=='.') continue;
    oname = de->d_name;
    olen = strlen(oname);
    if(oname[0]!='U' && oname[0]!='H' || oname[1]!=' ' || olen<5) { printf("ERROR: bad file %s/%s\n",dir,de->d_name); cc->err++; continue; }
    if(strcmp(oname+(olen-4),".sav") &&
       strcmp(oname+(olen-4),".bin")) { printf("ERROR: bad file %s/%s\n",dir,de->d_name); cc->err++; continue; }
    if(olen>=sizeof(sub)) abort();
    update_file(dir, dd, oname, cc);
  }
  close(dd);
  closedir(dp);
}

/* ret <0  fail; 0 = ok */
static int dynbuf_init_file(t_dynbuf * dynbuf, int dd, char const * fn, struct timespec * mt) {
  int fd;
  off_t sz;
  ssize_t sz2, rsz;
  size_t sz3, pos;
  unsigned char * buf;
  struct stat st;
  
  dynbuf_init(dynbuf, 0, 0, 0); /* prevent garbage on return -1 */
  if((fd=openat(dd, fn, O_RDONLY))<0) return -1;
  if(fstat(fd, &st)<0) { close(fd); return -2; }
  *mt = st.st_mtim;
  if((sz=lseek(fd,0,SEEK_END))<0) { close(fd); return -3; }
  if(lseek(fd,0,SEEK_SET)<0) { close(fd); return -4; }
  sz2 = sz;
  if(sz2!=sz) { close(fd); return -5; }
  sz3 = sz2;
  if(!(buf=fcl_malloc(sz3))) { close(fd); return -6; }
  free(dynbuf->buf);
  dynbuf->buf = buf;
  dynbuf->alloc = sz3;
  if(dynbuf->max<sz3) dynbuf->max=sz3;

  for(pos=0; pos<sz3; pos+=rsz) {
    while((rsz = read(fd, buf+pos, sz3-pos))<0) if(errno!=EINTR) { close(fd); return -7; }
    if(rsz<=0) { close(fd); return -8; }
    pos += rsz;
    dynbuf->len = pos;
  }
  close(fd);
  return 0;
}

extern int write_all(int fd, void const * buf, size_t sz) {
  char const * b;
  ssize_t wsz;
  size_t csz;
  b = buf;
  while(sz) {
    csz = sz;
    while(((ssize_t)csz)<=0) csz/=2;
    while((wsz = write(fd, b, csz))<0) if(errno!=EINTR && errno!=EWOULDBLOCK && errno!=EAGAIN) return -1;
    assert(wsz>0 && wsz<=(ssize_t)csz && csz<=sz);
    sz -= wsz;
    b += wsz;
  }
  return 0;
}

/*
  error: cc->err++;
  already V78: cc->alr++;
  converted to new unique data: cc->tot++, cc->new++;
  converted to already presetdata: cc->tot++; cc->hit++;
  converted redirect: cc->tot++; cc->redir++;
 */
extern int store_by_hash(int cache, t_dynbuf * content, char * hashp);
static void update_file(char const * dir, int dd, char const * fn, counters * cc) {
  t_dynbuf src, data;
  unsigned char * x;
  size_t xlen, xpos, pp, jj;
  int fd, rr;
  char hashp[80];
  struct timespec tvv[2];
  
  /* read source, check if it needs converting */
  if(dynbuf_init_file(&src, dd, fn, tvv)<0) { dynbuf_clear(&src); printf("ERROR: can't read file (%s/%s)\n", dir, fn); cc->err++; return; }
  tvv[1] = tvv[0];
  x = src.buf; xlen = src.len;
  if(xlen>=4 && !bcmp(x,"V78\n",4)) { dynbuf_clear(&src); if(v) printf("already converted file (%s/%s)\n", dir, fn); cc->alr++; return; }

  /* conversion needed */
  pp = 0;
  xpos = pp; while(pp<xlen && x[pp]!='\n') pp++; if(pp>=xlen) goto lfail; pp++;
  if(pp-xpos>=6 && !bcmp(x+xpos,"HOST=",5)) {
    xpos = pp; while(pp<xlen && x[pp]!='\n') pp++; if(pp>=xlen) goto lfail; pp++;
  }
  if(pp-xpos>=5 && !bcmp(x+xpos,"URL=",4)) {
    xpos = pp; while(pp<xlen && x[pp]!='\n') pp++; if(pp>=xlen) goto lfail; pp++;
  }
  if(pp-xpos>=5 && !bcmp(x+xpos,"ACAO\n",5)) {
    xpos = pp; while(pp<xlen && x[pp]!='\n') pp++; if(pp>=xlen) goto lfail; pp++;
  }
  if(pp-xpos>=7 && !bcmp(x+xpos,"REDIR:",6)) {
    if(!demo) {
      if((fd = openat(dd, fn, O_WRONLY|O_TRUNC))<0) { printf("ERROR: can't open file for writing (%s/%s) error %d (%s)\n", dir, fn, errno, strerror(errno)); dynbuf_clear(&src); cc->err++; return; }
      if(write_all(fd, "V78\n", 4)<0 || write_all(fd, x, xlen)<0 || futimens(fd,tvv)<0 || fsync(fd)<0) {
        printf("ERROR: write error (%s/%s) %d (%s)\n", dir, fn, errno, strerror(errno));
        close(fd);
        dynbuf_clear(&src); cc->err++; return;
      }
      close(fd);
    }
    if(v) printf("redirect converted file (%s/%s)\n", dir, fn);
    dynbuf_clear(&src);
    cc->tot++; cc->redir++; return;
  }
  for(jj=xpos; jj<pp-1; jj++) if(x[jj]<'0' || x[jj]>'9') goto lfail; /* status */
  xpos = pp; while(pp<xlen && x[pp]!='\n') pp++; if(pp>=xlen) goto lfail; pp++;
  /* now xpos points to content-type, pp points to data */

  data = (t_dynbuf) { .buf = x+pp, .len = xlen-pp };
  rr = store_by_hash(is_cache, &data, hashp);
  if(rr<0) { 
    printf("ERROR: data-write error for (%s/%s)\n", dir, fn);
    dynbuf_clear(&src); cc->err++; return;
  }

  if(!demo) {
    if((fd = openat(dd, fn, O_WRONLY|O_TRUNC))<0) { printf("ERROR: can't open file for writing (%s/%s) error %d (%s)\n", dir, fn, errno, strerror(errno)); dynbuf_clear(&src); cc->err++; return; }
    if(write_all(fd, "V78\n", 4)<0 || write_all(fd, x, pp)<0 || write_all(fd, hashp+16, strlen(hashp+16))<0 || futimens(fd,tvv)<0 || fsync(fd)<0) {
      printf("ERROR: write error (%s/%s) %d (%s)\n", dir, fn, errno, strerror(errno));
      close(fd);
      dynbuf_clear(&src); cc->err++; return;
    }
    close(fd);
  }
  if(v) printf("%s converted file (%s/%s)\n", (rr==2)?"dup":"new", dir, fn);
  dynbuf_clear(&src);
  if(rr==2) cc->hit++; else cc->new++;
  cc->tot++; return;
lfail: dynbuf_clear(&src); printf("ERROR: malformed file (%s/%s)\n", dir, fn); cc->err++; return;
}


/* from handle/cache.c */
#include <fcrypt/sha256.h>

static void b64hash(char * dest, uchar const * hash) {
  char const b64[]="ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz0123456789_-";
  size_t p;
  for(p=0; p<10; p++,dest+=4,hash+=3) {
    dest[0] = b64[hash[0] & 63];
    dest[1] = b64[(hash[0] >> 6) | ((hash[1] & 15) << 2)];
    dest[2] = b64[(hash[1] >> 4) | ((hash[2] & 3) << 4)];
    dest[3] = b64[(hash[2] >> 2) & 63];
  }
}

static int openat_nointr(int d, char const * name) {
  int fd;
  while((fd=openat(d,name,O_RDONLY))<0) if(errno!=EINTR) return -1;
  return fd;
}

#define IF_ERR_NOINTR(A) while((A)<0) if(errno!=EINTR)
/* always closes fd before return;
   ret. -1=i/o error; 0=file mismateches content; 1=file matches content */
static int compare_data(char const * fn, int fd, t_dynbuf * content, int gzip) {
  size_t p;
  ssize_t r;
  int rr, zerr;
  static char buf[65536];

  IF_ERR_NOINTR(flock(fd, LOCK_SH)) { error_logf("saving cache: flock(SH) file failed, error %d (%s), file (%s)", errno, strerror(errno), fn); close(fd); return -1; }
  if(!gzip) {
    for(p=0; p<=content->len; p+=r) {
      IF_ERR_NOINTR(r = read(fd,buf,sizeof(buf))) { error_logf("saving cache: file read error %d (%s), file (%s)", errno, strerror(errno), fn); close(fd); return -1; }
      if(p==content->len && !r) { close(fd); return 1; }
      if(p>=content->len || !r || (size_t)r>content->len-p || bcmp(buf, content->buf+p, r)) break;
    }
  } else {
    rr = gzip_decompressfd_check(content, fd, &zerr);
    if(rr == -3) { close(fd); return 0; }
    if(rr == 0) { close(fd); return 1; }
    close(fd); return -1;
  }
  close(fd);
  return 0;
}

extern int store_by_hash(int cache, t_dynbuf * content, char * hashp) {
  uint16 hashi;
  int fd, dd, rr, gz;
  uchar hash[32];
  t_dynbuf comp;

  sha256_hash(hash, (char*)content->buf, content->len);
  strcpy(hashp, cache?"./cache_data/":"./saved_data/");
  b64hash(hashp+16, hash);
  hashp[13] = hashp[16];
  hashp[14] = hashp[17];
  hashp[15] = 0;
  hashp[56] = '.';
  hashi = 0;
  mkdir(cache?"./cache_data/":"./saved_data/", 0700);
  mkdir(hashp, 0700);
  IF_ERR_NOINTR(dd = open(hashp, O_DIRECTORY)) { error_logf("saving cache: open directory failed, error %d (%s), dir (%s)", errno, strerror(errno), hashp); return -1; }
  hashp[15] = '/';
  
  do {
    IF_ERR_NOINTR(flock(dd, LOCK_EX)) { hashp[15]=0; error_logf("saving cache: flock() directory failed, error %d (%s), dir (%s)", errno, strerror(errno), hashp); close(dd); return -1; }

    /* try raw */
    sprintf(hashp+57,"%u",(uint)hashi);
    if((fd = openat_nointr(dd, hashp+16))>=0) {
      flock(dd,LOCK_UN);
      if(rr = compare_data(hashp, fd, content, 0)) { close(dd); return (rr<0)?-1:2; }
      continue;
    }
    if(errno!=ENOENT) { error_logf("saving cache: open file for reading failed, error %d (%s), file (%s)", errno, strerror(errno), hashp); close(dd); return -1; }

    /* try .gz */
    sprintf(hashp+57,"%u.gz",(uint)hashi);
    if((fd = openat_nointr(dd, hashp+16))>=0) {
      flock(dd,LOCK_UN);
      if(rr = compare_data(hashp, fd, content, 1)) { close(dd); return (rr<0)?-1:2; }
      continue;
    }
    if(errno!=ENOENT) { error_logf("saving cache: open file for reading failed, error %d (%s), file (%s)", errno, strerror(errno), hashp); close(dd); return -1; }

    /* create new */
    gz = 0;
    if(calc_randomless(content->buf, content->len)>=140) {
    // TODO: check first 1024 bytes, and 1024 bytes from middle and 1024 from the end
      dynbuf_init(&comp, 1000, 1000, 1000000000);
      if(gzip_compress(&comp, content, 0, &rr)!=0) { dynbuf_clear(&comp); error_logf("saving cache: gzip error %d, file (%s)", rr, hashp); close(dd); return -1; }
      if(comp.len>content->len-content->len/3) dynbuf_clear(&comp);
      else gz = 1;
    }
    if(!gz) sprintf(hashp+57,"%u",(uint)hashi);
    IF_ERR_NOINTR(fd = openat(dd, hashp+16, O_CREAT|O_EXCL|O_WRONLY, 0644)) { error_logf("saving cache: create file for writing failed, error %d (%s), file (%s)", errno, strerror(errno), hashp); close(dd); if(gz) dynbuf_clear(&comp); return -1; }
    IF_ERR_NOINTR(flock(fd, LOCK_EX)) { error_logf("saving cache: flock(EX) file failed, error %d (%s), file (%s)", errno, strerror(errno), hashp); close(fd); unlink(hashp); close(dd); if(gz) dynbuf_clear(&comp); return -1; }
    flock(dd,LOCK_UN);
    if(gz) content = &comp;
    if(write_all(fd,content->buf,content->len)<0 || fsync(fd)<0) {
      if(gz) dynbuf_clear(&comp);
      error_logf("saving cache: can't write file properly, file (%s)", hashp);
      ftruncate(fd, 0); /* do not unlink; this will lead to race problems */
      close(fd);
      close(dd);
      return -1;
    }
    if(gz) dynbuf_clear(&comp);
    close(fd); close(dd); return 1;
  } while(++hashi);

  error_logf("saving cache: collision counter overflow for file (%s/%s)", hashp, hashp+16);
  close(dd);
  return -1;
}

