/* (c) firk 2012-2016,2018,2019,2020,2021
   please respect the authorship
   feel free to use this code for anything */
/* 2016-11-27 incompatible change: addrs now stored in
   proper host byte order, not in reversed */
/* 2020-09-25 incompatible change: timeouts no longer
   capped at 999 ms */
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#ifdef WIN32
#include <winsock2.h>
#else
#include <unistd.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <sys/select.h>
#include <sys/param.h>
#include <netinet/in.h>
#include <netinet/tcp.h>
#include <arpa/inet.h>
#ifdef LST_SOCK_UNIX_SUPPORT
#include <sys/un.h>
#endif
#include <errno.h>
#include <fcntl.h>
#endif
#ifdef USE_POLL_H
#include <poll.h>
#ifndef USE_POLL
#define USE_POLL
#endif
#endif
#ifdef USE_SYS_POLL_H
#include <sys/poll.h>
#ifndef USE_POLL
#define USE_POLL
#endif
#endif
#include "libsimpletcp.h"

#ifndef HAVE_SOCKLEN_T
/* old BSD systems (and winsock) used int, new POSIX uses socklen_t=uint */
typedef int socklen_t;
#endif

extern int    lst_get_errcode(void) {
#ifdef WIN32
  switch(WSAGetLastError()) {
  case WSAEINTR:            return LST_EINTR;
  case WSAEBADF:            return LST_EBADF;
  case WSAEACCES:           return LST_EADDRINUSE;
  case WSAEFAULT:           return LST_EFAULT;
  case WSAEINVAL:           return LST_EINVAL;
  case WSAEMFILE:           return LST_ELIMIT;
  case WSAEWOULDBLOCK:      return LST_EWOULDBLOCK;
  case WSAEINPROGRESS:      return LST_EWOULDBLOCK;
  case WSAENOTSOCK:         return LST_EBADF;
  case WSAEDESTADDRREQ:
  case WSAEMSGSIZE:
  case WSAEPROTOTYPE:
  case WSAENOPROTOOPT:
  case WSAEPROTONOSUPPORT:
  case WSAESOCKTNOSUPPORT:
  case WSAEOPNOTSUPP:
  case WSAEPFNOSUPPORT:
  case WSAEAFNOSUPPORT:     return LST_EINVAL;
  case WSAEADDRINUSE:       return LST_EADDRINUSE;
  case WSAEADDRNOTAVAIL:    return LST_EINVAL;
  case WSAENETDOWN:
  case WSAENETUNREACH:
  case WSAENETRESET:
  case WSAEHOSTDOWN:
  case WSAEHOSTUNREACH:
  case WSAECONNABORTED:
  case WSAECONNRESET:
  case WSAETIMEDOUT:
  case WSAECONNREFUSED:
  case WSAENOTCONN:         return LST_ECONNRESET;
  case WSAENOBUFS:          return LST_ENOMEM;
  default:                  return LST_EUNKNOWN;
  }
#else
  int errn2;
  errn2 = errno; /* cache it, errno is not always just variable */
  /* switch not always ok because EWOULDBLOCK and EAGAIN
     may be "duplicate case" for example */
  if(errn2==EMFILE) return LST_ELIMIT;
  if(errn2==EINTR) return LST_EINTR;
  if(errn2==EWOULDBLOCK ||
     errn2==EINPROGRESS || errn2==EAGAIN) return LST_EWOULDBLOCK;
  if(errn2==EINVAL) return LST_EINVAL;
  if(errn2==ENOTSOCK || errn2==EBADF) return LST_EBADF;
  if(errn2==EFAULT) return LST_EFAULT;
  if(errn2==ENOBUFS || errn2==ENOMEM) return LST_ENOMEM;
  switch(errn2) {
  case EADDRNOTAVAIL:   return LST_EINVAL;
  case ENETDOWN:
  case ENETUNREACH:
  case ENETRESET:
  case EHOSTDOWN:
  case EHOSTUNREACH:
  case ECONNRESET:
  case ECONNREFUSED:
  case ECONNABORTED:
  case ETIMEDOUT:
  case ENOTCONN:
  case EPIPE:           return LST_ECONNRESET;
  case EDESTADDRREQ:
  case EMSGSIZE:
  case EPROTOTYPE:
  case ENOPROTOOPT:
  case EPROTONOSUPPORT:
  case ESOCKTNOSUPPORT:
  case EOPNOTSUPP:
  case EPFNOSUPPORT:
  case EAFNOSUPPORT:    return LST_EINVAL;
  case EADDRINUSE:      return LST_EADDRINUSE;
  default:              return LST_EUNKNOWN;
  }
#endif
}

static int parse_byte(char const * * src, uint8 * dest) {
  uint8 r, t, c;
  char const * s;
  int i;

  r = 0;
  i = 0;
  s = *src;
  while(c = (uchar)s[i]) {
    if(c<'0' || c>'9') break;
    c -= '0';
    t = r*10;
    if(t/10 != r) return 0;
    r = t+c;
    if(r<t) return 0;
    i++;
  }
  *dest = r;
  *src = s+i;
  return i;
}

static int parse_uint16(char const * * src, uint16 * dest) {
  uint16 r, t, c;
  char const * s;
  int i;

  r = 0;
  i = 0;
  s = *src;
  while(c = (uchar)s[i]) {
    if(c<'0' || c>'9') break;
    c -= '0';
    t = r*10;
    if(t/10 != r) return 0;
    r = t+c;
    if(r<t) return 0;
    i++;
  }
  *dest = r;
  *src = s+i;
  return i;
}

extern uint32 lst_parse_ipstr(char const * ipstr) {
  uint8 ip[4];
  if(!ipstr) return LST_INADDR_NONE;
  if(!parse_byte(&ipstr, ip+0)) return LST_INADDR_NONE;
  if((*(ipstr++))!='.') return LST_INADDR_NONE;
  if(!parse_byte(&ipstr, ip+1)) return LST_INADDR_NONE;
  if((*(ipstr++))!='.') return LST_INADDR_NONE;
  if(!parse_byte(&ipstr, ip+2)) return LST_INADDR_NONE;
  if((*(ipstr++))!='.') return LST_INADDR_NONE;
  if(!parse_byte(&ipstr, ip+3)) return LST_INADDR_NONE;
  if(*ipstr) return LST_INADDR_NONE;
  return ip[0]*0x1000000+ip[1]*0x10000+ip[2]*0x100+ip[3];
}

extern int lst_parse_addrstr(char const * st, uint32 * ip_r, uint16 * port_r) {
  uint8  ip[4];
  uint16 port;
  int res;
  if(!st) return -1;
  if(st[0]==':') { res = 0; goto parse_port; }
  if(!parse_byte(&st, ip+0)) return -1;
  if((*(st++))!='.') return -1;
  if(!parse_byte(&st, ip+1)) return -1;
  if((*(st++))!='.') return -1;
  if(!parse_byte(&st, ip+2)) return -1;
  if((*(st++))!='.') return -1;
  if(!parse_byte(&st, ip+3)) return -1;
  if(!st[0]) { *ip_r = ip[0]*0x1000000+ip[1]*0x10000+ip[2]*0x100+ip[3]; return 1; }
  if(st[0]!=':') return -1;
  res = 1;
parse_port:
  st++;
  if(st[0]==0 && res==0) return 0;
  if(!parse_uint16(&st, &port)) return -1;
  if(st[0]) return -1;
  if(res) *ip_r = ip[0]*0x1000000+ip[1]*0x10000+ip[2]*0x100+ip[3];
  *port_r = port;
  return res+2;
}

extern uint32 lst_parse_domainstr(char const * ipstr) {
  return LST_INADDR_NONE;
}

extern char * lst_print_ip(char * dest, uint32 addr) {
  if(dest)
/* this is just to avoid warnings; limit of 32 bytes will be never reached */
#ifdef USE_SNPRINTF
    snprintf(dest, 32,
#else
    sprintf(dest,
#endif
    "%u.%u.%u.%u", (unsigned)((addr>>24)&0xFF), (unsigned)((addr>>16)&0xFF), (unsigned)((addr>>8)&0xFF), (unsigned)(addr&0xFF));
  return dest;
}

extern char * lst_print_ip_port(char * dest, uint32 addr, uint16 port) {
  if(dest)
/* this is just to avoid warnings; limit of 32 bytes will be never reached */
#ifdef USE_SNPRINTF
    snprintf(dest, 32,
#else
    sprintf(dest,
#endif
    "%u.%u.%u.%u:%u", (unsigned)((addr>>24)&0xFF), (unsigned)((addr>>16)&0xFF), (unsigned)((addr>>8)&0xFF), (unsigned)(addr&0xFF), (unsigned)port);
  return dest;
}

extern int    lst_init(void) {
#ifdef WIN32
  WSADATA wsaData;
  if(WSAStartup(MAKEWORD(1,1), &wsaData)) return -1;
#endif
  return 0;
}

extern int    lst_conn_wait_connection(int sock, unsigned int ms) {
  int res, so_err;
  socklen_t so_sz;

  /* waiting for writable according to SUSv3 connect() and select() pages
     it says that socket will be marked as writable if:
      1) connection established
      2) connection failed and left a pending error;
     for winsock, it says that socket will be marked as except if error
      pending and says nothing about writable in this case;
     anyway we can check for error without getting signal from select()
      about it, probably in windows it will be with worst time granularity
     TODO: check exceptfds for windows?
   */
  res = lst_conn_wait_writable(sock, ms);
  if(res<0) return -1;
  so_sz = sizeof(so_err);
  if(getsockopt(sock, SOL_SOCKET, SO_ERROR, &so_err, &so_sz)<0) return -1;
  if(so_err) {
#ifdef WIN32
    WSASetLastError(so_err);
#else
    errno = so_err;
#endif
    return -1;
  }
  return res;
}

static int    _conn_wait_connection(int sock) {
  int res = 0;
  while(!res) res = lst_conn_wait_connection(sock, 1000);
  if(res<0) return -1;
  return 0;
}

extern int    lst_conn_open(uint32 addr, uint16 port, uint32 flags) {
  return lst_conn_open_from(0, 0, addr, port, flags);
}

extern int    lst_conn_open_from(uint32 b_addr, uint16 b_port, uint32 addr, uint16 port, uint32 flags) {
  struct sockaddr_in ba, sa;
  int sock;

  memset(&ba, 0, sizeof(ba));
  ba.sin_family = AF_INET;
  ba.sin_addr.s_addr = htonl(b_addr);
  ba.sin_port = htons(b_port);
  memset(&sa, 0, sizeof(sa));
  sa.sin_family = AF_INET;
  sa.sin_addr.s_addr = htonl(addr);
  sa.sin_port = htons(port);
  sock = socket(PF_INET, SOCK_STREAM, IPPROTO_TCP);
  if(sock<0) return -1;
  if((b_addr || b_port) && bind(sock, (struct sockaddr*)&ba, sizeof(ba))<0) {
    lst_conn_close(sock, LST_FLAG_IGNORE_ERRORS);
    return -1;
  }

  if(flags & LST_FLAG_NONBLOCK_BEFORE) {
    if(lst_conn_set_nonblock(sock)<0) {
      lst_conn_close(sock, LST_FLAG_IGNORE_ERRORS);
      return -1;
    }
    if(!connect(sock, (struct sockaddr*)&sa, sizeof(sa))) return sock;
    switch(lst_get_errcode()) {
    case LST_EINTR: case LST_EWOULDBLOCK: return sock;
    }
    lst_conn_close(sock, LST_FLAG_IGNORE_ERRORS);
    return -1;
  }
  if(connect(sock, (struct sockaddr*)&sa, sizeof(sa))) {
#ifndef WIN32
    /* WIN32 WSAEINTR has different meaning and not used win winsock2,
         but anyway can be ignored */
    if(lst_get_errcode()==LST_EINTR) {
      if(_conn_wait_connection(sock)>=0) goto fin;
    }
#endif
    lst_conn_close(sock, LST_FLAG_IGNORE_ERRORS);
    return -1;
  }
fin:
  if(flags & LST_FLAG_NONBLOCK_AFTER) {
    if(lst_conn_set_nonblock(sock)<0) {
      lst_conn_close(sock, LST_FLAG_IGNORE_ERRORS);
      return -1;
    }
  }
  return sock;
}

extern int    lst_conn_listen(uint32 addr, uint16 port, uint16 backlog, uint32 flags) {
  struct sockaddr_in sa;
  int sock;

  memset(&sa, 0, sizeof(sa));
  sa.sin_family = AF_INET;
  sa.sin_addr.s_addr = htonl(addr);
  sa.sin_port = htons(port);
  sock = socket(PF_INET, SOCK_STREAM, IPPROTO_TCP);
  if(sock<0) return -1;

  if(flags & LST_FLAG_REUSE_ADDR) {
    (void)lst_conn_set_reuseaddr(sock); /* non-critical if failed */
  }
  if(bind(sock, (struct sockaddr*)&sa, sizeof(sa))<0) {
    lst_conn_close(sock, LST_FLAG_IGNORE_ERRORS);
    return -1;
  }

  if(listen(sock, backlog)<0) {
    lst_conn_close(sock, LST_FLAG_IGNORE_ERRORS);
    return -1;
  }

  if(flags & LST_FLAG_NONBLOCK_AFTER) {
    if(lst_conn_set_nonblock(sock)<0) {
      lst_conn_close(sock, LST_FLAG_IGNORE_ERRORS);
      return -1;
    }
  }
  return sock;
}

extern int    lst_conn_accept(int lsock, uint32 * addr, uint16 * port, uint32 flags) {
  struct sockaddr_in sa;
  socklen_t sa_size;
  int sock;

  sa_size = sizeof(sa);
  memset(&sa, 0, sizeof(sa));
  sock = accept(lsock, (struct sockaddr*)&sa, &sa_size);
  if(sock>=0) {
    *addr = ntohl(sa.sin_addr.s_addr);
    *port = ntohs(sa.sin_port);
    if(flags & LST_FLAG_NONBLOCK_AFTER) {
      if(lst_conn_set_nonblock(sock)<0) {
        lst_conn_close(sock, LST_FLAG_IGNORE_ERRORS);
        return -1;
      }
    }
  }
  return sock;
}

extern int  lst_conn_open_unix(char const * path, uint32 flags) {
#ifdef WIN32
  WSASetLastError(WSAEPFNOSUPPORT); return -1;
#else
#ifndef LST_SOCK_UNIX_SUPPORT
  errno = EPFNOSUPPORT; return -1;
#else
  struct sockaddr_un sa;
  int sock;

  memset(&sa, 0, sizeof(sa));
  sa.sun_family = AF_UNIX;
  if(strlen(path)<1 || strlen(path)>=sizeof(sa.sun_path)) {
    errno = EADDRNOTAVAIL; return -1;
  }
  strncpy(sa.sun_path, path, sizeof(sa.sun_path)-1);
  sock = socket(AF_UNIX, SOCK_STREAM, 0);
  if(sock<0) return -1;

  if(flags & LST_FLAG_NONBLOCK_BEFORE) {
    if(lst_conn_set_nonblock(sock)<0) {
      lst_conn_close(sock, LST_FLAG_IGNORE_ERRORS);
      return -1;
    }
    if(!connect(sock, (struct sockaddr*)&sa, sizeof(sa))) return sock;
    switch(lst_get_errcode()) {
    case LST_EINTR: case LST_EWOULDBLOCK: return sock;
    }
    lst_conn_close(sock, LST_FLAG_IGNORE_ERRORS);
    return -1;
  }
  if(connect(sock, (struct sockaddr*)&sa, sizeof(sa))) {
#ifndef WIN32
    /* WIN32 WSAEINTR has different meaning and not used win winsock2,
         but anyway can be ignored */
    if(lst_get_errcode()==LST_EINTR) {
      if(_conn_wait_connection(sock)>=0) goto fin;
    }
#endif
    lst_conn_close(sock, LST_FLAG_IGNORE_ERRORS);
    return -1;
  }
fin:
  if(flags & LST_FLAG_NONBLOCK_AFTER) {
    if(lst_conn_set_nonblock(sock)<0) {
      lst_conn_close(sock, LST_FLAG_IGNORE_ERRORS);
      return -1;
    }
  }
  return sock;
#endif
#endif
}

extern int  lst_conn_listen_unix(char const * path, uint32 backlog, uint32 flags) {
#ifdef WIN32
  WSASetLastError(WSAEPFNOSUPPORT); return -1;
#else
#ifndef LST_SOCK_UNIX_SUPPORT
  errno = EPFNOSUPPORT; return -1;
#else
  struct sockaddr_un sa;
  int sock;

  memset(&sa, 0, sizeof(sa));
  sa.sun_family = AF_UNIX;
  if(strlen(path)<1 || strlen(path)>=sizeof(sa.sun_path)) { errno = EADDRNOTAVAIL; return -1; }
  strncpy(sa.sun_path, path, sizeof(sa.sun_path)-1);
  sock = socket(AF_UNIX, SOCK_STREAM, 0);
  if(sock<0) return -1;

  if(bind(sock, (struct sockaddr*)&sa, sizeof(sa))<0) {
    lst_conn_close(sock, LST_FLAG_IGNORE_ERRORS);
    return -1;
  }

  if(listen(sock, backlog)<0) {
    lst_conn_close(sock, LST_FLAG_IGNORE_ERRORS);
    return -1;
  }

  if(flags & LST_FLAG_NONBLOCK_AFTER) {
    if(lst_conn_set_nonblock(sock)<0) {
      lst_conn_close(sock, LST_FLAG_IGNORE_ERRORS);
      return -1;
    }
  }
  return sock;
#endif
#endif
}

extern int  lst_conn_accept_unix(int lsock, uint32 flags) {
#ifdef WIN32
  WSASetLastError(WSAEPFNOSUPPORT); return -1;
#else
#ifndef LST_SOCK_UNIX_SUPPORT
  errno = EPFNOSUPPORT; return -1;
#else
  struct sockaddr_un sa;
  socklen_t    sa_size;
  int    sock;

  sa_size = sizeof(sa);
  memset(&sa, 0, sizeof(sa));
  sock = accept(lsock, (struct sockaddr*)&sa, &sa_size);
  if(sock>=0) {
    if(flags & LST_FLAG_NONBLOCK_AFTER) {
      if(lst_conn_set_nonblock(sock)<0) {
        lst_conn_close(sock, LST_FLAG_IGNORE_ERRORS);
        return -1;
      }
    }
  }
  return sock;
#endif
#endif
}

extern int    lst_conn_set_nonblock(int sock) {
#ifdef WIN32
  u_long val;
  val = 1;
  if(!ioctlsocket(sock, FIONBIO, &val)) return 0;
  return -1;
#else
  int fl;
  fl = fcntl(sock, F_GETFL);
  if(fl==-1) return -1;
  if(!fcntl(sock, F_SETFL, fl|O_NONBLOCK)) return 0;
  return -1;
#endif
}

extern int    lst_conn_set_reuseaddr(int sock) {
  int v;
  v = 1;
  return setsockopt(sock, SOL_SOCKET, SO_REUSEADDR, (char*)&v, sizeof(v));
}

extern int    lst_conn_set_keepalive(int sock) {
  int v;
  v = 1;
  return setsockopt(sock, SOL_SOCKET, SO_KEEPALIVE, (char*)&v, sizeof(v));
}

extern int    lst_conn_set_tcp_nodelay(int sock) {
  int v;
  v = 1;
  return setsockopt(sock, IPPROTO_TCP, TCP_NODELAY, (char*)&v, sizeof(v));
}


extern int    lst_conn_is_readable(int sock) {
  return lst_conn_wait_readable(sock, 0);
}

extern int    lst_conn_is_writable(int sock) {
  return lst_conn_wait_writable(sock, 0);
}

extern int    lst_conn_wait_readable(int sock, unsigned int ms) {
#ifndef USE_POLL
  fd_set            fds;
  struct timeval    tv;
  int               res;
  FD_ZERO(&fds);
#ifdef WIN32
  if(sock<0) { WSASetLastError(WSAEBADF); return -1; }
  FD_SET((unsigned int)sock, &fds);
#else
  if(sock<0 || sock>=(int)FD_SETSIZE) { errno=EBADF; return -1; }
  /* TODO: more error reporting for this special error? */
  FD_SET(sock, &fds);
#endif
  tv.tv_sec = ms/1000;
  tv.tv_usec = ms%1000;
  tv.tv_usec *= 1000;
  res = select(sock+1, &fds, NULL, NULL, &tv);
  if(res==0) return 0;
  if(res>0) {
    if(FD_ISSET(sock, &fds)) return 1;
    return 0;
  }
  if(lst_get_errcode()==LST_EINTR) return 0;
  return -1;
#else
  struct pollfd fds;
  int res;
  while(0>(int)ms) ms/=2;
  fds.fd = sock;
  fds.events = POLLIN;
  fds.revents = 0;
  res = poll(&fds, 1, ms);
  if(!res) return 0;
  if(res>0) {
    /* tested on Linux: closed socket sets POLLHUP but not POLLIN */
    if(fds.revents & (POLLIN|POLLHUP)) return 1;
    return 0; /* can this happen? */
  }
  switch(lst_get_errcode()) {
  /* Linux's man poll says:
    On some other UNIX systems, poll() can fail with the error EAGAIN
    if the system fails to allocate kernel-internal resources, rather
    than ENOMEM as Linux does. POSIX permits this behavior. Portable
    programs may wish to check for EAGAIN and loop, just as with EINTR */
  case LST_EINTR:
  case LST_EWOULDBLOCK:
  case LST_ENOMEM:
    return 0;
  }
  return -1;
#endif
}

extern int    lst_conn_wait_writable(int sock, unsigned int ms) {
#ifndef USE_POLL
  fd_set            fds;
  struct timeval    tv;
  int               res;
  FD_ZERO(&fds);
#ifdef WIN32
  if(sock<0) { WSASetLastError(WSAEBADF); return -1; }
  FD_SET((unsigned int)sock, &fds);
#else
  if(sock<0 || sock>=(int)FD_SETSIZE) { errno=EBADF; return -1; }
  /* TODO: more error reporting for this special error? */
  FD_SET(sock, &fds);
#endif
  tv.tv_sec = ms/1000;
  tv.tv_usec = ms%1000;
  tv.tv_usec *= 1000;
  res = select(sock+1, NULL, &fds, NULL, &tv);
  if(res==0) return 0;
  if(res>0) {
    if(FD_ISSET(sock, &fds)) return 1;
    return 0;
  }
  if(lst_get_errcode()==LST_EINTR) return 0;
  return -1;
#else
  struct pollfd fds;
  int res;
  while(0>(int)ms) ms/=2;
  fds.fd = sock;
  fds.events = POLLOUT;
  fds.revents = 0;
  res = poll(&fds, 1, ms);
  if(!res) return 0;
  if(res>0) {
    /* tested on Linux: closed socket sets POLLOUT+POLLERR 
       but it is unclear if POLLOUT guaranted */
    if(fds.revents & (POLLOUT|POLLERR)) return 1;
    return 0; /* can this happen? */
  }
  switch(lst_get_errcode()) {
  /* Linux's man poll says:
    On some other UNIX systems, poll() can fail with the error EAGAIN
    if the system fails to allocate kernel-internal resources, rather
    than ENOMEM as Linux does. POSIX permits this behavior. Portable
    programs may wish to check for EAGAIN and loop, just as with EINTR */
  case LST_EINTR:
  case LST_EWOULDBLOCK:
  case LST_ENOMEM:
    return 0;
  }
  return -1;
#endif
}

extern void   lst_conn_close(int sock, uint32 flags) {
  /* TODO: error reporting from close() or
     application shouldn't rely on socket-level errors here?
     anyway it shouldn't rely on any error codes while checking
     if data succesfully received by remote program */
  int n;
#ifdef WIN32
  n = WSAGetLastError();
  if(flags & LST_FLAG_GRACEFUL) shutdown(sock, SD_BOTH);
  closesocket(sock);
  if(flags & LST_FLAG_IGNORE_ERRORS) WSASetLastError(n);
#else
  n = errno;
  if(flags & LST_FLAG_GRACEFUL) shutdown(sock, SHUT_RDWR);
  close(sock);
  if(flags & LST_FLAG_IGNORE_ERRORS) errno = n;
#endif
}

extern int    lst_conn_send(int sock, void const * buf, int bufsize) {
  int res;

  if(bufsize<1) return 0;
#ifdef WIN32
  res = send(sock, buf, bufsize, 0);
#else
  res = send(sock, buf, bufsize, MSG_NOSIGNAL);
#endif
  if(!res || res>bufsize) abort();
  if(res>0) return res;
  switch(lst_get_errcode()) {
  case LST_EINTR:
  case LST_EWOULDBLOCK:
  case LST_ENOMEM:
    return 0;
  }
  return -1;
}

extern int    lst_conn_send_all(int sock, void const * buf, int bufsize) {
  char const * b;
  int res;
  b = (char const*)buf;
  while(bufsize>0) {
    res = lst_conn_send(sock, b, bufsize);
    if(res<0) return -1;
    if(!res) { lst_conn_wait_writable(sock, 1000); continue; }
    b += res;
    bufsize -= res;
  }
  return 0;
}

extern int    lst_conn_recv(int sock, void * buf, int bufsize) {
  int res;

  if(bufsize<1) return 0;
  res = recv(sock, buf, bufsize, 0);
  if(res>0) return res;
  if(!res) {
#ifdef WIN32
    WSASetLastError(WSAECONNRESET);
#else
    errno = ECONNRESET;
#endif
    return -1;
  }
  switch(lst_get_errcode()) {
  case LST_EINTR:
  case LST_EWOULDBLOCK:
  case LST_ENOMEM:
    return 0;
  }
  return -1;
}

extern int    lst_conn_recv_all(int sock, void * buf, int bufsize) {
  int res, res2;

  if(bufsize<1) return 0;
  res2 = 0;
  while(res2<bufsize) {
    res = recv(sock, ((char*)buf)+res2, bufsize-res2, 0);
    if(res>0) {
      res2 += res;
      continue;
    }
    if(res2>0) return res2;
    if(!res) {
#ifdef WIN32
      WSASetLastError(WSAECONNRESET);
#else
      errno = ECONNRESET;
#endif
      return -1;
    }
    switch(lst_get_errcode()) {
    case LST_EINTR:
    case LST_EWOULDBLOCK:
    case LST_ENOMEM:
      return 0;
    }
    return -1;
  }
  return res2;
}

#ifndef HAVE_NOT_SIZE_T
#ifdef HAVE_NOT_SSIZE_T
typedef int ssize_t;
#endif
static ssize_t to_ssize(size_t x) {
  ssize_t sx;
  if(!x) return 0;
  sx = (ssize_t)x;
  while(sx<=0 || ((size_t)sx)!=x) { x/=2; sx=(ssize_t)x; }
  return sx;
}

extern int    lst_conn_s_send(int sock, void const * buf, size_t bufsize, size_t * rsize) {
  ssize_t res, wsize;

  *rsize = 0;
  if(bufsize<1) return 0;
  wsize = to_ssize(bufsize);
#ifdef WIN32
  res = send(sock, buf, wsize, 0);
#else
  res = send(sock, buf, wsize, MSG_NOSIGNAL);
#endif
  if(!res || res>wsize) abort();
  if(res>0) { *rsize = res; return 1; }
  switch(lst_get_errcode()) {
  case LST_EINTR:
  case LST_EWOULDBLOCK:
  case LST_ENOMEM:
    return 0;
  }
  return -1;
}

extern int    lst_conn_s_send_all(int sock, void const * buf, size_t bufsize) {
  char const * b;
  int res;
  size_t wsize;
  b = (char const*)buf;
  while(bufsize>0) {
    res = lst_conn_s_send(sock, b, bufsize, &wsize);
    if(res<0) return -1;
    if(!wsize) { lst_conn_wait_writable(sock, 1000); continue; }
    b += wsize;
    bufsize -= wsize;
  }
  return 0;
}

extern int    lst_conn_s_recv(int sock, void * buf, size_t bufsize, size_t * rsize) {
  ssize_t res, wsize;

  *rsize = 0;
  if(bufsize<1) return 0;
  wsize = to_ssize(bufsize);
  res = recv(sock, buf, wsize, 0);
  if(res>wsize) abort();
  if(res>0) { *rsize = res; return 1; }
  if(!res) {
#ifdef WIN32
    WSASetLastError(WSAECONNRESET);
#else
    errno = ECONNRESET;
#endif
    return -1;
  }
  switch(lst_get_errcode()) {
  case LST_EINTR:
  case LST_EWOULDBLOCK:
  case LST_ENOMEM:
    return 0;
  }
  return -1;
}

extern int    lst_conn_s_recv_all(int sock, void * buf, size_t bufsize, size_t * rsize) {
  char * b;
  ssize_t res, wsize;
  size_t res2;

  b = (char*)buf;
  *rsize = 0;
  if(bufsize<1) return 0;
  res2 = 0;
  while(bufsize>0) {
    wsize = to_ssize(bufsize);
    res = recv(sock, b, wsize, 0);
    if(res>wsize) abort();
    if(res>0) {
      res2 += res;
      b += res;
      bufsize -= res;
      continue;
    }
    if(res2>0) { *rsize = res2; return 1; }
    if(!res) {
#ifdef WIN32
      WSASetLastError(WSAECONNRESET);
#else
      errno = ECONNRESET;
#endif
      return -1;
    }
    switch(lst_get_errcode()) {
    case LST_EINTR:
    case LST_EWOULDBLOCK:
    case LST_ENOMEM:
      return 0;
    }
    return -1;
  }
  *rsize = res2; return 1;
}
#endif


